/**
 * @file RobotDimensions.h
 *
 * Description of the Dimensions of the Kondo Robot
 *
 * @author Cord Niehaus
 * Edited by Sam Barrett
 */

#ifndef __RobotDimensions_H__
#define __RobotDimensions_H__

#include <math/Vector3.h>
#include <math/Vector2.h>
#include <common/RobotInfo.h>

class RobotDimensions
{
public:

  float lengthBetweenLegs;         //!<length between leg joints LL1 and LR1
  float upperLegLength;            //!<length between leg joints LL2 and LL3 in z-direction
  float lowerLegLength;            //!<length between leg joints LL3 and LL4 in z-direction
 
  
	float zLegJoint1ToHeadPan;       //!<height offset between LL1 and head pan joint 
  
	Vector3<float> armOffset;             //! The offset of the first left arm joint relative to the middle between the hip joints
  float upperArmLength;            //! The length between the shoulder and the elbow in y-direction
  float lowerArmLength;            //!< height off lower arm starting at arm2/arm3
  float elbowOffsetY;
  
  float footHeight;

  // head stuff
  float headTiltOffset;
  float headPanOffset;

  float xHeadTiltToBottomCamera;
  float zHeadTiltToBottomCamera;
  float tiltOffsetToBottomCamera;
  float rollOffsetToBottomCamera;
  float xHeadTiltToTopCamera;
  float zHeadTiltToTopCamera;
  float tiltOffsetToTopCamera;
  float rollOffsetToTopCamera;

	float motionCycleTime;

  // foot stuff
  float footLength;
  float footWidth;
  float backOfFootToAnkle;
  float ankleToFootCenter;

  // fsrs - giving for left
  Vector3<float> leftFSROffsets[4];


	// Nao 3.3
  RobotDimensions() {
    //setNao33();
    setNaoH21();
  }

  void setNaoH21() {
    // Nao v4 with no motorized hands
		lengthBetweenLegs = 100;
		upperLegLength = 100;
		lowerLegLength = 102.90;
    zLegJoint1ToHeadPan = 211.5; // hipOffsetZ + neckOffsetZ
		armOffset = Vector3<float>(0,98,185); // z = hipOffsetZ + sholderOffsetZ
    upperArmLength = 105.0;
    lowerArmLength = 120.0; // estimated
    elbowOffsetY = 15.0;
    footHeight = 45.19;
    headTiltOffset = 0;
    headPanOffset = 0;
    xHeadTiltToBottomCamera = 48.8;
    zHeadTiltToBottomCamera = 23.81;
    //tistOffsetToBottomCamera(0.698132), // 40 degrees
    //tiltOffsetToBottomCamera = 0.837758; // 48 degrees
    tiltOffsetToBottomCamera = 0.767944871; // 44 degrees
    rollOffsetToBottomCamera = 0;
    xHeadTiltToTopCamera = 53.9;
    zHeadTiltToTopCamera = 67.9;
    tiltOffsetToTopCamera = 0; // 0 degrees
    rollOffsetToTopCamera = 0;
		motionCycleTime = 0.010;
    footLength = 160.0;
    footWidth = 75.0;
    backOfFootToAnkle = 0.33 * footLength;
    ankleToFootCenter = (footLength/2.0)-backOfFootToAnkle;
    
    leftFSROffsets[fsrLFL - fsrLFL] = Vector3<float>( 70.25f, 29.9f,0);
    leftFSROffsets[fsrLFR - fsrLFL] = Vector3<float>( 70.25f,-23.1f,0);
    leftFSROffsets[fsrLRL - fsrLFL] = Vector3<float>(-30.25f, 29.9f,0);
    leftFSROffsets[fsrLRR - fsrLFL] = Vector3<float>(-29.65f,-19.1f,0);
  }

  void setNao33() {
		lengthBetweenLegs = 100;
		upperLegLength = 100;
		lowerLegLength = 102.90;
    zLegJoint1ToHeadPan = 211.5;
		armOffset = Vector3<float>(0,98,185);
    upperArmLength = 105.0;
    lowerArmLength = 120.0; // non-upgraded and estimated
    elbowOffsetY = 0.0;
    footHeight = 45.11;
    headTiltOffset = 0;
    headPanOffset = 0;
    xHeadTiltToBottomCamera = 48.8;
    zHeadTiltToBottomCamera = 23.81;
    //tistOffsetToBottomCamera(0.698132), // 40 degrees
    //tiltOffsetToBottomCamera = 0.837758; // 48 degrees
    tiltOffsetToBottomCamera = 0.767944871; // 44 degrees
    rollOffsetToBottomCamera = 0;
    xHeadTiltToTopCamera = 53.9;
    zHeadTiltToTopCamera = 67.9;
    tiltOffsetToTopCamera = 0; // 0 degrees
    rollOffsetToTopCamera = 0;
		motionCycleTime = 0.010;
    footLength = 160.0;
    footWidth = 75.0;
    backOfFootToAnkle = 0.33 * footLength;
    ankleToFootCenter = (footLength/2.0)-backOfFootToAnkle;
	}
 
    // Nao 3,2
  void setNao32() {
		lengthBetweenLegs = 100;
		upperLegLength = 100;
		lowerLegLength = 102.75;
		zLegJoint1ToHeadPan = 211.5;
		armOffset = Vector3<float>(0,98,185);
		upperArmLength = 90;
		lowerArmLength = 120.0; // estimated
    elbowOffsetY = 0.0;
    footHeight = 45.11;
    headTiltOffset = 0;
    headPanOffset = 0;
    xHeadTiltToBottomCamera = 48.8;
    zHeadTiltToBottomCamera = 23.81;
    //tistOffsetToBottomCamera(0.698132), // 40 degrees
    //tiltOffsetToBottomCamera = 0.837758; // 48 degrees
    tiltOffsetToBottomCamera = 0.767944871; // 44 degrees
    rollOffsetToBottomCamera = 0;
    xHeadTiltToTopCamera = 53.9;
    zHeadTiltToTopCamera = 67.9;
    tiltOffsetToTopCamera = 0; // 0 degrees
    rollOffsetToTopCamera = 0;
		motionCycleTime = 0.010;
    footLength = 160.0;
    footWidth = 75.0;
    backOfFootToAnkle = 0.33 * footLength;
    ankleToFootCenter = (footLength/2.0)-backOfFootToAnkle;
	}

  void setCameraParameters(float tiltBotCam, float rollBotCam, float tiltTopCam, float rollTopCam) {
    tiltOffsetToBottomCamera = tiltBotCam;
    rollOffsetToBottomCamera = rollBotCam;
    tiltOffsetToTopCamera = tiltTopCam;
    rollOffsetToTopCamera = rollTopCam;
  }

  void setHeadOffsets(float tilt, float pan) {
    headTiltOffset = tilt;
    headPanOffset = pan;
  }

//# forward offset head tilt to camera
//48.8

//# height offset head tilt to camera
//23.81

//# head tilt to camera tilt (in radians)
//0.698132

//# forward offset head tilt to upper camera
//53.9

//# height offset head tilt to upper camera
//67.9

//# head tilt to upper camera tilt (in radians)
//0.0

//# image recording time (in s, for simulation)
//0

//# image recording delay (in s, for simulation)
//0
};

#endif
