#ifndef _WORLD_OBJECT_H
#define _WORLD_OBJECT_H

#include <math/Geometry.h>

enum {   // Types of objects
  WO_BALL = 0, 

  // Self and Teammates
  WO_TEAM1 = 1,
  WO_TEAM2 = 2,
  WO_TEAM3 = 3,
  WO_TEAM4 = 4,

  // Opponents?
  WO_OPPONENT1 = 5,
  WO_OPPONENT2 = 6,
  WO_OPPONENT3 = 7,
  WO_OPPONENT4 = 8,

  // center cirlce
  WO_CENTER_CIRCLE = 9,

  // 2 Landmark goals
  WO_OWN_GOAL = 10,
  WO_OPP_GOAL = 11,

  WO_OWN_LEFT_GOALPOST = 12,
  WO_OPP_LEFT_GOALPOST = 13,

  WO_OWN_RIGHT_GOALPOST = 14,
  WO_OPP_RIGHT_GOALPOST = 15,

  // Unknown goals and goal posts
  WO_UNKNOWN_GOAL = 16,
  WO_UNKNOWN_LEFT_GOALPOST = 17,
  WO_UNKNOWN_RIGHT_GOALPOST = 18,
  WO_UNKNOWN_GOALPOST = 19,

  // Field Corners
  WO_UNKNOWN_L_1 = 20,
  WO_UNKNOWN_L_2 = 21,
  WO_UNKNOWN_T_1 = 22,
  WO_UNKNOWN_T_2 = 23,

  // L intersections (yellow to blue, top to bottom)
  WO_OPP_FIELD_LEFT_L = 24,
  WO_OPP_FIELD_RIGHT_L = 25,
  WO_OPP_PEN_LEFT_L = 26,
  WO_OPP_PEN_RIGHT_L = 27,
  WO_OWN_PEN_RIGHT_L = 28,
  WO_OWN_PEN_LEFT_L = 29,
  WO_OWN_FIELD_RIGHT_L = 30,
  WO_OWN_FIELD_LEFT_L = 31,

  // T intersections (yellow to blue, top to bottom)
  WO_OPP_PEN_LEFT_T = 32,
  WO_OPP_PEN_RIGHT_T = 33,
  WO_CENTER_TOP_T = 34,
  WO_CENTER_BOTTOM_T = 35,
  WO_OWN_PEN_RIGHT_T = 36,
  WO_OWN_PEN_LEFT_T = 37,

  WO_UNKNOWN_FIELD_LINE_1 = 38,
  WO_UNKNOWN_FIELD_LINE_2 = 39,
  WO_UNKNOWN_FIELD_LINE_3 = 40,
  WO_UNKNOWN_FIELD_LINE_4 = 41,


  // Horizontal Lines (right to left)
  WO_OPP_GOAL_LINE = 42,          
  WO_OPP_PENALTY = 43,
  WO_CENTER_LINE = 44,
  WO_OWN_PENALTY = 45,
  WO_OWN_GOAL_LINE = 46,       

  // Vertical Lines (top to bottom)
  WO_TOP_SIDE_LINE = 47,    
  WO_PENALTY_TOP_OPP = 48,   
  WO_PENALTY_TOP_OWN = 49,
  WO_PENALTY_BOTTOM_OPP = 50, 
  WO_PENALTY_BOTTOM_OWN = 51,       
  WO_BOTTOM_SIDE_LINE = 52,   

  // penalty crosses
  WO_UNKNOWN_PENALTY_CROSS = 53,
  WO_OWN_PENALTY_CROSS = 54,
  WO_OPP_PENALTY_CROSS = 55,

  // cluster o' robots
  WO_ROBOT_CLUSTER = 56,

  NUM_WORLD_OBJS = 57,

  WO_INVALID = 58
};

// some constants to easily see which class each type is
const int LANDMARK_OFFSET = WO_CENTER_CIRCLE;
const int NUM_LANDMARKS = WO_OPP_RIGHT_GOALPOST - LANDMARK_OFFSET + 1;
const int INTERSECTION_OFFSET =  WO_OPP_FIELD_LEFT_L;
const int NUM_INTERSECTIONS = WO_OWN_PEN_LEFT_T - INTERSECTION_OFFSET + 1;
const int LINE_OFFSET =  WO_OPP_GOAL_LINE;
const int NUM_LINES = WO_BOTTOM_SIDE_LINE - LINE_OFFSET + 1;

const int L_INT_OFFSET = WO_OPP_FIELD_LEFT_L;
const int T_INT_OFFSET = WO_OPP_PEN_LEFT_T;

const int NUM_L_INTS = 8;
const int NUM_T_INTS = 6;

const int CROSS_OFFSET = WO_OWN_PENALTY_CROSS;
const int NUM_CROSSES = 2;

const int NUM_UNKNOWN_L = 2;
const int NUM_UNKNOWN_T = 2;
const int NUM_UNKNOWN_LINES = 4;
const int NUM_UNKNOWN_POSTS = 3;


class WorldObject {
public:

  WorldObject();
  ~WorldObject();

  // quick functions to determine which class of object this is
  bool isLine();
  bool isLandmark();
  bool isGoal();
  bool isGoalCenter();
  bool isGoalPost();
  bool isUnknownPost();
  bool isUnknownGoalCenter();
  bool isUnknownGoal();
  bool isBall();
  bool isIntersection();
  bool isT();
  bool isL();
  bool isTeammate();
  bool isOpponent();
  bool isUnknownIntersection();
  bool isUnknownLine();
  bool isCenterCircle();
  bool isUnknownPenaltyCross();
  bool isKnownPenaltyCross();
  bool isUnique();
  bool isAmbiguous();
  bool isPenaltySideline();
  bool isOwnGoal();
  bool isOppGoal();
  bool isKnownLine();
  bool isUnknownT();
  bool isUnknownL();

  float calcGroundDistance();
  void reset();

  bool seen;
  int frameLastSeen;
  int type;


  // These are relative to the robot and come directly from vision
  float visionDistance;
  float visionElevation;
  AngRad visionBearing;
  float visionConfidence;

  // These are relative to the robot but come from the filtered localization
  float distance;
  float elevation;
  AngRad bearing;  
  //float sdDistance;   // Commented out ones are things we havn't decided on
  //float sdElevation;
  //AngRad sdBearing;

  Vector2D relVel;
  
  //Vector2D sdRelVel;

  // These are the global position after localization (including confidence)
  // or permanent locations for landmarks and unmovable objects
  Point2D loc;
  Point2D endLoc;

  // for lines
  // for unknown lines this is the relative x,y (x in front, y to left) of the end points of the observed line segment
  // for known lines this is the location on the field
  Line2D lineLoc;

  // for seen lines
  Point2D visionPt1;
  Point2D visionPt2;
  Line2D visionLine;

  // heights of beacons/goals/etc in the world
  float lowerHeight; 
  float upperHeight;

  AngRad orientation;
  Point2D sd;
  AngRad sdOrientation;
  Vector2D absVel;

  Point2D relPos;
  AngRad relOrientation;

  // Vision
  int imageCenterX;
  int imageCenterY;

  int fieldLineIndex;
  int ballBlobIndex;

  bool fromTopCamera; // true if observed from top cam, false if from bottom
};


#endif
