#include "KinematicsModule.h"

#include <kinematics/ForwardKinematics.h>
#include <common/RobotInfo.h>

#include <memory/BodyModelBlock.h>
#include <memory/JointBlock.h>
#include <memory/SensorBlock.h>
#include <memory/RobotInfoBlock.h>

void KinematicsModule::specifyMemoryDependency() {
  requiresMemoryBlock("processed_sensors");
  requiresMemoryBlock("processed_joint_angles");
  requiresMemoryBlock("body_model");
  requiresMemoryBlock("robot_info");
}

void KinematicsModule::specifyMemoryBlocks() {
  getMemoryBlock(sensors_,"processed_sensors");
  getMemoryBlock(joints_,"processed_joint_angles");
  getMemoryBlock(body_model_,"body_model");
  getMemoryBlock(robot_info_,"robot_info");

  frames_in_air_ = 0;
}

void KinematicsModule::calculatePose() {
  ForwardKinematics::calculateRelativePose(joints_->values_, sensors_->values_[angleX], sensors_->values_[angleY], body_model_->rel_parts_, robot_info_->dimensions_);
  ForwardKinematics::calculateRelativePose(joints_->values_, 0, 0, body_model_->rel_parts_no_rotations_, robot_info_->dimensions_);
  ForwardKinematics::calculateAbsolutePose(body_model_->rel_parts_, body_model_->abs_parts_);
  ForwardKinematics::calculateCoM(body_model_->abs_parts_,body_model_->center_of_mass_, robot_info_->mass_calibration_);
  body_model_->is_calculated_ = true;
  
  // Fill body model with some different estimates of body tilt and roll
  body_model_->left_foot_body_tilt_roll_ = ForwardKinematics::calculateTiltRollFromLeg(true,joints_->values_, robot_info_->dimensions_);
  body_model_->right_foot_body_tilt_roll_ = ForwardKinematics::calculateTiltRollFromLeg(false,joints_->values_, robot_info_->dimensions_);
  body_model_->sensors_tilt_roll_.roll_ = sensors_->values_[angleX];
  body_model_->sensors_tilt_roll_.tilt_ = sensors_->values_[angleY];

  // See if feet are on ground
  float totalForce = 0;
  for (int i=fsrLFL; i<=fsrRRR; i++) {
    totalForce += sensors_->values_[i];
  }
  if (totalForce < 1.0) {
    frames_in_air_++;
    body_model_->feet_on_ground_inst_ = false;
  } else {
    frames_in_air_=0;
    body_model_->feet_on_ground_inst_ = true;
  }
  if (frames_in_air_ > 25) {
    if (body_model_->feet_on_ground_) std::cout << "BodyModel: I'm flying !\n";
    body_model_->feet_on_ground_=false;
  } else {
    body_model_->feet_on_ground_=true;
  }

  calcRelCenterOfMassFromFSRs();
}

void KinematicsModule::calcRelCenterOfMassFromFSRs() {
  body_model_->zmpFromFSRs.x = 0;
  body_model_->zmpFromFSRs.y = 0;
  Vector3<float> point;
  float totalForce = 0;
  int ind;
  BodyPart::Part foot = BodyPart::left_bottom_foot;
  int sign = 1;
  for (int i = fsrLFL; i <= fsrRRR; i++) {
    ind = (i - fsrLFL) % 4;
    if ((i - fsrLFL) == 4) {
      //std::cout << " | ";
      foot = BodyPart::right_bottom_foot;
      sign = -1;
    }
    Pose3D &startPos = body_model_->rel_parts_[foot];
    Vector3<float> offset = robot_info_->dimensions_.leftFSROffsets[ind];
    offset.y *= sign;

    point = Pose3D(startPos).translate(offset).translation;
    if (sensors_->values_[i] > 0.1) {
      body_model_->zmpFromFSRs.x += point.x * sensors_->values_[i];
      body_model_->zmpFromFSRs.y += point.y * sensors_->values_[i];
      totalForce += sensors_->values_[i];
    }
    //std::cout << sensors_->values_[i] << " ";
    //std::cout << point.x << " " << point.y << " | ";
    //std::cout << startPos.translation.y << " ";
  }
  //std::cout << std::endl;
  // normalize for the total force
  body_model_->zmpFromFSRs /= totalForce;

  //std::cout << body_model_->zmpFromFSRs.x << " " << body_model_->zmpFromFSRs.y << std::endl;
}
