/**
   7 states calculation
   Xhat[0][0]=x
   Xhat[1][0]=y
   Xhat[2][0]=orientation
   Xhat[3][0]=ballX
   Xhat[4][0]=ballY
   Xhat[5][0]=ballVeocityX
   Xhat[6][0]=ballVeocityY
**/

#include "UKF.h"
#include <iostream>

#include <memory/TextLogger.h>

UKF::UKF()
{
  /**************************Initialization**************************/
  /*This is where values can be adjusted*/

  alpha = 1.0; // Accuracy of model (0.0 -> 1.0)
  isActive = false; // Model currently in use.
  toBeActivated = false; // Model to be in use.
  framesSinceAmbig = 0;
  ballFramesSinceAmbig = 0;

  // Set Update Uncertainty
  updateUncertainties = NMatrix(7, 7, true);
  updateUncertainties[5][5] = 0.975; // Ball velocity x
  updateUncertainties[6][6] = 0.975; // Ball velocity y
  
  // Todd: update these in timeUpdate now (since timePassed may vary)
  //updateUncertainties[3][5] = 1.0f/frameRate; // [ballX][ballXvelocity]
  //updateUncertainties[4][6] = 1.0f/frameRate; // [ballY][ballYvelocity]

  //Initialisation of Xhat and S
  init();

  // Process Noise - NMatrix Square Root of Q
  // For regular update
  // Todd: I doubled these
  sqrtOfProcessNoise = NMatrix(7,7,true);
  sqrtOfProcessNoise[0][0] = 2.0* 2.0; // Robot X coord.
  sqrtOfProcessNoise[1][1] = 2.0* 2.0; // Robot Y coord.
  sqrtOfProcessNoise[2][2] = 2.0* 0.0316228; // Robot Theta.
  sqrtOfProcessNoise[3][3] = 2.0* 4.0; // Ball X.
  sqrtOfProcessNoise[4][4] = 2.0* 4.0; // Ball Y.
  sqrtOfProcessNoise[5][5] = 2.0* 3.16228; // Ball X Velocity.
  sqrtOfProcessNoise[6][6] = 2.0* 3.16228; // Ball Y Velocity.

  //RHM 7/7/08: Extra matrix for resetting
  // Process noise for used for reset.
  //  sqrtOfProcessNoiseReset = NMatrix(7,7,false);
  //  sqrtOfProcessNoiseReset[0][0] = 50.0; // extra 50cm sd when kidnapped?
  //  sqrtOfProcessNoiseReset[1][1] = 50.0; // extra 50cm sd when kidnapped?
  //sqrtOfProcessNoiseReset[2][2] = 0.25; // extra 15deg shift when kidnapped? 0.25
  //  sqrtOfProcessNoiseReset[2][2] = 0.5; // extra 15deg shift when kidnapped? 0.25
  //  sqrtOfProcessNoiseReset[3][3] = 20.0; // ball itself shouldn't have moved much?
  //  sqrtOfProcessNoiseReset[4][4] = 20.0; // just being cautious

  // For reset
  // NOTE: These were changed because the robot was unable to reset to the correct
  // side of the field if it managed to localise to the incorrect half.
  sqrtOfProcessNoiseReset = NMatrix(7,7,false);
  sqrtOfProcessNoiseReset[0][0] = 20.0; // extra 10cm sd when kidnapped
  sqrtOfProcessNoiseReset[1][1] = 20.0; // extra 10cm sd when kidnapped
  sqrtOfProcessNoiseReset[2][2] = 2.0; // extra 2.0 rad shift when kidnapped
  sqrtOfProcessNoiseReset[3][3] = 5.0; // ball itself shouldn't have moved much?
  sqrtOfProcessNoiseReset[4][4] = 5.0; // just being cautious


  nStates = stateEstimates.getm(); // number of states.

  // Create square root of W matrix
  // This is used to weight the sigma points.
  sqrtOfTestWeightings = NMatrix(1,2*nStates+1,false);
  sqrtOfTestWeightings[0][0] = sqrtf(1.0/(nStates+1.0));
  float outerWeighting = sqrtf(1.0/(2*(nStates+1.0)));
  for(int i=1; i <= 2*nStates; i++){
    sqrtOfTestWeightings[0][i] = (outerWeighting);
  }
  return;
}


void UKF::setTextLogger(TextLogger* tf){
  textlogger = tf;
}

void UKF::init()
{
  alpha = 1.0;
  framesSinceAmbig = 0;
  ballFramesSinceAmbig = 0;
  // Initial state estimates
  stateEstimates= NMatrix(7,1,true);
  //  stateEstimates[2][0]=3+3.1416/2.0; // 0 for all values but robot bearing = 3. Not sure why 3?
  // S = Standard deviation matrix.
  // Initial Uncertainty
  stateStandardDeviations = NMatrix(7,7,false);
  stateStandardDeviations[0][0] = 150; // 150 cm
  stateStandardDeviations[1][1] = 100; // 100 cm
  stateStandardDeviations[2][2] = 2;   // 2 radians
  stateStandardDeviations[3][3] = 150; // 150 cm
  stateStandardDeviations[4][4] = 100; // 100 cm
  stateStandardDeviations[5][5] = 10;  // 10 cm/s
  stateStandardDeviations[6][6] = 10;  // 10 cm/s
}

void UKF::reInit(){
  for (int i = 0; i < 7; i++){
    stateEstimates[i][0]  = 0.0;
  } 
  alpha = 1.0;
}

void UKF::setParams(SmallUKFParams u){
  ukfParams = u;

  // set stuff that uses these params
  updateUncertainties[5][5] = ukfParams.ball_decay_rate; // ball vel x
  updateUncertainties[6][6] = ukfParams.ball_decay_rate; // Ball velocity y

  stateStandardDeviations[0][0] = ukfParams.init_sd_x;
  stateStandardDeviations[1][1] = ukfParams.init_sd_y;
  stateStandardDeviations[2][2] = ukfParams.init_sd_theta;
  stateStandardDeviations[3][3] = ukfParams.init_sd_x;
  stateStandardDeviations[4][4] = ukfParams.init_sd_y;
  stateStandardDeviations[5][5] = ukfParams.init_sd_vel;
  stateStandardDeviations[6][6] = ukfParams.init_sd_vel;

  sqrtOfProcessNoise[0][0] = ukfParams.robot_pos_noise;    // Robot X coord.
  sqrtOfProcessNoise[1][1] = ukfParams.robot_pos_noise;    // Robot Y coord.
  sqrtOfProcessNoise[2][2] = ukfParams.robot_orient_noise; // Robot Theta.
  sqrtOfProcessNoise[3][3] = ukfParams.ball_pos_noise;     // Ball X.
  sqrtOfProcessNoise[4][4] = ukfParams.ball_pos_noise;     // Ball Y.
  sqrtOfProcessNoise[5][5] = ukfParams.ball_vel_noise;     // Ball X Velocity.
  sqrtOfProcessNoise[6][6] = ukfParams.ball_vel_noise;     // Ball Y Velocity.

  sqrtOfProcessNoiseReset[0][0] = ukfParams.reset_sd_pos;     // extra 150cm sd when kidnapped
  sqrtOfProcessNoiseReset[1][1] = ukfParams.reset_sd_pos;     // extra 100cm sd when kidnapped
  sqrtOfProcessNoiseReset[2][2] = ukfParams.reset_sd_theta; // extra 2.0 rad shift when kidnapped
  sqrtOfProcessNoiseReset[3][3] = ukfParams.reset_sd_ball; // ball itself shouldn't have moved much?
  sqrtOfProcessNoiseReset[4][4] = ukfParams.reset_sd_ball; // just being cautious

  sqrtOfTestWeightings = NMatrix(1,2*nStates+1,false);
  sqrtOfTestWeightings[0][0] = sqrt(ukfParams.kappa/(nStates+ukfParams.kappa));

  double outerWeighting = sqrt(1.0/(2*(nStates+ukfParams.kappa)));
  for(int i=1; i <= 2*nStates; i++){
    sqrtOfTestWeightings[0][i] = (outerWeighting);
  }

}


// we assume these odom measures are amount moved since last frame (not velocities)
void UKF::timeUpdate(float odometryForward, float odometryLeft, float odometryTurn, float timePassed, bool bump, bool fallen, bool standing)
{

  //ukfLog((45, "Before time update, ball at (%5.1f, %5.1f), sd (%5.1f, %5.1f)", stateEstimates[3][0], stateEstimates[4][0], stateStandardDeviations[3][3], stateStandardDeviations[4][4]));

  //-----------------------Update for locomotion
  // Calculate uncertainties -> using the estimated robot heading, the error in forward/sideways measurements are converted
  // to errors in the X and Y position of the robot.
  updateUncertainties[0][2] = -odometryForward*sin(stateEstimates[2][0]) - odometryLeft*cos(stateEstimates[2][0]); //??
  updateUncertainties[1][2] = odometryForward*cos(stateEstimates[2][0]) - odometryLeft*sin(stateEstimates[2][0]); //??

  updateUncertainties[3][5] = timePassed; // [ballX][ballXvelocity]
  updateUncertainties[4][6] = timePassed; // [ballY][ballYvelocity]


  // Movement along X-axis from odometry.
  stateEstimates[0][0] = stateEstimates[0][0] + odometryForward*cos(stateEstimates[2][0]) - odometryLeft*sin(stateEstimates[2][0]);
  // Movement along Y-axis from odometry.
  stateEstimates[1][0] = stateEstimates[1][0] + odometryForward*sin(stateEstimates[2][0]) + odometryLeft*cos(stateEstimates[2][0]);
  // Rotational movement in theta (turn) from odometry.
  stateEstimates[2][0] = stateEstimates[2][0] + odometryTurn;

  //-----------------------Update for ball velocity
  // Estimated velocity and position, combined with an estimated decay on velocity (friction) is used to predict the new
  // position of the ball.
  stateEstimates[3][0] = stateEstimates[3][0] + stateEstimates[5][0]*timePassed; // Update ball x position by ball x velocity.
  stateEstimates[4][0] = stateEstimates[4][0] + stateEstimates[6][0]*timePassed; // Update ball y position by ball y velocity.
  stateEstimates[5][0] = ukfParams.ball_decay_rate*stateEstimates[5][0]; // Reduce ball x velocity assuming deceleration
  stateEstimates[6][0] = ukfParams.ball_decay_rate*stateEstimates[6][0]; // Reduce ball y velocity assuming deceleration

  if (bump){
    // increase sd
    //sqrtOfProcessNoise[0][0] *= ukfParams.robot_bump_sd_factor;
    //sqrtOfProcessNoise[1][1] *= ukfParams.robot_bump_sd_factor;
    sqrtOfProcessNoise[2][2] *= ukfParams.robot_bump_sd_factor;
  }

  // if fallen, less movement in x,y noise
  if (fallen){
    sqrtOfProcessNoise[0][0] /= ukfParams.fallen_robot_sd_factor;
    sqrtOfProcessNoise[1][1] /= ukfParams.fallen_robot_sd_factor;
    sqrtOfProcessNoise[2][2] /= ukfParams.fallen_robot_sd_factor;
    sqrtOfProcessNoise[3][3] /= ukfParams.fallen_ball_sd_factor;
    sqrtOfProcessNoise[4][4] /= ukfParams.fallen_ball_sd_factor;

  }

  // if standing, less movement
  if (standing){
    sqrtOfProcessNoise[0][0] /= ukfParams.standing_robot_sd_factor;
    sqrtOfProcessNoise[1][1] /= ukfParams.standing_robot_sd_factor;
    sqrtOfProcessNoise[2][2] /= ukfParams.standing_robot_sd_factor;
  }

  // if ball is moving... more ball pos uncertainty
  if (fabs(stateEstimates[5][0]) > 20.0 || fabs(stateEstimates[6][0]) > 20.0){
    sqrtOfProcessNoise[3][3] *= ukfParams.moving_ball_sd_factor;
    sqrtOfProcessNoise[4][4] *= ukfParams.moving_ball_sd_factor;
    ukfLog((20, "Moving ball sd factor applied"));
  }

  // Householder transform. Unscented KF algorithm. Takes a while.
  stateStandardDeviations=HT(horzcat(updateUncertainties*stateStandardDeviations, sqrtOfProcessNoise));

  // revert the default values back

  // if ball is moving... more ball pos uncertainty
  if (fabs(stateEstimates[5][0]) > 20.0 || fabs(stateEstimates[6][0]) > 20.0){
    sqrtOfProcessNoise[3][3] /= ukfParams.moving_ball_sd_factor;
    sqrtOfProcessNoise[4][4] /= ukfParams.moving_ball_sd_factor;
  }

  // if standing, less movement
  if (standing){
    sqrtOfProcessNoise[0][0] *= ukfParams.standing_robot_sd_factor;
    sqrtOfProcessNoise[1][1] *= ukfParams.standing_robot_sd_factor;
    sqrtOfProcessNoise[2][2] *= ukfParams.standing_robot_sd_factor;
  }

  if (bump){
    // set it back
    //sqrtOfProcessNoise[0][0] /= ukfParams.robot_bump_sd_factor;
    //sqrtOfProcessNoise[1][1] /= ukfParams.robot_bump_sd_factor;
    sqrtOfProcessNoise[2][2] /= ukfParams.robot_bump_sd_factor;
  }

  // set it back
  if (fallen){
    sqrtOfProcessNoise[0][0] *= ukfParams.fallen_robot_sd_factor;
    sqrtOfProcessNoise[1][1] *= ukfParams.fallen_robot_sd_factor;
    sqrtOfProcessNoise[2][2] *= ukfParams.fallen_robot_sd_factor;
    sqrtOfProcessNoise[3][3] *= ukfParams.fallen_ball_sd_factor;
    sqrtOfProcessNoise[4][4] *= ukfParams.fallen_ball_sd_factor;
  }

  stateEstimates[2][0] = NORMALISE(stateEstimates[2][0]); // unwrap the robots angle to keep within -pi < theta < pi.

  //ukfLog((45, "After time update, ball at (%5.1f, %5.1f), sd (%5.1f, %5.1f)", stateEstimates[3][0], stateEstimates[4][0], stateStandardDeviations[3][3], stateStandardDeviations[4][4]));

  if (framesSinceAmbig > 0) framesSinceAmbig--;

  return;
}



// RHM 7/7/08: Additional function for resetting
void UKF::Reset()
{
  // Add extra uncertainty
  stateStandardDeviations = HT(horzcat(stateStandardDeviations, sqrtOfProcessNoiseReset));
}



ukfUpdateResult UKF::ballmeas(float obs_dist, float obs_theta)
// Specialised ball measurment update.
// This is different to the regular measurement update since the ball is mobile and has its own seperate states.
{
  float ballX_rel = obs_dist * cos(obs_theta);
  float ballY_rel = obs_dist * sin(obs_theta);

  ukfLog((40, "ball at dist %5.1f, bear %5.1f, relative (%5.1f, %5.1f)",obs_dist, obs_theta*RAD_T_DEG,ballX_rel,ballY_rel));

  // RHM 8/7/08:
  float R_range = ukfParams.R_ball_range_offset + ukfParams.R_ball_range_relative*powf(obs_dist,2);
  float R_bearing = ukfParams.R_ball_theta;

  ukfLog((40, "ball at r_range %5.5f, r_bear %5.5f, range offset %5.5f, range rel %5.5f, ball theta %5.5f", R_range, R_bearing, ukfParams.R_ball_range_offset, ukfParams.R_ball_range_relative, ukfParams.R_ball_theta));


  // Calculate update uncertainties - S_ball_rel & R_ball_rel.
  NMatrix S_ball_rel = NMatrix(2,2,false);
  S_ball_rel[0][0] = cos(obs_theta) * sqrtf(R_range);
  S_ball_rel[0][1] = -sin(obs_theta) * obs_dist * sqrtf(R_bearing);
  S_ball_rel[1][0] = sin(obs_theta) * sqrtf(R_range);
  S_ball_rel[1][1] = cos(obs_theta) * obs_dist * sqrtf(R_bearing);

  ukfLog((40, "S_ball_rel %5.5f, %5.5f, %5.5f, %5.5f", S_ball_rel[0][0], S_ball_rel[0][1], S_ball_rel[1][0], S_ball_rel[1][1] ));

  NMatrix R_ball_rel = S_ball_rel * S_ball_rel.transp();  // R = S^2
 

  NMatrix yBar;                                    //reset
  NMatrix Py;
  NMatrix Pxy = NMatrix(7,2,false);                    //Pxy=[0;0;0];

  NMatrix scriptX = NMatrix(stateEstimates.getm(),2 * nStates + 1,false);
  scriptX.setCol(0,stateEstimates);                         //scriptX(:,1)=Xhat; Current state.

  //----------------Saturate ScriptX angle sigma points to not wrap
  float sigmaAngleMax = 1.2;
  for(int i=1;i<nStates+1;i++){//hack to make test points distributed
    // Addition Portion.
    scriptX.setCol(i, stateEstimates + sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    scriptX[2][i] = CROP(scriptX[2][i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));
    // Subtraction Portion.
    scriptX.setCol(nStates + i,stateEstimates - sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    scriptX[2][nStates + i] = CROP(scriptX[2][nStates + i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));
  }
  //----------------------------------------------------------------

  NMatrix scriptY = NMatrix(2, 2 * nStates + 1, false);
  NMatrix temp = NMatrix(2,1,false);
  for(int i = 0; i < 2 * nStates + 1; i++){
    temp[0][0] = (scriptX[3][i] - scriptX[0][i]) * cos(scriptX[2][i]) + (scriptX[4][i] - scriptX[1][i]) * sin(scriptX[2][i]);
    temp[1][0] = -(scriptX[3][i] - scriptX[0][i]) * sin(scriptX[2][i]) + (scriptX[4][i] - scriptX[1][i]) * cos(scriptX[2][i]);
    //ukfLog((40, "sigma point %i has robot (%5.1f, %5.1f), r %5.1f, ball at (%5.1f, %5.1f) predicts rel ball (%5.1f, %5.1f)",i, scriptX[0][i], scriptX[1][i], RAD_T_DEG*scriptX[2][i], scriptX[3][i], scriptX[4][i], temp[0][0], temp[1][0]));

    scriptY.setCol(i,temp.getCol(0));
  }

  NMatrix Mx = NMatrix(scriptX.getm(), 2 * nStates + 1, false);
  NMatrix My = NMatrix(scriptY.getm(), 2 * nStates + 1, false);
  for(int i = 0; i < 2 * nStates + 1; i++){
    Mx.setCol(i, sqrtOfTestWeightings[0][i] * scriptX.getCol(i));
    My.setCol(i, sqrtOfTestWeightings[0][i] * scriptY.getCol(i));
  }

  NMatrix M1 = sqrtOfTestWeightings;
  yBar = My * M1.transp(); // Predicted Measurement
  
  // hack: just predict measurement based on mean point
  yBar = scriptY.getCol(0);

  Py = (My - yBar * M1) * (My - yBar * M1).transp();
  Pxy = (Mx - stateEstimates * M1) * (My -yBar * M1).transp();

  NMatrix K = Pxy * Invert22(Py + R_ball_rel);   // Kalman Filter Gain.

  NMatrix y = NMatrix(2,1,false); // Measurement.
  y[0][0] = ballX_rel;
  y[1][0] = ballY_rel;

  //RHM: (3) Outlier rejection.
  // (i) Compute
  // float innovation2 = (yBar-temp).transp()*Invert22(Py+R_ball_rel)*(yBar-temp);
  float innovation2 = convDble((yBar - y).transp()*Invert22(Py + R_ball_rel)*(yBar - y));
  // (ii) Now compare this with a 'magic number' (threshold2) for outliers, and also vision
  // object confidence, something like if ((innovation2> threshold2) && (visionObject.confidence is small) )
  
  // Update Alpha
  float innovation2measError = convDble((yBar - y).transp() * Invert22(R_ball_rel) * (yBar - y));
  alpha *= 1 / (1 + innovation2measError);
  

  ukfLog((10, "ball update innov2 %5.3f, innov2measerror %5.3f, thresh %5.3f", innovation2, innovation2measError, ukfParams.seen_ball_outlier_rejection_thresh));

  ukfLog((70, "Y: %5.3f, %5.3f, Ybar: %5.3f, %5.3f, yBar-y: %5.3f, %5.3f", y[0][0], y[1][0], yBar[0][0], yBar[1][0], yBar[0][0]-y[0][0], yBar[1][0]-y[1][0]));

  NMatrix change = NMatrix(7,1,false) - K*(yBar - y);
  ukfLog((70, "Change: %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f", change[0][0], change[1][0], change[2][0], change[3][0], change[4][0], change[5][0], change[6][0]));

  // always update ball
  if (innovation2 > ukfParams.seen_ball_outlier_rejection_thresh){
    ukfLog((15, "Ball rejected as outlier! %5.3f", innovation2));
    //std::cout<<"not update"<<std::endl;
    return UKF7_OUTLIER; // Outlier, so chuck it all out...
  }

  /*
  if (innovation2 < 0.02) {
    yBar = scriptY.getCol(0);
    ukfLog((70, "Low innov, now: Y: %5.3f, %5.3f, Ybar: %5.3f, %5.3f, yBar-y: %5.3f, %5.3f", y[0][0], y[1][0], yBar[0][0], yBar[1][0], yBar[0][0]-y[0][0], yBar[1][0]-y[1][0]));
  }
  */

  stateStandardDeviations = HT( horzcat(Mx - stateEstimates*M1 - K*My + K*yBar* M1, K*S_ball_rel) ); // Update uncertainties.

  stateEstimates = stateEstimates - K * (yBar - y); // Update States.
  return UKF7_OK; // It all worked OK!
}



ukfUpdateResult UKF::ballmeasWithBearing(float obs_dist, float obs_theta)
// Specialised ball measurment update.
// This is different to the regular measurement update since the ball is mobile and has its own seperate states.
{

  ukfLog((40, "ball at dist %5.1f, bear %5.1f",obs_dist, obs_theta*RAD_T_DEG));

  // RHM 8/7/08:
  float R_range = ukfParams.R_ball_range_offset + ukfParams.R_ball_range_relative*powf(obs_dist,2);
  float R_bearing = ukfParams.R_ball_theta;

  ukfLog((40, "ball at r_range %5.5f, r_bear %5.5f, range offset %5.5f, range rel %5.5f, ball theta %5.5f", R_range, R_bearing, ukfParams.R_ball_range_offset, ukfParams.R_ball_range_relative, ukfParams.R_ball_theta));


  // Calculate update uncertainties - S_ball_rel & R_ball_rel.
  NMatrix S_ball_rel = NMatrix(2,2,false);
  // Todd: convert dist/bearing error to dist/bearing error
  S_ball_rel[0][0] = sqrtf(R_range);  // drange/drange   * r_range
  S_ball_rel[0][1] = 0.0;             // drange/dbear    * r_bear
  S_ball_rel[1][0] = 0.0;             // dbear/drange    * r_range
  S_ball_rel[1][1] = sqrtf(R_bearing); // dbear/dbear   * r_bear

  ukfLog((40, "S_ball_rel %5.5f, %5.5f, %5.5f, %5.5f", S_ball_rel[0][0], S_ball_rel[0][1], S_ball_rel[1][0], S_ball_rel[1][1] ));

  NMatrix R_ball_rel = S_ball_rel * S_ball_rel.transp();  // R = S^2

  NMatrix yBar;                                    //reset
  NMatrix Py;
  NMatrix Pxy = NMatrix(7,2,false);                    //Pxy=[0;0;0];

  NMatrix scriptX = NMatrix(stateEstimates.getm(),2 * nStates + 1,false);
  scriptX.setCol(0,stateEstimates);                         //scriptX(:,1)=Xhat; Current state.

  //----------------Saturate ScriptX angle sigma points to not wrap
  float sigmaAngleMax = 1.2;
  for(int i=1;i<nStates+1;i++){//hack to make test points distributed
    // Addition Portion.
    scriptX.setCol(i, stateEstimates + sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    scriptX[2][i] = CROP(scriptX[2][i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));
    // Subtraction Portion.
    scriptX.setCol(nStates + i,stateEstimates - sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    scriptX[2][nStates + i] = CROP(scriptX[2][nStates + i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));
  }
  //----------------------------------------------------------------

  NMatrix scriptY = NMatrix(2, 2 * nStates + 1, false);
  NMatrix temp = NMatrix(2,1,false);
  for(int i = 0; i < 2 * nStates + 1; i++){
    Point2D sigmaBotLoc(scriptX[0][i], scriptX[1][i]);
    Point2D sigmaBallLoc(scriptX[3][i], scriptX[4][i]);
    temp[0][0] = sigmaBotLoc.getDistanceTo(sigmaBallLoc);
    temp[1][0] = sigmaBotLoc.getBearingTo(sigmaBallLoc, scriptX[2][i]);
    temp[1][0] = normalizeAngleAroundVal(temp[1][0], obs_theta);
    //ukfLog((90, "sigma point %i has robot (%5.1f, %5.1f), r %5.1f, ball at (%5.1f, %5.1f) predicts ball dist %5.1f, bear %5.1f",i, scriptX[0][i], scriptX[1][i], RAD_T_DEG*scriptX[2][i], scriptX[3][i], scriptX[4][i], temp[0][0], temp[1][0]*RAD_T_DEG));

    scriptY.setCol(i,temp.getCol(0));
  }

  NMatrix Mx = NMatrix(scriptX.getm(), 2 * nStates + 1, false);
  NMatrix My = NMatrix(scriptY.getm(), 2 * nStates + 1, false);
  for(int i = 0; i < 2 * nStates + 1; i++){
    Mx.setCol(i, sqrtOfTestWeightings[0][i] * scriptX.getCol(i));
    My.setCol(i, sqrtOfTestWeightings[0][i] * scriptY.getCol(i));
  }

  NMatrix M1 = sqrtOfTestWeightings;
  yBar = My * M1.transp(); // Predicted Measurement

  // hack: just predict measurement based on mean point
  yBar = scriptY.getCol(0);

  Py = (My - yBar * M1) * (My - yBar * M1).transp();
  Pxy = (Mx - stateEstimates * M1) * (My -yBar * M1).transp();

  NMatrix K = Pxy * Invert22(Py + R_ball_rel);   // Kalman Filter Gain.

  NMatrix y = NMatrix(2,1,false); // Measurement.
  y[0][0] = obs_dist;
  y[1][0] = obs_theta;

  //RHM: (3) Outlier rejection.
  // (i) Compute
  // float innovation2 = (yBar-temp).transp()*Invert22(Py+R_ball_rel)*(yBar-temp);
  float innovation2 = convDble((yBar - y).transp()*Invert22(Py + R_ball_rel)*(yBar - y));
  // (ii) Now compare this with a 'magic number' (threshold2) for outliers, and also vision
  // object confidence, something like if ((innovation2> threshold2) && (visionObject.confidence is small) )
  
  // Update Alpha
  float innovation2measError = convDble((yBar - y).transp() * Invert22(R_ball_rel) * (yBar - y));
  alpha *= 1 / (1 + innovation2measError);

  ukfLog((10, "ball update innov2 %5.3f, thresh %5.3f", innovation2, ukfParams.seen_ball_outlier_rejection_thresh));

  ukfLog((70, "Y: %5.3f, %5.3f, Ybar: %5.3f, %5.3f, yBar-y: %5.3f, %5.3f", y[0][0], y[1][0], yBar[0][0], yBar[1][0], yBar[0][0]-y[0][0], yBar[1][0]-y[1][0]));

  NMatrix change = NMatrix(7,1,false) - K*(yBar - y);
  ukfLog((70, "Change: %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f", change[0][0], change[1][0], change[2][0], change[3][0], change[4][0], change[5][0], change[6][0]));

  if (innovation2 > ukfParams.seen_ball_outlier_rejection_thresh){
    //std::cout<<"not update"<<std::endl;
    return UKF7_OUTLIER; // Outlier, so chuck it all out...
  }

  stateStandardDeviations = HT( horzcat(Mx - stateEstimates*M1 - K*My + K*yBar* M1, K*S_ball_rel) ); // Update uncertainties.
  stateEstimates = stateEstimates - K * (yBar - y); // Update States.
  return UKF7_OK; // It all worked OK!
}




// RHM 7/7/08: Change for resetting (return int)
ukfUpdateResult UKF::fieldObjectmeas(float distance,float bearing,float objX, float objY, float distanceErrorOffset, float distanceErrorRelative, float bearingError)
// Update for a field landmarks
{
  float objX_rel = distance * cos(bearing);
  float objY_rel = distance * sin(bearing);

  ukfLog((70, "Rel x,y: (%5.1f, %5.1f), dist: %5.1f, bear: %5.1f", objX_rel, objY_rel, distance, RAD_T_DEG*bearing));

  float R_range = distanceErrorOffset + distanceErrorRelative * powf(distance, 2);
  float R_bearing = bearingError;
  //if(not_goal && INGORE_RANGE) R_range= 22500;  //150^2

  // Calculate update uncertainties - S_obj_rel & R_obj_rel
  NMatrix S_obj_rel = NMatrix(2,2,false);
  // convert dist/bearing error to rel x/y error
  S_obj_rel[0][0] = cos(bearing) * sqrtf(R_range);  // dx/drange * r_range
  S_obj_rel[0][1] = -sin(bearing) * distance * sqrtf(R_bearing); // dx/dbear *r_bear
  S_obj_rel[1][0] = sin(bearing) * sqrtf(R_range);  // dy/drange * r_range
  S_obj_rel[1][1] = cos(bearing) * distance * sqrtf(R_bearing); // dy/dbear *r_bear

  NMatrix R_obj_rel = S_obj_rel * S_obj_rel.transp(); // R = S^2

  // Unscented KF Stuff.
  NMatrix yBar;                                    //reset
  NMatrix Py;
  NMatrix Pxy=NMatrix(7, 2, false);                    //Pxy=[0;0;0];
  NMatrix scriptX=NMatrix(stateEstimates.getm(), 2 * nStates + 1, false);
  scriptX.setCol(0, stateEstimates);                         //scriptX(:,1)=Xhat;

  //----------------Saturate ScriptX angle sigma points to not wrap
  float sigmaAngleMax = 1.2;
  for(int i=1;i<nStates+1;i++){//hack to make test points distributed
    // Addition Portion.
    scriptX.setCol(i, stateEstimates + sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    scriptX[2][i] = CROP(scriptX[2][i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));
    // Subtraction Portion.
    scriptX.setCol(nStates + i,stateEstimates - sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    scriptX[2][nStates + i] = CROP(scriptX[2][nStates + i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));
  }
  //----------------------------------------------------------------
  NMatrix scriptY = NMatrix(2, 2 * nStates + 1, false);
  NMatrix temp = NMatrix(2, 1, false);

  float dX,dY,Cc,Ss;

  for(int i = 0; i < 2 * nStates + 1; i++){
    // expected values for each sigma point (relx and rely)
    dX = objX-scriptX[0][i];
    dY = objY-scriptX[1][i];
    Cc = cos(scriptX[2][i]);
    Ss = sin(scriptX[2][i]);
    temp[0][0] = dX * Cc + dY * Ss;
    temp[1][0] = -dX * Ss + dY * Cc;
    scriptY.setCol(i, temp.getCol(0));
  }
  NMatrix Mx = NMatrix(scriptX.getm(), 2 * nStates + 1, false);
  NMatrix My = NMatrix(scriptY.getm(), 2 * nStates + 1, false);
  for(int i = 0; i < 2 * nStates + 1; i++){
    Mx.setCol(i, sqrtOfTestWeightings[0][i] * scriptX.getCol(i));
    My.setCol(i, sqrtOfTestWeightings[0][i] * scriptY.getCol(i));
  }

  NMatrix M1 = sqrtOfTestWeightings;
  yBar = My * M1.transp(); // Predicted Measurement.
  Py = (My - yBar * M1) * (My - yBar * M1).transp();
  Pxy = (Mx - stateEstimates * M1) * (My - yBar * M1).transp();

  NMatrix K = Pxy * Invert22(Py + R_obj_rel); // K = Kalman filter gain.

  NMatrix y = NMatrix(2,1,false); // Measurement. I terms of relative (x,y).
  y[0][0] = objX_rel;
  y[1][0] = objY_rel;
  //end of standard ukf stuff
  //RHM: 20/06/08 Outlier rejection.
  float innovation2 = convDble((yBar - y).transp() * Invert22(Py + R_obj_rel) * (yBar - y));

  // Update Alpha
  float innovation2measError = convDble((yBar - y).transp() * Invert22(R_obj_rel) * (yBar - y));
  alpha *= 1 / (1 + innovation2measError);

  ukfLog((10, "innov2 %5.3f, innov2measerror %5.3f, thresh %5.3f", innovation2, innovation2measError, ukfParams.outlier_rejection_thresh));

  ukfLog((70, "Y: %5.3f, %5.3f, Ybar: %5.3f, %5.3f, yBar-y: %5.3f, %5.3f", y[0][0], y[1][0], yBar[0][0], yBar[1][0], yBar[0][0]-y[0][0], yBar[1][0]-y[1][0]));


  if (innovation2 > ukfParams.outlier_rejection_thresh){
    return UKF7_OUTLIER;
  }
  if (innovation2 < 0){
    //std::cout<<"**********************************"<<std::endl;
    //Py.print();
    //R_obj_rel.print();
    //std::cout<<"**********************************"<<std::endl;
  }

  ukfLog((70, "Y: %5.3f, %5.3f, Ybar: %5.3f, %5.3f, yBar-y: %5.3f, %5.3f", y[0][0], y[1][0], yBar[0][0], yBar[1][0], yBar[0][0]-y[0][0], yBar[1][0]-y[1][0]));

  NMatrix change = NMatrix(7,1,false) - K*(yBar - y);
  //  ukfLog((70, "Change: %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f", change[0][0], change[1][0], change[2][0], change[3][0], change[4][0], change[5][0], change[6][0]));

  stateStandardDeviations = HT( horzcat(Mx - stateEstimates*M1 - K*My + K*yBar*M1, K*S_obj_rel) );
  stateEstimates = stateEstimates - K*(yBar - y);
  return UKF7_OK;
}


// RHM 7/7/08: Change for resetting (return int)
ukfUpdateResult UKF::fieldObjectMeasWithBearing(float distance,float bearing,float objX, float objY, float distanceErrorOffset, float distanceErrorRelative, float bearingError)
// Update for a field landmarks
{
  
  ukfLog((10, "Saw obs at dist %5.3f, bear %5.3f", distance, bearing *RAD_T_DEG));
  float R_range = distanceErrorOffset + distanceErrorRelative * powf(distance, 2);
  float R_bearing = bearingError;
  //if(not_goal && INGORE_RANGE) R_range= 22500;  //150^2

  // Calculate update uncertainties - S_obj_rel & R_obj_rel
  NMatrix S_obj_rel = NMatrix(2,2,false);

  // Todd: convert dist/bearing error to dist/bearing error
  S_obj_rel[0][0] = sqrtf(R_range);  // drange/drange   * r_range
  S_obj_rel[0][1] = 0.0;             // drange/dbear    * r_bear
  S_obj_rel[1][0] = 0.0;             // dbear/drange    * r_range
  S_obj_rel[1][1] = sqrtf(R_bearing); // dbear/dbear   * r_bear

  NMatrix R_obj_rel = S_obj_rel * S_obj_rel.transp(); // R = S^2

  // Unscented KF Stuff.
  NMatrix yBar;                                    //reset
  NMatrix Py;
  NMatrix Pxy=NMatrix(7, 2, false);                    //Pxy=[0;0;0];
  NMatrix scriptX=NMatrix(stateEstimates.getm(), 2 * nStates + 1, false);
  scriptX.setCol(0, stateEstimates);                         //scriptX(:,1)=Xhat;

  //----------------Saturate ScriptX angle sigma points to not wrap
  float sigmaAngleMax = 1.2;
  for(int i=1;i<nStates+1;i++){//hack to make test points distributed
    // Addition Portion.
    scriptX.setCol(i, stateEstimates + sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    scriptX[2][i] = CROP(scriptX[2][i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));
    // Subtraction Portion.
    scriptX.setCol(nStates + i,stateEstimates - sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    scriptX[2][nStates + i] = CROP(scriptX[2][nStates + i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));
  }
  //----------------------------------------------------------------
  NMatrix scriptY = NMatrix(2, 2 * nStates + 1, false);
  NMatrix temp = NMatrix(2, 1, false);

  for(int i = 0; i < 2 * nStates + 1; i++){
    // expected values for each sigma point (dist and bearing)
    Point2D sigmaLoc(scriptX[0][i], scriptX[1][i]);
    temp[0][0] = sigmaLoc.getDistanceTo(Point2D(objX, objY));
    temp[1][0] = sigmaLoc.getBearingTo(Point2D(objX, objY), scriptX[2][i]);
    temp[1][0] = normalizeAngleAroundVal(temp[1][0], bearing);
    ukfLog((90, "Robot at (%5.3f, %5.3f), %5.3f, dist %5.3f,bear %5.3f", scriptX[0][i], scriptX[1][i], scriptX[2][i]*RAD_T_DEG, temp[0][0], temp[1][0]*RAD_T_DEG));
    scriptY.setCol(i, temp.getCol(0));
  }
  NMatrix Mx = NMatrix(scriptX.getm(), 2 * nStates + 1, false);
  NMatrix My = NMatrix(scriptY.getm(), 2 * nStates + 1, false);
  for(int i = 0; i < 2 * nStates + 1; i++){
    Mx.setCol(i, sqrtOfTestWeightings[0][i] * scriptX.getCol(i));
    My.setCol(i, sqrtOfTestWeightings[0][i] * scriptY.getCol(i));
  }

  NMatrix M1 = sqrtOfTestWeightings;
  yBar = My * M1.transp(); // Predicted Measurement.
  Py = (My - yBar * M1) * (My - yBar * M1).transp();
  Pxy = (Mx - stateEstimates * M1) * (My - yBar * M1).transp();

  NMatrix K = Pxy * Invert22(Py + R_obj_rel); // K = Kalman filter gain.

  NMatrix y = NMatrix(2,1,false); // Measurement. I terms of relative (x,y).
  y[0][0] = distance;
  y[1][0] = bearing; 
  //end of standard ukf stuff
  //RHM: 20/06/08 Outlier rejection.
  float innovation2 = convDble((yBar - y).transp() * Invert22(Py + R_obj_rel) * (yBar - y));

  // Update Alpha
  float innovation2measError = convDble((yBar - y).transp() * Invert22(R_obj_rel) * (yBar - y));
  alpha *= 1 / (1 + innovation2measError);

  ukfLog((10, "innov2 %5.3f, innov2measerror %5.3f, thresh %5.3f", innovation2, innovation2measError, ukfParams.outlier_rejection_thresh));
  ukfLog((70, "Y: %5.3f, %5.3f, Ybar: %5.3f, %5.3f, yBar-y: %5.3f, %5.3f", y[0][0], y[1][0], yBar[0][0], yBar[1][0], yBar[0][0]-y[0][0], yBar[1][0]-y[1][0]));


  if (innovation2 > ukfParams.outlier_rejection_thresh){
    return UKF7_OUTLIER;
  }
  if (innovation2 < 0){
    //std::cout<<"**********************************"<<std::endl;
    //Py.print();
    //R_obj_rel.print();
    //std::cout<<"**********************************"<<std::endl;
  }

  NMatrix change = NMatrix(7,1,false) - K*(yBar - y);
  ukfLog((70, "Change: %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f", change[0][0], change[1][0], change[2][0], change[3][0], change[4][0], change[5][0], change[6][0]));

  stateStandardDeviations = HT( horzcat(Mx - stateEstimates*M1 - K*My + K*yBar*M1, K*S_obj_rel) );
  stateEstimates = stateEstimates - K*(yBar - y);
  return UKF7_OK;
}


// RHM 7/7/08: Change for resetting (return int)
ukfUpdateResult UKF::lineObjectMeas(float distance, float bearing, float objX, float objY, float lineOrient, float distanceErrorOffset, float distanceErrorRelative, float bearingError)
// Update for a field landmarks
{
  
  ukfLog((10, "Line seen at dist %5.3f, bear %5.3f, line pt at (%5.3f, %5.3f), orient %5.3f", distance, bearing*RAD_T_DEG, objX, objY, lineOrient*RAD_T_DEG));

  float R_range = distanceErrorOffset + distanceErrorRelative * powf(distance, 2);
  float R_bearing = bearingError;
  //if(not_goal && INGORE_RANGE) R_range= 22500;  //150^2

  // Calculate update uncertainties - S_obj_rel & R_obj_rel
  NMatrix S_obj_rel = NMatrix(2,2,false);

  // Todd: convert dist/bearing error to dist/bearing error
  S_obj_rel[0][0] = sqrtf(R_range);  // drange/drange   * r_range
  S_obj_rel[0][1] = 0.0;             // drange/dbear    * r_bear
  S_obj_rel[1][0] = 0.0;             // dbear/drange    * r_range
  S_obj_rel[1][1] = sqrtf(R_bearing); // dbear/dbear   * r_bear

  NMatrix R_obj_rel = S_obj_rel * S_obj_rel.transp(); // R = S^2

  // Unscented KF Stuff.
  NMatrix yBar;                                    //reset
  NMatrix Py;
  NMatrix Pxy=NMatrix(7, 2, false);                    //Pxy=[0;0;0];
  NMatrix scriptX=NMatrix(stateEstimates.getm(), 2 * nStates + 1, false);
  scriptX.setCol(0, stateEstimates);                         //scriptX(:,1)=Xhat;

  //----------------Saturate ScriptX angle sigma points to not wrap
  // Todd: sigmamax should be less than 90 degrees
  float sigmaAngleMax = 1.2;
  for(int i=1;i<nStates+1;i++){//hack to make test points distributed
    // Addition Portion.
    scriptX.setCol(i, stateEstimates + sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    float origX2 = scriptX[2][i];
    scriptX[2][i] = CROP(scriptX[2][i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));

    if (origX2 != scriptX[2][i]) 
      ukfLog((20, "had to crop scriptX from %5.3f to %5.3f", RAD_T_DEG*origX2, RAD_T_DEG * scriptX[2][i]));

    // Subtraction Portion.
    scriptX.setCol(nStates + i,stateEstimates - sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    origX2 = scriptX[2][nStates+i];
    scriptX[2][nStates + i] = CROP(scriptX[2][nStates + i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));

    if (origX2 != scriptX[2][nStates+i]) 
      ukfLog((20, "had to crop scriptX from %5.3f to %5.3f", RAD_T_DEG*origX2, RAD_T_DEG * scriptX[2][nStates+i]));
  }
  //----------------------------------------------------------------
  NMatrix scriptY = NMatrix(2, 2 * nStates + 1, false);
  NMatrix temp = NMatrix(2, 1, false);

  for(int i = 0; i < 2 * nStates + 1; i++){
    // expected values for each sigma point (dist and orient to us)
    if (lineOrient > 0.01) {
      // horizontal line, diff is change in y, closest x is my x
      temp[0][0] = fabs(scriptX[1][i] - objY);
      objX = scriptX[0][i];
    } else {
      // vertical line, diff is change in x, closest y is my y
      temp[0][0] = fabs(scriptX[0][i] - objX);
      objY = scriptX[1][i];
    } 
    Point2D sigmaLoc(scriptX[0][i], scriptX[1][i]);
    float bearToPt = sigmaLoc.getBearingTo(Point2D(objX, objY), scriptX[2][i]);
    temp[1][0] = bearToPt;
    temp[1][0] = normalizeAngleNinetyAroundVal(temp[1][0], bearing);

    //    if (temp[1][0] != bearToPt) {
    //ukfLog((10, "re-normalized bearing to be near seen bearing, orig bear %5.3f, new bear %5.3f, seen %5.3f", bearToPt*RAD_T_DEG, temp[1][0]*RAD_T_DEG, bearing*RAD_T_DEG));
    //}

    // if short distance and bearings are way off, try other side of the line
    float bearError = fabs(temp[1][0] - bearing);
    if (bearError > (DEG_T_RAD*110) && (temp[0][0] < 35 || distance < 35)) {
      temp[0][0] = -temp[0][0];
      temp[1][0] = normalizeAngleNinetyAroundVal(temp[1][0] + M_PI, bearing);
      //ukfLog((10, "small dist seen %5.3f, exp %5.3f, huge bearing error %5.3f, assuming other side of line. new expected bearing %5.3f", distance, temp[0][0], bearError*RAD_T_DEG, RAD_T_DEG*temp[1][0]));
    }

    ukfLog((90, "Robot at (%5.3f, %5.3f), %5.3f, dist %5.3f, lineOrient %5.3f, bear %5.3f", scriptX[0][i], scriptX[1][i], scriptX[2][i]*RAD_T_DEG, temp[0][0], lineOrient*RAD_T_DEG, temp[1][0]*RAD_T_DEG));
    scriptY.setCol(i, temp.getCol(0));
  }
  NMatrix Mx = NMatrix(scriptX.getm(), 2 * nStates + 1, false);
  NMatrix My = NMatrix(scriptY.getm(), 2 * nStates + 1, false);
  for(int i = 0; i < 2 * nStates + 1; i++){
    Mx.setCol(i, sqrtOfTestWeightings[0][i] * scriptX.getCol(i));
    My.setCol(i, sqrtOfTestWeightings[0][i] * scriptY.getCol(i));
  }

  NMatrix M1 = sqrtOfTestWeightings;
  yBar = My * M1.transp(); // Predicted Measurement.
  Py = (My - yBar * M1) * (My - yBar * M1).transp();
  Pxy = (Mx - stateEstimates * M1) * (My - yBar * M1).transp();

  NMatrix K = Pxy * Invert22(Py + R_obj_rel); // K = Kalman filter gain.

  NMatrix y = NMatrix(2,1,false); // Measurement. I terms of relative (x,y).
  y[0][0] = distance;
  y[1][0] = bearing; 
  //end of standard ukf stuff
  //RHM: 20/06/08 Outlier rejection.
  float innovation2 = convDble((yBar - y).transp() * Invert22(Py + R_obj_rel) * (yBar - y));

  // Update Alpha
  float innovation2measError = convDble((yBar - y).transp() * Invert22(R_obj_rel) * (yBar - y));
  alpha *= 1 / (1 + innovation2measError);

  ukfLog((10, "innov2 %5.3f, innov2measerror %5.3f, thresh %5.3f", innovation2, innovation2measError, ukfParams.outlier_rejection_thresh));

  ukfLog((70, "Y: %5.3f, %5.3f, Ybar: %5.3f, %5.3f, yBar-y: %5.3f, %5.3f", y[0][0], y[1][0]*RAD_T_DEG, yBar[0][0], yBar[1][0]*RAD_T_DEG, yBar[0][0]-y[0][0], (yBar[1][0]-y[1][0])*RAD_T_DEG));

  NMatrix change = NMatrix(7,1,false) - K*(yBar - y);

  ukfLog((70, "Change: %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f", change[0][0], change[1][0], change[2][0], change[3][0], change[4][0], change[5][0], change[6][0]));

  if (innovation2 > ukfParams.outlier_rejection_thresh){
    return UKF7_OUTLIER;
  }
  if (innovation2 < 0){
    //std::cout<<"**********************************"<<std::endl;
    //Py.print();
    //R_obj_rel.print();
    //std::cout<<"**********************************"<<std::endl;
  }

  stateStandardDeviations = HT( horzcat(Mx - stateEstimates*M1 - K*My + K*yBar*M1, K*S_obj_rel) );
  stateEstimates = stateEstimates - K*(yBar - y);
  return UKF7_OK;
}


// kick velocity update
ukfUpdateResult UKF::ballKickUpdate(float velocity, float relBearing, float velError, float bearingError)
{
  
  // convert to velocity in rel x/y
  float relvelx = velocity * cosf(relBearing);
  float relvely = velocity * sinf(relBearing);

  ukfLog((70, "Kick at vel %5.3f and bearing %5.3f, rel (%5.3f, %5.3f)", velocity, relBearing*RAD_T_DEG, relvelx, relvely));

  float R_vel = velError;
  float R_bearing = bearingError;

  // Calculate update uncertainties - S_obj_rel & R_obj_rel
  NMatrix S_obj_rel = NMatrix(2,2,false);

  // Todd: convert vel/bear error to rel vel x/y error
  S_obj_rel[0][0] = cos(relBearing) * sqrtf(R_vel);  // dx/drange * r_range
  S_obj_rel[0][1] = -sin(relBearing) * velocity * sqrtf(R_bearing); // dx/dbear *r_bear
  S_obj_rel[1][0] = sin(relBearing) * sqrtf(R_vel);  // dy/drange * r_range
  S_obj_rel[1][1] = cos(relBearing) * velocity * sqrtf(R_bearing); // dy/dbear *r_b

  NMatrix R_obj_rel = S_obj_rel * S_obj_rel.transp(); // R = S^2

  // Unscented KF Stuff.
  NMatrix yBar;                                    //reset
  NMatrix Py;
  NMatrix Pxy=NMatrix(7, 2, false);                    //Pxy=[0;0;0];
  NMatrix scriptX=NMatrix(stateEstimates.getm(), 2 * nStates + 1, false);
  scriptX.setCol(0, stateEstimates);                         //scriptX(:,1)=Xhat;

  //----------------Saturate ScriptX angle sigma points to not wrap
  float sigmaAngleMax = 1.2;
  for(int i=1;i<nStates+1;i++){//hack to make test points distributed
    // Addition Portion.
    scriptX.setCol(i, stateEstimates + sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    scriptX[2][i] = CROP(scriptX[2][i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));
    // Subtraction Portion.
    scriptX.setCol(nStates + i,stateEstimates - sqrtf((float)nStates + ukfParams.kappa) * stateStandardDeviations.getCol(i - 1));
    scriptX[2][nStates + i] = CROP(scriptX[2][nStates + i], (-sigmaAngleMax + stateEstimates[2][0]), (sigmaAngleMax + stateEstimates[2][0]));
  }
  //----------------------------------------------------------------
  NMatrix scriptY = NMatrix(2, 2 * nStates + 1, false);
  NMatrix temp = NMatrix(2, 1, false);

  for(int i = 0; i < 2 * nStates + 1; i++){
    // expected values for each sigma point (relvelx and relvely)
    float dX = scriptX[5][i];
    float dY = scriptX[6][i];
    float Cc = cos(scriptX[2][i]);
    float Ss = sin(scriptX[2][i]);
    temp[0][0] = dX * Cc + dY * Ss;
    temp[1][0] = -dX * Ss + dY * Cc;
    scriptY.setCol(i, temp.getCol(0));
    
    //ukfLog((10, "Sigma point %i has vel in dir (%5.3f, %5.3f), robot orient is %5.3f, relative vel (%5.3f, %5.3f)", i, scriptX[5][i], scriptX[6][i], RAD_T_DEG*scriptX[2][i], temp[0][0], temp[1][0]));
    
    scriptY.setCol(i, temp.getCol(0));
  }
  NMatrix Mx = NMatrix(scriptX.getm(), 2 * nStates + 1, false);
  NMatrix My = NMatrix(scriptY.getm(), 2 * nStates + 1, false);
  for(int i = 0; i < 2 * nStates + 1; i++){
    Mx.setCol(i, sqrtOfTestWeightings[0][i] * scriptX.getCol(i));
    My.setCol(i, sqrtOfTestWeightings[0][i] * scriptY.getCol(i));
  }

  NMatrix M1 = sqrtOfTestWeightings;
  yBar = My * M1.transp(); // Predicted Measurement.
  Py = (My - yBar * M1) * (My - yBar * M1).transp();
  Pxy = (Mx - stateEstimates * M1) * (My - yBar * M1).transp();

  NMatrix K = Pxy * Invert22(Py + R_obj_rel); // K = Kalman filter gain.

  NMatrix y = NMatrix(2,1,false); // Measurement. I terms of relative (x,y).
  y[0][0] = relvelx;
  y[1][0] = relvely; 
  //end of standard ukf stuff

  //Todd: no outlier rejection on ball velocity update

  //ukfLog((70, "Y: %5.3f, %5.3f, Ybar: %5.3f, %5.3f, yBar-y: %5.3f, %5.3f", y[0][0], y[1][0], yBar[0][0], yBar[1][0], yBar[0][0]-y[0][0], yBar[1][0]-y[1][0]));

  NMatrix change = NMatrix(7,1,false) - K*(yBar - y);
  //ukfLog((70, "Change: %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f", change[0][0], change[1][0], change[2][0], change[3][0], change[4][0], change[5][0], change[6][0]));

  stateStandardDeviations = HT( horzcat(Mx - stateEstimates*M1 - K*My + K*yBar*M1, K*S_obj_rel) );
  stateEstimates = stateEstimates - K*(yBar - y);
  return UKF7_OK;
}



//
//RHM: 26/6/08 New function, needed for shared ball stuff (and maybe others....)
//======================================================================
//
ukfUpdateResult UKF::linear2MeasurementUpdate(float Y1,float Y2, float SR11, float SR12, float SR22, int index1, int index2, int mate, bool weAreKeeper)
//
// KF update (called for example by shared ball) where a pair of measurements, [Y1;Y2];
// with variance Square Root of Covariance (SR) [SR11 SR12; 0 SR22]
// are predicted to be Xhat[index1][0] and Xhat[index2][0] respectively
// This is based on the function fieldObjectmeas, but simplified.
//
// Example Call (given data from wireless: ballX, ballY, SRballXX, SRballXY, SRballYY)
//      linear2MeasurementUpdate( ballX, ballY, SRballXX, SRballXY, SRballYY, 3, 4 )
{
  NMatrix SR = NMatrix(2,2,false);
  SR[0][0] = SR11;
  SR[0][1] = SR12;
  SR[1][1] = SR22;

  NMatrix R = NMatrix(2,2,false);
  R = SR * SR.transp();

  NMatrix Py = NMatrix(2, 2, false);
  NMatrix Pxy = NMatrix(7, 2, false);

  NMatrix CS = NMatrix(2, 7, false);
  CS.setRow(0, stateStandardDeviations.getRow(index1));
  CS.setRow(1, stateStandardDeviations.getRow(index2));

  Py = CS * CS.transp();
  Pxy = stateStandardDeviations * CS.transp();

  NMatrix K = Pxy * Invert22(Py + R);   //Invert22

  NMatrix y = NMatrix(2, 1, false);
  y[0][0] = Y1;
  y[1][0] = Y2;

  NMatrix yBar = NMatrix(2, 1, false); //Estimated values of the measurements Y1,Y2
  yBar[0][0] = stateEstimates[index1][0];
  yBar[1][0] = stateEstimates[index2][0];
  //RHM: (3) Outlier rejection.
  float innovation2 = convDble( (yBar - y).transp() * Invert22(Py + SR * SR.transp()) * (yBar - y) );
  //std::cout<<innovation2<<std::endl;
  // Update Alpha
  float innovation2measError = convDble((yBar - y).transp() * Invert22(SR * SR.transp()) * (yBar - y));
  alpha *= 1 / (1 + innovation2measError);
  
  // higher threshold if from keeper, so we trust their ball obs more
  float reject_thresh = ukfParams.shared_ball_outlier_rejection_thresh;
  if (mate == 1) reject_thresh += ukfParams.keeper_outlier_rejection_thresh_diff;
  
  // lower thresh if we're keeper, we trust everyone else less
  if (weAreKeeper) reject_thresh -= ukfParams.keeper_outlier_rejection_thresh_diff;

  ukfLog((20, "Shared update from %i, we're keeper? %i, rejection thresh %5.3f, innovation2 %5.3f", mate, weAreKeeper, reject_thresh, innovation2));

  // always update ball!
  if (innovation2 > reject_thresh){
    //std::cout<<"+++++++++++++++++not update+++++++++++++++++"<<std::endl;
    return UKF7_OUTLIER;
  }
  if(innovation2 < 0){
    //std::cout<<"**********************************"<<std::endl;
    //Py.print();
    //std::cout<<"**********************************"<<std::endl;
  }

  stateStandardDeviations = HT(horzcat(stateStandardDeviations - K * CS, K * SR));
  stateEstimates = stateEstimates - K * (yBar - y);
  return UKF7_OK;
}
//======================================================================================================
//
// RHM: 26/6/08 New Pseudo code for ball information to send over wireless
// Information to add to the broadcast to other robots:
// (along with other information)
// * Frame Number {timing indicator)
// * Frames since I last saw the ball
// * Xhat[3][0] (est ball X)
// * Xhat[4][0] (est ball Y)
// * SRxx, SRxy, SRyy (3 elements of 2x2 sqrt matrix formed as: HT(vertical_cat(S.getRow[3], S.getRow[4])))
// * Xhat[0][0]
// * Xhat[1][0] (could be used in behaviour)
// * other stuff probably
//
//



float UKF::variance(int Xi){
  return convDble(stateStandardDeviations.getRow(Xi)*stateStandardDeviations.getRow(Xi).transp());
}



float UKF::sd(int Xi){
  return sqrtf(variance(Xi));
}



float UKF::getState(int stateID){
  return stateEstimates[stateID][0];
}



NMatrix UKF::GetBallSR(){
  return HT(vertcat(stateStandardDeviations.getRow(3), stateStandardDeviations.getRow(4)));
}



float UKF::getDistanceToPosition(float posX, float posY){
  float selfX = stateEstimates[0][0];
  float selfY = stateEstimates[1][0];
  float diffX = selfX - posX;
  float diffY = selfY - posY;
  float distance = sqrtf( diffX * diffX + diffY * diffY );
  return distance;
}



float UKF::getBearingToPosition(float posX, float posY){
  float selfX = stateEstimates[0][0];
  float selfY = stateEstimates[1][0];
  float selfHeading = stateEstimates[2][0];
  float diffX = posX - selfX;
  float diffY = posY - selfY;
  if( (diffX == 0) && (diffY == 0)) diffY = 0.0001;
  float positionHeading = atan2(diffY, diffX);
  float bearing = NORMALISE(positionHeading - selfHeading);
  return bearing;
}



bool UKF::clipState(int stateIndex, float minValue, float maxValue){
  bool clipped = false;
  if(stateEstimates[stateIndex][0] > maxValue){
    float mult, Pii;
    NMatrix Si;
    Si = stateStandardDeviations.getRow(stateIndex);
    Pii = convDble(Si * Si.transp());
    mult = (stateEstimates[stateIndex][0] - maxValue) / Pii;
    stateEstimates = stateEstimates - mult * stateStandardDeviations * Si.transp();
    stateEstimates[stateIndex][0] = maxValue;
    clipped = true;
  }
  if(stateEstimates[stateIndex][0] < minValue){
    float mult, Pii;
    NMatrix Si;
    Si = stateStandardDeviations.getRow(stateIndex);
    Pii = convDble(Si * Si.transp());
    mult = (stateEstimates[stateIndex][0] - minValue) / Pii;
    stateEstimates = stateEstimates - mult * stateStandardDeviations * Si.transp();
    stateEstimates[stateIndex][0] = minValue;
    clipped = true;
  }
  stateEstimates[2][0] = NORMALISE(stateEstimates[2][0]);
  return clipped;
}
