#include "UKFModule.h"
#include <string>
#include <stdlib.h>
#include <iostream>
#include <fstream>


void UKFModule::specifyMemoryDependency() {
  requiresMemoryBlock("world_objects");
  requiresMemoryBlock("localization");
  requiresMemoryBlock("team_packets");
  requiresMemoryBlock("vision_frame_info");
  requiresMemoryBlock("vision_odometry");
  requiresMemoryBlock("robot_state");
  requiresMemoryBlock("game_state");
  requiresMemoryBlock("vision_joint_angles");
  requiresMemoryBlock("behavior");
  requiresMemoryBlock("vision_processed_sonar");
}

void UKFModule::specifyMemoryBlocks() {
  getOrAddMemoryBlock(worldObjects,"world_objects");
  getOrAddMemoryBlock(localizationMem,"localization");
  getOrAddMemoryBlock(teamPacketsMem,"team_packets");
  getOrAddMemoryBlock(frameInfo,"vision_frame_info");
  getOrAddMemoryBlock(odometry,"vision_odometry");
  getOrAddMemoryBlock(robotState,"robot_state");
  getOrAddMemoryBlock(gameState,"game_state");
  getOrAddMemoryBlock(jointAngles,"vision_joint_angles");
  getOrAddMemoryBlock(behaviorMem,"behavior");
  getOrAddMemoryBlock(processedSonar,"vision_processed_sonar");
}


void UKFModule::initSpecificModule(){
  // lets log all of these

  /*
    worldObjects->log_block = true;
    localizationMem->log_block = true;
    teamPacketsMem->log_block = true;
    frameInfo->log_block = true;
    odometry->log_block = true;
    robotState->log_block = true;
    gameState->log_block = true;
    sensors->log_block = true;
  */

  // Todd: also need:
  // robot state (i.e. playing, ready, etc)
  // odometry info
  // kick info
  // fallen info
  // self robot index
  // current head angles (to check if lines are in fov)

  cout << "UKFModule init" << endl;
  timePassed=0.0;
  timeLast=0.0;

  models[0].isActive = true; // Start with just the first model model
  models[0].toBeActivated = false;
  models[0].framesSinceAmbig = 0;
  models[0].ballFramesSinceAmbig = 0;

  wasPreviouslyPenalised = false;
  penaltyInitialTime = 0;
  isManualPenalty = false;
  
  numMateFlippedBalls = 0;
  numBadBallUpdates = 0;

  timeLast = frameInfo->seconds_since_start;
  previousGameState = INITIAL;
  previousDive = Dive::NONE;
  previousGetUp = Getup::NONE;
  lastKeeperObservationTime = 0;
  //  previousGameState = STATE_INITIAL;

  ukfLog((0, "done setting indiv ukf memory"));

  // RHM 7/7/08: Extra array for resetting algorithm
  for(int m = 0; m < c_MAX_MODELS; m++){
    models[m].setTextLogger(textlogger);
    for (int i=0; i<NUM_WORLD_OBJS; i++) modelObjectErrors[m][i] = 0.0;
  }

  ballKF.init(frameInfo,localizationMem,odometry,robotState,worldObjects,behaviorMem);
  ballKF.setTextLogger(textlogger);
}


void UKFModule::reInit(){
  wasPreviouslyPenalised = false;
  penaltyInitialTime = 0;
  timeLast = frameInfo->seconds_since_start;

  isManualPenalty = false;
  numMateFlippedBalls = 0;
  numBadBallUpdates = 0;


  for (int i = 0; i < c_MAX_MODELS; i++){
    models[i].isActive = false; // Start with just the first model model
    models[i].toBeActivated = false;
    models[i].framesSinceAmbig = 0;
    models[i].ballFramesSinceAmbig = 0;
  }
  models[0].isActive = true;

  SmallUKFParams modelParams;
  modelParams.R_ball_theta = ukfParams.R_ball_theta;
  modelParams.R_ball_range_offset = ukfParams.R_ball_range_offset;
  modelParams.R_ball_range_relative = ukfParams.R_ball_range_relative;
  modelParams.kappa = ukfParams.kappa;
  modelParams.ball_decay_rate = ukfParams.ball_decay_rate;
  modelParams.outlier_rejection_thresh = ukfParams.outlier_rejection_thresh;
  modelParams.kick_outlier_rejection_thresh = ukfParams.kick_outlier_rejection_thresh;
  modelParams.seen_ball_outlier_rejection_thresh = ukfParams.seen_ball_outlier_rejection_thresh;
  modelParams.shared_ball_outlier_rejection_thresh = ukfParams.shared_ball_outlier_rejection_thresh;
  modelParams.keeper_outlier_rejection_thresh_diff = ukfParams.keeper_outlier_rejection_thresh_diff;
  modelParams.robot_pos_noise = ukfParams.robot_pos_noise;
  modelParams.robot_orient_noise = ukfParams.robot_orient_noise;
  modelParams.ball_pos_noise = ukfParams.ball_pos_noise;

  if (robotState->WO_SELF == KEEPER){
    modelParams.ball_vel_noise = ukfParams.keeper_ball_vel_noise;
  } else {
    modelParams.ball_vel_noise = ukfParams.ball_vel_noise;
  }
  modelParams.robot_bump_sd_factor = ukfParams.robot_bump_sd_factor;
  modelParams.fallen_robot_sd_factor = ukfParams.fallen_robot_sd_factor;
  modelParams.fallen_ball_sd_factor = ukfParams.fallen_ball_sd_factor;
  modelParams.standing_robot_sd_factor = ukfParams.standing_robot_sd_factor;
  modelParams.moving_ball_sd_factor = ukfParams.moving_ball_sd_factor;

  modelParams.init_sd_x = ukfParams.init_sd_x;
  modelParams.init_sd_y = ukfParams.init_sd_y;
  modelParams.init_sd_theta = ukfParams.init_sd_theta;
  modelParams.init_sd_vel = ukfParams.init_sd_vel;

  modelParams.reset_sd_pos = ukfParams.reset_sd_pos;
  modelParams.reset_sd_theta = ukfParams.reset_sd_theta;
  modelParams.reset_sd_ball = ukfParams.reset_sd_ball;

  for (int m = 0; m < c_MAX_MODELS; m++){
    models[m].init();
    models[m].setParams(modelParams);
  }

  for(int m = 0; m < c_MAX_MODELS; m++){
    for (int i=0; i<NUM_WORLD_OBJS; i++) modelObjectErrors[m][i] = 0.0;
  }

  // in sim, accept ints that are farther away
  if (frameInfo->source == MEMORY_SIM){
    ukfParams.ignore_int_dist = 7000.0;
  }

  doPlayerReset();
  ukfLog((0, "done setting ukf lua params %f %f", ukfParams.ignore_int_dist, ukfParams.robot_pos_noise));
}


//--------------------------------- MAIN FUNCTIONS  ---------------------------------//


void UKFModule::resetTimeLast(){
  timePassed = 0;
  timeLast = frameInfo->seconds_since_start;
  previousGameState = gameState->state;
  lastKeeperObservationTime = timeLast;
}

void UKFModule::processFrame(){
  if (!isMemorySatisfied()) return;

  /*
    worldObjects->log_block = true;
    localizationMem->log_block = true;
    teamPacketsMem->log_block = true;
    frameInfo->log_block = true;
    odometry->log_block = true;
    robotState->log_block = true;
    gameState->log_block = true;
    sensors->log_block = true;
  */

  //std::cout << "UKF ProcessFrame()\n" << std::flush;
  ukfLog((70, "7-UKF processFrame start %5.3f, nModels: %i", frameInfo->seconds_since_start, getNumActiveModels()));

  logModels();

  // convert wo to cm for ukf
  mmTocm();

  CheckGameState();

  // compute time from last frame (for odom update)
  timePassed = frameInfo->seconds_since_start - timeLast;
  timeLast = frameInfo->seconds_since_start;

  //varianceCheckAll();
  doKickUpdates();
  runTimeUpdates();

  logModels();

  // updates from observations
  checkSharedBallForFlips();
  processSharedBall();

  // only do observations if not getting up and not diving
  if (odometry->getting_up_side_ == Getup::NONE && odometry->fall_direction_ == Fall::NONE && behaviorMem->keeperDiving == Dive::NONE && gameState->state != PENALISED){
    processObservations();
  }

  // check for outliers, reset sd if so
  CheckForOutlierResets();

  NormaliseAlphas();
  CheckForAmbigModels();

  // no merging while fallen... want to keep separate fall models
  if (odometry->getting_up_side_ == Getup::NONE && odometry->fall_direction_ == Fall::NONE && behaviorMem->keeperDiving == Dive::NONE){
    MergeModels(ukfParams.max_models_after_merge);
  }

  clipActiveModelsToField();

  // get best model and use to update world objects
  UKF* bestModel = getBestModel();
  populateWOSelf(bestModel);
  populateBall(bestModel);
  updateDistAndBearToObjects();

  // update memory for debug
  updateMemory();

  // convert wo back to mm
  cmTomm();

  ukfLog((70, "7-UKF processFrame complete %5.3f, nModels: %i", frameInfo->seconds_since_start, getNumActiveModels()));

  logModels();

  // relative ball filter
  ballKF.processFrame();
}

/*
  void UKFModule::ProcessObjects(int frameNumber, FieldObject* ourfieldObjects, StoredGamePackets* mostRecentPackets)
  {
  int numUpdates = 0;
  int updateResult;
  currentFrameNumber = frameNumber;
  objects = ourfieldObjects;
  // RHM 26/6/08: Modified below
  CheckGameState();

  if(balanceFallen) return;

  #if LOCWM_VERBOSITY > 2
  if(numUpdates == 0 ){
  debug_out  <<"[" << currentFrameNumber << "]: Update Starting." << endl;
  for(int i = 0; i < c_MAX_MODELS; i++){
  if(models[i].isActive == false) continue;
  debug_out  << "[" << currentFrameNumber << "]: Model[" << i << "]";
  debug_out  << " [alpha = " << models[i].alpha << "]";
  debug_out  << " Robot X: " << models[i].getState(0);
  debug_out  << " Robot Y: " << models[i].getState(1);
  debug_out  << " Robot Theta: " << models[i].getState(2) << endl;
  }
  }
  #endif

  varianceCheckAll(); // Correct orientation to face a goal if you can see it and are unsure which way you are facing.


  // Do Time update.
  doTimeUpdate(odomForward, odomLeft, odomTurn);
  #if LOCWM_VERBOSITY > 2
  debug_out  << "[" << currentFrameNumber << "]: Time update - odomForward = " << odomForward << " odomLeft = " << odomLeft << " odomTurn = " << odomTurn << endl;
  #endif
  // Proccess the Known Field Objects
  for(int objID = 0; objID < NUM_FIELD_OBJECTS; objID++){
  if(ourfieldObjects[objID].seen == false) continue; // Skip objects that were not seen.
  if( objID == FO_BALL ){ // The ball is special since it moves.
  updateResult = doBallMeasurementUpdate(ourfieldObjects[objID]);
  numUpdates++;
  }
  else if(ourfieldObjects[objID].isFixed()){
  updateResult = doKnownLandmarkMeasurementUpdate(ourfieldObjects[objID]);
  numUpdates++;
  }
  }

  NormaliseAlphas();

  #if SHARED_BALL_ON
  // Check the game packets.
  // We only want to do the shared ball updates if we can't see the ball ourselves.
  // there were problems where the team will keep sharing the previous position of their ball
  // and updates in vision would not supercede the shared data.
  int myPlayerNumber = GameController::getInstance().getPlayerNumber();
  if(ourfieldObjects[FO_BALL].framesSinceLastSeen > 3){ // TODO: Change to a SD value?
  for(int robotNum = 0; robotNum < NUM_ROBOTS; robotNum++){
  if(myPlayerNumber == (robotNum+1)) continue;
  if(mostRecentPackets[robotNum].processedWM == false){
  if(mostRecentPackets[robotNum].packet.ball.seen == true){
  #if LOCWM_VERBOSITY > 2
  debug_out  << "[" << currentFrameNumber << "]: Doing Shared ball update from robot " << robotNum+1 <<  endl;
  #endif
  doSharedBallUpdate(mostRecentPackets[robotNum].packet.ball);
  }
  else {
  #if LOCWM_VERBOSITY > 2
  debug_out  << "[" << currentFrameNumber << "]: Skipping shared ball update from robot " << robotNum+1 << endl;
  #endif
  }
  mostRecentPackets[robotNum].processedWM = true;
  }
  }
  }
  #endif // SHARED_BALL_ON

  NormaliseAlphas();

  #if MULTIPLE_MODELS_ON
  // Do Ambiguous objects.
  for(int objID = 0; objID < NUM_FIELD_OBJECTS; objID++){
  if(ourfieldObjects[objID].seen == false) continue; // Skip objects that were not seen.
  if(ourfieldObjects[objID].isAmbiguous()){
  updateResult = doAmbiguousLandmarkMeasurementUpdate(ourfieldObjects[objID]);
  NormaliseAlphas();
  }
  numUpdates++;
  }

  MergeModels(c_MAX_MODELS_AFTER_MERGE);
  #endif // MULTIPLE_MODELS_ON

  // Check for model reset. -> with multiple models just remove if not last one??
  // Need to re-do reset to be model specific.
  int numReset = CheckForOutlierResets();
  // clip models back on to field.
  clipActiveModelsToField();

  // Store WM Data in Field Objects.
  int bestModelID = getBestModelID(); // Get the best model to use.
  for(int objID = 0; objID < NUM_FIELD_OBJECTS; objID++){
  if(ourfieldObjects[objID].isFixed()){
  ourfieldObjects[objID].wmDistance = models[bestModelID].getDistanceToPosition(ourfieldObjects[objID].wmX, ourfieldObjects[objID].wmY);
  ourfieldObjects[objID].wmBearing = models[bestModelID].getBearingToPosition(ourfieldObjects[objID].wmX, ourfieldObjects[objID].wmY);
  }
  else if(objID == FO_BALL){
  ourfieldObjects[objID].wmX = models[bestModelID].getState(3);
  ourfieldObjects[objID].sdx = models[bestModelID].sd(3);
  ourfieldObjects[objID].wmY = models[bestModelID].getState(4);
  ourfieldObjects[objID].sdy = models[bestModelID].sd(4);
  ourfieldObjects[objID].vX = models[bestModelID].getState(5);
  ourfieldObjects[objID].vY = models[bestModelID].getState(6);
  ourfieldObjects[objID].wmDistance = models[bestModelID].getDistanceToPosition(ourfieldObjects[objID].wmX, ourfieldObjects[objID].wmY);
  ourfieldObjects[objID].wmBearing = models[bestModelID].getBearingToPosition(ourfieldObjects[objID].wmX, ourfieldObjects[objID].wmY);
  }
  else if(objID == FO_TEAM1){
  ourfieldObjects[objID].wmX = models[bestModelID].getState(0);
  ourfieldObjects[objID].sdx = models[bestModelID].sd(0);
  ourfieldObjects[objID].wmY = models[bestModelID].getState(1);
  ourfieldObjects[objID].sdy = models[bestModelID].sd(1);
  ourfieldObjects[objID].wmOrientation = models[bestModelID].getState(2);
  ourfieldObjects[objID].sdtheta = models[bestModelID].sd(2);
  }
  }
  #if LOCWM_VERBOSITY > 2
  if(numUpdates > 0){
  for (int i = 0; i < c_MAX_MODELS; i++){
  if(models[i].isActive == false) continue;
  debug_out  << "[" << currentFrameNumber << "]: Model[" << i << "]";
  debug_out  << " [alpha = " << models[i].alpha << "]";
  debug_out  << " Robot X: " << models[i].getState(0);
  debug_out  << " Robot Y: " << models[i].getState(1);
  debug_out  << " Robot Theta: " << models[i].getState(2) << endl;
  }
  debug_out  << "[" << currentFrameNumber << "]: Best Model";
  debug_out  << " [alpha = " << models[bestModelID].alpha << "]";
  debug_out  << " Robot X: " << models[bestModelID].getState(0);
  debug_out  << " Robot Y: " << models[bestModelID].getState(1);
  debug_out  << " Robot Theta: " << models[bestModelID].getState(2) << endl;
  }
  #endif
  }
*/


void UKFModule::CheckGameState()
{
  bool isPenalised = gameState->state == PENALISED;
  int currentState = gameState->state;

  // just got penalised, keep track of this
  if (isPenalised and !wasPreviouslyPenalised){
    penaltyInitialTime = frameInfo->seconds_since_start;

    // see if it was a manual penalty
    if ((frameInfo->frame_id - gameState->frameReceived) > 3) {
      // no gc
      isManualPenalty = true;
    } else {
      isManualPenalty = false;
    }

  }

  // If robot has been penalised and penalty is lifted, reset to the penalty replacement positions.
  // if playing -> if its been at least 15 seconds since penalty so we don't get messed up by accidental clicks by game controller
  if(!isPenalised && wasPreviouslyPenalised && !gameState->isPenaltyKick && ((frameInfo->seconds_since_start - penaltyInitialTime) > 15.0 || gameState->state != PLAYING || frameInfo->seconds_since_start < 20.0)){
    doPenaltyReset();
  }
  // ...otherwise if the player has come out of initial, Reset Localisation.
  else if (previousGameState != currentState && (previousGameState == INITIAL || previousGameState == FINISHED) && !gameState->isPenaltyKick){
    ukfLog((10, "Changed state from initial, do player reset"));
    doPlayerReset();
  }
  // also if we changed teams
  else if (previousTeam != robotState->team_){
    ukfLog((10, "Team changed, do player reset"));
    doPlayerReset();
  }

  // anything other than playing in PK, then we know exactly where we are
  if (gameState->isPenaltyKick && previousGameState != currentState){
    doPenaltyKickReset();
  }


  // in set... make sure we're on correct side of field (just flip to other side symettrically)
  if (gameState->state == SET && !gameState->isPenaltyKick){
    for (int modelNumber = 0; modelNumber < c_MAX_MODELS; modelNumber++){
      if(models[modelNumber].isActive == false) continue;
      if (models[modelNumber].stateEstimates[0][0] > 0){
        ukfLog((10, "in set, model %i believes its on wrong side of field with x %5.3f", modelNumber, models[modelNumber].stateEstimates[0][0]));
        for (int j = 0; j < 7; j++){
          if (j != 2)
            models[modelNumber].stateEstimates[j][0] = -models[modelNumber].stateEstimates[j][0];
          else
            models[modelNumber].stateEstimates[j][0] = normalizeAngle(models[modelNumber].stateEstimates[j][0] + M_PI);
        }
        models[modelNumber].Reset();
      }
    }
  }

  // update that we know ball is in center of field
  if (gameState->state == SET && ukfParams.USE_BALL_IN_SET && !gameState->isPenaltyKick){
    for (int modelNumber = 0; modelNumber < c_MAX_MODELS; modelNumber++){
      if(models[modelNumber].isActive == false) continue;
      float ball_set_sd = 50.0;
      ukfLog((10, "Set ball location at (0,0) with sd %5.3f", ball_set_sd));
      models[modelNumber].linear2MeasurementUpdate(0, 0, ball_set_sd, ball_set_sd, ball_set_sd, 3, 4, 0, false);
    }
  }

  // in set, spawn some models for manual positioning
  // every second (manual positioning takes a while) if we don't already have multiple models
  if (gameState->state == SET && !gameState->isPenaltyKick && frameInfo->frame_id % 30 == 0 && getNumActiveModels() == 1){
    spawnManualPositionModel();
    NormaliseAlphas();
  }

  if(!isPenalised && wasPreviouslyPenalised && isManualPenalty && !gameState->isPenaltyKick && ((frameInfo->seconds_since_start - penaltyInitialTime) > 40.0)){
    // this might really be set if it was manual
    ukfLog((10, "manual penalty over, may actually be in set"));
    spawnManualPositionModel();
    NormaliseAlphas();
  }

  // for PK -> update that we know ball is on penalty cross
  if (gameState->state != PLAYING && ukfParams.USE_BALL_IN_SET && gameState->isPenaltyKick){
    for (int modelNumber = 0; modelNumber < c_MAX_MODELS; modelNumber++){
      if(models[modelNumber].isActive == false) continue;
      float ball_set_sd = 50.0;
      float ball_x = PENALTY_CROSS_X/10.0;
      if (robotState->WO_SELF == KEEPER) ball_x = -ball_x;
      ukfLog((10, "Set ball location at (%5.3f,0) with sd %5.3f", ball_x, ball_set_sd));
      models[modelNumber].linear2MeasurementUpdate(ball_x, 0, ball_set_sd, ball_set_sd, ball_set_sd, 3, 4, 0, false);
    }
  }

  // check keeper observations
  if (robotState->role_ == KEEPER && !gameState->isPenaltyKick) {
    checkKeeperObservations();
  }


  // keeper spawn models of ball coming quickly at it
  if ((currentState == TESTING || currentState == PLAYING) && robotState->role_ == KEEPER){
    // if ball has moved closer by a significant amount
    WorldObject* ball = &(worldObjects->objects_[WO_BALL]);
    ukfLog((10, "keeper spawn, seen %i, dist %5.3f, vis %5.3f", ball->seen, ball->distance, ball->visionDistance));
    if (ball->seen && ball->visionDistance < (ball->distance - 2.5) && fabs(ball->visionDistance - ball->distance) < 30.0)
      spawnKickedBallModel();
  }

  // Increase uncertainty of ball position if it has gone out.. Cause it has probably been moved. // or if we just transitioned into ready or set
  if (gameState->dropInTime == 0 || (previousGameState != currentState && (currentState == READY || currentState == SET))){
    ukfLog((20, "ball drop in, or move to ready/set, reset ball sd and vel"));
    for (int modelNumber = 0; modelNumber < c_MAX_MODELS; modelNumber++){
      if(models[modelNumber].isActive == false) continue;
      models[modelNumber].stateStandardDeviations[3][3] = ukfParams.reset_ball_sd_pos;
      models[modelNumber].stateStandardDeviations[4][4] = ukfParams.reset_ball_sd_pos;
      models[modelNumber].stateStandardDeviations[5][5] = ukfParams.init_sd_vel;
      models[modelNumber].stateStandardDeviations[6][6] = ukfParams.init_sd_vel;
      // and velocity zero
      models[modelNumber].stateEstimates[5][0] = 0;
      models[modelNumber].stateEstimates[6][0] = 0;
    }
  }

  // split into models for turning each way when falling
  // appropriate split based on fall dir
  if (previousGetUp != odometry->getting_up_side_ && (odometry->getting_up_side_ == Getup::FRONT || odometry->getting_up_side_ == Getup::BACK)){
    splitModelsOnFall();
    // we'll have used fall dir here and can reset it
    odometry->fall_direction_ = Fall::NONE;
  }

  // Increase heading uncertainty if fallen.
  // We really have no idea which way we are facing after we get up
  // but hopefully haven't moved too far around the field.
  // set number of frames until we're allowed to converge to 1 model
  if (odometry->getting_up_side_ != Getup::NONE){
    for (int modelNumber = 0; modelNumber < c_MAX_MODELS; modelNumber++){
      // set number of frames until we're allowed to converge to 1 model
      models[modelNumber].framesSinceAmbig = ukfParams.ambig_model_min_nframes;

      // set number of ball sightings until we're allowed to converge to 1 model
      // if we are near middle of field
      if ((fabs(models[modelNumber].stateEstimates[0][0]) + fabs(models[modelNumber].stateEstimates[1][0])) < 80) {
        models[modelNumber].ballFramesSinceAmbig = ukfParams.ambig_model_min_ball_frames;
      }

    }
  }
  // Store historical variables.
  wasPreviouslyPenalised = isPenalised;
  previousGameState = currentState;
  previousTeam = robotState->team_;
  previousDive = behaviorMem->keeperDiving;
  previousGetUp = odometry->getting_up_side_;
}


void UKFModule::checkKeeperObservations(){

  // check if keeper has seen circle/ball/line directly in front of it
  WorldObject* circle = &(worldObjects->objects_[WO_CENTER_CIRCLE]);
  WorldObject* ball = &(worldObjects->objects_[WO_BALL]);

  if (circle->seen || ball->seen){
    lastKeeperObservationTime = frameInfo->seconds_since_start;
    return;
  }

  // check for far post
  WorldObject* post = &(worldObjects->objects_[WO_UNKNOWN_GOALPOST]);
  if (post->seen && post->visionDistance > 180){
    lastKeeperObservationTime = frameInfo->seconds_since_start;
    return;
  }

  // if near post and all outliered and keeper... maybe we're flipped
  if (post->isUnknownPost() && post->visionDistance < 140 && fabs(post->visionBearing) < 75.0*DEG_T_RAD && robotState->role_ == KEEPER && !gameState->isPenaltyKick){
    // check if there are models where this makes sense
    bool sensibleModel = false;
    for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
      if(models[modelID].isActive == false) continue;
      ukfLog((20, "model %i has loc (%5.3f, %5.3f), ori %5.3f", modelID, models[modelID].stateEstimates[0][0], models[modelID].stateEstimates[1][0], models[modelID].stateEstimates[2][0]*RAD_T_DEG));
      if (models[modelID].stateEstimates[0][0] < -FIELD_X/20.0 || fabs(models[modelID].stateEstimates[2][0]) > 15.0*DEG_T_RAD){
        sensibleModel = true;
        break;
      }
    }
    
    if (!sensibleModel){
      ukfLog((10, "Keeper near post seen in front of us with no valid models, flipped?"));
      splitKeeperModelTurn();
    }
  }

  // see if keeper has seen anything other than his own goal
  bool backwardLine = false;
  for (int i = WO_UNKNOWN_FIELD_LINE_1; i <= WO_UNKNOWN_FIELD_LINE_4; i++){
    WorldObject* wo = &(worldObjects->objects_[i]);
    if (!wo->seen) continue;
    if (fabs(wo->visionBearing - M_PI) < 0.25 && wo->visionDistance > 50 && wo->visionDistance < 100){
      ukfLog((10, "keeper saw line BEHIND it at bearing %5.3f, dist %5.3f, possibly front of goal box", wo->visionBearing, wo->visionDistance));
      backwardLine = true;
    }
    if (fabs(wo->visionBearing) < 0.25 && wo->visionDistance > 10 && wo->visionDistance < 70){
      ukfLog((10, "keeper saw line at bearing %5.3f, dist %5.3f, possibly front of goal box", wo->visionBearing, wo->visionDistance));
      lastKeeperObservationTime = frameInfo->seconds_since_start;
      return;
    }
  }

  // check how long its been since keeper has seen something
  float timeSinceObs = frameInfo->seconds_since_start - lastKeeperObservationTime;
  ukfLog((10, "keeper time since obs %5.3f", timeSinceObs));

  if ((timeSinceObs > 20.0 || (timeSinceObs > 10.0 && backwardLine)) && gameState->state == PLAYING) {
    ukfLog((10, "Keeper has not seen circle/ball/penbox in %5.3f seconds", timeSinceObs));
    lastKeeperObservationTime = frameInfo->seconds_since_start;
    // split a model where we're flipped
    splitKeeperModelTurn();
  }
}

void UKFModule::splitKeeperModelTurn(){
  // create one model for being turned 180
  int numFreeModels = getNumFreeModels();
  int numActiveModels = getNumActiveModels();
  int numRequiredModels = numActiveModels * 2;

  if(numFreeModels < numRequiredModels){
    int maxActiveAfterMerge = c_MAX_MODELS / 2;
    ukfLog((30, "Only %i free models, need %i models. merging to %i models",
            numFreeModels, numRequiredModels, maxActiveAfterMerge));
    MergeModels(maxActiveAfterMerge);
    if(getNumFreeModels() < (getNumActiveModels() * 2)){
      ukfLog((10, "Only %i free models. not enough. aborting update", getNumFreeModels()));
      return;
    }
  }

  for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == false) continue; // Skip inactive models.

    // Copy initial model to the temporary model.
    tempModel = models[modelID];
    tempModel.isActive = false;
    tempModel.toBeActivated = true;

    // now for kick success
    int newModelID = FindNextFreeModel();

    // If an invalid modelID has been returned, something has gone horribly wrong, so stop here.
    if(newModelID < 0){
      ukfLog((10, "keeper update, bad model id returned (%i), aborted.", newModelID));

      for(int m = 0; m < c_MAX_MODELS; m++) models[m].toBeActivated = false;
      return;
    }

    ukfLog((0, "keeper update. split model %i (turned 180) from model %i (normal)", modelID, newModelID));
    models[newModelID] = tempModel; // Get the new model from the temp

    // Copy outlier history from the current model.
    for (int i=0; i<NUM_WORLD_OBJS; i++){
      modelObjectErrors[newModelID][i] = modelObjectErrors[modelID][i];
    }

    // close to goal line, may be flipped around it
    // in middle of box, probably just flipped around
    if (models[newModelID].stateEstimates[0][0] < -270){
      models[newModelID].stateEstimates[0][0] = -300 + (-300 - models[newModelID].stateEstimates[0][0]);
    }
    models[newModelID].stateEstimates[1][0] = -models[newModelID].stateEstimates[1][0];
    models[newModelID].stateEstimates[2][0] = normalizeAngle(models[newModelID].stateEstimates[2][0]+M_PI);
    // 70 % chance of us being turned wrong way
    models[newModelID].alpha *= 0.7;
    models[modelID].alpha *= 0.3;

  } // per orig model

    // Split alpha between choices and also activate models
  for (int i=0; i< c_MAX_MODELS; i++) {
    if (models[i].toBeActivated) {
      models[i].isActive=true;
    }
    models[i].toBeActivated=false; // Turn off activation flag
  }
}

// based on current penalty reset rules
void UKFModule::doPenaltyReset()
{
  ukfLog((10, "Performing penalty reset"));

  // save where we think ball is
  int best = getBestModelID();
  float ballx = models[best].stateEstimates[3][0];
  float bally = models[best].stateEstimates[4][0];


  // Reset all of the models
  for(int m = 0; m < c_MAX_MODELS; m++){
    // reset outlier error count
    for (int i=0; i<NUM_WORLD_OBJS; i++)
      modelObjectErrors[m][i] = 0;

    // Disable models
    models[m].isActive = false;
    models[m].toBeActivated = false;
    models[m].framesSinceAmbig = 0;
    models[m].ballFramesSinceAmbig = 0;

  }

  // setup model 0 at one possible location
  models[0].isActive = true;
  models[0].framesSinceAmbig = ukfParams.ambig_model_min_nframes;
  models[0].ballFramesSinceAmbig = 0;

  // slight preference towards opposite side of field from ball
  if (bally < 0){
    models[0].alpha = 0.55;
  } else {
    models[0].alpha = 0.45;
  }

  // Set the state estimates
  models[0].stateEstimates[0][0] = -PENALTY_CROSS_X/10.0;  // Robot x - even with pen cross
  models[0].stateEstimates[1][0] = FIELD_Y/20.0; // Robot y (divide by 2 and convert to cm)
  models[0].stateEstimates[2][0] = -M_PI/2.0;    // Robot heading
  models[0].stateEstimates[3][0] = ballx;       // Ball x (from before reset)
  models[0].stateEstimates[4][0] = bally;       // Ball y (from before reset)
  models[0].stateEstimates[5][0] = 0.0;       // Ball vx
  models[0].stateEstimates[6][0] = 0.0;       // Ball vy

  // Set the uncertainties, with uncertain x, certian y and theta
  resetSdMatrixForSideline(0);

  // setup model 1 as the other sideline
  models[1].isActive = true;
  models[1].framesSinceAmbig = ukfParams.ambig_model_min_nframes;
  models[1].ballFramesSinceAmbig = 0;

  // slight preference towards opposite side of field from ball
  if (bally < 0){
    models[1].alpha = 0.45;
  } else {
    models[1].alpha = 0.55;
  }

  // Set the state estimates
  models[1].stateEstimates[0][0] = -PENALTY_CROSS_X/10.0;  // Robot x - even with pen cross
  models[1].stateEstimates[1][0] = -FIELD_Y/20.0; // Robot y (divide by 2 and convert to cm)
  models[1].stateEstimates[2][0] = M_PI/2.0;   // Robot heading
  models[1].stateEstimates[3][0] = ballx;       // Ball x (from before reset)
  models[1].stateEstimates[4][0] = bally;       // Ball y (from before reset)
  models[1].stateEstimates[5][0] = 0.0;       // Ball vx
  models[1].stateEstimates[6][0] = 0.0;       // ball vy

  // Set the uncertainties
  // Set the uncertainties, with uncertain x, certian y and theta
  resetSdMatrixForSideline(1);

  return;
}

// we know starting positions for PK
void UKFModule::doPenaltyKickReset(){
  ukfLog((10, "performing penalty kick reset"));

  // Reset all of the models
  for(int m = 0; m < c_MAX_MODELS; m++){
    // reset outlier error count
    for (int i=0; i<NUM_WORLD_OBJS; i++)
      modelObjectErrors[m][i] = 0;

    // Disable models
    models[m].isActive = false;
    models[m].toBeActivated = false;
    models[m].framesSinceAmbig = 0;
    models[m].ballFramesSinceAmbig = 0;

  }

  // set model 0 to be the truth
  models[0].isActive = true;
  models[0].alpha = 1.0;

  models[0].stateEstimates[0][0] = 0.0;         // Robot X
  models[0].stateEstimates[1][0] = 0.0;         // Robot y
  models[0].stateEstimates[2][0] = 0.0;          // Robot heading
  models[0].stateEstimates[3][0] = 0.0;       // Ball x
  models[0].stateEstimates[4][0] = 0.0;       // Ball y
  models[0].stateEstimates[5][0] = 0.0;       // Ball vx
  models[0].stateEstimates[6][0] = 0.0;       // Ball vy

  if (robotState->WO_SELF == KEEPER){
    models[0].stateEstimates[0][0] = -FIELD_X /20.0 + 10.0;  // our x
    models[0].stateEstimates[3][0] = -PENALTY_CROSS_X/10.0; // ball x
  } else {
    models[0].stateEstimates[0][0] = 0.0;  // our x
    models[0].stateEstimates[3][0] = PENALTY_CROSS_X/10.0; // ball x
  }

  initSdMatrix(0);
  models[0].stateStandardDeviations[0][0] = ukfParams.init_sd_x/10.0;
  models[0].stateStandardDeviations[1][1] = ukfParams.init_sd_y/10.0;
  models[0].stateStandardDeviations[2][2] = ukfParams.init_sd_theta/10.0;
  models[0].stateStandardDeviations[3][3] = ukfParams.init_sd_x/10.0;
  models[0].stateStandardDeviations[4][4] = ukfParams.init_sd_y/10.0;
  models[0].stateStandardDeviations[5][5] = ukfParams.init_sd_vel/10.0;
  models[0].stateStandardDeviations[6][6] = ukfParams.init_sd_vel/10.0;

}

void UKFModule::initSdMatrix(int modelID){
  // Set the uncertainties
  for (int i = 0; i < 7; i++){
    for (int j = 0; j < 7; j++){
      models[modelID].stateStandardDeviations[i][j] = 0;
    }
  }
}

// uses specific starting position for each # robot to help with ambig goals
void UKFModule::doPlayerReset()
{
  ukfLog((10, "performing player reset"));

  // Reset all of the models
  for(int m = 0; m < c_MAX_MODELS; m++){
    // reset outlier error count
    for (int i=0; i<NUM_WORLD_OBJS; i++)
      modelObjectErrors[m][i] = 0;

    // Disable models
    models[m].isActive = false;
    models[m].toBeActivated = false;
    models[m].framesSinceAmbig = 0;
    models[m].ballFramesSinceAmbig = 0;

  }

  // new code... each player has a specific spot
  models[0].isActive = true;
  models[0].alpha = 1.0;

  // keeper starts on left sideline, 1 meter from end
  if (robotState->WO_SELF == 1){
    models[0].stateEstimates[0][0] = -200;          // Robot X
    models[0].stateEstimates[1][0] = FIELD_Y/20.0;  // Robot y
    models[0].stateEstimates[2][0] = -M_PI/2.0;     // Robot heading
  }
  // player 2 starts on left side line, 2 meters from end
  else if (robotState->WO_SELF == 2){
    models[0].stateEstimates[0][0] = -100;          // Robot X
    models[0].stateEstimates[1][0] = FIELD_Y/20.0;  // Robot y
    models[0].stateEstimates[2][0] = -M_PI/2.0;     // Robot heading
  }
  // player 3 starts on right sideline, 1 meter from end
  else if (robotState->WO_SELF == 3){
    models[0].stateEstimates[0][0] = -200;          // Robot X
    models[0].stateEstimates[1][0] = -FIELD_Y/20.0; // Robot y
    models[0].stateEstimates[2][0] = M_PI/2.0;      // Robot heading
  }
  // player 4 starts on right side line, 2 meters from end
  else if (robotState->WO_SELF == 4){
    models[0].stateEstimates[0][0] = -100;          // Robot X
    models[0].stateEstimates[1][0] = -FIELD_Y/20.0; // Robot y
    models[0].stateEstimates[2][0] = M_PI/2.0;      // Robot heading
  } else {
    cout << "ERROR in ukf player reset: invalid player number! " << robotState->WO_SELF << endl;
    models[0].stateEstimates[0][0] = -200;          // Robot X
    models[0].stateEstimates[1][0] = FIELD_Y/20.0;  // Robot y
    models[0].stateEstimates[2][0] = -M_PI/2.0;     // Robot heading
  }

  // assume stationary ball at center of field
  models[0].stateEstimates[3][0] = 0.0;       // Ball x
  models[0].stateEstimates[4][0] = 0.0;       // Ball y
  models[0].stateEstimates[5][0] = 0.0;       // Ball vx
  models[0].stateEstimates[6][0] = 0.0;       // Ball vy

  // Set the uncertainties
  // more uncertainty in x, while y and theta should be pretty certain
  resetSdMatrixForSideline(0);

  return;
}


void UKFModule::resetSdMatrix(int modelNumber)
{
  // Set the uncertainties
  initSdMatrix(modelNumber);
  models[modelNumber].stateStandardDeviations[0][0] = ukfParams.init_sd_x;
  models[modelNumber].stateStandardDeviations[1][1] = ukfParams.init_sd_y;
  models[modelNumber].stateStandardDeviations[2][2] = ukfParams.init_sd_theta;
  models[modelNumber].stateStandardDeviations[3][3] = ukfParams.init_sd_x;
  models[modelNumber].stateStandardDeviations[4][4] = ukfParams.init_sd_y;
  models[modelNumber].stateStandardDeviations[5][5] = ukfParams.init_sd_vel;
  models[modelNumber].stateStandardDeviations[6][6] = ukfParams.init_sd_vel;
  return;
}


void UKFModule::resetSdMatrixForSideline(int modelNumber)
{
  // Set the uncertainties, but confidence in sideline/angle
  initSdMatrix(modelNumber);
  models[modelNumber].stateStandardDeviations[0][0] = ukfParams.init_sd_x;
  models[modelNumber].stateStandardDeviations[1][1] = ukfParams.init_sd_y / 2.5;
  models[modelNumber].stateStandardDeviations[2][2] = ukfParams.init_sd_theta / 2.5;
  models[modelNumber].stateStandardDeviations[3][3] = ukfParams.init_sd_x;
  models[modelNumber].stateStandardDeviations[4][4] = ukfParams.init_sd_y;
  models[modelNumber].stateStandardDeviations[5][5] = ukfParams.init_sd_vel;
  models[modelNumber].stateStandardDeviations[6][6] = ukfParams.init_sd_vel;
  return;
}

bool UKFModule::clipModelToField(int modelID)
{

  const float fieldXLength = GRASS_X/10.0;
  const float fieldYLength = GRASS_Y/10.0;
  const float fieldXMax = fieldXLength / 2.0;
  const float fieldXMin = - fieldXLength / 2.0;
  const float fieldYMax = fieldYLength / 2.0;
  const float fieldYMin = - fieldYLength / 2.0;

  bool wasClipped = false;
  bool clipped;
  float prevX, prevY, prevTheta;
  prevX = models[modelID].getState(0);
  prevY = models[modelID].getState(1);
  prevTheta = models[modelID].getState(2);

  clipped = models[modelID].clipState(0, fieldXMin, fieldXMax);   // Clipping for robot's X
  if(clipped){
    ukfLog((20, "Model %i, Alpha: %5.3f, Clipped 0, prev: (%5.1f, %5.1f), %5.1f, new: ((%5.1f, %5.1f), %5.1f", modelID, models[modelID].alpha, prevX, prevY, prevTheta, models[modelID].getState(0), models[modelID].getState(1), models[modelID].getState(2)));
  }

  wasClipped = wasClipped || clipped;

  prevX = models[modelID].getState(0);
  prevY = models[modelID].getState(1);
  prevTheta = models[modelID].getState(2);

  clipped = models[modelID].clipState(1, fieldYMin, fieldYMax);   // Clipping for robot's Y

  if(clipped){
    ukfLog((20, "Model %i, Alpha: %5.3f, Clipped 1, prev: (%5.1f, %5.1f), %5.1f, new: ((%5.1f, %5.1f), %5.1f", modelID, models[modelID].alpha, prevX, prevY, prevTheta, models[modelID].getState(0), models[modelID].getState(1), models[modelID].getState(2)));
  }

  wasClipped = wasClipped || clipped;
  prevX = models[modelID].getState(3);
  prevY = models[modelID].getState(4);

  clipped = models[modelID].clipState(3, fieldXMin, fieldXMax);   // Clipping for ball's X

  if(clipped){
    ukfLog((20, "Model %i, Alpha: %5.3f, Clipped 3, prev: (%5.1f, %5.1f), new: ((%5.1f, %5.1f)", modelID, models[modelID].alpha, prevX, prevY, models[modelID].getState(3), models[modelID].getState(4)));
  }

  wasClipped = wasClipped || clipped;

  prevX = models[modelID].getState(3);
  prevY = models[modelID].getState(4);

  clipped = models[modelID].clipState(4, fieldYMin, fieldYMax);   // Clipping for ball's Y

  if(clipped){
    ukfLog((20, "Model %i, Alpha: %5.3f, Clipped 4, prev: (%5.1f, %5.1f), new: ((%5.1f, %5.1f)", modelID, models[modelID].alpha, prevX, prevY, models[modelID].getState(3), models[modelID].getState(4)));
  }

  wasClipped = wasClipped || clipped;

  return wasClipped;
}



bool UKFModule::clipActiveModelsToField()
{
  bool wasClipped = false;
  bool modelClipped = false;
  for(int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == true){
      modelClipped = clipModelToField(modelID);
      wasClipped = wasClipped || modelClipped;
    }
  }
  return wasClipped;
}


void UKFModule::processSharedBall(){

  // if we're doing run core in the tool, set the updated flag back to true
  // so we'll process the shared packets again
#ifdef TOOL
  for (int i = WO_TEAM1; i <= WO_TEAM4; i++){
    if ((frameInfo->frame_id-1) == (unsigned)teamPacketsMem->frameReceived[i])
      teamPacketsMem->ballUpdated[i] = true;
  }
#endif

  // see if we have any team packets we haven't updated yet
  for (int i = WO_TEAM1; i <= WO_TEAM4; i++){
    if (i == robotState->WO_SELF) continue;
    if (teamPacketsMem->ballUpdated[i]){
      locStruct* tp = &(teamPacketsMem->tp[i].locData); // Struct of localization data from a team mate

      // only ready, set, playing
      if (teamPacketsMem->tp[i].bvrData.state != READY && teamPacketsMem->tp[i].bvrData.state != SET &&  teamPacketsMem->tp[i].bvrData.state != PLAYING){
        ukfLog((10, "ignore message from robot %i in state %i", i, teamPacketsMem->tp[i].bvrData.state));
        teamPacketsMem->ballUpdated[i] = false;
        continue;
      }

      // only if ball seen recently
      if (teamPacketsMem->tp[i].bvrData.ballMissed > 4) {
        ukfLog((10, "ignore message from robot %i who has missed ball for %i frames", i, teamPacketsMem->tp[i].bvrData.ballMissed));
        teamPacketsMem->ballUpdated[i] = false;
        continue;
      }

      // check that packet is ok
      if (isnan(tp->ballX) || isnan(tp->ballY) || tp->ballSDX == 0 || tp->ballSDXY == 0 || tp->ballSDY == 0 || (frameInfo->frame_id-teamPacketsMem->frameReceived[i]) > 30){
        ukfLog((20, "Bad teammate ball pkt, either nan, 0 sd, or bad frame #"));
        teamPacketsMem->ballUpdated[i] = false;
        continue;
      }

      ukfLog((20, "Do teammate ball update from mate %i recv frame %i", i, teamPacketsMem->frameReceived[i]));

      // no conversion between cm/mm because teammates using ukf send it in cm
      float sdFactor = ukfParams.shared_ball_sd_factor;
      doSharedBallUpdate(tp->ballX, tp->ballY, sdFactor * tp->ballSDX, sdFactor * tp->ballSDXY, sdFactor * tp->ballSDY, i);
      // set as updated now
      teamPacketsMem->ballUpdated[i] = false;
    }
  }

}


void UKFModule::checkSharedBallForFlips(){

  // if we're doing run core in the tool, set the updated flag back to true
  // so we'll process the shared packets again
#ifdef TOOL
  for (int i = WO_TEAM1; i <= WO_TEAM4; i++){
    if ((frameInfo->frame_id-1) == (unsigned)teamPacketsMem->frameReceived[i])
      teamPacketsMem->ballUpdated[i] = true;
  }
#endif

  // see if we have any team packets we haven't updated yet
  for (int i = WO_TEAM1; i <= WO_TEAM4; i++){
    if (i == robotState->WO_SELF) continue;
    if (teamPacketsMem->ballUpdated[i]){
      locStruct* tp = &(teamPacketsMem->tp[i].locData); // Struct of localization data from a team mate

      // only ready, set, playing
      if (teamPacketsMem->tp[i].bvrData.state != READY && teamPacketsMem->tp[i].bvrData.state != SET &&  teamPacketsMem->tp[i].bvrData.state != PLAYING){
        ukfLog((10, "ignore message from robot %i in state %i", i, teamPacketsMem->tp[i].bvrData.state));
        teamPacketsMem->ballUpdated[i] = false;
        continue;
      }

      // only if ball seen recently
      if (teamPacketsMem->tp[i].bvrData.ballMissed > 4) {
        ukfLog((10, "ignore message from robot %i who has missed ball for %i frames", i, teamPacketsMem->tp[i].bvrData.ballMissed));
        teamPacketsMem->ballUpdated[i] = false;
        continue;
      }

      // check that packet is ok
      if (isnan(tp->ballX) || isnan(tp->ballY) || tp->ballSDX == 0 || tp->ballSDXY == 0 || tp->ballSDY == 0 || (frameInfo->frame_id-teamPacketsMem->frameReceived[i]) > 30){
        ukfLog((20, "Bad teammate ball pkt, either nan, 0 sd, or bad frame #"));
        teamPacketsMem->ballUpdated[i] = false;
        continue;
      }

      ukfLog((20, "Do teammate ball flip check from mate %i recv frame %i", i, teamPacketsMem->frameReceived[i]));

      // no conversion between cm/mm because teammates using ukf send it in cm
      doSharedBallCheck(tp->ballX, tp->ballY, i);

    }
  }


  // we'll want to have seen at least 2 teammates tell us our ball is flipped at least 2 times
  if (numMateFlippedBalls > ukfParams.mate_flipped_ball_thresh && robotState->WO_SELF != KEEPER){
    ukfLog((5, "Mate flipped count up to %i, create flipped model",numMateFlippedBalls));
    spawnFlippedModel(getBestModelID(), false);
    numMateFlippedBalls = 0;
  }
}

void UKFModule::processObservations(){

  // loop through all wo from first non-robot
  for (int i = WO_CENTER_CIRCLE; i < NUM_WORLD_OBJS; i++){
    WorldObject* wo = &(worldObjects->objects_[i]);
    if (!wo->seen) continue;

    // unique objects
    if (wo->isUnique()){
      ukfLog((10, "Attempt model update on unique obs %i", i));
      // loop over active models and do update
      for(int modelID = 0; modelID < c_MAX_MODELS; modelID++){
        if(!models[modelID].isActive) continue;
        int result = doKnownLandmarkMeasurementUpdate(modelID, i);
        if (result == UKF7_OUTLIER){
          ukfLog((70, "Model update on obs %i returned OUTLIER", i));
        } else if (result == UKF7_OFF){
          ukfLog((70, "No update on obs %i because it is turned off in params", i));
        }
      }
    }

    // ambiguous objects
    if (ukfParams.USE_AMBIG && wo->isAmbiguous()){
      ukfLog((10, "Attempt update from ambig landmark %i", i));

      int result = UKF7_OK;

      // split on lines only if outliers, we need to figure out value
      //float sumErrors = 0;
      bool split_on_lines = false;
      if (wo->isLine() && !ukfParams.SPLIT_ON_LINES && ukfParams.SPLIT_LINES_ON_HIGH_SD){
        for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
          if (models[modelID].isActive == false) continue;
          ukfLog((20, "model %i, has sd (%5.3f, %5.3f) %5.3f", modelID, models[modelID].stateStandardDeviations[0][0], models[modelID].stateStandardDeviations[1][1], models[modelID].stateStandardDeviations[2][2] * RAD_T_DEG));
          if (models[modelID].stateStandardDeviations[0][0] > ukfParams.SPLIT_LINE_SD_POS_THRESH || models[modelID].stateStandardDeviations[1][1] > ukfParams.SPLIT_LINE_SD_POS_THRESH || models[modelID].stateStandardDeviations[2][2] > ukfParams.SPLIT_LINE_SD_THETA_THRESH){
            ukfLog((10, "model %i, has high sd (%5.3f, %5.3f) %5.3f, splitting on lines activated", modelID, models[modelID].stateStandardDeviations[0][0], models[modelID].stateStandardDeviations[1][1], models[modelID].stateStandardDeviations[2][2] * RAD_T_DEG));
            split_on_lines = true;
            break;
          }
        }
      }

      // normal ambig update where we split for each hypothesis
      // posts: only split on completely unknown posts, not if known left/right
      if ((i == WO_UNKNOWN_GOALPOST && ukfParams.SPLIT_ON_POSTS) ||
          (wo->isLine() && ukfParams.SPLIT_ON_LINES) ||
          (wo->isIntersection() && ukfParams.SPLIT_ON_INTS) ||
          (wo->isLine() && split_on_lines)){
        result = doAmbiguousLandmarkMeasurementUpdate(i);
      }
      // alternative where we keep one model and update on most likely landmark
      else {
        result = doAmbiguousUpdateWithMostLikelyLM(i);
      }

      // print result
      if (result == UKF7_OUTLIER){
        ukfLog((70, "Model update on obs %i returned OUTLIER", i));
      } else if (result == UKF7_OFF){
        ukfLog((70, "No update on obs %i because it is turned off in params", i));
      }
      NormaliseAlphas();
    }

  } // all wo

  // ball observation
  WorldObject* ball = &(worldObjects->objects_[WO_BALL]);
  if (ball->seen){
    ukfLog((10, "Update ukf with ball obs"));
    int goodUpdates = 0;

    // update all models
    for(int modelID = 0; modelID < c_MAX_MODELS; modelID++){
      if(!models[modelID].isActive) continue;

      ukfLog((20, "Model %i original estimates: (%5.3f, %5.3f) r: %5.3f, ball: (%5.3f, %5.3f), alpha: %5.8f", modelID, models[modelID].stateEstimates[0][0], models[modelID].stateEstimates[1][0], RAD_T_DEG*models[modelID].stateEstimates[2][0], models[modelID].stateEstimates[3][0], models[modelID].stateEstimates[4][0], models[modelID].alpha));

      // update ball observation
      goodUpdates += doBallMeasurementUpdate(modelID);

      ukfLog((20, "Model %i post-update estimates: (%5.3f, %5.3f) r: %5.3f, ball: (%5.3f, %5.3f), alpha: %5.8f", modelID, models[modelID].stateEstimates[0][0], models[modelID].stateEstimates[1][0], RAD_T_DEG*models[modelID].stateEstimates[2][0], models[modelID].stateEstimates[3][0], models[modelID].stateEstimates[4][0],models[modelID].alpha));

    }

    // no good updates!
    if (goodUpdates == 0){
      numBadBallUpdates++;
      ukfLog((10, "No good ball updates, count at %i!!", numBadBallUpdates));
    } else if (goodUpdates > 0 && numBadBallUpdates > -1){
      numBadBallUpdates--;
    }
  } // ball updates
  else {
    numBadBallUpdates = 0;
  }

  if (numBadBallUpdates > 2){
    int modelID = getBestModelID();
    Point2D relBall(ball->visionDistance, ball->visionBearing + models[modelID].stateEstimates[2][0], POLAR);
    Point2D seenBall = Point2D(models[modelID].stateEstimates[0][0], models[modelID].stateEstimates[1][0]) + relBall;
    ukfLog((10, "Many bad ball updates, spawn model with new ball location"));
    spawnNewBallModel(modelID, seenBall);
    numBadBallUpdates = 0;
  }
      

  NormaliseAlphas();

}

/** Time update for all models */
void UKFModule::runTimeUpdates(){

  // calculate odom update
  // convert from mm to cm
  float odomForward = odometry->displacement.translation.x/10.0;
  float odomLeft    = odometry->displacement.translation.y/10.0;
  float odomTurn    = odometry->displacement.rotation;

  ukfLog((30, "Update UKF odometry with fwd %f, left %f, turn %f, standing %i, timePassed %5.3f", odomForward, odomLeft, odomTurn, odometry->standing, timePassed));

  bool bump = processedSonar->bump_left_ || processedSonar->bump_right_ || odometry->getting_up_side_ != Getup::NONE;

  // if not walking.. no odom update
  if (odometry->standing){
    ukfLog((10, "Standing, no odom update"));
    odomForward = 0.0;
    odomLeft = 0.0;
    odomTurn = 0.0;
  }

  // bad odometry
  if (fabs(odomTurn) > 1.0 || fabs(odomForward) > 100 || fabs(odomLeft) > 100.0){
    ukfLog((10, "Bad odom fwd %5.3f, side %5.3f, turn %5.3f, ignoring", odomForward, odomLeft, odomTurn));
    odomForward = 0.0;
    odomLeft = 0.0;
    odomTurn = 0.0;
  }



  for(int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == false) continue; // Skip Inactive models.
    bool fallen = odometry->getting_up_side_ != Getup::NONE || odometry->fall_direction_ != Fall::NONE;
    models[modelID].timeUpdate(odomForward,odomLeft,odomTurn,timePassed,bump,fallen,odometry->standing);

  }
}


void UKFModule::doSharedBallCheck(float x, float y, int mate){

  ukfLog((20, "Shared Ball Check. SharedBall: (%5.1f, %5.1f)", x, y));

  float closestBallDist = 10000;
  bool possibleFlip = false;
  for(int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == false) continue; // Skip Inactive models.
    Point2D ourBall(models[modelID].stateEstimates[3][0], models[modelID].stateEstimates[4][0]);
    float ballDist = Point2D(x,y).getDistanceTo(ourBall);
    ukfLog((10, "Model %i has ball at (%5.3f, %5.3f), mate %i says ball at (%5.3f, %5.3f), dist %5.3f", modelID, models[modelID].stateEstimates[3][0], models[modelID].stateEstimates[4][0], mate, x, y, ballDist));
    if (ballDist < closestBallDist)
      closestBallDist = ballDist;
    if (ballDist > 120){
      Point2D flippedMateBall(-x, -y);
      if (ourBall.getDistanceTo(flippedMateBall) < 80){
        possibleFlip = true;
        ukfLog((10, "Possible flip from mate %i", mate));
      }
    }
  }

  ukfLog((10, "Closest ball match was distance %5.3f", closestBallDist));

  // no good updates, check if flipped ball
  // only track this if we've seen ball recently as well
  if ((frameInfo->frame_id - worldObjects->objects_[WO_BALL].frameLastSeen) < 10){
    if (closestBallDist > 120 && possibleFlip){
      numMateFlippedBalls++;
      ukfLog((10, "Possible flip, increase mate flip count to %i", numMateFlippedBalls));
    }
  
    // if the teammate agrees with us... then not a mate flipped ball mesg
    if (closestBallDist < 50){
      if (numMateFlippedBalls > -1)
        numMateFlippedBalls--;
      ukfLog((10, "Mate %i reported ball %5.3f of our ball location, agreeing with our ball estimate, mate flip count to %i", mate, closestBallDist, numMateFlippedBalls));
    }
  } else {
    // reset flipped count to 0 since we havent seen the ball ourselves
    numMateFlippedBalls = 0;
  }
  
 
}


int UKFModule::doSharedBallUpdate(float x, float y, float sdxx, float sdxy, float sdyy, int mate){
  int kf_return;
  int numSuccessfulUpdates = 0;

  ukfLog((20, "Shared Ball Update. SharedBall: (%5.1f, %5.1f) SRXX: %5.1f, SRXY %5.1f, SRYY %5.1f", x, y, sdxx, sdxy, sdyy));

  for(int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == false) continue; // Skip Inactive models.
    kf_return = UKF7_OK;
    // tell who this is from and whether we are the keeper
    kf_return = models[modelID].linear2MeasurementUpdate(x, y, sdxx, sdxy, sdyy, 3, 4, mate, robotState->WO_SELF == KEEPER);
    if(kf_return == UKF7_OK) numSuccessfulUpdates++;

    ukfLog((10, "New ball estimate: (%5.3f, %5.3f), alpha: %5.3f", models[modelID].stateEstimates[3][0], models[modelID].stateEstimates[4][0], models[modelID].alpha));
  }

  return numSuccessfulUpdates;
  
}




void UKFModule::spawnFlippedModel(int modelID, bool robotOnly){
  int numFreeModels = getNumFreeModels();
  int numActiveModels = getNumActiveModels();
  int numRequiredModels = 1;

  if(numFreeModels < numRequiredModels){
    int maxActiveAfterMerge = c_MAX_MODELS - numRequiredModels;
    ukfLog((30, "Only %i free models, need %i models. merging to %i models",
            numFreeModels, numRequiredModels, maxActiveAfterMerge));
    MergeModels(maxActiveAfterMerge);
    if(getNumFreeModels() < numRequiredModels){
      ukfLog((10, "Only %i free models. not enough. aborting update", getNumFreeModels()));
      return;
    }
  }

  // Copy initial model to the temporary model.
  tempModel = models[modelID];
  tempModel.isActive = false;
  tempModel.toBeActivated = true;

  // now for flipped model
  int newModelID = FindNextFreeModel();

  // If an invalid modelID has been returned, something has gone horribly wrong, so stop here.
  if(newModelID < 0){
    return;
  }

  ukfLog((10, "Flipped update. split model %i into model %i (flipped pos)", modelID, newModelID));
  models[newModelID] = tempModel; // Get the new model from the temp

  // Copy outlier history from the current model.
  for (int i=0; i<NUM_WORLD_OBJS; i++){
    modelObjectErrors[newModelID][i] = modelObjectErrors[modelID][i];
  }

  // update new modelID with flipped state
  models[newModelID].stateEstimates[0][0] = -models[modelID].stateEstimates[0][0];
  models[newModelID].stateEstimates[1][0] = -models[modelID].stateEstimates[1][0];
  models[newModelID].stateEstimates[2][0] = normalizeAngle(M_PI+models[modelID].stateEstimates[2][0]);
  if (!robotOnly){
    models[newModelID].stateEstimates[3][0] = -models[modelID].stateEstimates[3][0];
    models[newModelID].stateEstimates[4][0] = -models[modelID].stateEstimates[4][0];
    models[newModelID].stateEstimates[5][0] = -models[modelID].stateEstimates[5][0];
    models[newModelID].stateEstimates[6][0] = -models[modelID].stateEstimates[6][0];
  }

  if (models[newModelID].toBeActivated) {
    models[newModelID].alpha *= 0.5; // Divide each models alpha by the numbmer of splits.
    models[newModelID].isActive=true;
    models[newModelID].toBeActivated=false; // Turn off activation flag
  }
}

void UKFModule::spawnNewBallModel(int modelID, Point2D newBall){
  int numFreeModels = getNumFreeModels();
  int numActiveModels = getNumActiveModels();
  int numRequiredModels = 1;

  if(numFreeModels < numRequiredModels){
    int maxActiveAfterMerge = c_MAX_MODELS - numRequiredModels;
    ukfLog((30, "Only %i free models, need %i models. merging to %i models",
            numFreeModels, numRequiredModels, maxActiveAfterMerge));
    MergeModels(maxActiveAfterMerge);
    if(getNumFreeModels() < numRequiredModels){
      ukfLog((10, "Only %i free models. not enough. aborting update", getNumFreeModels()));
      return;
    }
  }

  // Copy initial model to the temporary model.
  tempModel = models[modelID];
  tempModel.isActive = false;
  tempModel.toBeActivated = true;

  // now for flipped model
  int newModelID = FindNextFreeModel();

  // If an invalid modelID has been returned, something has gone horribly wrong, so stop here.
  if(newModelID < 0){
    return;
  }

  ukfLog((10, "Flipped update. split model %i into model %i (flipped pos)", modelID, newModelID));
  models[newModelID] = tempModel; // Get the new model from the temp

  // Copy outlier history from the current model.
  for (int i=0; i<NUM_WORLD_OBJS; i++){
    modelObjectErrors[newModelID][i] = modelObjectErrors[modelID][i];
  }

  models[newModelID].stateEstimates[3][0] = newBall.x;
  models[newModelID].stateEstimates[4][0] = newBall.y;

  if (models[newModelID].toBeActivated) {
    models[newModelID].alpha *= 0.25; // Divide each models alpha by the numbmer of splits.
    models[newModelID].isActive=true;
    models[newModelID].toBeActivated=false; // Turn off activation flag
  }
}





int UKFModule::doBallMeasurementUpdate(int modelNum)
{

  WorldObject* ball = &(worldObjects->objects_[WO_BALL]);

  int kf_return;
  int numSuccessfulUpdates = 0;

  ukfLog((20, "Model %i Ball Update. Dist: %5.1f, Bear: %5.1f", modelNum, ball->visionDistance, RAD_T_DEG*ball->visionBearing));

  if (ball->visionDistance > 110 && models[modelNum].ballFramesSinceAmbig > 0){
    models[modelNum].ballFramesSinceAmbig--;
    ukfLog((10, "Ball seen at distance %5.3f, so ball ambig frames reduced to %i", ball->visionDistance, models[modelNum].ballFramesSinceAmbig));
  }

  kf_return = UKF7_OK;
  kf_return = models[modelNum].ballmeas(ball->visionDistance, ball->visionBearing);
  //kf_return = models[modelNum].ballmeasWithBearing(ball->visionDistance, ball->visionBearing);

  if(kf_return == UKF7_OK) numSuccessfulUpdates++;
  return numSuccessfulUpdates;
}


/** Update model modelNum based on observation of worldObject objID */
int UKFModule::doKnownLandmarkMeasurementUpdate(int modelNum, int objID)
{
  WorldObject* landmark = &(worldObjects->objects_[objID]);

  int kf_return;
  int numSuccessfulUpdates = 0;

  // normal
  float distanceOffsetError = ukfParams.R_goal_range_offset;
  float distanceRelativeError = ukfParams.R_goal_range_relative;
  float bearingError = ukfParams.R_goal_theta;

  if (landmark->isGoal() && !ukfParams.USE_GOAL) return UKF7_OFF;
  if (landmark->isGoalCenter() && !ukfParams.USE_WHOLE_GOAL) return UKF7_OFF;

  // goal
  if (landmark->isGoal()){
    if (!ukfParams.USE_GOAL) return UKF7_OFF;
    if (landmark->visionDistance > ukfParams.ignore_goal_dist){
      ukfLog((10, "Goal thrown out on distance %5.1f", landmark->visionDistance));
      return UKF7_OFF;
    }
    distanceOffsetError = ukfParams.R_goal_range_offset;
    distanceRelativeError = ukfParams.R_goal_range_relative;
    bearingError = ukfParams.R_goal_theta;
  }
  // intersections
  else if (landmark->isIntersection()){
    if (!ukfParams.USE_INTERSECTIONS) return UKF7_OFF;
    if (landmark->visionDistance > ukfParams.ignore_int_dist){
      ukfLog((10, "Int thrown out on distance %5.1f", landmark->visionDistance));
      return UKF7_OFF;
    }
    distanceOffsetError = ukfParams.R_int_range_offset;
    distanceRelativeError = ukfParams.R_int_range_relative;
    bearingError = ukfParams.R_int_theta;
  }
  // circle
  else if (landmark->isCenterCircle()){
    if (!ukfParams.USE_CIRCLE) return UKF7_OFF;
    if (landmark->visionDistance > ukfParams.ignore_circle_dist) {
      ukfLog((10, "Circle thrown out on distance %5.1f", landmark->visionDistance));
      return UKF7_OFF;
    }
    distanceOffsetError = ukfParams.R_circle_range_offset;
    distanceRelativeError = ukfParams.R_circle_range_relative;
    bearingError = ukfParams.R_circle_theta;
  }
  // line
  else if (landmark->isLine()){
    if (!ukfParams.USE_LINES) return UKF7_OFF;
    if (landmark->visionDistance > ukfParams.ignore_line_dist) {
      ukfLog((10, "Line thrown out on distance %5.1f", landmark->visionDistance));
      return UKF7_OFF;
    }
    distanceOffsetError = ukfParams.R_line_range_offset;
    distanceRelativeError = ukfParams.R_line_range_relative;
    bearingError = ukfParams.R_line_theta;
  }
  // ball
  else {
    ukfLog((0, "invalid landmark for known update id %i", objID));
  }

  Point2D lmLoc = landmark->loc;
  float lineOrient = 0;

  if (landmark->isLine()){
    // convert to mm since lineloc is still in cm
    Point2D myLoc(models[modelNum].stateEstimates[0][0]*10.0, models[modelNum].stateEstimates[1][0]*10.0);
    lmLoc = landmark->lineLoc.getPointOnLineClosestTo(myLoc);
    // convert lmLoc back to cm
    lmLoc /= 10.0;
    ukfLog((80, "Line update, myLoc (%5.1f, %5.1f), nearestPoint (%5.1f, %5.1f), orig landmasrkLoc (%5.1f, %5.1f)", myLoc.x, myLoc.y, lmLoc.x, lmLoc.y, landmark->loc.x, landmark->loc.y));

    // figure out line orientation
    float a = landmark->lineLoc.getACoefficient();
    if (a < 0.01) {
      // vertical line
      lineOrient = 0;
    } else {
      lineOrient = M_PI/2.0;
    }
  }

  ukfLog((20, "Model %i LM %i Update. Dist: %5.1f, Bear: %5.1f, Loc: (%5.1f, %5.1f)", modelNum, objID, landmark->visionDistance, RAD_T_DEG*landmark->visionBearing, lmLoc.x, lmLoc.y));

  // if image was cropped, increase distance R
  if (landmark->visionConfidence < 0.1 && !landmark->isLine()){
    ukfLog((20, "Object %i cropped (vc %5.3f), increase distance R by factor of %5.3f", objID, landmark->visionConfidence, ukfParams.R_crop_factor));
    distanceOffsetError *= ukfParams.R_crop_factor;
    distanceRelativeError *= ukfParams.R_crop_factor;
  }

  if(landmark->visionBearing != landmark->visionBearing){
    ukfLog((0, "Model %i LM %i NaN bearing", modelNum, objID));
    return 0;
  }
  kf_return = UKF7_OK;

  ukfLog((20, "Model %i original estimates: (%5.3f, %5.3f) r: %5.3f, ball: (%5.3f, %5.3f), alpha: %5.8f", modelNum, models[modelNum].stateEstimates[0][0], models[modelNum].stateEstimates[1][0], RAD_T_DEG*models[modelNum].stateEstimates[2][0], models[modelNum].stateEstimates[3][0], models[modelNum].stateEstimates[4][0], models[modelNum].alpha));

  if (landmark->isLine()){
    kf_return = models[modelNum].lineObjectMeas(landmark->visionDistance, landmark->visionBearing, lmLoc.x, lmLoc.y, lineOrient, distanceOffsetError, distanceRelativeError, bearingError);
  }
  // dont do dist/bear error for really close objects... bearing doesnt make much sense here
  else if (landmark->visionDistance > 50){
    kf_return = models[modelNum].fieldObjectMeasWithBearing(landmark->visionDistance, landmark->visionBearing, lmLoc.x, lmLoc.y, distanceOffsetError, distanceRelativeError, bearingError);
  } else {
    kf_return = models[modelNum].fieldObjectmeas(landmark->visionDistance, landmark->visionBearing, lmLoc.x, lmLoc.y, distanceOffsetError, distanceRelativeError, bearingError);
  }

  ukfLog((20, "Model %i post-update estimates: (%5.3f, %5.3f) r: %5.3f, ball: (%5.3f, %5.3f), alpha: %5.8f", modelNum, models[modelNum].stateEstimates[0][0], models[modelNum].stateEstimates[1][0], RAD_T_DEG*models[modelNum].stateEstimates[2][0], models[modelNum].stateEstimates[3][0], models[modelNum].stateEstimates[4][0],models[modelNum].alpha));

  if(kf_return == UKF7_OUTLIER) modelObjectErrors[modelNum][objID] += 1.0;

  ukfLog((70, "Update result %i", kf_return));

  if(kf_return == UKF7_OK) numSuccessfulUpdates++;

  return numSuccessfulUpdates;
}

/** Do an update more like a particle filter. Just use the most likely object. */
int UKFModule::doAmbiguousUpdateWithMostLikelyLM(int objID){
  ukfLog((10, "in ML ambig update %i", objID));
  int kf_return;


  WorldObject* wo = &(worldObjects->objects_[objID]);
  int firstPossible = 0;

  float distanceOffsetError = ukfParams.R_goal_range_offset;
  float distanceRelativeError = ukfParams.R_goal_range_relative;
  float bearingError = ukfParams.R_goal_theta;

  // figure out number of options
  int numOptions = 1;
  if (wo->isUnknownPost()){
    if (!ukfParams.USE_GOAL) return UKF7_OFF;
    if (wo->visionDistance > ukfParams.ignore_goal_dist) {
      ukfLog((10, "Goal thrown out on distance %5.1f", wo->visionDistance));
      return UKF7_OFF;
    }
    if (wo->type == WO_UNKNOWN_GOALPOST){
      numOptions = 4;
      firstPossible = WO_OWN_LEFT_GOALPOST;
    }
    else if (wo->type == WO_UNKNOWN_LEFT_GOALPOST){
      numOptions = 2;
      firstPossible = WO_OWN_LEFT_GOALPOST;
    } else {
      numOptions = 2;
      firstPossible = WO_OWN_RIGHT_GOALPOST;
    }
    distanceOffsetError = ukfParams.R_goal_range_offset;
    distanceRelativeError = ukfParams.R_goal_range_relative;
    bearingError = ukfParams.R_goal_theta;
  }
  else if (wo->isUnknownGoalCenter()){
    if (!ukfParams.USE_GOAL) return UKF7_OFF;
    if (!ukfParams.USE_WHOLE_GOAL) return UKF7_OFF;
    if (wo->visionDistance > ukfParams.ignore_goal_dist) {
      ukfLog((10, "Goal thrown out on distance %5.1f", wo->visionDistance));
      return UKF7_OFF;
    }
    numOptions = 2;
    firstPossible = WO_OWN_GOAL;

    distanceOffsetError = ukfParams.R_goal_range_offset;
    distanceRelativeError = ukfParams.R_goal_range_relative;
    bearingError = ukfParams.R_goal_theta;
  }
  else if (wo->isUnknownLine()){
    // using lines?
    if (!ukfParams.USE_LINES) return UKF7_OFF;
    if (wo->visionDistance > ukfParams.ignore_line_dist){
      ukfLog((10, "Line thrown out on distance %5.1f", wo->visionDistance));
      return UKF7_OFF;
    }
    numOptions = NUM_LINES;
    firstPossible = LINE_OFFSET;
    distanceOffsetError = ukfParams.R_line_range_offset;
    distanceRelativeError = ukfParams.R_line_range_relative;
    bearingError = ukfParams.R_line_theta;
  }
  else if (wo->isUnknownIntersection()){
    if (!ukfParams.USE_INTERSECTIONS) return UKF7_OFF;
    if (wo->visionDistance > ukfParams.ignore_int_dist) {
      ukfLog((10, "Int thrown out on distance %5.1f", wo->visionDistance));
      return UKF7_OFF;
    }
    numOptions = NUM_INTERSECTIONS;
    firstPossible = INTERSECTION_OFFSET;
    distanceOffsetError = ukfParams.R_int_range_offset;
    distanceRelativeError = ukfParams.R_int_range_relative;
    bearingError = ukfParams.R_int_theta;
  }
  else if (wo->isUnknownPenaltyCross()){
    if (!ukfParams.USE_CROSSES) return UKF7_OFF;
    if (wo->visionDistance > ukfParams.ignore_cross_dist) {
      ukfLog((10, "Cross thrown out on distance %5.1f", wo->visionDistance));
      return UKF7_OFF;
    }
    numOptions = NUM_CROSSES;
    firstPossible = WO_OWN_PENALTY_CROSS;
    distanceOffsetError = ukfParams.R_cross_range_offset;
    distanceRelativeError = ukfParams.R_cross_range_relative;
    bearingError = ukfParams.R_cross_theta;
  }
  else {
    ukfLog((0, "unknown type? really ambiguous? %i", objID));
    return UKF7_OUTLIER;
  }

  ukfLog((10, "Update from ambig lm %i, numOptions %i, starting from %i", objID, numOptions, firstPossible));

  // if image was cropped, increase distance R
  if (wo->visionConfidence < 0.1 && !wo->isLine()){
    ukfLog((20, "Object %i cropped (vc %5.3f), increase distance R by factor of %5.3f", objID, wo->visionConfidence, ukfParams.R_crop_factor));
    distanceOffsetError *= ukfParams.R_crop_factor;
    distanceRelativeError *= ukfParams.R_crop_factor;
  }


  // loop through all models
  for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == false) continue; // Skip inactive models.

    int bestMatch = -1;
    float bestSim = -1;
    Point2D bestLMLoc(0,0);
    float bestLineOrient = 0;

    // Now go through each of the possible options, and apply it to a copy of the model
    for(int possibleID = firstPossible; possibleID < (firstPossible+numOptions); possibleID++){

      WorldObject* possObj = &(worldObjects->objects_[possibleID]);

      Point2D lmLoc = possObj->loc;

      // get expected bearing and distance
      Point2D myLoc(models[modelID].stateEstimates[0][0], models[modelID].stateEstimates[1][0]);
      AngRad orient = models[modelID].stateEstimates[2][0];
      float lineOrient = 0;

      if (possObj->isLine()){

        // check that this line would even be visible from this location
        if (possObj->isPenaltySideline() && ukfParams.CHECK_LINES_IN_FOV){
          if (!lineInFOV(Point2D(models[modelID].stateEstimates[0][0], models[modelID].stateEstimates[1][0]),models[modelID].stateEstimates[2][0], possibleID, objID)){
            ukfLog((10, "Do no try match to line %i, not in FOV", possibleID));
            continue;
          }
        }

        // convert to mm since lineloc is still in cm
        Point2D myLLoc(models[modelID].stateEstimates[0][0]*10.0, models[modelID].stateEstimates[1][0]*10.0);
        lmLoc = possObj->lineLoc.getPointOnLineClosestTo(myLLoc);
        // convert lmLoc back to cm
        lmLoc /= 10.0;
        ukfLog((80, "Line update id %i, myLLoc (%5.1f, %5.1f), nearestPoint (%5.1f, %5.1f), orig landmarkLoc (%5.1f, %5.1f)", possibleID, myLLoc.x, myLLoc.y, lmLoc.x, lmLoc.y, possObj->loc.x, possObj->loc.y));

        float a = possObj->lineLoc.getACoefficient();
        if (a < 0.01) {
          // vertical line
          lineOrient = 0;
        } else {
          lineOrient = M_PI/2.0;
        }
      }


      float totalSim = 0;
      if (possObj->isLine()){
        totalSim = getLineSim(myLoc, orient, wo, lmLoc, lineOrient);
      } else {
        totalSim = getObsSim(myLoc, orient, wo, lmLoc);
      }
      if (totalSim > bestSim){
        bestSim = totalSim;
        bestMatch = possibleID;
        bestLMLoc = lmLoc;
        bestLineOrient = lineOrient;
        ukfLog((10, "Best match so far is %i, with sim %5.3f", possibleID, totalSim));
      }

    } // possible object loop

    ukfLog((10, "Update assuming landmark %i", bestMatch));
    ukfLog((20, "Model %i original estimates: (%5.3f, %5.3f) r: %5.3f, ball: (%5.3f, %5.3f), alpha: %5.8f", modelID, models[modelID].stateEstimates[0][0], models[modelID].stateEstimates[1][0], RAD_T_DEG*models[modelID].stateEstimates[2][0], models[modelID].stateEstimates[3][0], models[modelID].stateEstimates[4][0],models[modelID].alpha));

    // update assuming best match
    kf_return = UKF7_OK;
    if (wo->isLine()){
      kf_return = models[modelID].lineObjectMeas(wo->visionDistance, wo->visionBearing, bestLMLoc.x, bestLMLoc.y, bestLineOrient, distanceOffsetError, distanceRelativeError, bearingError);
    }
    else if (wo->visionDistance > 50){
      kf_return = models[modelID].fieldObjectMeasWithBearing(wo->visionDistance, wo->visionBearing, bestLMLoc.x, bestLMLoc.y, distanceOffsetError, distanceRelativeError, bearingError);
    } else {
      kf_return = models[modelID].fieldObjectmeas(wo->visionDistance, wo->visionBearing, bestLMLoc.x, bestLMLoc.y, distanceOffsetError, distanceRelativeError, bearingError);
    }

    ukfLog((20, "Model %i post-update estimates: (%5.3f, %5.3f) r: %5.3f, ball: (%5.3f, %5.3f), alpha: %5.8f", modelID, models[modelID].stateEstimates[0][0], models[modelID].stateEstimates[1][0], RAD_T_DEG*models[modelID].stateEstimates[2][0], models[modelID].stateEstimates[3][0], models[modelID].stateEstimates[4][0],models[modelID].alpha));

    if(kf_return == UKF7_OUTLIER) modelObjectErrors[modelID][bestMatch] += 1.0;

    ukfLog((70, "Update result %i", kf_return));

  } // model loop

  return UKF7_OK;
}

/** Multiple model UKF update. Split for each possible lm for ambiguous observations. */
int UKFModule::doAmbiguousLandmarkMeasurementUpdate(int objID){
  ukfLog((10, "in splitting ambig update %i", objID));
  int kf_return;


  WorldObject* wo = &(worldObjects->objects_[objID]);
  int firstPossible = 0;

  float distanceOffsetError = ukfParams.R_goal_range_offset;
  float distanceRelativeError = ukfParams.R_goal_range_relative;
  float bearingError = ukfParams.R_goal_theta;

  // figure out number of options
  int numOptions = 1;
  if (wo->isUnknownPost()){
    if (!ukfParams.USE_GOAL) return UKF7_OFF;
    if (wo->visionDistance > ukfParams.ignore_goal_dist) {
      ukfLog((10, "Goal thrown out on distance %5.1f", wo->visionDistance));
      return UKF7_OFF;
    }
    if (wo->type == WO_UNKNOWN_GOALPOST){
      numOptions = 4;
      firstPossible = WO_OWN_LEFT_GOALPOST;
    }
    else if (wo->type == WO_UNKNOWN_LEFT_GOALPOST){
      numOptions = 2;
      firstPossible = WO_OWN_LEFT_GOALPOST;
    } else {
      numOptions = 2;
      firstPossible = WO_OWN_RIGHT_GOALPOST;
    }
    distanceOffsetError = ukfParams.R_goal_range_offset;
    distanceRelativeError = ukfParams.R_goal_range_relative;
    bearingError = ukfParams.R_goal_theta;
  }
  else if (wo->isUnknownGoalCenter()){
    if (!ukfParams.USE_GOAL) return UKF7_OFF;
    if (wo->visionDistance > ukfParams.ignore_goal_dist) {
      ukfLog((10, "Goal thrown out on distance %5.1f", wo->visionDistance));
      return UKF7_OFF;
    }
    numOptions = 2;
    firstPossible = WO_OWN_GOAL;

    distanceOffsetError = ukfParams.R_goal_range_offset;
    distanceRelativeError = ukfParams.R_goal_range_relative;
    bearingError = ukfParams.R_goal_theta;
  }
  else if (wo->isUnknownLine()){
    // using lines?
    if (!ukfParams.USE_LINES) return UKF7_OFF;
    if (wo->visionDistance > ukfParams.ignore_line_dist){
      ukfLog((10, "Line thrown out on distance %5.1f", wo->visionDistance));
      return UKF7_OFF;
    }
    numOptions = NUM_LINES;
    firstPossible = LINE_OFFSET;
    distanceOffsetError = ukfParams.R_line_range_offset;
    distanceRelativeError = ukfParams.R_line_range_relative;
    bearingError = ukfParams.R_line_theta;
  }
  else if (wo->isUnknownIntersection()){
    if (!ukfParams.USE_INTERSECTIONS) return UKF7_OFF;
    if (wo->visionDistance > ukfParams.ignore_int_dist) {
      ukfLog((10, "Int thrown out on distance %5.1f", wo->visionDistance));
      return UKF7_OFF;
    }
    numOptions = NUM_INTERSECTIONS;
    firstPossible = INTERSECTION_OFFSET;
    distanceOffsetError = ukfParams.R_int_range_offset;
    distanceRelativeError = ukfParams.R_int_range_relative;
    bearingError = ukfParams.R_int_theta;
  }
  else if (wo->isUnknownPenaltyCross()){
    if (!ukfParams.USE_CROSSES) return UKF7_OFF;
    if (wo->visionDistance > ukfParams.ignore_cross_dist) {
      ukfLog((10, "Cross thrown out on distance %5.1f", wo->visionDistance));
      return UKF7_OFF;
    }
    numOptions = NUM_CROSSES;
    firstPossible = WO_OWN_PENALTY_CROSS;
    distanceOffsetError = ukfParams.R_cross_range_offset;
    distanceRelativeError = ukfParams.R_cross_range_relative;
    bearingError = ukfParams.R_cross_theta;
  }
  else {
    ukfLog((0, "unknown type? really ambiguous? %i", objID));
    return UKF7_OUTLIER;
  }

  ukfLog((10, "Update from ambig lm %i, numOptions %i, starting from %i", objID, numOptions, firstPossible));

  // if image was cropped, increase distance R
  if (wo->visionConfidence < 0.1 && !wo->isLine()){
    ukfLog((20, "Object %i cropped (vc %5.3f), increase distance R by factor of %5.3f", objID, wo->visionConfidence, ukfParams.R_crop_factor));
    distanceOffsetError *= ukfParams.R_crop_factor;
    distanceRelativeError *= ukfParams.R_crop_factor;
  }


  int numFreeModels = getNumFreeModels();
  int numActiveModels = getNumActiveModels();
  int numRequiredModels = numActiveModels * (numOptions+1); // An extra base model.

  if(numFreeModels < numRequiredModels){
    int maxActiveAfterMerge = c_MAX_MODELS /  (numOptions + 1);
    ukfLog((30, "Only %i free models, need %i models. merging to %i models",
            numFreeModels, numRequiredModels, maxActiveAfterMerge));
    MergeModels(maxActiveAfterMerge);
    if(getNumFreeModels() < (getNumActiveModels() * (numOptions+1))){
      ukfLog((10, "Only %i free models. not enough. aborting update", getNumFreeModels()));
      return UKF7_OUTLIER;
    }
  }

  ukfLog((20, "Doing ambig object %i update. Dist: %5.1f, Bear: %5.2f", objID, wo->visionDistance, RAD_T_DEG * wo->visionBearing));
  int numValid = 0;

  for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == false) continue; // Skip inactive models.

    // Copy initial model to the temporary model.
    tempModel = models[modelID];
    tempModel.isActive = false;
    tempModel.toBeActivated = true;

    // Now go through each of the possible options, and apply it to a copy of the model
    for(int possibleID = firstPossible; possibleID < (firstPossible+numOptions); possibleID++){

      int newModelID = FindNextFreeModel();

      // If an invalid modelID has been returned, something has gone horribly wrong, so stop here.
      if(newModelID < 0){
        ukfLog((10, "Ambig update, bad model id returned (%i), aborted.", newModelID));

        for(int m = 0; m < c_MAX_MODELS; m++) models[m].toBeActivated = false;
        return -1;
      }

      models[newModelID] = tempModel; // Get the new model from the temp

      // Copy outlier history from the current model.
      for (int i=0; i<NUM_WORLD_OBJS; i++){
        modelObjectErrors[newModelID][i] = modelObjectErrors[modelID][i];
      }

      WorldObject* possObj = &(worldObjects->objects_[possibleID]);

      Point2D lmLoc = possObj->loc;
      float lineOrient = 0;

      if (possObj->isLine()){

        // check that this line would even be visible from this location
        if (possObj->isPenaltySideline() && ukfParams.CHECK_LINES_IN_FOV){
          if (!lineInFOV(Point2D(models[newModelID].stateEstimates[0][0], models[newModelID].stateEstimates[1][0]),models[newModelID].stateEstimates[2][0], possibleID, objID)){
            ukfLog((10, "Do no split on line %i, not in FOV", possibleID));
            continue;
          }
        }

        // convert to mm since lineloc is still in cm
        Point2D myLoc(models[newModelID].stateEstimates[0][0]*10.0, models[newModelID].stateEstimates[1][0]*10.0);
        lmLoc = possObj->lineLoc.getPointOnLineClosestTo(myLoc);
        // convert lmLoc back to cm
        lmLoc /= 10.0;
        ukfLog((80, "Line update, myLoc (%5.1f, %5.1f), nearestPoint (%5.1f, %5.1f), orig landmarkLoc (%5.1f, %5.1f)", myLoc.x, myLoc.y, lmLoc.x, lmLoc.y, possObj->loc.x, possObj->loc.y));

        float a = possObj->lineLoc.getACoefficient();
        if (a < 0.01) {
          // vertical line
          lineOrient = 0;
        } else {
          lineOrient = M_PI/2.0;
        }
      }

      ukfLog((20, "Splitting model %i (alpha %5.3f) to model %i on obs of %i assuming its %i at loc: (%5.0f, %5.0f)", modelID, models[modelID].alpha, newModelID, objID, possibleID, lmLoc.x, lmLoc.y));

      ukfLog((20, "Model %i original estimates: (%5.3f, %5.3f) r: %5.3f, ball: (%5.3f, %5.3f), alpha: %5.8f", newModelID, models[newModelID].stateEstimates[0][0], models[newModelID].stateEstimates[1][0], RAD_T_DEG*models[newModelID].stateEstimates[2][0], models[newModelID].stateEstimates[3][0], models[newModelID].stateEstimates[4][0],models[newModelID].alpha));

      // Do the update.
      if (wo->isLine()){
        kf_return =  models[newModelID].lineObjectMeas(wo->visionDistance, wo->visionBearing, lmLoc.x, lmLoc.y, lineOrient, distanceOffsetError, distanceRelativeError, bearingError);
      }
      else  if (wo->visionDistance > 50){
        kf_return =  models[newModelID].fieldObjectMeasWithBearing(wo->visionDistance, wo->visionBearing, lmLoc.x, lmLoc.y, distanceOffsetError, distanceRelativeError, bearingError);
      } else {
        kf_return =  models[newModelID].fieldObjectmeas(wo->visionDistance, wo->visionBearing, lmLoc.x, lmLoc.y, distanceOffsetError, distanceRelativeError, bearingError);
      }

      ukfLog((20, "Model %i post-update estimates: (%5.3f, %5.3f) r: %5.3f, ball: (%5.3f, %5.3f), alpha: %5.8f", newModelID, models[newModelID].stateEstimates[0][0], models[newModelID].stateEstimates[1][0], RAD_T_DEG*models[newModelID].stateEstimates[2][0], models[newModelID].stateEstimates[3][0], models[newModelID].stateEstimates[4][0],models[newModelID].alpha));

      // reduce alpha by some amt if we're comparing T's with L's
      if ((possObj->isT() && wo->isUnknownL()) || (possObj->isL() && wo->isUnknownT())){
        ukfLog((30, "L and T mismatch, lower alpha %5.3f", models[newModelID].alpha));
        models[newModelID].alpha *= ukfParams.t_l_mismatch_degrade;
      }

      // If the update reult was an outlier rejection, the model need not be kept as the
      // information is already contained in the designated outlier model created earlier
      if (kf_return == UKF7_OUTLIER) {
        models[newModelID].toBeActivated=false;
      } else {
        numValid++;
      }

      ukfLog((70, "Output %i, alpha: %5.3f", kf_return, models[newModelID].alpha));
    } // per possible lm

    // original model is if no splits worked... meaning outlier
    modelObjectErrors[modelID][objID]+=1.0;
    models[modelID].alpha *= ukfParams.outlier_alpha_degrade;

  } // per orig model

  ukfLog((10, "Observation was updated into %i models", numValid));

  // Split alpha between choices and also activate models
  for (int i=0; i< c_MAX_MODELS; i++) {
    if (models[i].toBeActivated) {
      models[i].alpha *= 1.0/((float)(numOptions+1)); // Divide each models alpha by the numbmer of splits.
      models[i].isActive=true;
    }
    models[i].toBeActivated=false; // Turn off activation flag
  }
  
  return 1;
}


bool UKFModule::lineInFOV(Point2D kfLoc, AngRad orient, int objID, int obsID){

  // if seen line is way longer than real line could be, then it can't be right either
  float seenLength = (worldObjects->objects_[obsID].visionPt1.getDistanceTo(worldObjects->objects_[obsID].visionPt2))/ 10.0;
  float actualLength = worldObjects->objects_[objID].loc.getDistanceTo(worldObjects->objects_[objID].endLoc);

  if (seenLength > (actualLength * 1.5)){
    ukfLog((10, "Line seen is too big to be this line. seen length %5.3f, actual length %5.3f", seenLength, actualLength));
    return false;
  }

  // allow some error in our check
  float FOVfactor = 1.25;

  float maxFOV = jointAngles->values_[HeadPan] + (FOVx / 2.0) * FOVfactor;
  float minFOV = jointAngles->values_[HeadPan] - (FOVx / 2.0) * FOVfactor;

  // get bearing of line segment end points
  float b1 = kfLoc.getBearingTo(worldObjects->objects_[objID].loc, orient);

  ukfLog((60, "maxfov %5.3f, minfov %5.3f, b1 %5.3f", RAD_T_DEG*maxFOV, RAD_T_DEG*minFOV, RAD_T_DEG*b1));

  // is this point in fov
  if (b1 < minFOV || b1 > maxFOV){

    // b1 not in FOV, check b2
    float b2 = kfLoc.getBearingTo(worldObjects->objects_[objID].endLoc, orient);

    if (b2 < minFOV || b2 > maxFOV){

      // line segment not in view, reject
      ukfLog((61, "  Line %i thrown out because seg not in FOV(%5.5f to %5.5f), Seg bearing: %5.5f, %5.5f", objID, RAD_T_DEG*minFOV, RAD_T_DEG*maxFOV, RAD_T_DEG*b1, RAD_T_DEG*b2));
      return false;

    } // b2 bad

    else {
      ukfLog((60, "b2 %5.3f", RAD_T_DEG*b2));
    }

  } // b1 bad
  return true;
}

float UKFModule::getObsSim(Point2D myLoc, float orient, WorldObject* wo, Point2D lmLoc){
  AngRad expectedBearing = myLoc.getBearingTo(lmLoc, orient);
  float  expectedDist = myLoc.getDistanceTo(lmLoc);

  // calculate similarity of each to find most likely match
  float d = fabs(expectedBearing - wo->visionBearing) / M_PI;
  float bearingSim = 0;
  if ( d < 1 )
    bearingSim = exp( -10.0 * d * d );
  else
    bearingSim = exp( -10.0 * ( 2 - d ) * ( 2 - d ) );

  float d2 = fabs(expectedDist - wo->visionDistance) / 700.0;
  float distSim = exp( -50.0 * d2 *d2);

  float totalSim = bearingSim * distSim;

  ukfLog((10, "ExpDist: %5.3f, ObsDist: %5.3f, ExpBear %5.3f, ObsBear: %5.3f, bearSim %5.3f, distSim %5.3f, totalSim %5.3f", expectedDist, wo->visionDistance, expectedBearing *RAD_T_DEG, wo->visionBearing *RAD_T_DEG, bearingSim, distSim, totalSim));

  return totalSim;
}

float UKFModule::getLineSim(Point2D myLoc, float orient, WorldObject* wo, Point2D lmLoc, float lineOrient){
  AngRad expectedBearing = myLoc.getBearingTo(lmLoc, orient);
  float  expectedDist = myLoc.getDistanceTo(lmLoc);

  float seenDist = wo->visionDistance;
  float seenBear = wo->visionBearing;

  // kill warning
  lineOrient = lineOrient;

  // if short distance and bearings are way off, try other side of the line
  float bearError = fabs(seenBear - expectedBearing);
  if (bearError > (DEG_T_RAD*110) && (expectedDist < 35 || seenDist < 35)) {
    expectedDist = -expectedDist;
    expectedBearing = normalizeAngleNinetyAroundVal(expectedBearing + M_PI, seenBear);
    ukfLog((10, "small dist seen %5.3f, exp %5.3f, huge bearing error %5.3f, assuming other side of line. new expected bearing %5.3f", seenDist, expectedDist, bearError*RAD_T_DEG, RAD_T_DEG*expectedBearing));
  }

  // calculate similarity of each to find most likely match
  float d = fabs(expectedBearing - seenBear) / M_PI;
  float bearingSim = 0;
  if ( d < 1 )
    bearingSim = exp( -10.0 * d * d );
  else
    bearingSim = exp( -10.0 * ( 2 - d ) * ( 2 - d ) );

  float d2 = fabs(expectedDist - seenDist) / 700.0;
  float distSim = exp( -10.0 * d2 *d2);

  float totalSim = bearingSim * distSim;

  ukfLog((10, "ExpDist: %5.3f, ObsDist: %5.3f, ExpBear %5.3f, ObsBear: %5.3f, bearSim %5.3f, distSim %5.3f, totalSim %5.3f", expectedDist, seenDist, expectedBearing *RAD_T_DEG, seenBear *RAD_T_DEG, bearingSim, distSim, totalSim));

  return totalSim;
}


bool UKFModule::MergeTwoModels(int index1, int index2)
{
  // Merges second model into first model, then disables second model.
  bool success = true;
  if(index1 == index2) success = false; // Don't merge the same model.
  //    if((model[index1].active == false) || (model[index2].active == false)) success = false; // Both models must be active.
  if(success == false){

    ukfLog((10, "Merge models %i and %i FAILED", index1, index2));

    return success;
  }

  ukfLog((75, "Merge two models. Model %i has robot at (%5.3f, %5.3f) r: %5.3f alpha: %5.8f, model %i has robot at (%5.3f, %5.3f) r: %5.3f alpha: %5.8f", index1, models[index1].stateEstimates[0][0], models[index1].stateEstimates[1][0], models[index1].stateEstimates[2][0], models[index1].alpha ,index2,  models[index2].stateEstimates[0][0], models[index2].stateEstimates[1][0], models[index2].stateEstimates[2][0], models[index2].alpha));

  // Merge alphas
  float alphaMerged = models[index1].alpha + models[index2].alpha;
  float alpha1 = models[index1].alpha / alphaMerged;
  float alpha2 = models[index2].alpha / alphaMerged;

  NMatrix xMerged; // Merge State matrix

  // If one model is much more correct than the other, use the correct states.
  // This prevents drifting from continuouse splitting and merging even when one model is much more likely.
  if(models[index1].alpha > 10*models[index2].alpha){
    xMerged = models[index1].stateEstimates;
  }
  else if (models[index2].alpha > 10*models[index1].alpha){
    xMerged = models[index2].stateEstimates;
  }
  else {
    xMerged = (alpha1 * models[index1].stateEstimates + alpha2 * models[index2].stateEstimates);
    // Fix angle.
    float angleDiff = models[index2].stateEstimates[2][0] - models[index1].stateEstimates[2][0];
    angleDiff = NORMALISE(angleDiff);
    xMerged[2][0] = NORMALISE(models[index1].stateEstimates[2][0] + alpha2*angleDiff);
  }

  // Merge Covariance matrix (S = sqrt(P))
  NMatrix xDiff = models[index1].stateEstimates - xMerged;
  NMatrix p1 = (models[index1].stateStandardDeviations * models[index1].stateStandardDeviations.transp() + xDiff * xDiff.transp());

  xDiff = models[index2].stateEstimates - xMerged;
  NMatrix p2 = (models[index2].stateStandardDeviations * models[index2].stateStandardDeviations.transp() + xDiff * xDiff.transp());

  NMatrix sMerged = cholesky(alpha1 * p1 + alpha2 * p2); // P merged = alpha1 * p1 + alpha2 * p2.

  // weighted average of outlier values
  for (int i=0; i<NUM_WORLD_OBJS; i++)
    modelObjectErrors[index1][i] = alpha1 * modelObjectErrors[index1][i] + alpha2 * modelObjectErrors[index2][i]; // Reset outlier values.

  // Copy merged value to first model
  models[index1].alpha = alphaMerged;
  models[index1].stateEstimates = xMerged;
  models[index1].stateStandardDeviations = sMerged;

  // Disable second model
  models[index2].isActive = false;
  models[index2].toBeActivated = false;

  ukfLog((75, "New model has robot at (%5.3f, %5.3f) r: %5.3f alpha: %5.8f", models[index1].stateEstimates[0][0], models[index1].stateEstimates[1][0], models[index1].stateEstimates[2][0], models[index1].alpha));

  for (int i=0; i<NUM_WORLD_OBJS; i++)
    modelObjectErrors[index2][i] = 0; // Reset outlier values.

  return true;
}



int UKFModule::getNumActiveModels()
{
  int numActive = 0;
  for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == true) numActive++;
  }
  return numActive;
}



int UKFModule::getNumFreeModels()
{
  int numFree = 0;
  for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if((models[modelID].isActive == false) && (models[modelID].toBeActivated == false)) numFree++;
  }
  return numFree;
}



UKF* UKFModule::getBestModel()
{
  return &(models[getBestModelID()]);
}



int UKFModule::getBestModelID()
{
  // Return model with highest alpha value.
  int bestID = 0;
  if (!models[bestID].isActive) {
    ukfLog((0, "ERROR, no active ukf model at index 0"));
    models[bestID].alpha = 0;
  }
  for (int currID = 0; currID < c_MAX_MODELS; currID++){
    if(models[currID].isActive == false) continue; // Skip inactive models.
    if(models[currID].alpha > models[bestID].alpha) bestID = currID;
  }
  return bestID;
}


bool UKFModule::CheckModelForOutlierReset(int modelID)
{
  // RHM 7/7/08: Suggested incorporation of 'Resetting' for possibly 'kidnapped' robot
  //----------------------------------------------------------
  float sum = 0.0;
  int numObjects = 0;
  bool reset = false;
  for(int objID = 0; objID < NUM_WORLD_OBJS; objID++){
    if (worldObjects->objects_[objID].isGoalCenter()) continue;
    if (modelObjectErrors[modelID][objID] > 0)
      ukfLog((20, "check model %i outliers on obj %i: %5.3f", modelID, objID, modelObjectErrors[modelID][objID]));
    sum += modelObjectErrors[modelID][objID];
    if (modelObjectErrors[modelID][objID] > ukfParams.object_error_thresh) numObjects+=1;
    modelObjectErrors[modelID][objID] *= ukfParams.object_error_decay;
  }

  ukfLog((20, "Model %i reset sum %5.3f, num %i, sum_thresh %5.3f, num_thresh %i", modelID, sum, numObjects, ukfParams.reset_sum_thresh, ukfParams.reset_num_thresh));

  // Check if enough recent 'outliers' that we should reset ?
  if ((sum > ukfParams.reset_sum_thresh) && (numObjects >= ukfParams.reset_num_thresh)) {
    reset = true;
    //debug_out  <<">>>>>>>>>>>>>>>>>>>>RESET"<<std::endl;
    models[modelID].Reset(); //Reset KF varainces. Leave Xhat!

    ukfLog((20, "Reset model %i due to outliers. Sum %5.2f, num %i", modelID, sum, numObjects));

    for (int i=0; i<NUM_WORLD_OBJS; i++)
      modelObjectErrors[modelID][i] = 0.0; // Reset the outlier history
  }
  return reset;
}



int UKFModule::CheckForOutlierResets()
{
  bool numResets = 0;
  for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if (!models[modelID].isActive) continue;
    if(CheckModelForOutlierReset(modelID)) numResets++;
  }
  // if we did any resets, try reseed
  if (numResets > 0 && ukfParams.RESEED_ON_OUTLIERS){
    seedNewModel();
  }

  return numResets;
}


void UKFModule::doKickUpdates()
{
  if (!odometry->didKick)
    return;

  // no kicks while fallen
  if (odometry->getting_up_side_ != Getup::NONE || odometry->fall_direction_ != Fall::NONE)
    return;

  float vel = odometry->kickVelocity;
  float heading = odometry->kickHeading;
  ukfLog((30, "UKF update with kick vel %5.3f, heading %5.3f", vel, heading *RAD_T_DEG));
  vel /= 10.0;

  // create one model for kick fail, and one for kick success
  int numFreeModels = getNumFreeModels();
  int numActiveModels = getNumActiveModels();
  int numRequiredModels = numActiveModels * 2;

  if(numFreeModels < numRequiredModels){
    int maxActiveAfterMerge = c_MAX_MODELS / 2;
    ukfLog((30, "Only %i free models, need %i models. merging to %i models",
            numFreeModels, numRequiredModels, maxActiveAfterMerge));
    MergeModels(maxActiveAfterMerge);
    if(getNumFreeModels() < (getNumActiveModels() * 2)){
      ukfLog((10, "Only %i free models. not enough. aborting update", getNumFreeModels()));
      return;
    }
  }

  for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == false) continue; // Skip inactive models.

    // Copy initial model to the temporary model.
    tempModel = models[modelID];
    tempModel.isActive = false;
    tempModel.toBeActivated = true;

    // now for kick success
    int newModelID = FindNextFreeModel();

    // If an invalid modelID has been returned, something has gone horribly wrong, so stop here.
    if(newModelID < 0){
      ukfLog((10, "kick update, bad model id returned (%i), aborted.", newModelID));

      for(int m = 0; m < c_MAX_MODELS; m++) models[m].toBeActivated = false;
      return;
    }

    ukfLog((10, "Kick update. split model %i (kick fail) into model %i (kick success)", modelID, newModelID));
    models[newModelID] = tempModel; // Get the new model from the temp

    // Copy outlier history from the current model.
    for (int i=0; i<NUM_WORLD_OBJS; i++){
      modelObjectErrors[newModelID][i] = modelObjectErrors[modelID][i];
    }

    // update new modelID with kick success
    // give ball vel a high sd first
    models[newModelID].stateStandardDeviations[5][5] = ukfParams.kick_ball_vel_reset_noise;
    models[newModelID].stateStandardDeviations[6][6] = ukfParams.kick_ball_vel_reset_noise;


    float xvel = vel * cos(models[newModelID].stateEstimates[2][0]+heading);
    float yvel = vel * sin(models[newModelID].stateEstimates[2][0]+heading);
    ukfLog((10, "kick at vel %5.3f, bearing %5.3f, global (%5.3f, %5.3f)", vel, heading, xvel, yvel));
    models[newModelID].stateEstimates[5][0] = xvel;
    models[newModelID].stateEstimates[6][0] = yvel;

    //int result = models[newModelID].ballKickUpdate(vel, heading, ukfParams.R_kick_vel, ukfParams.R_kick_bearing);
    //ukfLog((10, "Kick update on model %i, result is %i", newModelID, result));
    //ukfLog((10, "Model %i, after kick, vel (%5.3f, %5.3f) orient (%5.3f)", newModelID, models[newModelID].stateEstimates[5][0],  models[newModelID].stateEstimates[6][0], models[newModelID].stateEstimates[2][0] * RAD_T_DEG));

    // update old model ID with kick fail (higher ball vel sd)
    // give ball vel a high sd first
    models[modelID].stateStandardDeviations[5][5] = ukfParams.kick_ball_vel_reset_noise;
    models[modelID].stateStandardDeviations[6][6] = ukfParams.kick_ball_vel_reset_noise;

  } // per orig model

  // Split alpha between choices and also activate models
  for (int i=0; i< c_MAX_MODELS; i++) {
    if (models[i].toBeActivated) {
      models[i].alpha *= 0.5; // Divide each models alpha by the numbmer of splits.
      models[i].isActive=true;
    }
    models[i].toBeActivated=false; // Turn off activation flag
  }
}


void UKFModule::spawnKickedBallModel(){

  // no spawning balls while fallen
  if (odometry->getting_up_side_ != Getup::NONE || odometry->fall_direction_ != Fall::NONE)
    return;

  ukfLog((20, "Keeper spawn kicked ball model"));
  for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == false) continue;
    ukfLog((20, "model %i has xvel %5.3f", modelID, models[modelID].stateEstimates[5][0]));
    // if there's already a model of ball coming at us, return
    if (models[modelID].stateEstimates[5][0] < -(ukfParams.keeper_ball_spawn_velocity / 2.0))
      return;
  }

  // create one model for status quo, and one for ball kicked at us
  int numFreeModels = getNumFreeModels();
  int numActiveModels = getNumActiveModels();
  int numRequiredModels = numActiveModels * 2;

  if(numFreeModels < numRequiredModels){
    int maxActiveAfterMerge = c_MAX_MODELS / 2;
    ukfLog((30, "Only %i free models, need %i models. merging to %i models",
            numFreeModels, numRequiredModels, maxActiveAfterMerge));
    MergeModels(maxActiveAfterMerge);
    if(getNumFreeModels() < (getNumActiveModels() * 2)){
      ukfLog((10, "Only %i free models. not enough. aborting update", getNumFreeModels()));
      return;
    }
  }

  int modelID = getBestModelID();

  // Copy initial model to the temporary model.
  tempModel = models[modelID];
  tempModel.isActive = false;
  tempModel.toBeActivated = true;

  // now for kicked ball
  int newModelID = FindNextFreeModel();
  
  // If an invalid modelID has been returned, something has gone horribly wrong, so stop here.
  if(newModelID < 0){
    ukfLog((10, "kick update, bad model id returned (%i), aborted.", newModelID));
    
    for(int m = 0; m < c_MAX_MODELS; m++) models[m].toBeActivated = false;
    return;
  }
  
  ukfLog((10, "Keeper kick update. split model %i (normal) into model %i (kicked at us)", modelID, newModelID));
  models[newModelID] = tempModel; // Get the new model from the temp
  
  // Copy outlier history from the current model.
  for (int i=0; i<NUM_WORLD_OBJS; i++){
    modelObjectErrors[newModelID][i] = modelObjectErrors[modelID][i];
  }
  
  // update new modelID with kick success
  // give ball vel a high sd first
  models[newModelID].stateStandardDeviations[5][5] = ukfParams.kick_ball_vel_reset_noise;
  models[newModelID].stateStandardDeviations[6][6] = ukfParams.kick_ball_vel_reset_noise;
  

  float vel = ukfParams.keeper_ball_spawn_velocity;
  float heading = worldObjects->objects_[WO_BALL].bearing + M_PI;

  float xvel = vel * cos(models[newModelID].stateEstimates[2][0]+heading);
  float yvel = vel * sin(models[newModelID].stateEstimates[2][0]+heading);
  ukfLog((10, "kick at vel %5.3f, bearing %5.3f, global (%5.3f, %5.3f)", vel, heading, xvel, yvel));
  models[newModelID].stateEstimates[5][0] = xvel;
  models[newModelID].stateEstimates[6][0] = yvel;
  
  // Split alpha between choices and also activate models
  for (int i=0; i< c_MAX_MODELS; i++) {
    if (models[i].toBeActivated) {
      models[i].alpha *= ukfParams.keeper_ball_spawn_likelihood; // Small chance of this happening
      models[i].isActive=true;
    }
    models[i].toBeActivated=false; // Turn off activation flag
  }
}



void UKFModule::spawnManualPositionModel(){
  int numFreeModels = getNumFreeModels();
  int numNeeded = 1;
  if (robotState->WO_SELF == 2 || robotState->WO_SELF == 3 || robotState->WO_SELF == 4)
    numNeeded = 3;

  if (numFreeModels < numNeeded){
    int maxActiveAfterMerge = c_MAX_MODELS - numNeeded;
    ukfLog((30, "Only %i free models, need %i models. merging to %i models",
            numFreeModels, numNeeded, maxActiveAfterMerge));
    MergeModels(maxActiveAfterMerge);
    if(getNumFreeModels() < numNeeded){
      ukfLog((10, "Only %i free models. not enough. aborting update", getNumFreeModels()));
      return;
    }
  }

  // spawn model for actual player num
  if (robotState->WO_SELF == 1)
    spawnManualPositionModel(robotState->WO_SELF);

  // extra model for players 2 and 3
  // since refs may swap them in manual positioning
  if (robotState->WO_SELF > 1){
    spawnManualPositionModel(4);
    spawnManualPositionModel(3);
    spawnManualPositionModel(2);
  }

}

void UKFModule::spawnManualPositionModel(int playerNum){

  int newModelID = FindNextFreeModel();

  // new code... each player has a specific spot
  models[newModelID].isActive = true;
  if (playerNum == 2 || playerNum == 3)
    models[newModelID].alpha = 0.25;
  else
    models[newModelID].alpha = 0.4;

  if (gameState->ourKickOff){
    if (playerNum == 1){
      models[newModelID].stateEstimates[0][0] = -FIELD_X/20.0;          // Robot X
      models[newModelID].stateEstimates[1][0] = 0;
    }
    else if (playerNum == 2){
      models[newModelID].stateEstimates[0][0] = -PENALTY_CROSS_X/10.0;  // Robot X
      models[newModelID].stateEstimates[1][0] = PENALTY_Y/20.0;  // Robot y
    }
    else if (playerNum == 3){
      models[newModelID].stateEstimates[0][0] = -FIELD_X/20.0 + PENALTY_X/10.0;
      models[newModelID].stateEstimates[1][0] = -PENALTY_Y/20.0;  // Robot y
    }
    else {
      models[newModelID].stateEstimates[0][0] = -CIRCLE_RADIUS/10.0;
      models[newModelID].stateEstimates[1][0] = 0;
    }

  }
  // their kickoff
  else {
    if (playerNum == 1){
      models[newModelID].stateEstimates[0][0] = -FIELD_X/20.0;          // Robot X
      models[newModelID].stateEstimates[1][0] = 0;
    }
    else if (playerNum == 2){
      models[newModelID].stateEstimates[0][0] = -FIELD_X/20.0 + PENALTY_X/10.0;
      models[newModelID].stateEstimates[1][0] = (PENALTY_Y + FIELD_Y)/40.0;  // Robot y
    }
    else if (playerNum == 3){
      models[newModelID].stateEstimates[0][0] = -FIELD_X/20.0 + PENALTY_X/10.0;
      models[newModelID].stateEstimates[1][0] = -(PENALTY_Y+FIELD_Y)/40.0;  // Robot y
    }
    else {
      models[newModelID].stateEstimates[0][0] = -FIELD_X/20.0 + PENALTY_X/10.0;
      models[newModelID].stateEstimates[1][0] = GOAL_Y/40.0;
    }
  }

  // assume stationary ball and facing forward
  models[newModelID].stateEstimates[2][0] = 0.0;       // Robot heading
  models[newModelID].stateEstimates[3][0] = 0.0;       // Ball x
  models[newModelID].stateEstimates[4][0] = 0.0;       // Ball y
  models[newModelID].stateEstimates[5][0] = 0.0;       // Ball vx
  models[newModelID].stateEstimates[6][0] = 0.0;       // Ball vy

  // Set the uncertainties
  resetSdMatrix(newModelID);

  ukfLog((30, "Manual positioning model %i spawned at %3.0f, %3.0f, var %3.3f, angle sd %3.3f", newModelID, models[newModelID].stateEstimates[0][0], models[newModelID].stateEstimates[1][0], models[newModelID].variance(2), models[newModelID].stateStandardDeviations[2][2]));

}


void UKFModule::splitModelsOnFall(){

  // figure out new orientation
  int orientChange = 0;
  // face right
  if ((odometry->fall_direction_ == Fall::RIGHT && odometry->getting_up_side_ == Getup::FRONT) || (odometry->fall_direction_ == Fall::LEFT && odometry->getting_up_side_ == Getup::BACK))
    orientChange = -1;
  else if ((odometry->fall_direction_ == Fall::LEFT && odometry->getting_up_side_ == Getup::FRONT) || (odometry->fall_direction_ == Fall::RIGHT && odometry->getting_up_side_ == Getup::BACK))
    orientChange = 1;

  ukfLog((10, "Fall dir %i, get up side %i, expected orient change %f", odometry->fall_direction_, odometry->getting_up_side_, orientChange*M_PI/2.0));

  // create one model for getting up same orient, left 90 and right 90
  int numFreeModels = getNumFreeModels();
  int numActiveModels = getNumActiveModels();
  int numRequiredModels = numActiveModels * 3;

  if(numFreeModels < numRequiredModels){
    int maxActiveAfterMerge = c_MAX_MODELS / 3;
    ukfLog((30, "Only %i free models, need %i models. merging to %i models",
            numFreeModels, numRequiredModels, maxActiveAfterMerge));
    MergeModels(maxActiveAfterMerge);
    if(getNumFreeModels() < (getNumActiveModels() * 3)){
      ukfLog((10, "Only %i free models. not enough. aborting update", getNumFreeModels()));
      return;
    }
  }

  for (int modelID = 0; modelID < c_MAX_MODELS; modelID++){
    if(models[modelID].isActive == false) continue; // Skip inactive models.

    for (int fallSide = 0; fallSide < 2; fallSide++){

      // Copy initial model to the temporary model.
      tempModel = models[modelID];
      tempModel.isActive = false;
      tempModel.toBeActivated = true;

      // now for kick success
      int newModelID = FindNextFreeModel();

      // If an invalid modelID has been returned, something has gone horribly wrong, so stop here.
      if(newModelID < 0){
        ukfLog((10, "fall update, bad model id returned (%i), aborted.", newModelID));

        for(int m = 0; m < c_MAX_MODELS; m++) models[m].toBeActivated = false;
        return;
      }

      ukfLog((10, "Fall update. split model %i (fall fwd) into model %i (fall %i)", modelID, newModelID, fallSide));
      models[newModelID] = tempModel; // Get the new model from the temp

      // Copy outlier history from the current model.
      for (int i=0; i<NUM_WORLD_OBJS; i++){
        modelObjectErrors[newModelID][i] = modelObjectErrors[modelID][i];
      }

      // update new modelID with turn one way or another
      if (fallSide == 0){
        models[newModelID].stateEstimates[2][0] = normalizeAngle(models[newModelID].stateEstimates[2][0] + M_PI/2.0);
        if (orientChange == 1)
          models[newModelID].alpha *= 0.9;
        else
          models[newModelID].alpha *= 0.05;
      }
      else {
        models[newModelID].stateEstimates[2][0] = normalizeAngle(models[newModelID].stateEstimates[2][0] - M_PI/2.0);
        if (orientChange == -1)
          models[newModelID].alpha *= 0.9;
        else
          models[newModelID].alpha *= 0.05;
      }

    }// per new model

    // if the original orient is expected, give it 0.9 alpha
    if (orientChange == 0)
      models[modelID].alpha *= 0.9;
    else
      models[modelID].alpha *= 0.05;

  } // per orig model

  // activate models
  for (int i=0; i< c_MAX_MODELS; i++) {
    if (models[i].toBeActivated) {
      models[i].isActive=true;
    }
    models[i].toBeActivated=false; // Turn off activation flag
  }
}


int UKFModule::varianceCheckAll()
{
  // only if we're using goals
  if (!ukfParams.USE_GOAL) return 0;

  int numModelsChanged = 0;
  bool changed;
  for (int currID = 0; currID < c_MAX_MODELS; currID++){
    if(models[currID].isActive == false) continue; // Skip inactive models.
    changed = varianceCheck(currID);
    if(changed) numModelsChanged++;
  }
  return numModelsChanged;
}

bool UKFModule::varianceCheck(int modelID){
  UKF* kf = &(models[modelID]);
  float var=kf->variance(2);  //angle variance

  // check if variance is ok
  if(var<(ukfParams.large_angle_sd * ukfParams.large_angle_sd))
    return false;

  WorldObject* wo;
  bool changed = false;

  // try unknown goal if no known ones
  for (int i=WO_UNKNOWN_LEFT_GOALPOST; i<=WO_UNKNOWN_GOALPOST; i++){
    wo=&(worldObjects->objects_[i]);
    if(wo->visionDistance>150 && wo->seen){
      Point2D myLoc(kf->stateEstimates[0][0], kf->stateEstimates[1][0]);
      AngRad expOrientOwnLeft, expOrientOwnRight;
      AngRad expOrientOppLeft, expOrientOppRight;
      expOrientOwnLeft = myLoc.getAngleTo(worldObjects->objects_[WO_OWN_LEFT_GOALPOST].loc) - wo->visionBearing;
      expOrientOwnRight = myLoc.getAngleTo(worldObjects->objects_[WO_OWN_RIGHT_GOALPOST].loc) - wo->visionBearing;
      expOrientOppLeft = myLoc.getAngleTo(worldObjects->objects_[WO_OPP_LEFT_GOALPOST].loc) - wo->visionBearing;
      expOrientOppRight = myLoc.getAngleTo(worldObjects->objects_[WO_OPP_RIGHT_GOALPOST].loc) - wo->visionBearing;

      // use one that's closer to our current orient
      float ownLeftDiff = fabs(normalizeAngle(expOrientOwnLeft - kf->stateEstimates[2][0]));
      float ownRightDiff = fabs(normalizeAngle(expOrientOwnRight - kf->stateEstimates[2][0]));
      float oppLeftDiff = fabs(normalizeAngle(expOrientOppLeft - kf->stateEstimates[2][0]));
      float oppRightDiff = fabs(normalizeAngle(expOrientOppRight - kf->stateEstimates[2][0]));

      if (ownLeftDiff <= ownRightDiff && ownLeftDiff <= oppLeftDiff && ownLeftDiff <= oppRightDiff){
        kf->stateEstimates[2][0]= expOrientOwnLeft;
      } else if (ownRightDiff <= ownLeftDiff && ownRightDiff <= oppLeftDiff && ownRightDiff <= oppRightDiff){
        kf->stateEstimates[2][0]= expOrientOwnRight;
      } else if (oppLeftDiff <= ownLeftDiff && oppLeftDiff <= ownRightDiff && oppLeftDiff <= oppRightDiff){
        kf->stateEstimates[2][0]= expOrientOppLeft;
      } else {
        kf->stateEstimates[2][0]= expOrientOppRight;
      }
      ukfLog((10, "Model %i Variance check %5.3f, reset orient to %5.1f based on obs %i matched to a real post", modelID, var, kf->stateEstimates[2][0]*RAD_T_DEG, i));
      changed = true;
      break;
    }
  }
  return changed;
}


void UKFModule::CheckForAmbigModels()
{

  // make sure models created from ambiguous situations (penalty & getup)
  // stick around with a minimum alpha for a certain number of frames
  bool modelsChanged = false;
  for (int i = 0; i < c_MAX_MODELS; i++) {
    if (models[i].isActive && (models[i].framesSinceAmbig > 0 || models[i].ballFramesSinceAmbig > 0)){
      ukfLog((40, "Model %i has alpha %f, and frames since ambig %i, ball frames %i, maintain min alpha %f", i, models[i].alpha, models[i].framesSinceAmbig, models[i].ballFramesSinceAmbig, ukfParams.ambig_min_model_alpha));
      if (models[i].alpha < ukfParams.ambig_min_model_alpha){
        models[i].alpha = ukfParams.ambig_min_model_alpha;
        modelsChanged = true;
      }
    }
  }

  if (modelsChanged) NormaliseAlphas();
}


void UKFModule::NormaliseAlphas()
{
  // Normalise all of the models alpha values such that all active models sum to 1.0
  float sumAlpha=0.0;
  for (int i = 0; i < c_MAX_MODELS; i++) {
    if (models[i].isActive) {
      sumAlpha+=models[i].alpha;
    }
  }
  ukfLog((80, "alpha sum is %5.3f", sumAlpha));
  if(sumAlpha == 1) return;
  if (sumAlpha == 0) sumAlpha = 1e-12;
  for (int i = 0; i < c_MAX_MODELS; i++) {
    if (models[i].isActive) {
      models[i].alpha=models[i].alpha/sumAlpha;
    }
  }
}



int UKFModule::FindNextFreeModel()
{
  for (int i=0; i<c_MAX_MODELS; i++) {
    if ((models[i].isActive == true) || (models[i].toBeActivated == true)) continue;
    else return i;
  }
  return -1; // NO FREE MODELS - This is very, very bad.
}


// Reset all of the models
void UKFModule::ResetAll()
{

  ukfLog((20, "Reset All Models"));

  for(int modelNum = 0; modelNum < c_MAX_MODELS; modelNum++){
    models[modelNum].init();
    for (int i=0; i<NUM_WORLD_OBJS; i++) modelObjectErrors[modelNum][i] = 0.0; // Reset outlier values.
  }
}



//**************************************************************************
//  This method begins the process of merging close models together

void UKFModule::MergeModels(int maxAfterMerge) {
  MergeModelsBelowThreshold(0.001);
  MergeModelsBelowThreshold(0.01);

  //  float threshold=0.04;
  float threshold=0.05;

  while (getNumActiveModels()>maxAfterMerge) {
    float min = MergeModelsBelowThreshold(threshold);
    //      threshold*=5.0;
    threshold += 0.05;
    if (min > threshold) threshold = min + 0.01;
    if (min > 99999) return;
  }
  return;
}



void UKFModule::PrintModelStatus(int modelID)
{
  ukfLog((20, "Model %i, Alpha: %5.3f, Active: %i, Activate: %i", modelID,models[modelID].alpha, models[modelID].isActive, models[modelID].toBeActivated));
  return;
}



float UKFModule::MergeModelsBelowThreshold(float MergeMetricThreshold)
{
  float mergeM;
  float minMetric = 100000;;
  for (int i = 0; i < c_MAX_MODELS; i++) {
    for (int j = i; j < c_MAX_MODELS; j++) {
      if(i == j) continue;
      if (!models[i].isActive || !models[j].isActive ) continue;
      mergeM = fabs( MergeMetric(i,j) );
      if (mergeM < minMetric) minMetric = mergeM;
      if (mergeM < MergeMetricThreshold) { //0.5
        ukfLog((20, "Merging models %i, alpha: %5.3f, into model %i, alpha: %5.3f, merge metric = %5.3f", j, models[j].alpha, i, models[i].alpha, mergeM));
        MergeTwoModels(i,j);
      }
    }
  }
  return minMetric;
}



//************************************************************************
// model to compute a metric for how 'far' apart two models are in terms of merging.
float UKFModule::MergeMetric(int index1, int index2)
{
  if (index1==index2) return 10000.0;
  if (!models[index1].isActive || !models[index2].isActive ) return 10000.0; //at least one model inactive
  NMatrix xdif = models[index1].stateEstimates - models[index2].stateEstimates;
  NMatrix p1 = models[index1].stateStandardDeviations * models[index1].stateStandardDeviations.transp();
  NMatrix p2 = models[index2].stateStandardDeviations * models[index2].stateStandardDeviations.transp();

  xdif[2][0] = NORMALISE(xdif[2][0]);

  float dij=0;
  for (int i=0; i<p1.getm(); i++) {
    dij+=(xdif[i][0]*xdif[i][0]) / (p1[i][i]+p2[i][i]);
  }
  return dij*( (models[index1].alpha*models[index2].alpha) / (models[index1].alpha+models[index2].alpha) );
}


void UKFModule::mmTocm() {
  WorldObject* objs = &(worldObjects->objects_[0]);
  for (int i=0; i<NUM_WORLD_OBJS; i++) {
    // skip teammates since this might interfere with listenTeamUDP
    if (i >= WO_TEAM1 && i <= WO_TEAM4 && i != robotState->WO_SELF) continue;
    objs[i].distance/=10.0;
    objs[i].visionDistance/=10.0;
    objs[i].loc.x/=10.0;
    objs[i].loc.y/=10.0;
    objs[i].endLoc.x/=10.0;
    objs[i].endLoc.y/=10.0;
    objs[i].sd.x/=10.0;
    objs[i].sd.y/=10.0;
    objs[i].absVel.x/=10.0;
    objs[i].absVel.y/=10.0;
    objs[i].relVel.x/=10.0;
    objs[i].relVel.y/=10.0;
  }
}

void UKFModule::cmTomm() {
  WorldObject* objs = &(worldObjects->objects_[0]);
  for (int i=0; i<NUM_WORLD_OBJS; i++) {
    // skip teammates since this might interfere with listenTeamUDP
    if (i >= WO_TEAM1 && i <= WO_TEAM4 && i != robotState->WO_SELF) continue;
    objs[i].distance*=10.0;
    objs[i].visionDistance*=10.0;
    objs[i].loc.x*=10.0;
    objs[i].loc.y*=10.0;
    objs[i].endLoc.x*=10.0;
    objs[i].endLoc.y*=10.0;
    objs[i].sd.x*=10.0;
    objs[i].sd.y*=10.0;
    objs[i].absVel.x*=10.0;
    objs[i].absVel.y*=10.0;
    objs[i].relVel.x*=10.0;
    objs[i].relVel.y*=10.0;
  }
}


void UKFModule::populateWOSelf(UKF* kf) {
  int WO_SELF=robotState->WO_SELF;
  WorldObject* self = &(worldObjects->objects_[WO_SELF]);
  //cout << "UKF="<<kf->stateEstimates[1][0] << "," << kf->stateEstimates[0][0] << endl << flush;
  self->loc.x=kf->stateEstimates[0][0];
  self->loc.y=kf->stateEstimates[1][0];
  self->orientation=kf->stateEstimates[2][0];
  self->sd.x=kf->sd(0);
  self->sd.y=kf->sd(1);
  self->sdOrientation=kf->sd(2);

  ukfLog((32, "Robot loc: (%5.1f, %5.1f), sd: (%5.1f, %5.1f)", self->loc.x, self->loc.y, self->sd.x, self->sd.y));

}


void UKFModule::populateBall(UKF* kf) {
  WorldObject* ball = &(worldObjects->objects_[WO_BALL]);

  //cout << "UKF="<<kf->StateEstimates[1][0] << "," << kf->StateEstimates[0][0] << endl << flush;
  ball->loc.x=kf->stateEstimates[3][0];
  ball->loc.y=kf->stateEstimates[4][0];
  ball->absVel.x=kf->stateEstimates[5][0];
  ball->absVel.y=kf->stateEstimates[6][0];

  ball->sd.x=kf->sd(3);
  ball->sd.y=kf->sd(4);

  ukfLog((32, "Populate ball, (%5.3f, %5.3f) vel (%5.3f, %5.3f) sd(%5.3f, %5.3f)", ball->loc.x, ball->loc.y, ball->absVel.x, ball->absVel.y, ball->sd.x, ball->sd.y));

  // calculate rel vel and rel location?
  int WO_SELF=robotState->WO_SELF;
  WorldObject* self = &(worldObjects->objects_[WO_SELF]);

  // rel pos
  ball->relPos = ball->loc - self->loc;
  ball->relPos.rotate(-self->orientation);
  ball->relPos *= 10.0; // to mm here

  // have to make a copy first because rotate rotates the actual point
  ball->relVel = ball->absVel;
  ball->relVel.rotate(-self->orientation);
}

void UKFModule::updateDistAndBearToObjects(){
  int WO_SELF=robotState->WO_SELF;
  WorldObject* self = &(worldObjects->objects_[WO_SELF]);

  for (int i = 0; i < NUM_WORLD_OBJS; i++){

    // calculate distance and bearing to each object
    worldObjects->objects_[i].distance =
      self->loc.getDistanceTo(worldObjects->objects_[i].loc);
    worldObjects->objects_[i].bearing =
      self->loc.getBearingTo(worldObjects->objects_[i].loc,
                             self->orientation);

  }

}


void UKFModule::initFromMemory() {
  // Reset all of the models
  for(int m = 0; m < c_MAX_MODELS; m++){
    // reset outlier error count
    for (int i=0; i<NUM_WORLD_OBJS; i++)
      modelObjectErrors[m][i] = 0;

    // Disable models
    models[m].isActive = false;
    models[m].toBeActivated = false;
    models[m].framesSinceAmbig = 0;
    models[m].ballFramesSinceAmbig = 0;

  }

  int tempCounter = 0;
  for (int i = 0; i< MAX_MODELS_IN_MEM; i++){
    if (localizationMem->alpha[i] <  0.001 || localizationMem->alpha[i] > 1.0) continue;
    ukfLog((10, "load model %i with alpha %5.3f was # %i into model %i",i, localizationMem->alpha[i], localizationMem->modelNumber[i], tempCounter));
    resetSdMatrix(tempCounter);
    models[tempCounter].isActive = true;
    models[tempCounter].alpha = localizationMem->alpha[i];
    models[tempCounter].stateEstimates[0][0] = localizationMem->X00[i];
    models[tempCounter].stateEstimates[1][0] = localizationMem->X10[i];
    models[tempCounter].stateEstimates[2][0] = localizationMem->X20[i];
    models[tempCounter].stateEstimates[3][0] = localizationMem->X30[i];
    models[tempCounter].stateEstimates[4][0] = localizationMem->X40[i];
    models[tempCounter].stateEstimates[5][0] = localizationMem->X50[i];
    models[tempCounter].stateEstimates[6][0] = localizationMem->X60[i];

    tempCounter++;
    if (tempCounter >= c_MAX_MODELS) break;
  }
  numMateFlippedBalls = localizationMem->numMateFlippedBalls;
  numBadBallUpdates = localizationMem->numBadBallUpdates;
}

void UKFModule::updateMemory() {
  localizationMem->numMateFlippedBalls = numMateFlippedBalls;
  localizationMem->numBadBallUpdates = numBadBallUpdates;
  WorldObject* self = &(worldObjects->objects_[robotState->WO_SELF]);

  localizationMem->bestModel = getBestModelID();
  localizationMem->bestAlpha = models[localizationMem->bestModel].alpha;
  localizationMem->oppositeModels = false;
  localizationMem->fallenModels = false;
  float bestTheta = models[localizationMem->bestModel].stateEstimates[2][0];
  int i=0;
  for (i=0; i < MAX_MODELS_IN_MEM; i++) {
    localizationMem->alpha[i]=-1000; // set the models inactive by defualt
  }
  int tempCounter=0;
  for (i=0; i<c_MAX_MODELS; i++) {
    if (models[i].isActive) {
      localizationMem->modelNumber[tempCounter]=i;
      localizationMem->alpha[tempCounter]=models[i].alpha;
      localizationMem->X00[tempCounter]=models[i].stateEstimates[0][0];
      localizationMem->X10[tempCounter]=models[i].stateEstimates[1][0];
      localizationMem->X20[tempCounter]=models[i].stateEstimates[2][0];
      localizationMem->X30[tempCounter]=models[i].stateEstimates[3][0];
      localizationMem->X40[tempCounter]=models[i].stateEstimates[4][0];
      localizationMem->X50[tempCounter]=models[i].stateEstimates[5][0];
      localizationMem->X60[tempCounter]=models[i].stateEstimates[6][0];

      localizationMem->P00[tempCounter]=models[i].sd(0);
      localizationMem->P11[tempCounter]=models[i].sd(1);
      localizationMem->P01[tempCounter]=fabs(models[i].stateStandardDeviations[0][1]);
      localizationMem->P10[tempCounter]=localizationMem->P01[tempCounter];
      localizationMem->P22[tempCounter]=models[i].sd(2);
      localizationMem->P33[tempCounter]=models[i].sd(3);
      localizationMem->P34[tempCounter]=fabs(models[i].stateStandardDeviations[3][4]);
      localizationMem->P44[tempCounter]=models[i].sd(4);
      tempCounter++;

      // check if any likely models say we're facing opposite dir of what best says
      if (!localizationMem->oppositeModels && i != localizationMem->bestModel && models[i].alpha > 0.03){
        if (fabs(normalizeAngle(bestTheta - models[i].stateEstimates[2][0])) > 2.8){
          localizationMem->oppositeModels = true;
        }
      }
      // check if we have any models off by 90 deg (post fall)
      if (!localizationMem->fallenModels && i != localizationMem->bestModel && models[i].alpha > 0.03){
        float angleDiff = fabs(normalizeAngle(bestTheta - models[i].stateEstimates[2][0]));
        if (angleDiff > 1.42 && self->loc.getDistanceTo(Point2D(models[i].stateEstimates[0][0], models[i].stateEstimates[1][0])) < 50){
          localizationMem->fallenModels = true;
        }
      }
      
    }
    if (tempCounter >= MAX_MODELS_IN_MEM) break;
  }

  ukfLog((10, "Save model.. opposite %i, fallen %i", localizationMem->oppositeModels, localizationMem->fallenModels));

  // update ball sr for best model
  NMatrix ballSR = models[0].GetBallSR();
  localizationMem->SRXX = ballSR[0][0];
  localizationMem->SRXY = ballSR[0][1];
  localizationMem->SRYY = ballSR[1][1];

}

/** Method to seed a new model based on observations. */
void UKFModule::seedNewModel(){

  // check that there is a free model
  int numFreeModels = getNumFreeModels();
  if (numFreeModels < 1) return;

  // make sure we've seen at least two things
  /*
    int seenCount = 0;
    for (int i = WO_CENTER_CIRCLE; i < NUM_WORLD_OBJS; i++){
    if (worldObjects->objects_[i].seen) seenCount++;
    }
    ukfLog((10, "Seed attempt, seen %i objects", seenCount));
    if (seenCount < 2) return;
  */

  // get new location
  float x, y, orient;
  bool gotLocation = getSeedLocation(&x, &y, &orient);

  if (!gotLocation) return;

  // init new model with location
  int newModelID = FindNextFreeModel();
  models[newModelID].isActive = true;
  models[newModelID].alpha = ukfParams.seed_model_alpha;

  models[newModelID].stateEstimates[0][0] = x;      // Robot X
  models[newModelID].stateEstimates[1][0] = y;      // Robot y
  models[newModelID].stateEstimates[2][0] = orient; // Robot heading

  // use current ball estimate
  WorldObject* ball = &(worldObjects->objects_[WO_BALL]);

  models[newModelID].stateEstimates[3][0] = ball->loc.x;    // Ball x
  models[newModelID].stateEstimates[4][0] = ball->loc.y;    // Ball y
  models[newModelID].stateEstimates[5][0] = ball->absVel.x;    // Ball vx
  models[newModelID].stateEstimates[6][0] = ball->absVel.y;    // Ball vy

  // Set the uncertainties
  resetSdMatrix(newModelID);
}

/** Calculate seed location based on observations */
bool UKFModule::getSeedLocation(float* x, float* y, float* orient){


  // if we've seen two unique obs, then its easy
  for (int i = WO_CENTER_CIRCLE; i < NUM_WORLD_OBJS; i++){
    WorldObject* wo = &(worldObjects->objects_[i]);
    if (wo->isGoalCenter() || !wo->seen || !wo->isUnique()) continue;
    if (wo->isGoal() && !ukfParams.USE_GOAL) continue;
    if (wo->isCenterCircle() && !ukfParams.USE_CIRCLE) continue;

    // got our first: i

    for (int j = i+1; j < NUM_WORLD_OBJS; j++){
      wo = &(worldObjects->objects_[j]);
      if (wo->isGoalCenter() || !wo->seen || !wo->isUnique()) continue;
      if (wo->isGoal() && !ukfParams.USE_GOAL) continue;
      if (wo->isCenterCircle() && !ukfParams.USE_CIRCLE) continue;

      // got our second: j

      // try 2 lm reseed
      bool gotPoint = doTwoLMSeed(i, j, x, y, orient);
      if (gotPoint) return true;
    }
  }

  // nothing happened (no two unique landmarks)
  return false;

}

bool UKFModule::doTwoLMSeed(int i, int j, float* x, float* y, float* orient){
  ukfLog((20, "Attempting 2LM triangulation between objects %i and %i",i,j));

  WorldObject *wo = &(worldObjects->objects_[0]);

  // return if observation angles are too close
  if (fabs(wo[i].visionBearing - wo[j].visionBearing) < ukfParams.min_reseed_bearing_diff){

    ukfLog((20, "Bearing too close for 2lm reseed"));
    return false;
  }

  // make two circles
  Circle circ1( wo[i].loc, wo[i].visionDistance );
  Circle circ2( wo[j].loc, wo[j].visionDistance );

  // find intersections
  Point2D pts[3];

  int numPoints = circ1.getIntersectionPoints(circ2, &pts[0], &pts[1]);

  // try making the smaller circle slightly bigger so they'll intersect
  if (numPoints == 0) {

    if (wo[i].visionDistance < wo[j].visionDistance){
      circ1 = Circle(wo[i].loc, wo[i].visionDistance + 20);
    } else {
      circ2 = Circle(wo[j].loc, wo[j].visionDistance + 20);
    }

    numPoints = circ1.getIntersectionPoints(circ2, &pts[0], &pts[1]);
  }

  // got no points
  if (numPoints == 0){
    ukfLog((20, "Circles don't intersect for 2lm reseed"));
    return false;
  }

  //cout << "C1: " << circ1 << " C2: " << circ2 << endl << flush;

  // make a third point thats the average of the other two
  //(two intersections are probably around real int point)
  if (numPoints == 2) {
    pts[2] = (pts[0] + pts[1]) / 2.0;
    numPoints++;
  }

  for (int i = 0; i < numPoints; i++){
    Point2D p = pts[i];

    ukfLog((20, "Point %i: (%5.3f, %5.3f)", i, p.x, p.y));

    // check that they're in the field
    if (!FIELD.isInside(p*10.0)){
      ukfLog((20, "Point not inside field for 2lm reseed"));
      continue;
    }

    // difference between angle to this landmark and
    // angle it was seen at
    AngRad bearing1 = p.getBearingTo(wo[i].loc,
                                     wo[i].visionBearing );

    AngRad bearing2 = p.getBearingTo(wo[j].loc,
                                     wo[j].visionBearing );

    AngRad newOrient = atan2f( sinf( bearing1 ) + sinf( bearing2 ),
                               cosf( bearing1 ) + cosf( bearing2 ) );

    ukfLog((20, "Seed loc (%5.3f, %5.3f), orient: %5.3f", p.x, p.y, newOrient*RAD_T_DEG));

    // set values and return
    *x = p.x;
    *y = p.y;
    *orient = newOrient;
    return true;
  }

  return false;

}


// print debug of models to log for later use
void UKFModule::logModels(){

  // for each, print estimate and alpha
  for (int i = 0; i < c_MAX_MODELS; i++) {
    if (models[i].isActive) {
      ukfLog((20, "Model %i: (%5.3f, %5.3f) r: %5.3f, ball: (%5.3f, %5.3f), alpha: %5.8f, sinceAmbig %i, ball %i", i, models[i].stateEstimates[0][0], models[i].stateEstimates[1][0], RAD_T_DEG*models[i].stateEstimates[2][0], models[i].stateEstimates[3][0], models[i].stateEstimates[4][0], models[i].alpha, models[i].framesSinceAmbig, models[i].ballFramesSinceAmbig));
      ukfLog((20, "  SD (%5.3f, %5.3f) r: %5.3f, ball: (%5.3f, %5.3f)", models[i].stateStandardDeviations[0][0], models[i].stateStandardDeviations[1][1], models[i].stateStandardDeviations[2][2], models[i].stateStandardDeviations[3][3], models[i].stateStandardDeviations[4][4], models[i].stateStandardDeviations[5][5], models[i].stateStandardDeviations[6][6]));
    }
  }
}


void UKFModule::filterCloseBallPosition(float offset_x, float offset_y, float scale_y) {
  WorldObject &ball = worldObjects->objects_[WO_BALL];
  // Todd: this was tuned for crazy rel ball filter
  Point2D relBall = behaviorMem->keeperRelBallPos;

  filtered_close_ball_.x = relBall.x + offset_x;
  filtered_close_ball_.y = scale_y * (relBall.y + offset_y);
  // no longer averaging over time
  //if (ball.seen)
  //filtered_close_ball_buffer_y_.add(ball.relPos.y);
  //filtered_close_ball_.y = 0.772 * (filtered_close_ball_buffer_y_.getAverage() + offset_y);
}
