#ifndef UKF_MODULE_H
#define UKF_MODULE_H

#include "UKF.h"
#include <Module.h>

#include <memory/Memory.h>

#include <memory/WorldObjectBlock.h>
#include <memory/LocalizationBlock.h>
#include <memory/TeamPacketsBlock.h>
#include <memory/FrameInfoBlock.h>
#include <memory/OdometryBlock.h>
#include <memory/RobotStateBlock.h>
#include <memory/GameStateBlock.h>
#include <memory/JointBlock.h>
#include <memory/BehaviorBlock.h>
#include <memory/ProcessedSonarBlock.h>

#include "BallKF.h"
#include <math/Geometry.h>
#include <common/WorldObject.h>
#include <common/RobotInfo.h>
#include <common/States.h>
#include <common/Field.h>
//#include <common/RingBufferWithSum.h>


// params for the ukf
struct UKFParams{

  // R values for measurements of different types
  float R_goal_theta;
  float R_goal_range_offset;
  float R_goal_range_relative;
  float ignore_goal_dist;

  float R_circle_theta;
  float R_circle_range_offset;
  float R_circle_range_relative;
  float ignore_circle_dist;

  float R_line_theta;
  float R_line_range_offset;
  float R_line_range_relative;
  float ignore_line_dist;

  float R_int_theta;
  float R_int_range_offset;
  float R_int_range_relative;
  float ignore_int_dist;

  float R_cross_theta;
  float R_cross_range_offset;
  float R_cross_range_relative;
  float ignore_cross_dist;

  float R_ball_theta;
  float R_ball_range_offset;
  float R_ball_range_relative;
  float ignore_ball_dist;

  float R_kick_vel;
  float R_kick_bearing;

  float kick_ball_vel_reset_noise;
  float shared_ball_sd_factor;


  float R_crop_factor;

  // reset thresholds
  float large_angle_sd;
  float object_error_thresh;
  float object_error_decay;
  float reset_sum_thresh;
  int reset_num_thresh;

  // ambig object param
  float outlier_alpha_degrade; // for orig model when splitting on ambig
  float t_l_mismatch_degrade; // degrade of alpha when matching wrong int type

  float kappa;
  float ball_decay_rate;
  float outlier_rejection_thresh;
  float kick_outlier_rejection_thresh;
  float seen_ball_outlier_rejection_thresh;
  float shared_ball_outlier_rejection_thresh;
  float keeper_outlier_rejection_thresh_diff;

  // process noise
  float robot_pos_noise;
  float robot_orient_noise;
  float ball_pos_noise;
  float ball_vel_noise;
  float keeper_ball_vel_noise;
  float robot_bump_sd_factor;
  float fallen_robot_sd_factor;
  float fallen_ball_sd_factor;
  float standing_robot_sd_factor;
  float moving_ball_sd_factor;

  // std's used when resetting things
  float init_sd_x;
  float init_sd_y;
  float init_sd_theta;
  float init_sd_vel;
  float reset_ball_sd_pos;

  float fallen_sd_theta;
  float reset_sd_pos;
  float reset_sd_theta;
  float reset_sd_ball;

  // num models
  int max_models_after_merge;

  // seeding
  float seed_model_alpha;
  float min_reseed_bearing_diff;

  // keeping models after pen/getup until we're sure which side we're on
  float ambig_min_model_alpha;
  int   ambig_model_min_nframes;
  int   ambig_model_min_ball_frames;

  // keeper ball spawning
  int keeper_ball_spawn_freq;
  float keeper_ball_spawn_likelihood;
  float keeper_ball_spawn_velocity;

  // flipped robot threshold
  int mate_flipped_ball_thresh;

  // which things to use
  bool USE_CIRCLE;
  bool USE_GOAL;
  bool USE_WHOLE_GOAL;
  bool USE_AMBIG;

  bool SPLIT_ON_POSTS;
  bool SPLIT_ON_LINES;
  bool SPLIT_LINES_ON_HIGH_SD;
  bool SPLIT_ON_INTS;

  float SPLIT_LINE_SD_POS_THRESH;
  float SPLIT_LINE_SD_THETA_THRESH;

  bool USE_INTERSECTIONS;
  bool USE_CROSSES;
  bool USE_BALL_IN_SET;
  bool USE_LINES;
  bool RESEED_ON_OUTLIERS;
  bool USE_SHARED_BALL;
  bool CHECK_LINES_IN_FOV;
  bool DO_RESEED;

};


class UKFModule: public Module 
{
 public:
  void specifyMemoryDependency();
  void specifyMemoryBlocks();
  void initSpecificModule();

  void reInit();
  void initFromMemory();

  void processFrame();
  void processObservations();
  void processSharedBall();
  void checkSharedBallForFlips();

  void resetTimeLast();

  void CheckGameState();
  int varianceCheckAll();
  bool varianceCheck(int modelID);
  void ResetAll();
  //void writeToLog();
  void runTimeUpdates();

  void logModels();

  bool clipModelToField(int modelID);
  bool clipActiveModelsToField();
  
  int doKnownLandmarkMeasurementUpdate(int modelNum, int objID);
  int doSharedBallUpdate(float x, float y, float sdxx, float sdxy, float sdyy, int mate);
  void doSharedBallCheck(float x, float y, int mate);

  int doBallMeasurementUpdate(int modelNum);
  int doAmbiguousLandmarkMeasurementUpdate(int objID);
  int doAmbiguousUpdateWithMostLikelyLM(int objID);
  void doKickUpdates();
  void spawnFlippedModel(int modelID, bool robotOnly);
  void spawnNewBallModel(int modelID, Point2D newball);

  bool lineInFOV(Point2D kfLoc, AngRad orient, int objID, int obsID);

  int getNumActiveModels();
  int getNumFreeModels();
  bool CheckModelForOutlierReset(int modelID);
  int  CheckForOutlierResets();
  UKF* getBestModel();
  int getBestModelID();
  void NormaliseAlphas();
  void CheckForAmbigModels();
  int FindNextFreeModel();
  bool MergeTwoModels(int index1, int index2);
  float MergeMetric(int index1, int index2);
  void MergeModels(int maxAfterMerge);
  float MergeModelsBelowThreshold(float MergeMetricThreshold);
  void PrintModelStatus(int modelID);

  // similarity for doing ml ambig update
  float getObsSim(Point2D loc, float orient, WorldObject* obs, Point2D lmLoc);
  float getLineSim(Point2D loc, float orient, WorldObject* obs, Point2D lmLoc, float lineOrient);

  // seeding
  void seedNewModel();
  bool getSeedLocation(float* x, float* y, float* orient);
  bool doTwoLMSeed(int i, int j, float* x, float* y, float* orient);
  
  // fill world obj with loc info
  void populateWOSelf(UKF* kf);
  void populateBall(UKF* kf);
  void updateDistAndBearToObjects();

  // Model Reset Functions
  void doPenaltyReset();
  void doPenaltyKickReset();
  void doPlayerReset();
  void initSdMatrix(int modelNumber);
  void resetSdMatrix(int modelNumber);
  void resetSdMatrixForSideline(int modelNumber);
  void incrementModelError(int model, int object){modelObjectErrors[model][object]+=1.0;};
  void splitModelsOnFall();
  void checkKeeperObservations();
  void splitKeeperModelTurn();
  void spawnManualPositionModel();
  void spawnManualPositionModel(int playerNum);
  void spawnKickedBallModel();

  // helper methods for lua
  bool isModelActive(int index) {return models[index].isActive;};
  UKF *getModelPtr(int index) {return &(models[index]);};
  void updateMemory();

  // convert between mm and cm
  void mmTocm();
  void cmTomm();

  // change ball function for close balls for final approach and kicking
  void filterCloseBallPosition(float offset_x, float offset_y, float scale_y);
  Vector2<float> filtered_close_ball_;
  //RingBufferWithSum<float,20> filtered_close_ball_buffer_y_;

  // Multiple Models Stuff
  static const int c_MAX_MODELS = 34; //(c_MAX_MODELS_AFTER_MERGE*8+2); // Total models
  UKF tempModel;
  UKF models[c_MAX_MODELS];

  UKFParams ukfParams;

  int currentFrameNumber;
  float modelObjectErrors[c_MAX_MODELS][NUM_WORLD_OBJS]; // Storage of outlier history.

  // Game state memory
  bool wasPreviouslyPenalised;
  int previousGameState;
  int previousTeam;
  int previousDive;
  bool isManualPenalty;
  Getup::GetupType previousGetUp;
  float lastKeeperObservationTime;
  float penaltyInitialTime;

  // keep these counts internally, just store to mem for debug
  int numMateFlippedBalls;
  int numBadBallUpdates;

  // timing info
  double timeLast;
  double timePassed;

  // memory info
  WorldObjectBlock* worldObjects;
  LocalizationBlock* localizationMem;
  TeamPacketsBlock* teamPacketsMem;
  FrameInfoBlock* frameInfo;
  RobotStateBlock* robotState;
  GameStateBlock* gameState;
  OdometryBlock* odometry;
  JointBlock* jointAngles;
  BehaviorBlock* behaviorMem;
  ProcessedSonarBlock* processedSonar;

  BallKF ballKF;
};

#endif
