DEG_T_RAD = math.pi / 180.0
RAD_T_DEG = 180.0 / math.pi

require 'percepts'
require 'cfgcam'
require 'cfgwalk'
require 'commands'
require 'config'
require 'cfgstiff'
require 'cfgkick'
require 'kicks'
require 'poses'
require 'behavior'

-- This function is run to initialise everything
function init()
  if core then
    print "Creating Lua link to C++ Core"  
    robot = core.VisionCore_inst_
  else
    print "***ERROR*** Can't create Lua link to C++ Core" 
  end

  visionC = robot.vision_
  localizationC = robot.localization_
  perfectC = robot.perfect_localization_
  opponentsC = robot.opponents_ 
  ledsC = robot.leds_

  initMemory()

  -- init localization params
  config.configLocalization()
  localizationC:reInit()
  opponentsC:reInit()
  world_objects:init(robot_state.team_)

  -- init which side of the field blue is one
  localization_mem.blueSide = core.Sides_NOSIDES;


  kicks.checkKicks() -- must be before setKickAngles
  
  -- set these here so it works for behavior simulation
  setRobotSpecificParams() -- accomplishes the sadness of setting the robot specific parameters
  commands.setKickParameters(cfgStraightKick,cfgStraightKickSpline)
  --commands.setKickParameters(cfgSplineStraightKick,cfgSplineStraightKickSpline)
  -- al walk
  commands.setWalkMode(core.WalkMode_SLOW,true)
  commands.saveCurrentHTWKWalkParams()

  first_frame = true
  announced_vision = false
end

function initMemory()
  luaC = robot.lua_
  
  behavior_mem = luaC.behavior_
  camera_block = luaC.camera_block_
  game_state = luaC.game_state_
  kick_request = luaC.kick_request_
  odometry = luaC.odometry_
  robot_state = luaC.robot_state_
  sensors = luaC.sensors_
  vision_frame_info = luaC.vision_frame_info_
  walk_param = luaC.walk_param_
  kick_params = luaC.kick_params_
  walk_request = luaC.walk_request_
  world_objects = luaC.world_objects_
  team_packets = luaC.team_packets_
  opponent_mem = luaC.opponents_  
  joint_commands = luaC.joint_commands_
  processed_sonar = luaC.vision_processed_sonar_
  al_walk_param = luaC.al_walk_param_
  walk_info = luaC.walk_info_
  robot_vision = luaC.robot_vision_
  body_model = luaC.body_model_
  robot_info = luaC.robot_info_
  speech = luaC.speech_
  localization_mem = luaC.localization_

  camera_offset = robot_info.dimensions_.tiltOffsetToBottomCamera - robot_info.dimensions_.tiltOffsetToTopCamera

  text_logger = robot.textlog_

  --joint_angles = luaC.joint_angles_
  joint_angles = percepts.joint_angles

end

function errHandler(x)
  UTdebug.log(0,'LUA ERROR:')
  UTdebug.log(0,x)
  UTdebug.log(0,debug.traceback())
  speech:say("lua")
end

-- This is where lua gets called from the vision thread
function processFrame()
  robot:preVision() -- get the vision lock and copy over necessary data from both the motion process and the image capture thread  

  if first_frame then
    first_frame = false
    world_objects:init(robot_state.team_)

    -- init camera parameters
    commands.setCameraParams(core.Camera_BOTTOM,cfgCamRoboCup)
    commands.setCameraParams(core.Camera_TOP,cfgCamRoboCup)

    setRobotSpecificParams() -- accomplishes the sadness of setting the robot specific parameters

    commands.setKickParameters(cfgStraightKick,cfgStraightKickSpline)

    -- init walk parameters
    -- al walk
    commands.setWalkMode(core.WalkMode_SLOW,true)
    cfgwalk.initWalk()

    -- vision stuff
    if (visionC ~= nil) then
      visionC:initSpecificModule()
    end

  end

  if not(announced_vision) and camera_block.cameras_tested_ then
    speech:say("Vision")
    announced_vision = true
  end


  UTdebug.log(10, "odom, dive, walk_req", odometry.getting_up_side_, odometry.fall_direction_, walk_request.motion_)
  -- not getting up, not falling, and not diving
  if (odometry.getting_up_side_ == core.Getup_NONE and odometry.fall_direction_ == core.Fall_NONE and game_state.state ~= core.TEST_ODOMETRY) then 
    if (visionC ~= nil) then
      visionC:processFrame()
    end
  else
    -- reset vision of everything
    world_objects:reset()
  end
  
  robot:postVision() -- release the lock

  percepts.populateLocalCopy()
  commands.resetTiltOffset()

  --perfectC:processFrame()
  localizationC:processFrame()
  opponentsC:processFrame()
  local camOffsets = getCamOffsets()
  localizationC:filterCloseBallPosition(camOffsets.rel_ball_fwd,camOffsets.rel_ball_side,camOffsets.scale_side)

  -- do this after setting other stuff
  behavior.processFrame()

  commands.handleTiltOffset()

  --checkSonarStatus();
  --walk_request:setWalk(1.0,0.0,0)

  doLights()

  robot:publishData() -- copy commands back to the motion process and copy camera info back to the image capture thread
  --robot:logMemory()
end


function doLights()
  doFootLights()
  doStateLights()
  doEyeLights()
  doEarLights() 
end

function doEarLights()
  ledsC:allLeftEar(0)
  ledsC:allRightEar(0)
end

function doStateLights()
  -- do chest led based on state
  state = game_state.state
  if (state == core.INITIAL) then 
    ledsC:chest(0,0,0)
  elseif (state == core.READY) then
    ledsC:chest(0,0,1)
  elseif (state == core.SET) then
    ledsC:chest(1,1,0)
  elseif (state == core.PLAYING) then
    ledsC:chest(0,1,0)
  elseif (state == core.PENALISED) then
    ledsC:chest(1,0,0)
  elseif (state == core.TESTING) then
    ledsC:chest(1,0,1)
  elseif (state == core.FINISHED) then
    ledsC:chest(1,1,1)
  end
end


function doFootLights()
  -- do feet leds
  if (robot_state.team_== 0) then --blue
    ledsC:rightFoot(0,0,1)
  elseif (robot_state.team_==1) then --red
    ledsC:rightFoot(1,0,0)
  end
  
  if (game_state.ourKickOff) then
    ledsC:leftFoot(1,1,1)
  else
    ledsC:leftFoot(0,0,0)
  end
end


function doEyeLights()
  doLeftEyePower()
  doRightEyeHeat()
end

function doLeftEyePower()
  local battery = sensors:getValue(core.battery)
  if (battery > 0.95) then
    ledsC:allLeftEye(1,1,1)
  elseif (battery > 0.8) then
    ledsC:allLeftEye(0,1,0)
  elseif (battery > 0.65) then
    ledsC:allLeftEye(0,0,1)
  else
    ledsC:allLeftEye(1,0,0)
  end
end

function doRightEyeHeat()
  local max_temp = getMaxTemp()
  if (max_temp > 74) then
    ledsC:allRightEye(1,0,0)
    return true
  elseif (max_temp > 64) then
    ledsC:allRightEye(1,1,0)
    return true
  elseif (max_temp > 54) then
    ledsC:allRightEye(1,1,1)
    return true
  else
    return false
  end
end

function getMaxTemp()
  local max_temp = 0
  for i = 0, core.NUM_JOINTS-1 do
    local temp = sensors:getJointTemperature(i)
    if (temp > max_temp) then
      max_temp = temp
    end
  end
  return max_temp
end

function setRobotSpecificParams()
  --setRobotSpecificWalkParams()
  setRobotSpecificKickParams()
  setRobotSpecificSittingPose()
end

-- code that runs on require
init()

