#include <math/YaMatrix.h>
#include <vector>
#ifndef GAUSS_NEWTON_OPTIMIZER_H
#define GAUSS_NEWTON_OPTIMIZER_H

using namespace std;

  template <class M, class C>
  class GaussNewtonOptimizer
  {
  private:
    const unsigned int numOfMeasurements; /**< The number of measurements. */
    const unsigned int numOfParameters; /**< The number of parameters. */
    YaMatrix<float> currentParameters; /**< The vector (Nx1-matrix) containing the current parameters. */
    YaMatrix<float> currentValues; /**< The vector (Nx1-matrix) containing the current error values for all measurements. */
    const vector<M>& measurements; /**< A reference to the vector containing all measurements. */
    C& object; /**< The object used to call the error function. */
    float(C::*pFunction)(const M& measurement, const vector<float>& parameters);  /**< A pointer to the error function. */
    const float delta; /**< The delta used to approximate the partial derivatives of the Jacobian. */

  public:
	GaussNewtonOptimizer();
    GaussNewtonOptimizer(const vector<float>& parameters, const vector<M>& measurements, C& object, float(C::*pFunction)(const M& measurement, const vector<float>& parameters));

    /**
     * This method executes one iteration of the Gauss-Newton algorithm.
     * The new parameter vector is computed by a_i+1 = a_i - (D^T * D)^-1 * D^T * r
     * where D is the Jacobian, a is the parameter vector and r is the vector containing the current error values.
     * @return The sum of absolute differences between the old and the new parameter vector.
     */
	 float iterate();

    /**
     * The method returns the current parameter vector.
     * @return The current parameter vector.
     */
    vector<float> getParameters() const;
  };

	template <class M, class C>
    GaussNewtonOptimizer<M,C>::GaussNewtonOptimizer(const vector<float>& parameters, const vector<M>& measurements, C& object, float(C::*pFunction)(const M& measurement, const vector<float>& parameters))
      : numOfMeasurements(measurements.size()), numOfParameters(parameters.size()), currentParameters(numOfParameters, 1), currentValues(numOfMeasurements, 1), measurements(measurements), object(object), pFunction(pFunction), delta(0.1f)
    {
      for(unsigned int i = 0; i < numOfParameters; ++i)
      {
        currentParameters[i][0] = parameters[i];
      }
      for(unsigned int i = 0; i < numOfMeasurements; ++i)
      {
        currentValues[i][0] = (object.*pFunction)(measurements[i], currentParameters.transpose()[0]);
      }
    }

	template <class M, class C>
	float GaussNewtonOptimizer<M,C>::iterate()
    {
      // build jacobi matrix
      YaMatrix<float> jacobiMatrix(numOfMeasurements, numOfParameters);
      for(unsigned int j = 0; j < numOfParameters; ++j)
      {
        // the first derivative is approximated using values slightly above and below the current value
        const float oldParameter = currentParameters[j][0];
        const float parameterAbove = oldParameter + delta;
        const float parameterBelow = oldParameter - delta;
        for(unsigned int i = 0; i < numOfMeasurements; ++i)
        {
          // approximate first derivation numerically
          currentParameters[j][0] = parameterAbove;
          const float valueAbove = (object.*pFunction)(measurements[i], currentParameters.transpose()[0]);
          currentParameters[j][0] = parameterBelow;
          const float valueBelow = (object.*pFunction)(measurements[i], currentParameters.transpose()[0]);
          const float derivation = (valueAbove - valueBelow) / (2.0f * delta);
          jacobiMatrix[i][j] = derivation;
        }
        currentParameters[j][0] = oldParameter;
      }

      try
      {
        YaMatrix<float> result = (jacobiMatrix.transpose() * jacobiMatrix).inverse() * jacobiMatrix.transpose() * currentValues;
        currentParameters -= (result * .001f);

        for(unsigned int i = 0; i < numOfMeasurements; ++i)
        {
          currentValues[i][0] = (object.*pFunction)(measurements[i], currentParameters.transpose()[0]);
        }

        float sum = 0;
        for(unsigned int i = 0; i < numOfParameters; ++i)
        {
          sum += abs(result[i][0]);
        }
        return sum;
      }
      catch(...)
      {
        return 0.0f;
      }

    }

	template <class M, class C>
    vector<float> GaussNewtonOptimizer<M,C>::getParameters() const
    {
      return currentParameters.transpose()[0];
    }

#endif
