#include "LogReader.h"
#include "ImageBlock.h"

#define MAX_EXPECTED_MODULES_PER_MEMORY 40


LogReader::LogReader(const char *filename):
  using_buffer_(false),
  log_buffer_(NULL),
  buffer_size_(0)
{
  filename_ = filename;
  log_file_.open(filename,std::ios::binary);
  //std::cout << filename << std::endl;
  if (!good())
    std::cout << "problem opening log" << std::endl << std::flush;
}

LogReader::LogReader(char *buffer, unsigned long buffer_size):
  using_buffer_(true),
  log_buffer_(buffer),
  buffer_size_(buffer_size),
  buffer_ind_(0)
{
}

LogReader::~LogReader() {
  close();
}

#include <iostream>

bool LogReader::readMemoryHeader(MemoryHeader &header) {
  // get the number of blocks
  unsigned int num_blocks = 0;
  read((char *)&(num_blocks),sizeof(num_blocks));
  //std::cout << "reading " << num_blocks << " blocks" << std::endl;
  if (num_blocks > MAX_EXPECTED_MODULES_PER_MEMORY) {
    std::cout << "LogReader::readMemoryHeader: BAD NUMBER OF BLOCKS " << num_blocks << std::endl;
    return false;
  }
  // loop through and load the block_names
  char ch;
  //get(ch); // why is this necessary
  for (unsigned int i = 0; i < num_blocks; i++) {
    header.block_names.push_back("");
    read(&ch,1);
    while (ch != '\0'){
      header.block_names[i] += ch;
      read(&ch,1);
    }
  }
  return true;
}

bool LogReader::readMemory(Memory &memory, bool suppress_errors) {
  // get the header and see how many blocks we need to read
  MemoryHeader header;
  bool res;
  res = readMemoryHeader(header);
  if (!res)
    return false;
  //std::cout << "block names: ";
  //for (unsigned int i = 0; i < header.block_names.size(); i ++)
    //std::cout << header.block_names[i] << " ";
  //std::cout << std::endl;
  if (header.block_names.size()<1) return false;
  //std::cout << "Reading " << header.block_names.size() << " blocks" << std::endl;
  // loop through and read all blocks
  for (unsigned int i = 0; i < header.block_names.size(); i++) {
    //std::cout << "reading " << header.block_names[i] << std::endl;
    // first, read in header
    MemoryBlockHeader block_header;
    read((char *)&block_header,sizeof(block_header));
    if (!good()) {
      if(!suppress_errors)
        std::cout << "trouble reading header" << std::endl << std::flush;
      return false;
    }

    std::string &id = header.block_names[i];
    //std::cout << "trying to read block:" << id << std::endl;
    MemoryBlock *block = memory.getBlockPtrByName(id);
    if (block == NULL) {
      // the memory doesn't have that block
      //std::cerr << "Block " << id << " doesn't exist in memory, trying to add" << std::endl;
      bool res = memory.addBlockByName(id);
      //bool res = false;
      if (!res) {
        std::cout << "Adding block " << id << " failed, just skipping" << std::endl << std::flush;
        readAndIgnoreBlock(block_header);
        continue;
      }
      else {
        block = memory.getBlockPtrByName(id);
        assert(block != NULL); // really shouldn't happen
      }
    }
    res = readBlock(block_header,*block);
    if (!res) {
      std::cerr << "Failed to load block " << id << ", skipping" << std::endl;
    }
    if (id == "raw_image")
      ((ImageBlock*)block)->postRead();
  }
  return true;
}

bool LogReader::readBlock(const MemoryBlockHeader &header,MemoryBlock &block) {
  //MemoryBlockHeader header;
  //log_file_.read((char *)&header,sizeof(header));
  //std::cout << header.version << " " << header.size << std::endl << std::flush;
  // make sure the header we read matches what the block expects
  if ((header.version != block.header.version) || (header.size != block.header.size)) {
    std::cout << "failed to load block";
    if (header.version != block.header.version){
      std::cout << "Log version " << header.version << " current version: " << block.header.version << std::endl;
    }
    if (header.size != block.header.size){
        std::cout << "Log size " << header.size << " current size: " << block.header.size << std::endl;
    }
    // still, try to read the block out of the log so we fail nicely
    //std::cout << header.version << "," << header.size << " != " << block.header.version << "," << block.header.size << std::endl;
    readAndIgnoreBlock(header);
    return false;
  }

  read((char *)&block,header.size);
  if (!good()) {
    std::cout << "problem after reading block" << std::endl << std::flush;
    return false;
  }

  return true;
}

void LogReader::readAndIgnoreBlock(const MemoryBlockHeader &header) {
  char *buffer = new char[header.size];
  read(buffer,header.size);
  //std::cout << "type: " << typeid((MemoryBlock*)buffer).name() << std::endl;
  delete [] buffer;
}

void LogReader::close() {
  if (! using_buffer_)
    log_file_.close();
}

void LogReader::read(char *s, std::streamsize n) {
  if (using_buffer_) {
    if (buffer_ind_ + n > buffer_size_) {
      //std::cerr << "LogReader: tried to read past end of buffer" << std::endl;
      buffer_ind_ = buffer_size_ + 1;
      return;
    }
    memcpy(s,log_buffer_+buffer_ind_,n);
    buffer_ind_ += n;
  } else
    log_file_.read(s,n);
}

bool LogReader::good() {
  if (using_buffer_) {
    return buffer_ind_ <= buffer_size_;
  } else
    return log_file_.good();
}

Log* LogReader::readLog(int start, int finish) {
    Log* frames = new Log();
    bool ok = true;
    int count = 0;
    while (ok) {
        Memory frame(false,MemoryOwner::TOOL_MEM, 0, 1);
        ok = readMemory(frame, true);
        bool loadFrame = count >= start && (finish == -1 || count < finish);
        if (ok && loadFrame) frames->push_back(frame);
        count++;
    }
    close();
    frames->name = filename_;
    return frames;
}
