#include "Logger.h"
#include <iostream>
#include <ctime>
#include <memory/ImageBlock.h>

Logger::Logger(const char *filename, bool appendUniqueId):
  log_buffer_(NULL),
  buffer_size_(0),
  msg_size_(NULL),
  using_buffer_(false)
{
  if (filename) {
    open(filename, appendUniqueId);
  }
}

Logger::Logger (char *buffer,unsigned long *msg_size,unsigned long buffer_size):
  log_buffer_(buffer),
  buffer_size_(buffer_size),
  msg_size_(msg_size),
  using_buffer_(true)
{
}

Logger::~Logger() {
  close();
}

void Logger::writeMemoryHeader(const MemoryHeader &header) {
  unsigned int num_blocks = header.block_names.size();
  write((const char*)&(num_blocks),sizeof(num_blocks));
  // write each block name
  for (unsigned int i = 0; i < num_blocks; i++)
    write(header.block_names[i].c_str(),header.block_names[i].size()+1); // +1 so we write the \0
}

void Logger::writeMemory(Memory &memory) {
  if (using_buffer_ || log_file_.is_open()) {
    MemoryHeader header;
    MemoryBlock *block;
    // first get a list of the blocks we're loading
    memory.getBlockNames(header.block_names,true);
    // write the header
    writeMemoryHeader(header);
    // write each block
    for (unsigned int i = 0; i < header.block_names.size(); i++) {
      //std::cout << "writing block " << header.block_names[i] << " ";
      block = memory.getBlockPtrByName(header.block_names[i]);
      if (header.block_names[i] == "raw_image" && type_ != CORE_TOOL)
        ((ImageBlock*)block)->preWrite();
      writeBlock(*block);
    }
  }
}

void Logger::writeBlock(MemoryBlock &block) {
  //std::cout << block.header.version << " " << block.header.size << std::endl << std::flush;
  write((const char *)&block.header,sizeof(block.header)); // write out the header
  write((const char *)&block,block.header.size); // write the block
}

void Logger::open(const char *filename, bool appendUniqueId) {
  close(); // any previously opened logs
  if (appendUniqueId) {
    char buffer[100];
    generateUniqueFileName(buffer, filename, "log");
    std::cout << "Logging to file " << buffer << std::endl;
    log_file_.open(buffer, std::ios::binary);
  } else {
    std::cout << "Logging to file " << filename << std::endl;
    log_file_.open(filename, std::ios::binary);
  }
}

void Logger::close() {
  if (log_file_.is_open()) {
    std::cout << "Closing log file" << std::endl;
    log_file_.close();
  }
}

void Logger::setType(int t){
  type_ = t;
}

void Logger::generateUniqueFileName(char *fileName, const char *prefix, const char *ext) {
  std::string robotbase = "/home/nao/logs/";
  std::string base = robotbase;
  if (type_ == CORE_SIM) {
    std::string naohome = getenv("NAO_HOME");
    std::string simbase = naohome + "/logs/";
    base = simbase;
  } else if (type_ == CORE_TOOL) {
    std::string naohome = getenv("NAO_HOME");
    std::string toolbase = naohome + "/logs/";
    base = toolbase;
  }

  std::string logFile = base + std::string(prefix) + "_%y_%m_%d-%H_%M_%S." + std::string(ext);
  time_t rawtime;
  struct tm *timeInfo;
  time(&rawtime);
  timeInfo = localtime(&rawtime);
  strftime(fileName, 80, logFile.c_str(), timeInfo);
}

void Logger::write(const char *s, std::streamsize n) {
  if (using_buffer_) {
    if (*msg_size_ + n >= buffer_size_) {
      std::cerr << "Logger::write MESSAGE TO BIG FOR BUFFER" << std::endl;
      return;
    }
    memcpy(log_buffer_+*msg_size_,s,n);
    *msg_size_ += n;
  } else
    log_file_.write(s,n);
}
