#ifndef WALKREQUESTBLOCK_RE8SDRLN
#define WALKREQUESTBLOCK_RE8SDRLN

#include <math/Pose2D.h>
#include "MemoryBlock.h"


const std::string walkNames[] = {
  "WALK",
  "STAND",
  "FALLING",
  "GETUP",
  "FALL_PREVENTION",
  "NONE",
  "STEP_LEFT",
  "STEP_RIGHT",
  "WAIT"
};


struct WalkRequestBlock : public MemoryBlock {
public:
  enum Motion {
    WALK,
    STAND,
    FALLING,
    GETUP,
    FALL_PREVENTION,
    NONE,
    STEP_LEFT,
    STEP_RIGHT,
    WAIT,
    NUM_OF_MOTIONS
  };

  WalkRequestBlock():
    new_command_(false),
    motion_(NONE),
    speed_(),
    percentage_speed_(true),
    pedantic_walk_(false),
    walk_to_target_(false),
    target_walk_is_active_(false),
    rotate_around_target_(false),
    rotate_heading_(0),
    perform_kick_(false),
    kick_heading_(0),
    kick_distance_(3000),
    kick_with_left_(false),
    tilt_fallen_counter_(0),
    roll_fallen_counter_(0),
    getup_from_keeper_dive_(false),
    odometry_fwd_offset_(0),
    odometry_side_offset_(0),
    odometry_turn_offset_(0),
    keep_arms_out_(false),
    slow_stand_(false)
  {
    header.version = 14;
    header.size = sizeof(WalkRequestBlock);
  }
    
  void set(Motion m, Pose2D speed, bool percentage_speed, bool pedantic) {
    new_command_ = true;
    motion_ = m;
    speed_ = speed;
    percentage_speed_ = percentage_speed;
    pedantic_walk_ = pedantic;
    walk_to_target_ = false;
    rotate_around_target_ = false;
    perform_kick_ = false;
    kick_heading_ = 0;
    kick_distance_ = 3000;
    kick_with_left_ = false;
    rotate_distance_ = 0;
    rotate_heading_ = 0;
  }

  void noWalk() {
    set(NONE,Pose2D(0,0,0),true,false);
  }

  void stand() {
    set(STAND,Pose2D(0,0,0),true,false);
  }

  void wait() {
    set(WAIT,Pose2D(0,0,0),true,false);
  }

  void setStep(bool isLeft, float x, float y, float rotation) {
    Motion walk = STEP_RIGHT;
    if (isLeft) walk = STEP_LEFT;
    set(walk,Pose2D(rotation,x,y),false,false);
  }

  void setWalk(float x, float y, float rotation) {
    set(WALK,Pose2D(rotation,x,y),true,false);
  }

  void setPedanticWalk(float x, float y, float rotation) {
    set(WALK,Pose2D(rotation,x,y),true,true);
  }
  
  void setFalling() {
    set(FALLING,Pose2D(0,0,0),true,false);
  }

  void setKick(float distance, float heading, bool with_left){
    motion_ = WALK;
    perform_kick_ = true;
    kick_heading_ = heading;
    kick_distance_ = distance;
    kick_with_left_ = with_left;
  }

  void setOdometryOffsets(float fwd, float side, float turn) {
    odometry_fwd_offset_ = fwd;
    odometry_side_offset_ = side;
    odometry_turn_offset_ = turn;
  }

  void setWalkTarget(float relx, float rely, float relang, bool pedantic = false) {
    new_command_ = true;
    motion_ = WALK;
    walk_to_target_ = true;
    rotate_around_target_ = false;
    target_point_.translation.x = relx;
    target_point_.translation.y = rely;
    target_point_.rotation = relang;
    pedantic_walk_ = pedantic;
  }

  bool new_command_;
  Motion motion_; // what type of motion, walk/stand/kick
  Pose2D speed_; // the speed of the walk
  bool percentage_speed_; // true if speed is percentage rather than absolute vel
  bool pedantic_walk_; // true disables the step size stabilization.  "Set it when precision is indispensable"

  bool is_penalised_;

  // target point walk 
  Pose2D target_point_;
  bool walk_to_target_;
  bool target_walk_is_active_;
  
  bool rotate_around_target_;
  float rotate_distance_;
  float rotate_heading_;

  // for kicking from walk
  bool perform_kick_;
  float kick_heading_;
  float kick_distance_;
  bool kick_with_left_;

  int tilt_fallen_counter_;
  int roll_fallen_counter_;
  bool getup_from_keeper_dive_;
  
  // walk odometry offsets
  float odometry_fwd_offset_;
  float odometry_side_offset_;
  float odometry_turn_offset_;

  // goalie arms
  bool keep_arms_out_;

  bool slow_stand_; // true if we need a slow stand
};

#endif /* end of include guard: WALKREQUESTBLOCK_RE8SDRLN */
