/**
* @file RobotModel.h
*
* Declaration of class RobotModel
*
* @author <A href="mailto:allli@informatik.uni-bremen.de">Alexander Hrtl</A>
*/

#pragma once

#include "Tools/Math/Pose3D.h"
#include "Representations/Infrastructure/JointData.h"
#include "Representations/Configuration/RobotDimensions.h"
#include "Representations/Configuration/MassCalibration.h"

/**
 * @class RobotModel
 *
 * Contains information about extremities.
 */
class RobotModel : public Streamable
{
  /** Streaming */
  virtual void serialize(In* in, Out* out)
  {
    STREAM_REGISTER_BEGIN();
    STREAM(limbs);
    STREAM(centerOfMass);
    STREAM(totalMass);
    STREAM_REGISTER_FINISH();
  }

public:
  Pose3DBH limbs[MassCalibrationBH::numOfLimbs]; /**< Coordinate frame of the limbs of the robot relative to the robot's origin. */
  Vector3BH<> centerOfMass; /**< Position of the center of mass (center of gravity) relative to the robot's origin. */
  float totalMass; /**< The mass of the robot. */

  /** Constructor */
  RobotModel() : totalMass(0) {}

  /**
  * Constructs the RobotModel from given joint data.
  * @param joints The joint data.
  * @param robotDimensions The dimensions of the robot.
  * @param massCalibration The mass calibration of the robot.
  */
  RobotModel(const JointData& joints, const RobotDimensionsBH& robotDimensions, const MassCalibrationBH& massCalibration);

  /**
  * Recalculates the RobotModel from given joint data.
  * @param joints The joint data.
  * @param robotDimensions The dimensions of the robot.
  * @param massCalibration The mass calibration of the robot.
  */
  void setJointData(const JointData& joints, const RobotDimensionsBH& robotDimensions, const MassCalibrationBH& massCalibration);

  /** Creates a 3-D drawing of the robot model. */
  void draw();
};
