/**
 * @file RingBuffer.h
 *
 * Declaration of class RingBufferBH
 *
 * @author Max Risler
 */

#pragma once

/**
 * @class RingBufferBH
 *
 * template class for cyclic buffering of the last n values of Type V
 */
template <class V, int n> class RingBufferBH
{
public:
  /** Constructor */
  RingBufferBH() {init();}

  /**
   * initializes the Ringbuffer
   */
  inline void init() {current = n - 1; numberOfEntries = 0;}

  /**
   * adds an entry to the buffer
   * \param v value to be added
   */
  inline void add(const V& v)
  {
    add();
    buffer[current] = v;
  }

  /**
   * adds an entry to the buffer.
   * The new head is not initialized, but can be changed afterwards.
   */
  inline void add()
  {
    current++;
    current %= n;
    if(++numberOfEntries >= n) numberOfEntries = n;
  }

  /**
   * removes the first added entry to the buffer
   */
  inline void removeFirst()
  {
    --numberOfEntries;
  }

  /**
   * returns an entry
   * \param i index of entry counting from last added (last=0,...)
   * \return a reference to the buffer entry
   */
  inline V& getEntry(int i)
  {
    return buffer[(n + current - i) % n];
  }

  /**
   * returns an const entry
   * \param i index of entry counting from last added (last=0,...)
   * \return a reference to the buffer entry
   */
  inline const V& getEntry(int i) const
  {
    return buffer[(n + current - i) % n];
  }

  /**
   * returns an entry
   * \param i index of entry counting from last added (last=0,...)
   * \return a reference to the buffer entry
   */
  inline V& operator[](int i)
  {
    return buffer[(n + current - i) % n];
  }

  /**
   * returns a constant entry.
   * \param i index of entry counting from last added (last=0,...)
   * \return a reference to the buffer entry
   */
  inline const V& operator[](int i) const
  {
    return buffer[(n + current - i) % n];
  }

  /** Returns the number of elements that are currently in the ring buffer
  * \return The number
  */
  inline int getNumberOfEntries() const
  {
    return numberOfEntries;
  }

  /**
  * Returns the maximum entry count.
  * \return The maximum entry count.
  */
  inline int getMaxEntries() const
  {
    return n;
  }

private:
  int current;
  int numberOfEntries;
  V buffer[n];
};
