#include "BallDetector.h"

BallDetector::BallDetector(
                           WorldObjectBlock*& world_object_block,
                           BodyModelBlock*& body_model_block,
                           RobotVisionBlock*& robot_vision_block,
                           FrameInfoBlock*& frame_info_block,
                           RobotStateBlock*& robot_state_block,
                           Vision::TransformationProvider& transform,
                           Classifier*& classifier,
                           BlobDetector*& blob_detector,
                           Camera::Type& camera)
  :   world_objects_(world_object_block),
      body_model_(body_model_block),
      robot_vision_(robot_vision_block),
      vision_frame_info_(frame_info_block),
      robot_state_(robot_state_block),
      transform_(transform),
      classifier_(classifier),
      blob_detector_(blob_detector),
      camera_(camera)
{
  verticalBlob = blob_detector_->verticalBlob;
  horizontalBlob = blob_detector_->horizontalBlob;
  verticalBlobCount = blob_detector_->verticalBlobCount;
  horizontalBlobCount = blob_detector_->horizontalBlobCount;
  for(int i = 0; i < MAX_BALL_CANDS; i++)
    candidates[i] = new BallCandidate();
}

void BallDetector::findBall() {
  bestBallCandIndex = -1;
  blob_detector_->resetOrangeBlobs();
#ifdef TOOL
  robot_vision_->doHighResBallScan = true;
#endif

  if(camera_ == Camera::BOTTOM) {
    blob_detector_->formOrangeBlobs();
    blob_detector_->calculateOrangeBlobData();
    checkAndSelectBall();
  }
  else {
    ballCandCount = 0;
    if(classifier_->startHighResBallScan()) {
      classifier_->didHighResBallScan = true;
      blob_detector_->formOrangeBlobs();
      blob_detector_->calculateOrangeBlobData();
      checkAndSelectBall();
      classifier_->completeHighResScan();
    }
  }
  robot_vision_->doHighResBallScan = false;
}

bool BallDetector::checkAndSelectBall()
{
  ballCandCount = 0;

  // Note(piyushk): remove below because vision should not mess with reset
  // // reset ball world object
  // WorldObject *ball = &world_objects_->objects_[WO_BALL];
  // ball->reset();

  if (horizontalBlobCount[c_ORANGE] > MAX_ORANGE_BLOBS) {
    visionLog((34,  "Too many orange blobs in the scene"));
    return true;
  }

  if (!horizontalBlobCount[c_ORANGE]) {
    visionLog((34,  "No orange blobs in the scene"));
    return false;
  }

  // merge orange blobs
  uint16_t mergeIndex[MAX_ORANGE_BLOBS];
  blob_detector_->mergeHorizontalBlobs(c_ORANGE, mergeIndex, MAX_ORANGE_BLOBS);

  for(int i = 0; i < horizontalBlobCount[c_ORANGE]; i++) {
    Blob* blob = &horizontalBlob[c_ORANGE][i];
    visionLog((34, "merged blob %i: (xi:xf,yi:yf) = (%i:%i,%i:%i)", i, blob->xi, blob->xf, blob->yi, blob->yf));
  }

  // sort orange blobs by blob area size
  uint16_t sortIndex[MAX_ORANGE_BLOBS];
  sortOrangeBlobs(mergeIndex, sortIndex);

  for(int i = 0; i < horizontalBlobCount[c_ORANGE]; i++) {
    Blob* blob = &horizontalBlob[c_ORANGE][i];
    visionLog((34, "sorted blob %i: (xi:xf,yi:yf) = (%i:%i,%i:%i)", i, blob->xi, blob->xf, blob->yi, blob->yf));
  }

  // fit circles to ball candidates
  fitCircleToBallCands(sortIndex, mergeIndex);

  // sanity check ball candidates and return best
  WorldObject ballCand[MAX_BALL_CANDS];
  return sanityCheckBallCands(sortIndex, ballCand);

}

bool BallDetector::sanityCheckBallCands(uint16_t *sortIndex, WorldObject *ballCand)
{
  for (uint16_t i = 0; i < ballCandCount; i++) {
    WorldObject *ball = &ballCand[i];
    Blob *blob = &horizontalBlob[c_ORANGE][sortIndex[i]];
    ball->ballBlobIndex = sortIndex[i];

    visionLog((34,   "Ball candidate %i: (cx,cy) = (%f,%f), blob (xi:xf,yi:yf) = (%i:%i,%i:%i)",
               i, candidates[i]->centerX,candidates[i]->centerY, blob->xi,blob->xf,blob->yi,blob->yf));

    // Size Check
    int minXSize = 7, minYSize = 7;
    // allow smaller balls if we did high res scan
    if (classifier_->didHighResBallScan) {
      minXSize = 4;
      minYSize = 2;
    }
    if (blob->dx < minXSize || blob->dy < minYSize) {
      visionLog((34,  "bad minSize orange-blob[%i] dx %i, dy %i, minXSize %i minYSize %i",
                 sortIndex[i], blob->dx, blob->dy, minXSize,minYSize));
      ball->visionConfidence = -1;
      continue;
    }
    /*
    // Ratio check
    float ratio = blob->dy / (float)blob->dx;
    if (ratio < .5 || ratio > 2) {
    ball->visionConfidence = -1;
    continue;
    }
    */

    // pixel ratio check
    if (blob->correctPixelRatio < 0.25f ) {
      visionLog((34,  "bad pixel ratio for orange-blob[%i]=%f", sortIndex[i],blob->correctPixelRatio));
      ball->visionConfidence = -1;
      continue;
    }

    // Check circle fit results
    bool ballCenterXInBlob =
      (candidates[i]->centerX > blob->xi && candidates[i]->centerX < blob->xf);
    bool ballCenterYInBlob =
      (candidates[i]->centerY > blob->yi && candidates[i]->centerY < blob->yf);
    bool ballCenterInBlob = (ballCenterXInBlob && ballCenterYInBlob);
    bool decentCircleFit = (candidates[i]->stddev < .5);
    //bool ballCenterXOffScreen =
    //(candidates[i]->centerX < 0 || candidates[i]->centerX > IMAGE_WIDTH);
    //bool ballCenterYOffScreen =
    //(candidates[i]->centerY < 0 || candidates[i]->centerY > IMAGE_HEIGHT);
    float dist;
    if (ballCenterInBlob ||
        (decentCircleFit && ( // ball is most probably at the edge
                             // (ballCenterXInBlob && ballCenterYOffScreen) ||
                             // (ballCenterYInBlob && ballCenterXOffScreen)))) {
                             ballCenterXInBlob || ballCenterYInBlob))) {
      ball->imageCenterX = candidates[i]->centerX;
      ball->imageCenterY = candidates[i]->centerY;
      ball->visionConfidence -= candidates[i]->stddev;
      visionLog((34,   "Circle deviation[%i] = %f",sortIndex[i],candidates[i]->stddev));
      dist = GetDistanceByBlobWidth(2.0f*candidates[i]->radius);
      //float dist2 = GetDistanceByKinematics(ball->imageCenterX, ball->imageCenterY);
      //std::cout << "width dist:         " << dist << "\n";
      //std::cout << "kinematics dist:                                " << dist2 << "\n";

    } else { // ball center is not in blob
      // visionLog((20,   "bad circle center for orange-blob[%i]", sortIndex[i]));
      // ball->imageCenterX = (blob->xi + blob->xf) >> 1;
      // ball->imageCenterY = (blob->yi + blob->yf) >> 1;
      ball->visionConfidence = -1;
      // dist = GetDistanceByBlobWidth(blob->dx);
      visionLog((34,   "Ball center for candidate %i is not in blob", i));
      continue;
    }

    // check ball radius compared to blob width and height
    /* if (candidates[i]->radius < 0.33f * max(blob->dx,blob->dy)) {
       visionLog((20,  "Bad circle radius compared to blob size for orange-blob[%i], %f, %f",sortIndex[i], candidates[i]->radius, max(blob->dx,blob->dy)));
       continue;
       } */

    float bear = transform_.calculateBearing(ball->imageCenterX);
    // TODO remove this hack, added by sbarrett on 6-14-12 to account for differences in ball side positions when head is straight vs looking at the ball, help a bit, but still rough
    if(camera_ == Camera::BOTTOM) {
      bear *= 0.70;
    }
    float elev = transform_.calculateElevation(ball->imageCenterY);

    transform_.transformPolar(
                              dist, bear, elev, ball->visionDistance, ball->visionBearing, ball->visionElevation);
    ball->visionConfidence = 1.;

    // Horizon check because the Z height check doesn't work very well

    if(horizon_.exists && !horizon_.isAbovePoint(ball->imageCenterX, ball->imageCenterY + 10)) {
        visionLog((34, "Ball at %i is above horizon, throwing out."));
        continue;
    }

    // Height Check

    float ballZ = ball->visionDistance * tan(ball->visionElevation);
    if (ballZ > 500 || ballZ < -450) {
      ball->visionConfidence = -1;
      visionLog((34,  "Bad Z height %5.5f, distance = %f, elevation = %f", ballZ, ball->visionDistance, ball->visionElevation * RAD_T_DEG));
      continue;
    }

    // Elevation Check

    if (ball->visionElevation > 20 * DEG_T_RAD || ball->visionDistance > 8000) {
      ball->visionConfidence = -1;
      visionLog((34,  "Bad elevation %5.5f or distance %5.5f", ball->visionElevation, ball->visionDistance));
      continue;
    }

    // check if ball is in pink band
    if (!checkSurroundingPink(i)){
      ball->visionConfidence = -1;
      visionLog((34, "Too much pink in ball, throw out"));
      continue;
    }

    visionLog((34,"ball dx %i, dy %i", blob->dx, blob->dy));
    // check for green/white below ball
    // Todd: good balls always had at least 0.26 pct green/white
    // bad balls had as high as 0.125 pct
    // so being very conservative
    // also, this check will return that 1.0 if check was cutoff
    // so we'll accept all balls near bottom of image
    // Todd: only do this with small balls.. big ones have large shadows
    if (camera_ == Camera::TOP || (blob->dx < 55 && blob->dy < 55)){
      float belowGreenWhitePct = checkBelowGreenWhitePct(i);
      float minPct = .115;
      if (belowGreenWhitePct < minPct){
        ball->visionConfidence = -1;
        visionLog((34, "Ball had %5.3f pct green/white (minimum is %5.3f)", belowGreenWhitePct, minPct));
        continue;
      }
    }

    if (camera_ == Camera::TOP && (blob->dx > 85 || blob->dy > 85)) {
      ball->visionConfidence = -1;
      visionLog((34, "Ball is too big"));
      continue;
    }

    // Check percent of orange in fitted circle
    float circleOrangePct = checkColorsInCircleFit(i);
    // Todd: Even 0.1 will throw out most of the bad ones in legs/pink
    // But a few bands with a lot of orange in them were up around .45 and .57
    // And a few valid balls were as low as .45
    // I'm thinking we want to keep those valid ones, and just not let that
    // much pink get classified as orange.  But we could set this higher
    // (like to 0.6)
    if (circleOrangePct < 0.33f){
      ball->visionConfidence = -1;
      visionLog((34, "Bad orange pct in fitted circle: %5.5f", circleOrangePct));
      continue;
    }

    // if we get here, we have a ball!!!
    visionLog((34, "Candidate %i selected as ball", i));
    return setBestBallCandidate(sortIndex, ballCand, i);

  }

  return false;
}


/** checks the percent of pink between a given set of points */
float BallDetector::checkPinkPct(int x1, int x2, int y1, int y2){
  if (x1 < 0) x1 = 0;
  if (y1 < 0) y1 = 0;
  if (x2 >= IMAGE_WIDTH) x2 = IMAGE_WIDTH-1;
  if (y2 >= IMAGE_HEIGHT) y2 = IMAGE_HEIGHT-1;

  int numTotal = 0;
  int numPink = 0;
  for (int x = x1; x <= x2; x+=4){
    for (int y = y1; y <= y2; y+=2){
      numTotal++;
      int c = getSegPixelValueAt(x, y);
      if (c == c_PINK) numPink++;
    }
  }

  //visionLog((43, "check from x: %i, %i, and y: %i, %i, total: %i, pink: %i",
  //    x1, x2, y1, y2, numTotal, numPink));

  if (numTotal == 0)
    return 0.0;

  return (float)numPink/(float)numTotal;

}


/** Todd: this checks along 4 vectors from the ball for pink. */
bool BallDetector::checkSurroundingPink(int i){

  int roundX = 4;
  int roundY = 2;
  int pixToCheck = 10;

  int midY = ((int)(candidates[i]->centerY/roundY))*roundY;
  int midX = ((int)(candidates[i]->centerX/roundX))*roundX;

  int endX1 = ((int)((candidates[i]->centerX - candidates[i]->radius)/roundX))*roundX;
  int startX1 = endX1 - roundX * pixToCheck;

  float check1 = checkPinkPct(startX1, endX1, midY-roundY*2, midY+roundY*2);

  int startX2 = ((int)((candidates[i]->centerX + candidates[i]->radius)/roundX))*roundX;
  int endX2 = startX2 + roundX * pixToCheck;

  float check2 = checkPinkPct(startX2, endX2, midY-roundY*2, midY+roundY*2);

  int endY1 = ((int)((candidates[i]->centerY - candidates[i]->radius)/roundY))*roundY;
  int startY1 = endY1 - roundY * pixToCheck;

  float check3 = checkPinkPct(midX - roundX*2, midX+roundX*2, startY1, endY1);

  int startY2 = ((int)((candidates[i]->centerY + candidates[i]->radius)/roundY))*roundY;
  int endY2 = startY2 + roundY * pixToCheck;

  float check4 = checkPinkPct(midX - roundX*2, midX+roundX*2, startY2, endY2);

  visionLog((43, "Nearby pink %f, %f, %f, %f", check1, check2, check3, check4));

  // Todd: for good balls, these never got over 0.09
  // for orange in pink bands, there was always one over 0.16
  // going to set threshold at 0.15
  if (check1 > 0.15 || check2 > 0.15 || check3 > 0.15 || check4 > 0.15){
    return false;
  }

  return true;

}

/** Check the percantage of green/white below ball. */
float BallDetector::checkBelowGreenWhitePct(int i){
  int hstep, vstep;
  classifier_->getStepSize(hstep,vstep);
  int roundX = hstep;
  int roundY = vstep;

  int midX = ((int)(candidates[i]->centerX/roundX))*roundX;
  int midY = ((int)(candidates[i]->centerY/roundY))*roundY;

  int ymin = ((int)((candidates[i]->centerY + candidates[i]->radius)/roundY))*roundY;
  int ymax = ymin + candidates[i]->radius * 2;

  int xmin = ((int)((candidates[i]->centerX - candidates[i]->radius)/roundX))*roundX;
  int xmax = ((int)((candidates[i]->centerX + candidates[i]->radius)/roundX))*roundX;

  xmin = std::max(xmin, 0);
  xmax = std::min(xmax, IMAGE_WIDTH - 1);
  ymin = std::max(ymin, 0);
  ymax = std::min(ymax, IMAGE_HEIGHT - 1);

  int numTotal = 0;
  int numGreenWhite = 0;
  for (int x = xmin; x <= xmax; x += hstep){
    for (int y = ymin; y <= ymax; y += vstep){
      numTotal++;
      int c = getSegPixelValueAt(x, y);
      if (c == c_FIELD_GREEN || c == c_WHITE) numGreenWhite++;
    }
  }

  float pct = (float)numGreenWhite/(float)numTotal;
  if(!numTotal && ymin >= IMAGE_HEIGHT)
  // We can't see below the ball due to the height, so we assume 1. There
  // is no reason the robot should see anything else that's orange and that
  // extends vertically below the lower boundary of the image.
    pct = 1;

  visionLog((43, "check from x: %i, %i, and y: %i, %i, hstep: %i, vstep: %i, total: %i, greenwhite: %i, pct: %5.3f", xmin, xmax, ymin, ymax, hstep, vstep, numTotal, numGreenWhite, pct));

  return pct;
}


/** This checks the percentage of orange inside in the fitted ball circle
    (or, reality, inside the biggest square that can fit inside that circle.
    It helps prevent false balls when we fit orange to a small sliver of orange. */
float BallDetector::checkColorsInCircleFit(int i){
  // check pixels inside circle radius
  // whats the length of one side of the largest square inside the circle
  float squareWidth = sqrtf((4.0*candidates[i]->radius*candidates[i]->radius)/2.0);

  int roundX = 4;
  int roundY = 2;

  int startX = ((int)((candidates[i]->centerX - squareWidth/2.0)/roundX))*roundX;
  int startY = ((int)((candidates[i]->centerY - squareWidth/2.0)/roundY))*roundY;
  int endX = candidates[i]->centerX + squareWidth/2.0;
  int endY = candidates[i]->centerY + squareWidth/2.0;
  if (startX < 0) startX = 0;
  if (startY < 0) startY = 0;
  if (endX >= IMAGE_WIDTH) endX = IMAGE_WIDTH-1;
  if (endY >= IMAGE_HEIGHT) endY = IMAGE_HEIGHT-1;

  //visionLog((43, "ball radius scan from x: %i, %i, and y: %i, %i, radius %f, sqWidth %f",
  //         startX, endX, startY, endY, candidates[i]->radius, squareWidth));

  int totalPixelsChecked = 0;
  int numOrange = 0;
  //int numUndefined = 0;
  //int numPink = 0;
  for (int x = startX; x < endX; x+=4){
    for (int y = startY; y < endY; y+=2){
      totalPixelsChecked++;
      int c = getSegPixelValueAt(x, y);
      if (c == c_ORANGE) numOrange++;
      // if (c == c_PINK) numPink++;
      //if (c == c_UNDEFINED) numUndefined++;
    }
  }

  //visionLog((43, "ball radius scan total %i, orange %i, undef %i, pink %i",
  //          totalPixelsChecked, numOrange, numUndefined, numPink));
  if (totalPixelsChecked == 0) {
    return 0;
  }

  float pctOrange = (float)numOrange / (float)totalPixelsChecked;
  //visionLog((43, "pctOrange in circle fit %f", pctOrange));

  return pctOrange;
}


void BallDetector::fitCircleToPoints(
                                     uint16_t *x, uint16_t *y, uint16_t n,               // in
                                     float *cx, float *cy, float *radius, float *stddev) // out
{
  *cx = 0; *cy = 0; *radius = 0; *stddev = 1e6;
  if (n < 3) return; // ERROR: fewer than three points

  // computing centroids
  float centroidX = 0, centroidY = 0;
  for (uint16_t i = 0; i < n; i++) {
    centroidX += x[i]; centroidY += y[i]; }
  centroidX /= n; centroidY /= n;

  // computing moments
  float m_xx = 0, m_xy = 0, m_xz = 0, m_yy = 0, m_yz = 0, m_zz = 0;
  for (uint16_t i = 0; i < n; i++) {
    float xi = x[i] - centroidX, yi = y[i] - centroidY;
    float zi = xi * xi + yi * yi;
    m_xx += xi * xi; m_xy += xi * yi; m_xz += xi * zi;
    m_yy += yi * yi; m_yz += yi * zi; m_zz += zi * zi;
  }
  m_xx /= n; m_xy /= n; m_xz /= n; m_yy /= n; m_yz /= n; m_zz /= n;

  // computing the coefficients of the characteristic polynomial
  float m_z = m_xx + m_yy;
  float cov_xy = m_xx * m_yy - m_xy * m_xy;
  float a3 = 4 * m_z;
  float a2 = -3 * m_z * m_z - m_zz;
  float a1 = m_zz * m_z + 4 * cov_xy * m_z;
  a1 -= m_xz * m_xz + m_yz * m_yz + m_z * m_z * m_z;
  float a0 = m_xz * m_xz * m_yy + m_yz * m_yz * m_xx;
  a0 -= m_zz * cov_xy + 2 * m_xz * m_yz * m_xy - m_z * m_z * cov_xy;
  float a22 = a2 + a2;
  float a33 = a3 + a3 + a3;
  float newX = 0;
  float newY = 1e20;
  const float eps = 1e-12;
  const size_t MAX_ITER = 20;

  // newton's method starting at x = 0
  for (size_t i = 0; i < MAX_ITER; i++) {
    float oldY = newY;
    newY = a0 + newX * (a1 + newX * (a2 + newX * a3));
    if (fabs(newY) > fabs(oldY))
      return; // ERROR: newton-taubin going wrong dir.
    float dy = a1 + newX * (a22 + newX * a33);
    float oldX = newX;
    newX = oldX - newY / dy;
    if (fabs((newX - oldX) / newX) < eps)
      break; // converged!
    if (newX < 0)
      newX = 0; // ERROR: newton-taubin having neg. root
  }

  // computing the circle parameters
  float det = 2 * (newX * newX - newX * m_z + cov_xy);
  if (fabs(det) < 1e-3) return; // ERROR: zero determinant

  *cx = (m_xz * (m_yy - newX) - m_yz * m_xy) / det;
  *cy = (m_yz * (m_xx - newX) - m_xz * m_xy) / det;
  *radius = sqrtf(*cx * *cx + *cy * *cy + m_z);
  *cx += centroidX; *cy += centroidY;

  // compute standard deviation with a normalized circle (r = 1)
  float m = 0, v = 0;
  for (uint16_t i = 0; i < n; i++) {
    float dx = x[i] - *cx, dy = y[i] - *cy;
    float d = sqrtf(dx * dx + dy * dy) / *radius;
    v += d * d; m += d;
  }
  m /= n; v /= n;
  *stddev = sqrtf(v - m * m);
}

void BallDetector::getBlobContour(
                                  uint16_t blobIndex, uint16_t *mergeIndex,                       // in
                                  uint16_t *xinit, uint16_t *xfinal)                                     // out
{
  int hscale, vscale, hstep, vstep;
  classifier_->getStepScale(hscale,vscale);
  classifier_->getStepSize(hstep,vstep);
  // initialize output
  for (uint16_t y = 0; y < IMAGE_HEIGHT; y += vstep) {
    xinit[y >> vscale] = (uint16_t)-1; xfinal[y >> vscale] = 0; }

  // insert first blob's contour
  const Blob *bli = &horizontalBlob[c_ORANGE][blobIndex];
  for (uint16_t i = 0; i < bli->lpCount; i++) {
    uint16_t index = bli->lpIndex[i] & 0xffff;
    uint16_t y = bli->lpIndex[i] >> 16; // index
    VisionPoint *lp = &classifier_->horizontalPoint[c_ORANGE][y][index];
    xinit[y >> vscale] = std::min(xinit[y >> vscale], lp->xi);    // coord
    xfinal[y >> vscale] = std::max(xfinal[y >> vscale], lp->xf);
    //std::cout << "set xinit[" << (y >> vscale) << "] to " << xinit[y >> vscale] << ", xfinal[" << (y >> vscale) << "] to " << xfinal[y >> vscale] << "\n";

    }  // coord

  // insert merged blobs' contours
  for (uint16_t i = 0; i < horizontalBlobCount[c_ORANGE]; i++) {
    if (mergeIndex[i] != blobIndex)
      continue;
    Blob *blj = &horizontalBlob[c_ORANGE][i];
    for (uint16_t j = 0; j < blj->lpCount; j++) {
      uint16_t index = blj->lpIndex[j] & 0xffff;
      uint16_t y = blj->lpIndex[j] >> 16; // index
      VisionPoint *lp = &classifier_->horizontalPoint[c_ORANGE][y][index];
      xinit[y >> vscale] = std::min(xinit[y >> vscale], lp->xi);    // coord
      xfinal[y >> vscale] = std::max(xfinal[y >> vscale], lp->xf); }} // coord
}

void BallDetector::fitCircleToBallCand(
                                       uint16_t ballCandIndex, uint16_t blobIndex, uint16_t *mergeIndex)
{
  int hscale, vscale;
  classifier_->getStepScale(hscale,vscale);
  int size = IMAGE_HEIGHT >> vscale;
  // get ball candidate's contour
  uint16_t xi[size], xf[size];
  getBlobContour(blobIndex, mergeIndex, xi, xf);
  //for(int i = 0; i < size; i++)
    //std::cout << xi[i] << ",";
  //std::cout << "\n";

  //for(int i = 0; i < size; i++)
    //std::cout << xf[i] << ",";
  //std::cout << "\n";

  // format full, left, right contour for circle fitting
  uint16_t contFX[IMAGE_HEIGHT     ], contFY[IMAGE_HEIGHT     ]; // full
  uint16_t contFCount = 0;
  uint16_t contLX[size], contLY[size]; // left
  uint16_t contLCount = 0;
  uint16_t contRX[size], contRY[size]; // rite
  uint16_t contRCount = 0;
  const Blob *bl = &horizontalBlob[c_ORANGE][blobIndex];
  for (uint16_t y = 0; y < size; y++) {
    if (xi[y] == (uint16_t)-1 && xf[y] == 0) continue;
    if (bl->xi > 8) {
      contFX[contFCount] = contLX[contLCount] = xi[y];
      contFY[contFCount] = contLY[contLCount] = y << vscale;
      contFCount++; contLCount++; }
    if (bl->xf < IMAGE_WIDTH - 8) {
      contFX[contFCount] = contRX[contRCount] = xf[y];
      contFY[contFCount] = contRY[contRCount] = y << vscale;
      contFCount++; contRCount++; }}

  // only consider upper half of the contour if too close to body
  if (bl->dx > 40 && bl->yi > size) {
    contFCount = 3 * contFCount / 5;
    contLCount = 3 * contLCount / 5;
    contRCount = 3 * contRCount / 5; }

  // fit circle
  float cxF, cyF, rF, sdF;
  for(int i = 0; i < contFCount; i++) {
    //std::cout << "contfx,contfy: (" << contFX[i] << "," << contFY[i] << ")\n";
  }                              // full
  fitCircleToPoints(contFX, contFY, contFCount, &cxF, &cyF, &rF, &sdF);
  //std::cout << "fit points 1: (cx,cy), r, sd: " << "(" << cxF << "," << cyF << "), " << rF << ", " << sdF << "\n";
  if (2 * bl->dy > 3 * bl->dx) {
    float cxL, cyL, rL, sdL;                                       // left
    fitCircleToPoints(contLX, contLY, contLCount, &cxL, &cyL, &rL, &sdL);
    float cxR, cyR, rR, sdR;                                       // rite
    fitCircleToPoints(contRX, contRY, contRCount, &cxR, &cyR, &rR, &sdR);

    // set final ball candidate circle fit values
    //const uint16_t cxBlob = (bl->xi + bl->xf) >> 1;
    if (cxR > bl->xf) {
      visionLog((15,  "<<< LEFT contour chosen !! >>>"));
      candidates[ballCandIndex]->centerX = cxL;
      candidates[ballCandIndex]->centerY = cyL;
      candidates[ballCandIndex]->radius = rL;
      candidates[ballCandIndex]->stddev = sdL;
      return;
    } else if (cxL < bl->xi) {
      visionLog((15,   "<<< RITE contour chosen !! >>>"));
      candidates[ballCandIndex]->centerX = cxR;
      candidates[ballCandIndex]->centerY = cyR;
      candidates[ballCandIndex]->radius = rR;
      candidates[ballCandIndex]->stddev = sdR;
      return;
    }
  }
  visionLog((15,  "<<< FULL contour chosen: (cx,cy) = (%f,%f), rad = %f, stddev = %f !! >>>", cxF, cyF, rF, sdF));
  candidates[ballCandIndex]->centerX = cxF;
  candidates[ballCandIndex]->centerY = cyF;
  candidates[ballCandIndex]->radius = rF;
  candidates[ballCandIndex]->stddev = sdF;
}

void BallDetector::fitCircleToBallCands(
                                        uint16_t *sortIndex, uint16_t *mergeIndex) {

  for (uint16_t ballCandIndex = 0;
       ballCandIndex < ballCandCount; ballCandIndex++) {

    fitCircleToBallCand(
                        ballCandIndex, sortIndex[ballCandIndex], mergeIndex);

  }
}

bool BallDetector::setBestBallCandAsBall(
                                         uint16_t *sortIndex, WorldObject *ballCand) {

  WorldObject *ball = &world_objects_->objects_[WO_BALL];
  //ball->reset();

  // Find
  uint16_t maxConfIndex = (uint16_t)-1;
  float maxConf = -1e6;
  for (uint16_t i = 0; i < ballCandCount; i++) {
    float conf = ballCand[i].visionConfidence;
    if (conf > maxConf) {
      maxConf = conf;
      maxConfIndex = i;
    }
  }

  return setBestBallCandidate(sortIndex, ballCand, maxConfIndex);
}


bool BallDetector::setBestBallCandidate(uint16_t *sortIndex, WorldObject *ballCand, int index){

  //TODO get rid of warning
  sortIndex = sortIndex;

  WorldObject *ball = &world_objects_->objects_[WO_BALL];
  if(ball->seen && !ball->fromTopCamera && camera_ == Camera::TOP) {
    visionLog((34, "Ball was already selected by the bottom camera, throwing out."));
    return false; // Always take balls from the bottom camera when available
  }
  //ball->reset();

  // check if ball is way off field (like a kid in an orange shirt at the US open)
  WorldObject *bestBallCand = &ballCand[index];
  if (bestBallCand->visionDistance > 2000){
    WorldObject* self = &(world_objects_->objects_[robot_state_->WO_SELF]);
    Point2D relBall = Point2D(bestBallCand->visionDistance,bestBallCand->visionBearing, POLAR);
    Point2D absBall = relBall.relativeToGlobal(self->loc, self->orientation);
    if (fabs(absBall.x) > ((FIELD_X / 2.0) + 2000) ||
        fabs(absBall.y) > ((FIELD_Y / 2.0) + 2000)){
      visionLog((34, " Ball thrown out for being too far off field, %f, %f",
                 absBall.x, absBall.y));
      return false;
    }
  }

  // Set
  if (bestBallCand->visionConfidence > 0) {
    ball->reset();
    ball->imageCenterX = bestBallCand->imageCenterX;
    ball->imageCenterY = bestBallCand->imageCenterY;
    ball->visionDistance = bestBallCand->visionDistance;
    ball->visionBearing = bestBallCand->visionBearing;
    ball->visionElevation = bestBallCand->visionElevation;
    ball->seen = true;
    ball->visionConfidence = bestBallCand->visionConfidence;
    ball->frameLastSeen = vision_frame_info_->frame_id;
    ball->ballBlobIndex = bestBallCand->ballBlobIndex;
    ball->fromTopCamera = (camera_ == Camera::TOP);

    bestBallCandIndex = index;
    return true;
  }
  else
    return false;
}

void BallDetector::sortOrangeBlobs(uint16_t *mergeIndex, uint16_t *sortIndex) {
  ballCandCount = 0;
  for (uint16_t i = 0; i < horizontalBlobCount[c_ORANGE]; i++)
    if (mergeIndex[i] == MAX_ORANGE_BLOBS)
      sortIndex[ballCandCount++] = i;
  for (uint16_t i = 0; i < ballCandCount - 1; i++) {
    Blob *bli = &horizontalBlob[c_ORANGE][sortIndex[i]];
    uint32_t areai = bli->dx * bli->dy;
    for (uint16_t j = i + 1; j < ballCandCount; j++) {
      Blob *blj = &horizontalBlob[c_ORANGE][sortIndex[j]];
      uint32_t areaj = blj->dx * blj->dy;
      if (areaj > areai) {
        uint16_t temp = sortIndex[i];
        sortIndex[i] = sortIndex[j];
        sortIndex[j] = temp;
        areai = areaj;
      }
    }
  }
  ballCandCount = std::min(ballCandCount, (uint16_t)MAX_BALL_CANDS);
}

float BallDetector::GetDistanceByBlobWidth(float dx) {
  return 16765.0f * powf(dx, -0.769f);
}
// DO NOT REMOVE
// vim: expandtab:noai:sts=2:sw=2:ts=2
float BallDetector::GetDistanceByKinematics(int imageX, int imageY) {
    Point2D proj = transform_.forwardProject(imageX, imageY);
    float z = cameraZ();
    //if(camera_ == Camera::TOP) std::cout << "top\n";
    //else std::cout << "bottom:\n";
    //std::cout << "x,y,z: " << proj.x << "," << proj.y << "," << z << "\n";
    float distance = sqrt(pow(proj.x,2) + pow(proj.y,2) + pow(z,2));
    return distance;
}

float BallDetector::cameraZ() {
    if(camera_ == Camera::TOP)
        return body_model_->abs_parts_[BodyPart::top_camera].translation.z;
    else
        return body_model_->abs_parts_[BodyPart::bottom_camera].translation.z;
}

bool BallDetector::bestCandidateFound() {
    return (bestBallCandIndex != ((uint16_t)-1));
}

void BallDetector::setHorizon(HorizonLine horizon) {
    horizon_ = horizon;
}
