#include "BlobDetector.h"

BlobDetector::BlobDetector(
                           RobotVisionBlock*& robot_vision_block,
                           ImageBlock*& image_block,
                           Classifier*& classifier,
                           VisionParams& params,
                           Camera::Type& camera)
  :   robot_vision_(robot_vision_block),
      image_(image_block),
      classifier_(classifier),
      params_(params),
      camera_(camera) {

  verticalPoint = classifier_->verticalPoint;
  verticalPointCount = classifier_->verticalPointCount;
  horizontalPoint = classifier_->horizontalPoint;
  horizontalPointCount = classifier_->horizontalPointCount;

  //Blob horizontalBlob[NUM_COLORS][MAX_LINE_BLOBS];
  horizontalBlob = new Blob*[NUM_COLORS];
  for(int i = 0; i < NUM_COLORS; i++)
    horizontalBlob[i] = new Blob[MAX_LINE_BLOBS];

  //Blob verticalBlob[NUM_COLORS][MAX_LINE_BLOBS];
  verticalBlob = new Blob*[NUM_COLORS];
  for(int i = 0; i < NUM_COLORS; i++)
    verticalBlob[i] = new Blob[MAX_LINE_BLOBS];

  horizontalBlobCount = new uint32_t[NUM_COLORS];
  verticalBlobCount = new uint32_t[NUM_COLORS];
      }

void BlobDetector::setImagePointers() {
  if(camera_ == Camera::TOP) {
    segImg_ = robot_vision_->segImgTop;
    img_ = image_->img_top_.get();
    rawImageExists_ = (bool)image_->img_top_;
  }
  else {
    segImg_ = robot_vision_->segImgBottom;
    img_ = image_->img_bottom_.get();
    rawImageExists_ = (bool)image_->img_bottom_;
  }
}

void BlobDetector::formAllBlobs() {

  unsigned char horzColors[] = {
    c_WHITE,
    c_BLUE
  };
  formBlobsWithHorzLineSegs(horzColors, 2);

  // unsigned char horzColors[] = {
  //   c_WHITE,
  //   c_BLUE,
  //   c_YELLOW
  // };
  // formBlobsWithHorzLineSegs(horzColors, 3);

  unsigned char vertColors[] = {
    c_WHITE,
    c_BLUE,
    c_PINK
  };
  formBlobsWithVertLineSegs(vertColors, 3);

  //formOrangeBlobs();

}

void BlobDetector::resetOrangeBlobs() {
  horizontalBlobCount[c_ORANGE] = 0;
}

void BlobDetector::resetYellowBlobs() {
  horizontalBlobCount[c_YELLOW] = 0;
}

void BlobDetector::formYellowBlobs() {

  unsigned char horzColors[] = {
    c_YELLOW
  };

  formBlobsWithHorzLineSegs(horzColors, 1);
}

void BlobDetector::formBlobsWithHorzLineSegs(unsigned char * horzColors, int numColors) {
  int hstep, vstep;
  classifier_->getStepSize(hstep,vstep);
  for (int c = 0; c < numColors; c++) {
    unsigned char color = horzColors[c];
    horizontalBlobCount[color] = 0;
    for (uint32_t y = 0; y < IMAGE_HEIGHT - vstep; y += vstep) {
      for (uint16_t i = 0; i < horizontalPointCount[color][y]; i++) {
        if (!horizontalPoint[color][y][i].isValid)
          continue;
        // Check for overlapping line segments with similar width
        uint16_t j;
        for (j = 0; j < horizontalPointCount[color][y + vstep]; j++) {
          if (!horizontalPoint[color][y + vstep][j].isValid)
            continue;
          bool overlap =
            (horizontalPoint[color][y][i].xi < horizontalPoint[color][y + vstep][j].xf &&
             horizontalPoint[color][y][i].xf > horizontalPoint[color][y + vstep][j].xi);
          bool similarWidth = (
                               (horizontalPoint[color][y][i].dx >= horizontalPoint[color][y + vstep][j].dx &&
                                ((params_.blobWidthRatio[color] * horizontalPoint[color][y][i].dx <=
                                  8 * horizontalPoint[color][y + vstep][j].dx) ||
                                 (horizontalPoint[color][y][i].dx - horizontalPoint[color][y + vstep][j].dx <= params_.blobWidthAbs[color]))) ||
                               (horizontalPoint[color][y][i].dx <= horizontalPoint[color][y + vstep][j].dx &&
                                ((8 * horizontalPoint[color][y][i].dx >=
                                  params_.blobWidthRatio[color] * horizontalPoint[color][y + vstep][j].dx) ||
                                 (horizontalPoint[color][y + vstep][j].dx - horizontalPoint[color][y][i].dx <= params_.blobWidthAbs[color]))));
          if (overlap && (!params_.blobUseWidthRatioHorz[color] || similarWidth))
            break;
        }

        // If no overlapping line segments, continue
        if (j == horizontalPointCount[color][y + vstep])
          continue;

        // If line segment belongs to an existing field line blob
        if (y > 0 && horizontalPoint[color][y][i].lbIndex != (uint16_t)-1) {
          Blob *bl = &horizontalBlob[color][horizontalPoint[color][y][i].lbIndex];
          bl->lpIndex[bl->lpCount++] = (y + vstep) << 16 | j;
          horizontalPoint[color][y + vstep][j].lbIndex = horizontalPoint[color][y][i].lbIndex;

          // Otherwise, create a new field line blob
        } else if (horizontalBlobCount[color] < MAX_LINE_BLOBS - 1) {
          Blob *bl = &horizontalBlob[color][horizontalBlobCount[color]];
          bl->lpCount = 0;
          bl->lpIndex[bl->lpCount++] =  y      << 16 | i;
          bl->lpIndex[bl->lpCount++] = (y + vstep) << 16 | j;
          horizontalPoint[color][y    ][i].lbIndex =
            horizontalPoint[color][y + vstep][j].lbIndex = horizontalBlobCount[color]++;
          // bound check
        } else {
          visionLog((30, "****MAX_LINE_BLOBS exceeded while forming blobs. Please increase in Vision.h"));
        }

      }
    }
  }
}

void BlobDetector::formBlobsWithVertLineSegs(unsigned char * vertColors, int numColors) {
  int hstep, vstep;
  classifier_->getStepSize(hstep, vstep);
  for (uint16_t c = 0; c < numColors; c++) {

    unsigned char color = vertColors[c];
    verticalBlobCount[color] = 0;
    // Form the actual blob
    for (uint32_t x = 0; x < IMAGE_WIDTH - hstep; x += hstep) {
        //std::cout << "-- x:" << x / hstep << ";";
      for (uint16_t i = 0; i < verticalPointCount[color][x]; i++) {
        if (!verticalPoint[color][x][i].isValid)
          continue;
        uint16_t j = 0;
        // Check for overlapping line segments with similar width
        for (; j < verticalPointCount[color][x + hstep]; j++) {
            //std::cout << "1;";
          if (!verticalPoint[color][x + hstep][j].isValid)
            continue;
            //std::cout << "2;";
            //if(color == c_WHITE) std::cout << "(" << x / hstep << "," << i << ":" << verticalPoint[color][x][i].yi << "," << verticalPoint[color][x][i].yf << "):(" << (x / hstep + 1) << "," << j << ":" << verticalPoint[color][x + hstep][j].yi << "," << verticalPoint[color][x + hstep][j].yf << ");";
          bool overlap =
            (verticalPoint[color][x][i].yi < verticalPoint[color][x + hstep][j].yf &&
             verticalPoint[color][x][i].yf > verticalPoint[color][x + hstep][j].yi);
          bool similarWidth = (
                               (verticalPoint[color][x][i].dy >= verticalPoint[color][x + hstep][j].dy &&
                                ((params_.blobWidthRatio[color] * verticalPoint[color][x][i].dy <=
                                  8 * verticalPoint[color][x + hstep][j].dy) ||
                                 (verticalPoint[color][x][i].dy - verticalPoint[color][x + hstep][j].dy <= params_.blobWidthAbs[color]))) ||
                               (verticalPoint[color][x][i].dy <= verticalPoint[color][x + hstep][j].dy &&
                                ((8 * verticalPoint[color][x][i].dy >=
                                  params_.blobWidthRatio[color] * verticalPoint[color][x + hstep][j].dy) ||
                                 (verticalPoint[color][x + hstep][j].dy - verticalPoint[color][x][i].dy <= params_.blobWidthAbs[color]))));
                                 //std::cout << "sim: " << similarWidth << ";over: " << overlap << ";";
          if (overlap && (!params_.blobUseWidthRatioVert[color] || similarWidth))
            break;
        }
        // If no overlapping line segments, continue
        if (j == verticalPointCount[color][x + hstep]) {
          j = 0;
          continue;
        }
        //std::cout << "3;";
        // If line segment belongs to an existing field line blob
        if (x > 0 && verticalPoint[color][x][i].lbIndex != (uint16_t)-1) {
          Blob *lb =
            &verticalBlob[color][verticalPoint[color][x][i].lbIndex];
          lb->lpIndex[lb->lpCount++] = (x + hstep) << 16 | j;
          verticalPoint[color][x + hstep][j].lbIndex =
            verticalPoint[color][x][i].lbIndex;
            //std::cout << "4;";
          // Otherwise, create a new field line blob
        } else if (verticalBlobCount[color] < MAX_LINE_BLOBS - 1){
            //std::cout << "st:(" << x / hstep << "," << i << ") ,, ";
          Blob *lb = &verticalBlob[color][verticalBlobCount[color]];
          lb->lpCount = 0;
          lb->lpIndex[lb->lpCount++] = x << 16 | i;
          lb->lpIndex[lb->lpCount++] = (x + hstep) << 16 | j;
          verticalPoint[color][x][i].lbIndex =
            verticalPoint[color][x + hstep][j].lbIndex = verticalBlobCount[color]++;
        } else {
          visionLog((30, "****MAX_LINE_BLOBS exceeded while forming blobs. Please increase in Vision.h"));
        }
      }
    }
  }
  //std::cout << "\n";
}


void BlobDetector::CalHorzBlobData (unsigned char color) {

  // Calculating Data for each blob
  for (uint16_t i = 0; i < verticalBlobCount[color]; i++) {
    Blob *lbi = &verticalBlob[color][i];
    //if(color == c_WHITE) std::cout << i << ": " << (lbi->lpIndex[0] >> 18) << ",";
    lbi->invalid = false;

    // To Calculate the start and end slopes, we need to avergae out a few values to get correct results.
    uint16_t amountAverage = (2 * lbi->lpCount) / 3;
    amountAverage = (amountAverage > 5) ? 5 : amountAverage;

    // Start Slope Initial Point
    VisionPoint *lp0 = &verticalPoint[color][lbi->lpIndex[0] >> 16][lbi->lpIndex[0] & 0xfffful];
    uint16_t y0 = (lp0->yi + lp0->yf) / 2, x0 = lp0->xi;

    visionLog((45,  "Blob %i: (%i, %i)", i, x0, y0));

    // Start Slope Finish Point
    VisionPoint *lp1 = &verticalPoint[color][lbi->lpIndex[amountAverage] >> 16][lbi->lpIndex[amountAverage] & 0xfffful];
    uint16_t y1 = (lp1->yi + lp1->yf) / 2, x1 = lp1->xi;

    // End Slope Start Point
    VisionPoint *lp2 = &verticalPoint[color][lbi->lpIndex[lbi->lpCount - 1 - amountAverage] >> 16][lbi->lpIndex[lbi->lpCount - 1 - amountAverage] & 0xfffful];
    uint16_t y2 = (lp2->yi + lp2->yf) / 2, x2 = lp2->xi;

    // End Slope Finish Point
    VisionPoint *lp3 = &verticalPoint[color][lbi->lpIndex[lbi->lpCount - 1] >> 16][lbi->lpIndex[lbi->lpCount - 1] & 0xfffful];
    uint16_t y3 = (lp3->yi + lp3->yf) / 2, x3 = lp3->xi;

    // Calculate Final Values;
    lbi->diffStart = ((float) (y1 - y0)) / (x1 - x0);
    lbi->diffEnd = ((float) (y3 - y2)) / (x3 - x2);

    if (amountAverage == lbi->lpCount - 1) {
      lbi->doubleDiff = 0;
    } else {
      lbi->doubleDiff = (lbi->diffEnd - lbi->diffStart) / (x2 - x0);
    }

    // Calculate width at the end
    amountAverage = (lbi->lpCount > 5) ? 5 : lbi->lpCount - 1;
    uint16_t width = 0;
    for (uint16_t point = 0; point <= amountAverage; point++) {
      uint16_t yi = verticalPoint[color][lbi->lpIndex[point] >> 16][lbi->lpIndex[point] & 0xfffful].yi;
      uint16_t yf = verticalPoint[color][lbi->lpIndex[point] >> 16][lbi->lpIndex[point] & 0xfffful].yf;
      width += yf - yi;
    }
    width /= amountAverage + 1;
    lbi->widthStart = width;

    // Calculate width at the end
    width = 0;
    for (uint16_t point = lbi->lpCount - amountAverage - 1; point < lbi->lpCount; point++) {
      uint16_t yi = verticalPoint[color][lbi->lpIndex[point] >> 16][lbi->lpIndex[point] & 0xfffful].yi;
      uint16_t yf = verticalPoint[color][lbi->lpIndex[point] >> 16][lbi->lpIndex[point] & 0xfffful].yf;
      width += yf - yi;
    }
    width /= amountAverage + 1;
    lbi->widthEnd = width;

    lbi->xi = x0;
    lbi->yi = y0;
    lbi->xf = x3;
    lbi->yf = y3;

    visionLog((45,  "Blob %i:- Starts at (%i, %i), Ends at (%i, %i), Num Line Points: %i, Calculate Values = (%f, %f, %f), width = (%i, %i)", i, x0, y0, x3, y3, lbi->lpCount, lbi->diffStart, lbi->diffEnd, lbi->doubleDiff, lbi->widthStart, lbi->widthEnd));

  }
  //std::cout << "\n";

}

void BlobDetector::CalVertBlobData(unsigned char color) {

  // Calculating Data for each blob
  for (uint16_t i = 0; i < horizontalBlobCount[color]; i++) {

    Blob *lbi = &horizontalBlob[color][i];
    lbi->invalid = false;

    // To Calculate the start and end slopes, we need to avergae out a few values to get correct results.
    uint16_t amountAverage = (2 * lbi->lpCount) / 3;
    amountAverage = (amountAverage > 5) ? 5 : amountAverage;

    // Start Slope Initial Point
    VisionPoint *lp0 = &horizontalPoint[color][lbi->lpIndex[0] >> 16][lbi->lpIndex[0] & 0xfffful];
    uint16_t y0 = lp0->yi, x0 = (lp0->xi + lp0->xf) / 2;

    // Start Slope Finish Point
    VisionPoint *lp1 = &horizontalPoint[color][lbi->lpIndex[amountAverage] >> 16][lbi->lpIndex[amountAverage] & 0xfffful];
    uint16_t y1 = lp1->yi, x1 = (lp1->xi + lp1->xf) / 2;

    // End Slope Start Point
    VisionPoint *lp2 = &horizontalPoint[color][lbi->lpIndex[lbi->lpCount - 1 - amountAverage] >> 16][lbi->lpIndex[lbi->lpCount - 1 - amountAverage] & 0xfffful];
    uint16_t y2 = lp2->yi, x2 = (lp2->xi + lp2->xf) / 2;

    // End Slope Finish Point
    VisionPoint *lp3 = &horizontalPoint[color][lbi->lpIndex[lbi->lpCount - 1] >> 16][lbi->lpIndex[lbi->lpCount - 1] & 0xfffful];
    uint16_t y3 = lp3->yi, x3 = (lp3->xi + lp3->xf) / 2;

    // Calculate Final Values;
    lbi->diffStart = ((float) (x1 - x0)) / (y1 - y0);
    lbi->diffEnd = ((float) (x3 - x2)) / (y3 - y2);

    if (amountAverage == lbi->lpCount - 1) {
      lbi->doubleDiff = 0;
    } else {
      lbi->doubleDiff = (lbi->diffEnd - lbi->diffStart) / (y2 - y1);
    }

    // Calculate width at the end
    amountAverage = (lbi->lpCount > 5) ? 5 : lbi->lpCount - 1;
    uint16_t width = 0;
    for (uint16_t point = 0; point <= amountAverage; point++) {
      uint16_t xi = horizontalPoint[color][lbi->lpIndex[point] >> 16][lbi->lpIndex[point] & 0xfffful].xi;
      uint16_t xf = horizontalPoint[color][lbi->lpIndex[point] >> 16][lbi->lpIndex[point] & 0xfffful].xf;
      width += xf - xi;
    }
    width /= amountAverage + 1;
    lbi->widthStart = width;

    // Calculate width at the end
    width = 0;
    for (uint16_t point = lbi->lpCount - amountAverage - 1; point < lbi->lpCount; point++) {
      uint16_t xi = horizontalPoint[color][lbi->lpIndex[point] >> 16][lbi->lpIndex[point] & 0xfffful].xi;
      uint16_t xf = horizontalPoint[color][lbi->lpIndex[point] >> 16][lbi->lpIndex[point] & 0xfffful].xf;
      width += xf - xi;
    }
    width /= amountAverage + 1;
    lbi->widthEnd = width;

    lbi->xi = x0;
    lbi->yi = y0;
    lbi->xf = x3;
    lbi->yf = y3;

    visionLog((45,  "Blob %i: (%i, %i) - %f, %f, %f, %i, %i", i, x0, y0, lbi->diffStart, lbi->diffEnd, lbi->doubleDiff, lbi->widthStart, lbi->widthEnd));
  }
}


void BlobDetector::HorizontalBlobSort(unsigned char color) {
  // Selection Sort
  for (uint16_t i = 0; i < verticalBlobCount[color]; i++) {
    Blob *minIndex = &verticalBlob[color][i];
    VisionPoint *lp = &verticalPoint[color][minIndex->lpIndex[0] >> 16][minIndex->lpIndex[0] & 0xfffful];
    uint32_t minVal = lp->xi * IMAGE_HEIGHT + ((lp->yi + lp->yf) / 2);
    bool isOrigValue = true;
    for (uint16_t j = i + 1; j < verticalBlobCount[color]; j++) {
      Blob *tempIndex = &verticalBlob[color][j];
      lp = &verticalPoint[color][tempIndex->lpIndex[0] >> 16][tempIndex->lpIndex[0] & 0xfffful];
      uint32_t tempVal = lp->xi * IMAGE_HEIGHT + ((lp->yi + lp->yf) / 2);
      if (tempVal < minVal) {
        minIndex = tempIndex;
        minVal = tempVal;
        isOrigValue = false;
      }
    }
    if (!isOrigValue) {
      Blob temp = *minIndex;
      *minIndex = verticalBlob[color][i];
      verticalBlob[color][i] = temp;
    }
  }
}

void BlobDetector::VerticalBlobSort(unsigned char color) {
  // Selection Sort
  for (uint16_t i = 0; i < horizontalBlobCount[color]; i++) {
    Blob *minIndex = &horizontalBlob[color][i];
    VisionPoint *lp = &horizontalPoint[color][minIndex->lpIndex[0] >> 16][minIndex->lpIndex[0] & 0xfffful];
    uint32_t minVal = ((lp->xi + lp->xf) / 2) + (lp->yi) * IMAGE_WIDTH;
    bool isOrigValue = true;
    for (uint16_t j = i + 1; j < horizontalBlobCount[color]; j++) {
      Blob *tempIndex = &horizontalBlob[color][j];
      lp = &horizontalPoint[color][tempIndex->lpIndex[0] >> 16][tempIndex->lpIndex[0] & 0xfffful];
      uint32_t tempVal = ((lp->xi + lp->xf) / 2) + (lp->yi) * IMAGE_WIDTH;
      if (tempVal < minVal) {
        minIndex = tempIndex;
        minVal = tempVal;
        isOrigValue = false;
      }
    }
    if (!isOrigValue) {
      Blob temp = *minIndex;
      *minIndex = horizontalBlob[color][i];
      horizontalBlob[color][i] = temp;
    }
  }
}


void BlobDetector::formOrangeBlobs() {

  // print all the line segments
  for (int i = 0; i < IMAGE_HEIGHT; i++) {
    for (uint32_t j = 0; j < horizontalPointCount[c_ORANGE][i]; j++) {
      VisionPoint *lp = &horizontalPoint[c_ORANGE][i][j];
      //visionLog((20, "(%i, %i, %i)", lp->yi, lp->xi, lp->xf));
    }
  }
  /*
    unsigned char orangeColor = c_ORANGE;
    formBlobsWithHorzLineSegs(&orangeColor, 1);*/
  horizontalBlobCount[c_ORANGE] = 0;

  int hstep, vstep;
  classifier_->getStepSize(hstep,vstep);

  for (uint16_t y = 0; y < IMAGE_HEIGHT - vstep; y += vstep){
    for (uint16_t i = 0; i < horizontalPointCount[c_ORANGE][y]; i++) {
      VisionPoint *lpi = &horizontalPoint[c_ORANGE][y][i];
      //std::cout << lpi->xi << "," << lpi->xf << ",";
      for (uint16_t j = 0; j < horizontalPointCount[c_ORANGE][y + vstep]; j++) {
        VisionPoint *lpj = &horizontalPoint[c_ORANGE][y + vstep][j];

        // if line segments overlap
        if (lpi->xi < lpj->xf && lpi->xf > lpj->xi) {

          // if line segment belongs to an existing blob
          if (y > 0 && lpi->lbIndex != (uint16_t)-1) {
            Blob *bl = &horizontalBlob[c_ORANGE][lpi->lbIndex];
            if(bl->lpCount >= IMAGE_WIDTH) {
                visionLog((30, "****Max line points exceeded while forming orange blobs. Please increase in Blob.h"));
            }
            else {
                bl->lpIndex[bl->lpCount++] = (y + vstep) << 16 | j;
                lpj->lbIndex = lpi->lbIndex;
            }

            // otherwise, create a new blob
          } else if (horizontalBlobCount[c_ORANGE] < MAX_LINE_BLOBS - 1) {
            Blob *bl = &horizontalBlob[c_ORANGE][horizontalBlobCount[c_ORANGE]];
            bl->lpCount = 0;
            bl->lpIndex[bl->lpCount++] =  y      << 16 | i;
            bl->lpIndex[bl->lpCount++] = (y + vstep) << 16 | j;
            lpi->lbIndex = lpj->lbIndex = horizontalBlobCount[c_ORANGE]++;
            // bound check
          } else {
            visionLog((30, "****MAX_LINE_BLOBS exceeded while forming orange blobs. Please increase in Vision.h"));
          }
        }
      }
    }
  }
}

void BlobDetector::calculateOrangeBlobData() {

  // Compute each blob's xi, xf, dx, yi, yf, dy
  for (unsigned int i = 0; i < horizontalBlobCount[c_ORANGE]; i++) {
    Blob *bl = &horizontalBlob[c_ORANGE][i];
    int y = bl->lpIndex[0] >> 16;
    int k = bl->lpIndex[0] & 0xffff;
    bl->xi = horizontalPoint[c_ORANGE][y][k].xi;
    bl->xf = horizontalPoint[c_ORANGE][y][k].xf;
    bl->correctPixelRatio = horizontalPoint[c_ORANGE][y][k].dx;
    for (int j = 1; j < bl->lpCount; j++) {
      int y = bl->lpIndex[j] >> 16;
      int k = bl->lpIndex[j] & 0xffff;
      VisionPoint *lp = &horizontalPoint[c_ORANGE][y][k];
      bl->xi = bl->xi > lp->xi ? lp->xi : bl->xi;
      bl->xf = bl->xf < lp->xf ? lp->xf : bl->xf;
      bl->correctPixelRatio += lp->dx;
    }
    bl->dx = bl->xf - bl->xi + 1;
    bl->yi = bl->lpIndex[0] >> 16;
    bl->yf = bl->lpIndex[bl->lpCount - 1] >> 16;
    bl->dy = bl->yf - bl->yi;
    visionLog((30, "orange blob %i calculated: (xi:xf,yi:yf) = (%i:%i,%i:%i)", i, bl->xi, bl->xf, bl->yi, bl->yf));
    bl->correctPixelRatio /= 0.5f * bl->dx * bl->dy;
    visionLog((20,   "correctpixel[%i] = %f",i,bl->correctPixelRatio));
  }
}


void BlobDetector::calculateBandBlobData() {

  unsigned char vertColors[] = {
    c_BLUE,
    c_PINK
  };

  for (uint16_t c = 0; c < 2; c++) {

    unsigned char color = vertColors[c];
    for (uint32_t i = 0; i < verticalBlobCount[color]; i++) {

      Blob * blob = &verticalBlob[color][i];

      int j = 0;
      VisionPoint *lp = &verticalPoint[color][blob->lpIndex[j] >> 16][blob->lpIndex[j] & 0xfffful];
      blob->xi = lp->xi;

      uint32_t sumY = (lp->yi + lp->yf) >> 1 ;
      float sumWidth = lp->dy;

      uint16_t minY = IMAGE_HEIGHT - 1, maxY = 0;

      for (j = 1; j < blob->lpCount; j++) {
        lp = &verticalPoint[color][blob->lpIndex[j] >> 16][blob->lpIndex[j] & 0xfffful];
        sumY += (lp->yi + lp->yf) >> 1 ;
        sumWidth += lp->dy;
        minY = std::min(lp->yi, minY);
        maxY = std::max(lp->yf, maxY);
      }
      blob->yi = minY;
      blob->yf = maxY;
      blob->xf = lp->xi;


      blob->avgX = (blob->xf + blob->xi) >> 1;
      blob->avgY = sumY / blob->lpCount;
      blob->avgWidth = sumWidth / blob->lpCount;

    }
  }
}



void BlobDetector::SetColorTable(unsigned char* table) {
  color_table_ = table;
}

uint16_t BlobDetector::mergeHorizontalBlobs(Color color, uint16_t* mergeIndex, int mergeCount) {
  mergeBlobs(color, horizontalBlob, horizontalBlobCount, mergeIndex, mergeCount);
}

uint16_t BlobDetector::mergeVerticalBlobs(Color color, uint16_t* mergeIndex, int mergeCount) {
  mergeBlobs(color, verticalBlob, verticalBlobCount, mergeIndex, mergeCount);
}

uint16_t BlobDetector::mergeBlobs(Color color, Blob** blobs, uint32_t* blobCounts, uint16_t *mergeIndex, int mergeCount) {
  // Initialize
  for (uint16_t i = 0; i < blobCounts[color]; i++)
    mergeIndex[i] = mergeCount;

  // For each (blob[i], blob[j]) tuple
  for (uint16_t i = 0; i < blobCounts[color]; i++) {
    Blob *bli = &blobs[color][i];
    for (uint16_t j = 0; j < blobCounts[color]; j++) {
      if (i == j)
        continue;
      Blob *blj = &blobs[color][j];
      //visionLog((39, "comparing (%i,%i),(%i,%i) with (%i,%i),(%i,%i)", bli->xi, bli->xf, bli->yi, bli->yf, blj->xi, blj->xf, blj->yi, blj->yf));

      // Merge close and overlapping blobs
      bool closeX = (abs((int)bli->xf - (int)blj->xi) < 12);
      bool closeY = (abs((int)bli->yf - (int)blj->yi) <  6);
      bool overlapX = (bli->xi <= blj->xf && bli->xf >= blj->xi);
      bool overlapY = (bli->yi <= blj->yf && bli->yf >= blj->yi);
      if ((overlapX && closeY) ||
          (closeX && overlapY) ||
          (overlapX && overlapY)) {
            //visionLog((39,"blob %i overlaps with %i: closeX: %i closeY: %i overlapX: %i overlapY: %i", i, j, closeX, closeY, overlapX, overlapY));
        bli->xi = std::min(bli->xi, blj->xi);
        bli->xf = std::max(bli->xf, blj->xf);
        bli->yi = std::min(bli->yi, blj->yi);
        bli->yf = std::max(bli->yf, blj->yf);
        bli->correctPixelRatio = (bli->dx * bli->dy * bli->correctPixelRatio + blj->dy * blj->dx * blj->correctPixelRatio);
        // Update dx, dy
        bli->dx = bli->xf - bli->xi;
        bli->dy = bli->yf - bli->yi;
        bli->correctPixelRatio /= 0.5f * bli->dx * bli->dy;

        uint16_t minBlobIndex = std::min(i, j);
        uint16_t maxBlobIndex = std::max(i, j);
        mergeIndex[maxBlobIndex] =
          (mergeIndex[minBlobIndex] == mergeCount) ?
          minBlobIndex : mergeIndex[minBlobIndex];
        for (uint16_t k = 0; k < blobCounts[color]; k++)
          if (mergeIndex[k] == maxBlobIndex)
            mergeIndex[k] = mergeIndex[maxBlobIndex];
      }
    }

    // TODO:  This loop is incomplete.  After a blob merge, you have
    // to restart the entire loop which requires a runtime of O(n^4)
  }
  uint16_t sum = 0;
  for (uint16_t i = 0; i < horizontalBlobCount[color]; i++)
    sum += (mergeIndex[i] == MAX_ORANGE_BLOBS);
  return sum;
}
