#include "CameraCalibrator.h"
#define DEBUG_CALIBRATION 1
float CameraCalibrator::computeErrorParameterVector(const Sample& sample, const vector<float>& parameters) {
    RobotPose pose;
    CameraCalibration cal;
    translateParameters(parameters, cal, pose);
    if(sample.camera == Camera::TOP)
        top_transform_.calculateTransformation(
            top_transform_.getTorsoMatrix(body_model_block_),
            NECKOFFSETZ,
            getHeadPan() + cal.headPanOffset * DEG_T_RAD,
            getHeadTilt() - tilt_ * DEG_T_RAD,
            cal.topCameraXOffset * DEG_T_RAD,
            cal.topCameraZOffset * DEG_T_RAD,
            cal.topCameraTiltOffset * DEG_T_RAD,
            cal.topCameraRollOffset * DEG_T_RAD,
            cal.topCameraYawOffset * DEG_T_RAD,
            getFrontHeight(),
            cal.topCameraFOVx * DEG_T_RAD,
            cal.topCameraFOVy * DEG_T_RAD
        );
    else
        bottom_transform_.calculateTransformation(
            bottom_transform_.getTorsoMatrix(body_model_block_),
            NECKOFFSETZ,
            getHeadPan() + cal.headPanOffset * DEG_T_RAD,
            getHeadTilt() - tilt_ * DEG_T_RAD,
            cal.bottomCameraXOffset * DEG_T_RAD,
            cal.bottomCameraZOffset * DEG_T_RAD,
            cal.bottomCameraTiltOffset * DEG_T_RAD,
            cal.bottomCameraRollOffset * DEG_T_RAD,
            cal.bottomCameraYawOffset * DEG_T_RAD,
            getFrontHeight(),
            cal.bottomCameraFOVx * DEG_T_RAD,
            cal.bottomCameraFOVy * DEG_T_RAD
        );
    float error = getSampleErrorCameraFrame(sample, pose);
    //std::cout << "Error calculated as " << error << "\n";
    return error;
}

float CameraCalibrator::getSampleErrorGlobalFrame(const Sample& sample, const RobotPose& pose) const{
    Pose3D position(pose.translation.x, pose.translation.y, getFrontHeight());
    position.rotateZ(rotation_);
    float error = 3000;
    vector<LineSegment> segments = getFieldLinesGlobalFrame();
    const Vision::TransformationProvider& transform(sample.camera == Camera::TOP ? top_transform_ : bottom_transform_);
    for(vector<LineSegment>::iterator segI = segments.begin(); segI != segments.end(); segI++){
        LineSegment seg = *segI;
        Point2D projS = transform.forwardProject(sample.x, sample.y);
        Pose3D globS = Pose3D(projS.x, projS.y, 0).relativeToGlobal(position);
        Point2D globPtS = Point2D(globS.translation.x, globS.translation.y);
        float distance = seg.getDistanceToPoint(globPtS);
        if(distance < error)
            error = distance;
    }
    return error * error;
}

float CameraCalibrator::getSampleErrorCameraFrame(const Sample& sample, const RobotPose& pose) const {
    float error = 1000;
    vector<LineSegment> segments = getFieldLinesCameraFrame(pose, (sample.camera == Camera::TOP ? top_transform_ : bottom_transform_));
    LineSegment s;
    for(vector<LineSegment>::iterator segI = segments.begin(); segI != segments.end(); segI++){
        LineSegment seg = *segI;
        Line2D line(seg.start, seg.end);
        float distance = seg.getDistanceToPoint(sample);
        if(distance < error) {
            error = distance;
            s = seg;
        }
    }
    //std::cout << "closest seg to " << sample.x << "," << sample.y << " is " << s.start.x << "," << s.start.y << " - > " << s.end.x << "," << s.end.y << "\n";
    return error * error;
}

float CameraCalibrator::getHeadTilt() const {
  if (joint_block_ == NULL)
    return 0;
  return -joint_block_->values_[HeadTilt];
}

float CameraCalibrator::getHeadPan() const {
  if (joint_block_ == NULL)
    return 0;
  return joint_block_->values_[HeadPan];
}

float CameraCalibrator::getFrontHeight() const {
  if (body_model_block_ == NULL)
    return 0;
  return body_model_block_->abs_parts_[BodyPart::torso].translation.z;
}

void CameraCalibrator::translateParameters(const vector<float>& parameters, CameraCalibration& cameraCalibration, RobotPose& robotPose) const
{
  cameraCalibration.topCameraTiltOffset = parameters[topCameraTiltOffset];
  cameraCalibration.topCameraRollOffset = parameters[topCameraRollOffset];
  cameraCalibration.bottomCameraTiltOffset = parameters[bottomCameraTiltOffset];
  cameraCalibration.bottomCameraRollOffset = parameters[bottomCameraRollOffset];
  //cameraCalibration.headTiltOffset = parameters[headTiltOffset];
  cameraCalibration.headPanOffset = parameters[headPanOffset];

  robotPose.translation.x = parameters[poseX];
  robotPose.translation.y = parameters[poseY];
  //robotPose.rotation = parameters[poseRotation];
}

void CameraCalibrator::translateParameters(const CameraCalibration& cameraCalibration, const RobotPose& robotPose, vector<float>& parameters) const {

  parameters[topCameraTiltOffset] = cameraCalibration.topCameraTiltOffset;
  parameters[topCameraRollOffset] = cameraCalibration.topCameraRollOffset;
  parameters[bottomCameraTiltOffset] = cameraCalibration.bottomCameraTiltOffset;
  parameters[bottomCameraRollOffset] = cameraCalibration.bottomCameraRollOffset;
  //parameters[headTiltOffset] = cameraCalibration.headTiltOffset;
  parameters[headPanOffset] = cameraCalibration.headPanOffset;

  parameters[poseX] = robotPose.translation.x;
  parameters[poseY] = robotPose.translation.y;
  //parameters[poseRotation] = robotPose.rotation;
}

vector<LineSegment> CameraCalibrator::getFieldLinesGlobalFrame() const {
    vector<LineSegment> segments;
    for (int i = LINE_OFFSET; i < LINE_OFFSET + NUM_LINES; i++) {
        WorldObject &wo = world_object_block_->objects_[i];
        LineSegment segment;
        segment.start = wo.loc;
        segment.end = wo.endLoc;
        segments.push_back(segment);
    }
    return segments;
}

vector<LineSegment> CameraCalibrator::getFieldLinesCameraFrame(const RobotPose& pose, const Vision::TransformationProvider& transform) const {
    Pose3D position(pose.translation.x, pose.translation.y, 0);
    position.rotateZ(rotation_);
    return getFieldLinesCameraFrame(position, world_object_block_, transform);
}

vector<LineSegment> CameraCalibrator::getFieldLinesCameraFrame(Pose3D position, WorldObjectBlock* world_object_block, const Vision::TransformationProvider& transform) {
  vector<LineSegment> segments;
  for (int i = LINE_OFFSET; i < LINE_OFFSET + NUM_LINES; i++) {
    WorldObject &wo = world_object_block->objects_[i];

    Pose3D pt1 = Pose3D(wo.loc.x, wo.loc.y, 0).globalToRelative(position);
    Pose3D pt2 = Pose3D(wo.endLoc.x, wo.endLoc.y, 0).globalToRelative(position);
    float distance = (pt1.translation - pt2.translation).abs();

    // Split each line into a number of segments - helps prevent distortion
    int numSegments = ceil(distance / 500);
    if (ceil(distance/0.5) == floor(distance/500))
      numSegments++;

    for (int i = 0; i < numSegments; i++) {
      float r1 = (500 * i) / distance;
      float r2 = (500 * (i+1)) / distance;
      if (i == numSegments - 1)
        r2 = 1;
      Vector3<float> p1 = pt1.translation * (1-r1) + pt2.translation * r1, p1i;
      Vector3<float> p2 = pt1.translation * (1-r2) + pt2.translation * r2, p2i;
      transform.invertPoint(p1,p1i);
      transform.invertPoint(p2,p2i);
      if(p1i.x < 0 && p2i.x < 0) {
        continue;
      }
      LineSegment segment = transform.backProject(p1,p2);
      segments.push_back(segment);
    }
  }
  for (int i = 0; i < 24; i++) {
    Pose3D pt1 = Pose3D(CIRCLE_RADIUS * sinf(i * M_PI / 12.0), CIRCLE_RADIUS * cosf(i * M_PI / 12.0), 0).globalToRelative(position);
    Pose3D pt2 = Pose3D(CIRCLE_RADIUS * sinf((i + 1) * M_PI / 12.0), CIRCLE_RADIUS * cosf((i + 1) * M_PI / 12.0), 0).globalToRelative(position);
    Vector3<float> p1 = pt1.translation, p1i, p2 = pt2.translation, p2i;
    transform.invertPoint(p1, p1i);
    transform.invertPoint(p2, p2i);
    if(p1i.x < 0 && p2i.x < 0) {
        continue;
    }

    LineSegment segment = transform.backProject(pt1.translation,pt2.translation);
    if(segment.start.x < 0 && segment.end.x > IMAGE_X)
        continue;
    if(segment.start.x > IMAGE_X && segment.end.x < 0)
        continue;
    segments.push_back(segment);
  }
  return segments;
}

void CameraCalibrator::calibrate(CameraCalibration& cal, RobotPose& pose, vector<Sample> samples, int iterations) {
    type_ = cal.type;
    rotation_ = pose.rotation;
    tilt_ = cal.headTiltOffset;
	GaussNewtonOptimizer<Sample, CameraCalibrator>* optimizer;
	vector<float> parameters;
	parameters.resize(numOfParameterTranslations);
	translateParameters(cal, pose, parameters);
	optimizer = new GaussNewtonOptimizer<Sample, CameraCalibrator>(parameters, samples, *this, &CameraCalibrator::computeErrorParameterVector);
	if(DEBUG_CALIBRATION) std::cout << "\n";
	for(int i=0; i < iterations; i++) {
        float epsilon = optimizer->iterate();
        if(epsilon == 0.0f) {
            std::cout << "Optimizer diverged.\n";
            break;
        }
        parameters = optimizer->getParameters();
        translateParameters(parameters, cal, pose);
        if(DEBUG_CALIBRATION) {
            printf("Pose: (%.03f,%.03f,%.03f) Offsets: (%.03f,%.03f,%.03f,%.03f,%.03f,%.03f)",
                pose.translation.x,
                pose.translation.y,
                pose.rotation,
                cal.topCameraTiltOffset,
                cal.topCameraRollOffset,
                cal.bottomCameraTiltOffset,
                cal.bottomCameraRollOffset,
                cal.headTiltOffset,
                cal.headPanOffset);
            std::cout.flush();
            std::cout << "\r";
        }
	}
	if(DEBUG_CALIBRATION) std::cout << "\n";
    parameters = optimizer->getParameters();
    delete optimizer;
    translateParameters(parameters, cal, pose);
}
