#include "Classifier.h"
#include <iostream>

#define pointWasScanned(x,y) pointScanned[y * IMAGE_WIDTH + x]
#define focusMap(x,y) focusMap_[y * IMAGE_WIDTH + x]
//#define previousFocusMap(x,y) previousFocusMap_[y * IMAGE_WIDTH + x]
//#define fastFocusMap(x,y) fastFocusMap_[y * IMAGE_WIDTH + x]

Classifier::Classifier(
                       ImageBlock*& image,
                       RobotVisionBlock*& robot_vision_block,
                       VisionParams& params,
                       Camera::Type& camera)
  :
  image_(image),
  robot_vision_(robot_vision_block),
  params_(params),
  camera_(camera),
  initialized_(false),
  doingHighResScan_(false) {

  //VisionPoint horizontalPoint[NUM_COLORS][IMAGE_HEIGHT >> 1][IMAGE_WIDTH >> 2];
  horizontalPoint = new VisionPoint**[NUM_COLORS];
  for(int i=0;i<NUM_COLORS;i++) {
    horizontalPoint[i] = new VisionPoint*[IMAGE_HEIGHT];
    for(int j = 0; j < IMAGE_HEIGHT; j++)
      horizontalPoint[i][j] = new VisionPoint[IMAGE_WIDTH];
  }
  //VisionPoint verticalPoint[NUM_COLORS][IMAGE_WIDTH >> 2][IMAGE_HEIGHT >> 1];
  verticalPoint = new VisionPoint**[NUM_COLORS];
  for(int i=0;i<NUM_COLORS;i++) {
    verticalPoint[i] = new VisionPoint*[IMAGE_WIDTH];
    for(int j = 0; j < IMAGE_WIDTH; j++)
      verticalPoint[i][j] = new VisionPoint[IMAGE_HEIGHT];
  }
  //uint32_t horizontalPointCount[NUM_COLORS][IMAGE_HEIGHT >> 1];
  horizontalPointCount = new uint32_t*[NUM_COLORS];
  for(int i=0;i<NUM_COLORS;i++) {
    horizontalPointCount[i] = new uint32_t[IMAGE_HEIGHT];
  }
  //uint32_t verticalPointCount[NUM_COLORS][IMAGE_WIDTH >> 2];
  verticalPointCount = new uint32_t*[NUM_COLORS];
  for(int i=0;i<NUM_COLORS;i++)
    verticalPointCount[i] = new uint32_t[IMAGE_WIDTH];

    setStepScale(DEFAULT_HORIZONTAL_STEP_SCALE, DEFAULT_VERTICAL_STEP_SCALE);
}

void Classifier::setImagePointers() {
  if(camera_ == Camera::TOP) {
    segImg_ = robot_vision_->segImgTop;
    img_ = image_->img_top_.get();
  }
  else {
    segImg_ = robot_vision_->segImgBottom;
    img_ = image_->img_bottom_.get();
  }
  if(!initialized_) {
    bool imageLoaded = (camera_ == Camera::TOP ? image_->img_top_ : image_->img_bottom_);
    #ifdef TOOL
    if(imageLoaded)
    #endif
    memset(segImg_, c_UNDEFINED, sizeof(unsigned char) * IMAGE_SIZE);
    initialized_ = true;
  }
}

void Classifier::classifyImage(unsigned char* colorTable){
    colorTable_ = colorTable;
    setImagePointers();
    bool imageLoaded = (camera_ == Camera::TOP ? image_->img_top_ : image_->img_bottom_);
    if(!imageLoaded) {
        visionLog((20, "Classifying with no raw image"));
    }
    for (uint32_t y = 0; y < IMAGE_HEIGHT; y += vstep_) {
        for(uint32_t x = 0; x < IMAGE_WIDTH; x += hstep_) {
            if(doingHighResScan_ && !focusMap(x,y)) continue;

            Color c;
            #ifdef TOOL
                if (imageLoaded) // if a raw image is available
            #endif
                {
                    c = ColorTableMethods::xy2color(img_, colorTable, x, y);
                    segImg_[IMAGE_WIDTH * y + x] = c;
                }
        }
    }
}

bool Classifier::startHighResScan(Color c, int hStepScale, int vStepScale) {
    if(!prepareFocusMap(c)) return false;
    doingHighResScan_ = true;
    setStepScale(hStepScale, vStepScale);
    classifyImage(colorTable_);
    constructRuns(1 << c);
    return true;
}

bool Classifier::startFastHighResScan(Color c, int hStepScale, int vStepScale) {
    if(!prepareFastFocusMap(c)) return false;
    doingHighResScan_ = true;
    setStepScale(hStepScale, vStepScale);
    classifyImage(colorTable_);
    prepareFocusMap(c);
    constructRuns(1 << c);
    return true;
}

bool Classifier::startHighResBallScan() {
  didHighResBallScan = true;
  return startHighResScan(c_ORANGE);
}

bool Classifier::startHighResGoalScan() {
  return startHighResScan(c_YELLOW, 0, DEFAULT_VERTICAL_STEP_SCALE);
}

void Classifier::completeHighResScan() {
  setStepScale(DEFAULT_HORIZONTAL_STEP_SCALE, DEFAULT_VERTICAL_STEP_SCALE);
  // for(int y = 0; y < IMAGE_HEIGHT; y++) {
  //   for(int x = 0; x < IMAGE_WIDTH; x++) {
  //     if(focusMap(x,y))
  //       previousFocusMap(x,y) = true;
  //   }
  // }
  doingHighResScan_ = false;
}

void Classifier::setHorizon(HorizonLine horizon) {
  horizon_ = horizon;
}

void Classifier::constructRuns(int colorFlags) {

  uint32_t x = 0, y = 0, z;

  uint8_t vRecent[IMAGE_WIDTH];
  uint8_t vRunClr[IMAGE_WIDTH]; // Running line region color
  uint16_t vStartY[IMAGE_WIDTH]; // Start of running line region

  // Reset stuff used to track vertical line points in horizontal scan
  for (x = 0; x < IMAGE_WIDTH; x++) {
    vRecent[x] = c_UNDEFINED;
    vRunClr[x] = c_UNDEFINED;
    vGreenPosition[x] = (uint16_t)-1;
  }

  // Reset Vertical Point Counts
  for (z = 0; z < NUM_COLORS; z++) {
    if(!isInFlags(z, colorFlags)) continue;
    for (x = 0; x < IMAGE_WIDTH; x++) {
      verticalPointCount[z][x] = 0;
    }
  }

  // Reset Horizontal Point Counts
  for (z = 0; z < NUM_COLORS; z++) {
    if(!isInFlags(z, colorFlags)) continue;
    for (y = 0; y < IMAGE_HEIGHT; y++) {
      horizontalPointCount[z][y] = 0;
    }
  }

  bool inFocusArea = false;

  for (y = 0; y < IMAGE_HEIGHT; y += vstep_) {

    hGreenPosition[y] = (uint16_t)-1;
    uint8_t hRecent = c_UNDEFINED;
    uint8_t hRunClr = c_UNDEFINED;
    uint16_t hStartX = 0;

    for (x = 0; x < IMAGE_WIDTH; x += hstep_) {
      if(doingHighResScan_ && focusMap(x,y)) inFocusArea = true;
      if(doingHighResScan_ && !focusMap(x,y)) {
        if(inFocusArea && isInFlags(hRunClr, colorFlags)) {
          completeHorizontalRun(hRunClr, y, hStartX, x);
          inFocusArea = false;
        }
        continue;
      }

      uint8_t c = segImg_[IMAGE_WIDTH * y + x];

      // HORIZONTAL


      // Segment Finishes - when 2 consecutive pixels are different
      if (hRunClr != hRecent && hRunClr != c && isInFlags(hRunClr, colorFlags)) {
        if (hRunClr == c_FIELD_GREEN) {
          hGreenPosition[y] = hStartX;
        } else if (hRunClr == c_BLUE || hRunClr == c_YELLOW || hRunClr == c_ORANGE ||
                   (hRunClr == c_WHITE && hGreenPosition[y] != (uint16_t) -1)) {

          uint16_t valXI = hStartX;
          uint16_t valXF = (x - 2 * hstep_);

          switch(hRunClr) {

          case c_WHITE: {
            bool aboveHorizon = horizon_.exists && (horizon_.gradient * valXI + horizon_.offset > y);
            bool widthTooHigh = valXF - valXI > 75;
            bool onShoulder = bodyExclusionAvailable;
            uint16_t valAvgX = (valXI + valXF) / 2;
            uint16_t xIndex = ((NUM_BODY_EXCL_POINTS - 1) * valAvgX) / IMAGE_WIDTH;
            onShoulder &=
              bodyExclusionSlope[xIndex] * valAvgX + bodyExclusionOffset[xIndex] < y;
            if (aboveHorizon || widthTooHigh || onShoulder) {
              break;
            }
          }

          default: {
            // Save line segment
            VisionPoint *lp =
              &horizontalPoint[hRunClr][y][horizontalPointCount[hRunClr][y]++];
            lp->xi = valXI;
            lp->xf = valXF;
            lp->dx = lp->xf - lp->xi + 1;
            lp->yi = lp->yf = y;
            lp->dy = 1;
            lp->isValid = true;
            lp->lbIndex = (uint16_t)-1; // Will get this information later
          }

          }
        }

        hRunClr = c_UNDEFINED;
      }
      // Segment Starts - when 2 consecutive pixels are same
      if (c != hRunClr && c == hRecent) {
        hRunClr = c;
        hStartX = x - hstep_;
      }
      hRecent = c;

      // VERTICAL

      uint8_t runClr = vRunClr[x];
      uint8_t recent = vRecent[x];

      // Vision Point Segment finishes
      if (runClr != recent && runClr != c && isInFlags(runClr, colorFlags)) {
        if (runClr == c_FIELD_GREEN) {
          vGreenPosition[x] = vStartY[x];
        } else if (runClr == c_BLUE || runClr == c_PINK ||
                   (vGreenPosition[x] != (uint16_t)-1 && runClr == c_WHITE)) {

          uint16_t valX = x;
          uint16_t valYI = vStartY[x];
          uint16_t valYF = y - 4;

          // Introduce some special checks here
          switch(runClr) {

          case c_WHITE: {
            bool aboveHorizon = horizon_.exists && (horizon_.gradient * valX + horizon_.offset > valYI);
            bool widthTooHigh = valYF - valYI > 125;
            bool onShoulder = bodyExclusionAvailable;
            uint16_t xIndex = ((NUM_BODY_EXCL_POINTS - 1) * valX) / IMAGE_WIDTH;
            onShoulder &=
              bodyExclusionSlope[xIndex] * valX + bodyExclusionOffset[xIndex] < valYF;
            if (aboveHorizon || widthTooHigh || onShoulder) {
              break;
            }
          }

          default: {
            // Save Line Segment
            VisionPoint *lp =
              &verticalPoint[runClr][x][verticalPointCount[runClr][x]++];
            lp->xi = lp->xf = valX;
            lp->dx = 1;
            lp->yi = valYI;
            lp->yf = valYF;
            lp->dy = lp->yf - lp->yi + 1;
            lp->isValid = true;
            lp->lbIndex = (uint16_t)-1;
          }

          }
        }
        vRunClr[x] = c_UNDEFINED;
      }

      // Vision Point Segment Starts
      if (c != runClr && c == recent) {
        vRunClr[x] = c;
        vStartY[x] = y - 2;
      }
      vRecent[x] = c;
    }

    // Last Horizontal Segment
    // Don't save the white ones
    if(!isInFlags(hRunClr, colorFlags)) continue;
    if (hRunClr == c_FIELD_GREEN) {
      hGreenPosition[y] = hStartX;
    } else if (hRunClr == c_BLUE || hRunClr == c_YELLOW || hRunClr == c_ORANGE) {
      // Save line segment
      completeHorizontalRun(hRunClr, y, hStartX);
    }
  }

  // Last Vertical Segment
  // Don't save the white ones
  for (x = 0; x < IMAGE_WIDTH; x++) {
    uint8_t runClr = vRunClr[x];
    if(!isInFlags(runClr, colorFlags)) continue;
    if (runClr == c_FIELD_GREEN) {
      vGreenPosition[x] = vStartY[x];
    } else if (runClr == c_BLUE || runClr == c_PINK) {
      // Save line segment
      completeVerticalRun(runClr, x, vStartY[x]);
      //TODO: Fix for high res vertical scans
    }
  }
}

void Classifier::preProcessPoints() {

  // White Points without green at the end are removed
  for (uint16_t x = 0; x < IMAGE_WIDTH; x++) {
    for (uint16_t point = 0; point < verticalPointCount[c_WHITE][x]; point++) {
      if (verticalPoint[c_WHITE][x][point].yi > vGreenPosition[x]) {
        verticalPointCount[c_WHITE][x] = point;
        break;
      }
    }
  }

  // Horizontal white points without green at the end are also removed
  for (uint16_t y = 0; y < IMAGE_HEIGHT; y++) {
    for (uint16_t point = 0; point < horizontalPointCount[c_WHITE][y]; point++) {
      if (horizontalPoint[c_WHITE][y][point].xi > hGreenPosition[y]) {
        horizontalPointCount[c_WHITE][y] = point;
        break;
      }
    }
  }

  // This is done for the colored bands
  // Join 2 linepoints together which are in the same vertical line and very close.
  if (params_.ALLOW_BAND_MERGE_POINTS) {
    unsigned char bandColors[] = {
      c_BLUE,
      c_PINK
    };

    for (uint16_t c = 0; c < 2; c++) {
      unsigned char color = bandColors[c];
      for (uint16_t x = 0; x < IMAGE_WIDTH; x++) {
        uint16_t skip = 1;
        for (int16_t i = 0; i < ((int16_t)verticalPointCount[color][x]) - 1; i += skip) {
          skip = 1;
          for (uint16_t j = i + 1; j < verticalPointCount[color][x]; j++) {
            if (verticalPoint[color][x][j].yi - verticalPoint[color][x][i].yf <= params_.BAND_MERGE_POINT_DIST) {
              verticalPoint[color][x][i].yf = verticalPoint[color][x][j].yf;
              verticalPoint[color][x][i].dy = verticalPoint[color][x][i].yf - verticalPoint[color][x][i].yi + 1;
              verticalPoint[color][x][j].isValid = false;
              skip++;
            }
          }
        }
      }
    }
  }

}

void Classifier::preProcessGoalPoints() {

  // This is done for the goals
  // Join 2 linepoints together which are in the same vertical line and very close.
  if (params_.ALLOW_GOAL_MERGE_POINTS) {
    unsigned char goalColors[] = {
      c_YELLOW
    };

    for (uint16_t c = 0; c < 1; c++) {
      unsigned char color = goalColors[c];
      for (uint16_t x = 0; x < IMAGE_HEIGHT; x++) {
        uint16_t skip = 1;
        for (int16_t i = 0; i < ((int16_t)horizontalPointCount[color][x]) - 1; i += skip) {
          skip = 1;
          for (uint16_t j = i + 1; j < horizontalPointCount[color][x]; j++) {
            if (horizontalPoint[color][x][j].xi - horizontalPoint[color][x][i].xf <= params_.GOAL_MERGE_POINT_DIST) {
              horizontalPoint[color][x][i].xf = horizontalPoint[color][x][j].xf;
              horizontalPoint[color][x][i].dx = horizontalPoint[color][x][i].xf - horizontalPoint[color][x][i].xi + 1;
              horizontalPoint[color][x][j].isValid = false;
              skip++;
            }
          }
        }
      }
    }
  }

}

void Classifier::completeHorizontalRun(uint8_t &hRunClr, int y, int hStartX, int finX) {
  VisionPoint *lp = &horizontalPoint[hRunClr][y][horizontalPointCount[hRunClr][y]++];
  lp->xi = hStartX;
  lp->xf = ((finX >> hscale_) - 1) << hscale_;
  lp->dx = lp->xf - lp->xi + 1;
  lp->yi = lp->yf = y;
  lp->dy = 1;
  lp->isValid = true;
  lp->lbIndex = (uint16_t)-1;
  hRunClr = c_UNDEFINED;
}

void Classifier::completeVerticalRun(uint8_t &runClr, int x, int vStartY) {
  VisionPoint *lp =
    &verticalPoint[runClr][x][verticalPointCount[runClr][x]++];
  lp->xi = lp->xf = x;
  lp->dx = 1;
  lp->yi = vStartY;
  lp->yf = IMAGE_HEIGHT - (1 << vscale_);
  lp->dy = lp->yf - lp->yi + 1;
  lp->isValid = true;
  lp->lbIndex = (uint16_t)-1;
  runClr = c_UNDEFINED;
}

void Classifier::setStepScale(int h, int v){
    hstep_ = (1 << h);
    vstep_ = (1 << v);
    hscale_ = h;
    vscale_ = v;
}

void Classifier::getStepSize(int& h, int& v){
    h = hstep_;
    v = vstep_;
}

void Classifier::getStepScale(int& h, int& v){
    h = hscale_;
    v = vscale_;
}

bool Classifier::prepareFocusMap(Color c) {
  bool pixelFound = false;
  memset(focusMap_, false, IMAGE_SIZE);
  for(int y = 0; y < IMAGE_HEIGHT; y += vstep_) {
    int ymin = std::max(0, y - FOCUS_RANGE), ymax = std::min(IMAGE_HEIGHT, y + FOCUS_RANGE);
    for(int i = 0; i < horizontalPointCount[c][y]; i++) {
      VisionPoint* point = &horizontalPoint[c][y][i];
      int xmin = std::max(0, point->xi - FOCUS_RANGE), xmax = std::min(IMAGE_WIDTH, point->xf + FOCUS_RANGE);
      for(int x = xmin; x < xmax; x++) {
        for(int yy = ymin; yy < ymax; yy++) {
          focusMap(x,yy) = true;
          pixelFound = true;
        }
      }
    }
  }
  return pixelFound;
}

bool Classifier::prepareFastFocusMap(Color c) {
  bool pixelFound = false;
  memset(focusMap_, false, IMAGE_SIZE);
  for(int y = 0; y < IMAGE_HEIGHT; y += vstep_) {
    for(int i = 0; i < horizontalPointCount[c][y]; i++) {
      VisionPoint* point = &horizontalPoint[c][y][i];
      int xmin = std::max(0, point->xi - FOCUS_RANGE), xmax = std::min(IMAGE_WIDTH, point->xf + FOCUS_RANGE);
      for(int x = xmin; x < point->xi; x++) {
        focusMap(x,y) = true;
        pixelFound = true;
      }
      for(int x = point->xi; x <= point->xf; x++) {
        segImg_[IMAGE_WIDTH * y + x] = c;
      }
      for(int x = point->xf + 1; x < xmax; x++) {
        focusMap(x,y) = true;
        pixelFound = true;
      }
    }
  }
  return pixelFound;
}


void Classifier::clearPreviousHighResScans() {
  setImagePointers();
#ifdef TOOL
  bool imageLoaded = (camera_ == Camera::TOP ? image_->img_top_ : image_->img_bottom_);
  if (imageLoaded) // if a raw image is available
#endif
  {
    if (segImg_) {
      memset(segImg_, c_UNDEFINED, IMAGE_SIZE);
      visionLog((20, "clearing seg image"));
    }
  }
}
