#include "CrossDetector.h"

CrossDetector::CrossDetector(
  WorldObjectBlock*& world_objects,
  FrameInfoBlock*& vision_frame_info,
  Classifier*& classifier,
  BlobDetector*& blob_detector,
  Vision::TransformationProvider& transform)
  : world_objects_(world_objects),
    vision_frame_info_(vision_frame_info),
    classifier_(classifier),
    blob_detector_(blob_detector),
    transform_(transform) {

}

void CrossDetector::detectCross() {
  int hstep, vstep;
  classifier_->getStepSize(hstep, vstep);
  int wideXRange = 100, narrowXRange = 30, wideYRange = 25, narrowYRange = 15;
  int xmin, xmax, ymin, ymax;
  int count = blob_detector_->verticalBlobCount[c_WHITE];
  Blob* blobs = blob_detector_->verticalBlob[c_WHITE];
  int mergeCount = 100;
  uint16_t mergeIndex[mergeCount];
  blob_detector_->mergeVerticalBlobs(c_WHITE, mergeIndex, mergeCount);
  for(uint16_t i = 0; i < count; i++) {
    if(mergeIndex[i] != mergeCount) continue; // This blob was merged with another blob
    Blob* blob = &blobs[i];
    if(!crossHorizon_.isAbovePoint(blob->xi, blob->yi) || !crossHorizon_.isAbovePoint(blob->xf, blob->yi)) continue;
    int left = blob->xi, right = blob->xf, yval = blob->yi, top = blob->yf, bottom = blob->yi;
    if(right - left > 30) continue;
    if(right - left < 10) continue;
    if(top - bottom > 10) continue;

    float gtopleft = getGreenPercentage(left - wideXRange, left - narrowXRange, yval - wideYRange, yval - narrowYRange, hstep, vstep);
    float gtopmid = getGreenPercentage(left - narrowXRange, right + narrowXRange, yval - wideYRange, yval - narrowYRange, hstep, vstep);
    float gtopright = getGreenPercentage(right + narrowXRange, right + wideXRange, yval - wideYRange, yval - narrowYRange, hstep, vstep);

    float gleft = getGreenPercentage(left - wideXRange, left - narrowXRange, yval - narrowYRange, yval + narrowYRange, hstep, vstep);
    float gright = getGreenPercentage(right + narrowXRange, right + wideXRange, yval - narrowYRange, yval + narrowYRange, hstep, vstep);

    float gbottomleft = getGreenPercentage(left - wideXRange, left - narrowXRange, yval + narrowYRange, yval + wideYRange, hstep, vstep);
    float gbottommid = getGreenPercentage(left - narrowXRange, right + narrowXRange, yval + narrowYRange, yval + wideYRange, hstep, vstep);
    float gbottomright = getGreenPercentage(right + narrowXRange, right + wideXRange, yval + narrowYRange, yval + wideYRange, hstep, vstep);

    visionLog((34, "Blob %i chosen for cross detection", i));

    visionLog((34, "Green percentages: TL %5.3f TM %5.3f TR %5.3f L %5.3f R %5.3f BL %5.3f BM %5.3f BR %5.3f", gtopleft, gtopmid, gtopright, gleft, gright, gbottomleft, gbottommid, gbottomright));

    float minPercentage = .8;

    if(gtopleft < minPercentage) continue;
    if(gtopmid < minPercentage) continue;
    if(gtopright < minPercentage) continue;

    if(gleft < minPercentage) continue;
    if(gright < minPercentage) continue;

    if(gbottomleft < minPercentage) continue;
    if(gbottommid < minPercentage) continue;
    if(gbottomright < minPercentage) continue;

    visionLog((34, "Blob %i selected as cross", i));

    setCrossObject(blob);
    return;
  }
}

void CrossDetector::correctRanges(int& xmin, int& xmax, int& ymin, int& ymax, int hstep, int vstep) {
  xmin = std::max(0, xmin);
  xmax = std::min(IMAGE_WIDTH - 1, xmax);
  ymin = std::max(0, ymin);
  ymax = std::min(IMAGE_HEIGHT - 1, ymax);
  xmin = xmin - (xmin % hstep);
  ymin = ymin - (ymin % vstep);
}

float CrossDetector::getGreenPercentage(int xmin, int xmax, int ymin, int ymax, int hstep, int vstep) {
  visionLog((34, "Scanning from x:%i->%i, y:%i->%i", xmin, xmax, ymin, ymax));
  correctRanges(xmin,xmax,ymin,ymax,hstep,vstep);
  int numTotal = 0, numGreen = 0;
  for(int x = xmin; x <= xmax; x += hstep) {
    for(int y = ymin; y <= ymax; y += vstep) {
      Color c = classifier_->xy2color(x,y);
      numTotal++;
      if(c == c_FIELD_GREEN) numGreen++;
      if(c == c_WHITE) return 0.0; // There shouldn't be any white around the cross
    }
  }
  if(numTotal == 0) return 0.0; // No false positives, no regrets
  return (float)numGreen / numTotal;
}

void CrossDetector::setHorizon(HorizonLine horizon) {
  crossHorizon_ = horizon;
}

float CrossDetector::getDistanceByBlobWidth(float dx) {
  return .9f * 16765.0f * powf(dx, -0.769f);
}

void CrossDetector::setCrossObject(Blob* blob) {
  int left = blob->xi, right = blob->xf, yval = blob->yi;
  int centerX = left + (right - left) / 2;
  float bear = transform_.calculateBearing(centerX);
  float elev = transform_.calculateElevation(yval);
  float dist = getDistanceByBlobWidth(right - left);
  WorldObject* cross = &world_objects_->objects_[WO_OWN_PENALTY_CROSS];
  transform_.transformPolar(dist, bear, elev, cross->visionDistance, cross->visionBearing, cross->visionElevation);
  cross->seen = true;
  cross->fromTopCamera = true;
  cross->imageCenterX = centerX;
  cross->imageCenterY = yval;
  cross->visionConfidence = 1.0;
  cross->frameLastSeen = vision_frame_info_->frame_id;
  //cross->ballBlobIndex = fill this in;
  visionLog((34, "Cross identified at %i, %i", centerX, yval));
}
