#include "GoalDetector.h"

GoalDetector::GoalDetector(
                           WorldObjectBlock*& world_object_block,
                           FrameInfoBlock*& frame_info_block,
                           Vision::TransformationProvider& transform,
                           Classifier*& classifier,
                           BlobDetector*& blob_detector,
                           LineDetector*& line_detector,
                           Camera::Type& camera)
  :
  world_objects_(world_object_block),
  vision_frame_info_(frame_info_block),
  transform_(transform),
  classifier_(classifier),
  blob_detector_(blob_detector),
  line_detector_(line_detector),
  camera_(camera),
  color(c_YELLOW){

  YellowPostCounter = 0;

  yellowPosts = new FieldLine * [MAX_FIELDLINES];
  for (int i = 0; i < MAX_FIELDLINES; i++) {
    yellowPosts[i] = new FieldLine();
    yellowPosts[i]->id = i;
    yellowPosts[i]->PointsArray =
      new LinePoint * [MAX_POINTS_PER_LINE];

    for (int j = 0; j < MAX_POINTS_PER_LINE; j++)
      yellowPosts[i]->PointsArray[j] = new LinePoint();

  }
}

void GoalDetector::sanitizeGoalBlobs() {

  for (uint32_t j = 0; j < blob_detector_->horizontalBlobCount[color]; j++) {
    Blob * lb = &blob_detector_->horizontalBlob[color][j];
    float xDiff = (int16_t) lb->xf - (int16_t) lb->xi;
    float yDiff = (int16_t) lb->yf - (int16_t) lb->yi;
    float slope = yDiff / xDiff;
    lb->invalid = slope < 1 && slope > -1;
  }

}

float GoalDetector::estimateGoalDistance(FieldLine * goal) {

  //float x = goal->tL.x - goal->bR.x, y = goal->tL.y - goal->bR.y;

  //(piyushk) slightly better distance estimate, as the first/last point might be at a slightly wrong x location
  float y1 = goal->tL.y;
  float x1 = (int) ((y1 - goal->Offset) / goal->Slope);
  float y2 = goal->bR.y;
  float x2 = (int) ((y2 - goal->Offset) / goal->Slope);
  float x = x1 - x2, y = y1 - y2;

  float height = sqrtf(x * x + y * y);

  float width = fabs(goal->preciseWidth * sinf(goal->Angle));
  float oldWidth = fabs(goal->width * sinf(goal->Angle));

  float distanceByHeight = estimateGoalDistanceByHeight(height);
  float distanceByWidth = estimateGoalDistanceByWidth(width);

  float avgDistance = (distanceByHeight + distanceByWidth) / 2.0;

  visionLog((35,  "        Goal distance for Width(%0.3f),Height(%0.3f), from width: %0.3f, height: %0.3f, Average: %0.3f", width, height, distanceByWidth, distanceByHeight, avgDistance));
  //printf("        Goal distance for Width(%0.3f),Height(%0.3f), from width: %0.3f, height: %0.3f, Average: %0.3f, OldWidth %0.3f\n", width, height, distanceByWidth, distanceByHeight, avgDistance, oldWidth);

  // should we always use height distance (since tilt always the same and lots of blur horizontally)?? Todd: yes
  float returnDistance = distanceByWidth;//distanceByHeight;//avgDistance;
  //(distanceByWidth < distanceByHeight) ? distanceByWidth : distanceByHeight;

  // Should we ever use this?
  // estimateGoalDistanceByKinematics(goal);

  return returnDistance;
}

float GoalDetector::estimateGoalDistanceByKinematics(FieldLine * goal) {
  float x, y;
  //FastGetPointProjection(goal->bR.x, goal->bR.y, &x, &z);
  transform_.getPointProjection(goal->bR.x, goal->bR.y, x, y);
  return sqrtf(x * x + y * y);
}

float GoalDetector::estimateGoalDistanceByHeight(float height) {
  //return 316220.25 * powf(height, -0.88);   // Old Regression Function using Excel
  // New regression function. Todd 4/17/12
  //return 7490.60245 * powf(0.994199014280916, height);
  // sbarrett 5/22/12
  //return 178594.371031 * powf(height,-0.940758);
  // sbarrett + katie 6/12/12
  //return 392560.190411 * powf(height,-1.059754);
  // piyushk 6/14/12
  return 1243031.764766 * powf(height,-1.235303);

}

float GoalDetector::estimateGoalDistanceByWidth(float width) {
  // powf(x,y) = x^y
  //return 43431.24 * powf(fabs(goal->width * sinf(goal->Angle)), -0.87);   // Old Regression Function using Excel
  // New regression function. Todd 4/17/12
  //return 4270.81632 * powf(0.974517256449273, fabs(goal->width * sinf(goal->Angle)));
  //return 4270.81632 * powf(0.974517256449273, width);
  // sbarrett 5/22/12
  //return 18008.801780 * powf(width,-0.715858);
  // sbarrett + katie 6/12/12
  // return 26258.349912 * powf(width,-0.806607);
  // piyushk 6/14/12
  return 38438.155815 * powf(width,-0.905797);
}

void GoalDetector::sanitizeGoalCandidate(FieldLine *goal) {

  if (!goal->ValidLine) {
      visionLog((35, "    Reject goal %i (ValidLine is false)", goal->id));
      return;
  }

  goal->confidence = 1.0;

  // Remove if not too vertical
  if (goal->Slope < 2 && goal->Slope > -2) {
    visionLog((35, "    Reject goal %i on slope %f (not vertical enough)", goal->id, goal->Slope));
    goal->ValidLine = false;
    totalValidGoals--;
    return;
  }

  // sanitize goals based on left-right edge
  visionLog((35,"    Goal avgX: %i width: %i",goal->avgX,goal->width));
  float left  = goal->avgX - goal->width * 0.5;
  float right = goal->avgX + goal->width * 0.5;
  if (left < 10 || right > IMAGE_WIDTH - 10) {
    visionLog((35,  "    Reject Goal %i close to left-right edge %i", goal->id,goal->avgX));
    goal->ValidLine = false;
    totalValidGoals--;
    return;
  }

  // sanitize based on height to width
  float height = (float)goal->length;
  float width = fabs(goal->width * sinf(goal->Angle));
  float htw = height / width;

  if (htw > 4 * 7.5 || htw < 0.3 * 7.5) {
    visionLog((35,  "    Reject Goal %i htw %f", goal->id,htw));
    goal->ValidLine = false;
    totalValidGoals--;
    return;
  }

  // Sanitize based on elevation of lower most point

  if ((goal->bR.y < IMAGE_HEIGHT - 5) &&
      (goal->avgX > 10 && goal->avgY < IMAGE_WIDTH - 10)) {
    float distance = estimateGoalDistance(goal);
    float bearing = transform_.calculateBearing(goal->bR.x);
    float elevation = transform_.calculateElevation(goal->bR.y);
    //float dist, bear, elev;
    float x, y, z;
    //FastTransformPosition(distance, goal->bR.x, goal->bR.y, &dist, &bear, &elev);
    //FastTransformPoint(distance, goal->bR.x, goal->bR.y, &x, &y, &z);
    transform_.transform(distance, bearing, elevation, x, y, z);
    visionLog((35, "    Goal %i on elevation of bottom point %f, (%f, %5.0f, %5.0f)", goal->id, elevation, distance, goal->bR.x, goal->bR.y));
    visionLog((35, "    Goal %i: vision (%5.0f, %5.0f), transform(%f, %f, %f), distance %f", goal->id, goal->bR.x, goal->bR.y, x, y, z, distance));
    if (z > 100) {
      visionLog((35, "      Reject goal %i on height of bottom point %f", goal->id, z));
      goal->ValidLine = false;
      totalValidGoals--;
      return;
    }
  }

  float greenWhitePercent = greenWhitePercentBelow(goal);
  float minPercent = .1;
  if(greenWhitePercent < minPercent) {
    visionLog((35,  "    Reject Goal %i due to green/white percent %5.3f (minimum %5.3f)", goal->id, greenWhitePercent, minPercent));
    goal->ValidLine = false;
    totalValidGoals--;
    return;
  }
}

void GoalDetector::setGoalObject(int goalIndex, float distance, float bearing, float elevation, int centerX, int centerY, float confidence, int lineIndex) {
  world_objects_->objects_[goalIndex].seen = true;
  world_objects_->objects_[goalIndex].visionBearing = bearing;
  world_objects_->objects_[goalIndex].visionElevation = elevation;
  world_objects_->objects_[goalIndex].visionDistance = distance;
  world_objects_->objects_[goalIndex].imageCenterX = centerX;
  world_objects_->objects_[goalIndex].imageCenterY = centerY;
  world_objects_->objects_[goalIndex].frameLastSeen = vision_frame_info_->frame_id;
  world_objects_->objects_[goalIndex].visionConfidence = confidence;
  world_objects_->objects_[goalIndex].fieldLineIndex = lineIndex;
  world_objects_->objects_[goalIndex].fromTopCamera = (camera_ == Camera::TOP);

}

void GoalDetector::FormGoal() {
  // We saw too many valid goal posts, this looks suspicious, or we saw none at all
  visionLog((31, "FormGoal"));
  if (totalValidGoals > 2 || totalValidGoals < 1) {
    return;
  }

  int index1 = -1, index2 = -1;
  for (int goal = 0; goal < line_detector_->currentLineCounter; goal++) {
    if (currentLine[goal]->ValidLine) {
      if (index1 == -1) {
        index1 = goal;
        if (totalValidGoals != 2) {
          break;
        }
      } else {
        index2 = goal;
        break;
      }
    }
  }

  if (totalValidGoals == 1) { // Unknown Goal Post
    visionLog((32, "    Single Goal Post Detected:-"));
    FieldLine* post = currentLine[index1];

    // The distance estimate also gives us some confidence on its value
    // Values relative to camera
    float confidence = post->confidence;
    float headDistance = estimateGoalDistance(post);
    int headX = post->avgX;
    int headY = post->avgY;
    float headBearing = transform_.calculateBearing(headX);
    float headElevation = transform_.calculateElevation(headY);

    // Values relative to body
    float distance, elevation, bearing;
    //FastTransformPosition(headDistance, headX, headY, &distance, &bearing, &elevation);
    transform_.transformPolar(headDistance, headBearing, headElevation, distance, bearing, elevation);

    // Attempt to find which goal post it based on intersections
    bool canCalculate = false;
    bool isLeft = false;
    for (int corner = 0; corner < cornerPointCounter; corner++) {
      if (cornerPoints[corner]->CornerType) {
        canCalculate = true;
        isLeft = cornerPoints[corner]->PosX < post->avgX;
      }
    }
    int postIndex = WO_UNKNOWN_GOALPOST;
    if (canCalculate) {
      visionLog((32,  "        Post Location detected using T intersection next to goal post as (left = 1, right = 0): %i"));
      postIndex = (isLeft) ? WO_UNKNOWN_LEFT_GOALPOST : WO_UNKNOWN_RIGHT_GOALPOST;
    }

    // Set the unknown goal post
    setGoalObject(postIndex, distance, bearing, elevation,
                  headX, headY,
                  confidence, post->id);

  } else { // Both Goal Posts seen, Yay!

    // Information for post 1
    visionLog((32, "     Both Goal Posts Detected:-"));

    visionLog((32, "     Post 1:-"));
    float confidence1 = currentLine[index1]->confidence;
    float headDistance1 = estimateGoalDistance(currentLine[index1]);
    int headX1 = currentLine[index1]->avgX;
    int headY1 = currentLine[index1]->avgY;
    float headBearing1 = transform_.calculateBearing(headX1);
    float headElevation1 = transform_.calculateElevation(headY1);
    float distance1, bearing1, elevation1;
    //FastTransformPosition(headDistance1, headX1, headY1, &distance1, &bearing1, &elevation1);
    transform_.transformPolar(headDistance1, headBearing1, headElevation1, distance1, bearing1, elevation1);

    // Information for post 2
    visionLog((32, "     Post 2:-"));
    float confidence2 = currentLine[index2]->confidence;
    float headDistance2 = estimateGoalDistance(currentLine[index2]);
    int headX2 = currentLine[index2]->avgX;
    int headY2 = currentLine[index2]->avgY;
    float headBearing2 = transform_.calculateBearing(headX2);
    float headElevation2 = transform_.calculateElevation(headY2);
    float distance2, bearing2, elevation2;
    //FastTransformPosition(headDistance2, headX2, headY2, &distance2, &bearing2, &elevation2);
    transform_.transformPolar(headDistance2, headBearing2, headElevation2, distance2, bearing2, elevation2);

    // TODO(piyushk): Sanity Check distance between the 2 goals, clean up extra long distances

    // Information for goal
    float confidence = confidence1 * confidence2;
    float headBearing = (headBearing1 + headBearing2) / 2;
    float headElevation = (headElevation1 + headElevation2) / 2;

    // Accurate Distance calculation for center of goal
    float bearingDiff = headBearing2 - headBearing1;
    float goalPostProjection = headDistance1 * cosf(bearingDiff);
    float goalCenterProjection = (headDistance2 + goalPostProjection) / 2;
    float headDistance = goalCenterProjection / cosf(bearingDiff / 2);

    // Calculate actual values
    float distance, bearing, elevation;
    //TransformPosition(headDistance, headBearing, headElevation, &distance, &bearing, &elevation);
    transform_.transformPolar(headDistance, headBearing, headElevation, distance, bearing, elevation);

    // Decide which post is left and which post is right
    int post1 = WO_UNKNOWN_LEFT_GOALPOST;
    int post2 = WO_UNKNOWN_RIGHT_GOALPOST;
    if (currentLine[index2]->avgX < currentLine[index1]->avgX) {
      post1 = WO_UNKNOWN_RIGHT_GOALPOST;
      post2 = WO_UNKNOWN_LEFT_GOALPOST;
    }

    // Set Goals and Goal Posts
    setGoalObject(WO_UNKNOWN_GOAL, distance, bearing, elevation,
                  (currentLine[index1]->avgX + currentLine[index2]->avgX) / 2,
                  (currentLine[index1]->avgY + currentLine[index2]->avgY) / 2,
                  confidence, -1);
    setGoalObject(post1, distance1, bearing1, elevation1,
                  currentLine[index1]->avgX, currentLine[index1]->avgY,
                  confidence1, currentLine[index1]->id);
    setGoalObject(post2, distance2, bearing2, elevation2,
                  currentLine[index2]->avgX, currentLine[index2]->avgY,
                  confidence2, currentLine[index2]->id);
  }
}

void GoalDetector::resetYellowGoal() {
  visionLog((32, "resetYellowGoal()"));
  YellowPostCounter = 0;
  totalValidGoals = 0;
}

void GoalDetector::FormYellowGoal() {
  visionLog((32, "FormYellowGoal()"));

  currentLine = yellowPosts;
  color = c_YELLOW;
  line_detector_->currentLineCounter = 0;

  blob_detector_->VerticalBlobSort(color);
  blob_detector_->CalVertBlobData(color);
  sanitizeGoalBlobs();
  line_detector_->FindVertLines(currentLine, color);

  visionLog((32, "     %i initial detections", line_detector_->currentLineCounter));

  YellowPostCounter = line_detector_->currentLineCounter;
  totalValidGoals = line_detector_->currentLineCounter;

  // Calculate the details for the lines comprising of the goal
  for (short line = 0; line < YellowPostCounter; line++) {
    line_detector_->CalLineDetails(yellowPosts[line]);
  }
  for (short lineTo = 0; lineTo < YellowPostCounter; lineTo++) {
    if (!yellowPosts[lineTo]->ValidLine)
      continue;
    for (short lineFrom = lineTo + 1; lineFrom < YellowPostCounter; lineFrom++) {
      if (!yellowPosts[lineFrom]->ValidLine)
        continue;
      mergeLinesVertical(yellowPosts[lineTo], yellowPosts[lineFrom]);
    }
  }
  for (short line = 0; line < YellowPostCounter; line++) {
    sanitizeGoalCandidate(yellowPosts[line]);
  }

  FormGoal();

}


bool GoalDetector::mergeLinesVertical(FieldLine * lineTo, FieldLine * lineFrom) {

  if (!lineTo->ValidLine || !lineFrom->ValidLine) {
    return false;
  }

  // This is an extra sanitary check. If multiple detections are made,
  // the goals should always have similar widths or something is crazy
  if ((lineTo->width / lineFrom->width) > 6.0){
    lineFrom->ValidLine = false;
    totalValidGoals--;
    return false;
  } else if ((lineFrom->width / lineTo->width) > 6.0){
    lineTo->ValidLine = false;
    totalValidGoals--;
    return false;
  }

  float angleDiff = abs(lineTo->Angle - lineFrom->Angle);
  if (angleDiff < (M_PI / 6) || angleDiff > (M_PI - M_PI / 6)) {

    // Calculate Point of intersection
    float xInt = (lineTo->Offset - lineFrom->Offset) / (lineFrom->Slope - lineTo->Slope);
    float yInt = xInt * lineTo->Slope + lineTo->Offset;
    uint16_t avgDiff = abs(lineTo->avgX - lineFrom->avgX);
    float avgWidth = (lineTo->width + lineFrom->width) / 2.0;

    float yOffsetDiff = lineTo->Offset - lineFrom->Offset;
    yOffsetDiff *= cosf(lineTo->Angle);
    yOffsetDiff = fabs(yOffsetDiff);
    float xOffsetDiff = lineTo->Offset / lineTo->Slope - lineFrom->Offset / lineFrom->Slope;
    xOffsetDiff *= sinf(lineTo->Angle);
    xOffsetDiff = fabs(xOffsetDiff);

    if ((yInt > -IMAGE_HEIGHT / 10 && yInt < 11 * IMAGE_HEIGHT / 10) || avgDiff < avgWidth || xOffsetDiff < 64) {

      float posX[MAX_POINTS_PER_LINE];
      float posY[MAX_POINTS_PER_LINE];
      float width[MAX_POINTS_PER_LINE];

      uint16_t counter, counter1 = 0, counter2 = 0;
      for (counter = 0; counter < lineTo->Points + lineFrom->Points && counter < MAX_POINTS_PER_LINE; counter++) {
        if (counter2 == lineFrom->Points || (lineTo->PointsArray[counter1]->PosY <= lineFrom->PointsArray[counter2]->PosY && counter1 < lineTo->Points)) {
          posX[counter] = lineTo->PointsArray[counter1]->PosX;
          posY[counter] = lineTo->PointsArray[counter1]->PosY;
          width[counter] = lineTo->PointsArray[counter1]->Width;
          counter1++;
        } else {
          posX[counter] = lineFrom->PointsArray[counter2]->PosX;
          posY[counter] = lineFrom->PointsArray[counter2]->PosY;
          width[counter] = lineFrom->PointsArray[counter2]->Width;
          counter2++;
        }
      }

      for (counter = 0; counter < lineTo->Points + lineFrom->Points && counter < MAX_POINTS_PER_LINE; counter++) {
        lineTo->PointsArray[counter]->PosX = posX[counter];
        lineTo->PointsArray[counter]->PosY = posY[counter];
        lineTo->PointsArray[counter]->Width = width[counter];
      }

      lineTo->Points = lineTo->Points + lineFrom->Points;
      if (lineTo->Points > MAX_POINTS_PER_LINE)
        lineTo->Points = MAX_POINTS_PER_LINE;

      line_detector_->CalLineDetails(lineTo);
      lineFrom->ValidLine = false;
      totalValidGoals--;

      return true;

    }
  }

  return false;

}

void GoalDetector::setCornerPoints(CornerPoint** points){
  cornerPoints = new CornerPoint * [MAX_CORNERPOINTS];
  for (int i = 0; i < MAX_CORNERPOINTS; i++)
    cornerPoints[i] = points[i];
}


void GoalDetector::setCornerPointCounter(int value) {
  cornerPointCounter = value;
}

float GoalDetector::greenWhitePercentBelow(FieldLine* post){
  int hstep, vstep;
  classifier_->getStepSize(hstep,vstep);

  int
    xmin = post->avgX - post->width / 2.0,
    xmax = post->avgX + post->width / 2.0,
    ymin = post->avgY + post->length / 2.0 + post->length * .02, // rounding on the post increases the length by about 2%
    ymax = post->avgY + post->length / 2.0 + post->length * .2; // this worked in testing, could be increased or decreased

  xmin = std::max(xmin, 0) - (xmin % hstep);
  xmax = std::min(xmax, IMAGE_WIDTH - 1) - (xmax % hstep);
  ymin = std::max(ymin, 0) - (ymin % vstep);
  ymax = std::min(ymax, IMAGE_HEIGHT - 1) - (ymax % vstep);

  int numTotal = 0;
  int numGreen = 0;
  for (int x = xmin; x <= xmax; x += hstep){
    for (int y = ymin; y <= ymax; y += vstep){
      numTotal++;
      Color c = classifier_->xy2color(x,y);
      if (c == c_FIELD_GREEN || c == c_WHITE) numGreen++;
    }
  }

  float pct = (float)numGreen/(float)numTotal;

  if(!numTotal && ymin >= IMAGE_HEIGHT)
  // We can't see below the post due to the height, so we assume 1. There
  // is no reason the robot should see anything else that's yellow and that
  // extends vertically below the lower boundary of the image.
    pct = 1;

  visionLog((43, "check from x: %i, %i, and y: %i, %i, total: %i, green: %i, pct: %5.3f", xmin, xmax, ymin, ymax, numTotal, numGreen, pct));
  return pct;
}
