#include "ImageProcessor.h"
#include <iostream>

ImageProcessor::ImageProcessor(
                               WorldObjectBlock*& world_object_block,
                               BodyModelBlock*& body_model_block,
                               JointBlock*& joint_block,
                               ImageBlock*& image_block,
                               RobotVisionBlock*& robot_vision_block,
                               FrameInfoBlock*& frame_info_block,
                               RobotStateBlock*& robot_state_block,
                               Camera::Type camera)
  :   world_objects_(world_object_block),
      body_model_(body_model_block),
      joint_angles_(joint_block),
      image_(image_block),
      vision_frame_info_(frame_info_block),
      robot_vision_(robot_vision_block),
      robot_state_(robot_state_block),
      calibration_(NULL)
{
  camera_ = camera;
  classifier_ = new Classifier(image_, robot_vision_, visionParams, camera_);
  classifier_->didHighResBallScan = false;
  blob_detector_ = new BlobDetector(robot_vision_, image_, classifier_, visionParams, camera_);
  line_detector_ = new LineDetector(world_objects_, robot_vision_, body_model_, joint_angles_, frame_info_block, transform_, visionParams, classifier_, blob_detector_, camera_);
  goal_detector_ = new GoalDetector(world_objects_, vision_frame_info_, transform_, classifier_, blob_detector_, line_detector_, camera_);
  ball_detector_ = new BallDetector(world_objects_, body_model_, robot_vision_, vision_frame_info_, robot_state_, transform_, classifier_, blob_detector_, camera_);
  robot_detector_ = new RobotDetector(world_objects_, robot_vision_, transform_, classifier_, blob_detector_, camera_);
  cross_detector_ = new CrossDetector(world_objects_, vision_frame_info_, classifier_, blob_detector_, transform_);
}

void ImageProcessor::init(TextLogger* tl){
  textlogger = tl;
  visionParams.init();
  classifier_->init(tl);
  blob_detector_->init(tl);
  line_detector_->init(tl);
  goal_detector_->init(tl);
  ball_detector_->init(tl);
  robot_detector_->init(tl);
  cross_detector_->init(tl);
}

unsigned char* ImageProcessor::getImg() {
  if(camera_ == Camera::TOP)
    return image_->img_top_.get();
  return image_->img_bottom_.get();
}

unsigned char* ImageProcessor::getSegImg(){
  if(camera_ == Camera::TOP)
    return robot_vision_->segImgTop;
  return robot_vision_->segImgBottom;
}

unsigned char* ImageProcessor::getColorTable(){
  return color_table_;
}

const Vision::TransformationProvider& ImageProcessor::getTransform(){
  return transform_;
}

void ImageProcessor::updateTransform(){
  if(calibration_)
    transform_.calculateTransformation(
                                       transform_.getTorsoMatrix(body_model_),
                                       //(joint_angles_->values_[RHipPitchA] + joint_angles_->values_[LHipPitchA]) / 2,
                                       NECKOFFSETZ,
                                       getHeadPan() + calibration_->headPanOffset * DEG_T_RAD,
                                       getHeadTilt() - calibration_->headTiltOffset * DEG_T_RAD,
                                       (camera_ == Camera::TOP ? calibration_->topCameraXOffset : calibration_->bottomCameraXOffset),
                                       (camera_ == Camera::TOP ? calibration_->topCameraZOffset : calibration_->bottomCameraZOffset),
                                       (camera_ == Camera::TOP ? calibration_->topCameraTiltOffset : calibration_->bottomCameraTiltOffset) * DEG_T_RAD,
                                       (camera_ == Camera::TOP ? calibration_->topCameraRollOffset : calibration_->bottomCameraRollOffset) * DEG_T_RAD,
                                       (camera_ == Camera::TOP ? calibration_->topCameraYawOffset : calibration_->bottomCameraYawOffset) * DEG_T_RAD,
                                       getTrueFrontHeight(),
                                       (camera_ == Camera::TOP ? calibration_->topCameraFOVx : calibration_->bottomCameraFOVx) * DEG_T_RAD,
                                       (camera_ == Camera::TOP ? calibration_->topCameraFOVy : calibration_->bottomCameraFOVy) * DEG_T_RAD
                                       );
  else
    transform_.calculateTransformation(transform_.getCameraMatrix(body_model_,camera_), getTrueFrontHeight());

}


bool ImageProcessor::isRawImageLoaded() {
  if(camera_ == Camera::TOP)
    return image_->img_top_;
  return image_->img_bottom_;
}

int ImageProcessor::getImageHeight() {
  if(camera_ == Camera::TOP)
    return TOP_IMAGE_HEIGHT;
  return BOTTOM_IMAGE_HEIGHT;
}

int ImageProcessor::getImageWidth() {
  if(camera_ == Camera::TOP)
    return TOP_IMAGE_WIDTH;
  return BOTTOM_IMAGE_WIDTH;
}

double ImageProcessor::getCurrentTime() {
  return vision_frame_info_->seconds_since_start;
}

void ImageProcessor::setCalibration(CameraCalibration calibration){
  if(calibration_) delete calibration_;
  calibration_ = new CameraCalibration(calibration);
}

void ImageProcessor::processFrame(){
  visionLog((30, "Process Frame camera %i", camera_));

  visionLog((30, "Calculating horizon line"));
  updateTransform();
  HorizonLine horizon = HorizonLine::generate(transform_, 8000, getTrueFrontHeight());
  robot_vision_->horizon = horizon;
  classifier_->bodyExclusionAvailable = false;

  visionLog((30, "Classifying Image", camera_));
  classifier_->clearPreviousHighResScans();
  classifier_->classifyImage(color_table_);
  classifier_->setHorizon(horizon);
  classifier_->constructRuns();

  visionLog((30, "Preprocessing line points and forming blobs"));
  classifier_->preProcessPoints();
  blob_detector_->formAllBlobs();

  visionLog((30, "Detecting balls"));
  ball_detector_->setHorizon(horizon);
  ball_detector_->findBall();

  visionLog((30, "Detecting White Lines"));
  line_detector_->setHorizon(horizon);
  line_detector_->FormLines(c_WHITE);
  line_detector_->mergeLines();
  line_detector_->generateTransformedImage();
  line_detector_->differentiateCurves();

  line_detector_->FilterLines();

  line_detector_->FormCorners();
  line_detector_->DecodeCorners();

  // TODO: set corners for goal detector
  // only do goals on top camera
  if (camera_ == Camera::TOP){
    goal_detector_->resetYellowGoal();
    blob_detector_->resetYellowBlobs();
    if (fabs(getHeadChange()) > DEG_T_RAD * .5) {
      robot_vision_->reported_head_moving = true;
      visionLog((30,"Skipping yellow goal formation because of head movement:%0.3f ",getHeadChange()));
    } else {
      if (robot_vision_->reported_head_moving) {
        visionLog((30,"Head STOP reported"));
        robot_vision_->reported_head_stop_time = getCurrentTime();
        robot_vision_->reported_head_moving = false;
      }
      if (getCurrentTime() - robot_vision_->reported_head_stop_time > HEAD_STOP_THRESHOLD) {
        visionLog((30, "Enough time has passed since head stop (%f), forming yellow goals", getCurrentTime() - robot_vision_->reported_head_stop_time));
        classifier_->startHighResGoalScan();
        classifier_->preProcessGoalPoints();
        blob_detector_->formYellowBlobs();
        goal_detector_->FormYellowGoal();
        classifier_->completeHighResScan();
      } else {
        visionLog((30, "Too soon since head movement stop, skipping yellow goal formation (%f)", getCurrentTime() - robot_vision_->reported_head_stop_time));
      }
    }
  }

  line_detector_->DetermineEndPoints();
  line_detector_->IdentifyLines();
  line_detector_->DecodeLines();

  // only do robots on top camera
  if (camera_ == Camera::TOP){
    visionLog((30, "Detecting robots - calculating band data"));
    blob_detector_->calculateBandBlobData();
    #ifndef TOOL
    if (getTeamColor() == TEAM_BLUE)
    #endif
    {
      visionLog((30, "Detecting Pink Robots"));
      robot_detector_->detectPinkRobots();
    #ifndef TOOL
    } else {
    #endif
      visionLog((30, "Detecting Blue Robots"));
      robot_detector_->detectBlueRobots();
    }

    robot_detector_->detectRobotCluster();
  }
  robot_vision_->lookForCross = true; // Always enabled for now
  if(camera_ == Camera::TOP && robot_state_->role_ == KEEPER && robot_vision_->lookForCross) {
    visionLog((30, "This is the goalie, detecting the cross"));
    cross_detector_->setHorizon(HorizonLine::generate(transform_, 2700, getTrueFrontHeight()));
    cross_detector_->detectCross();
  }
  else visionLog((30, "Skipping cross detection"));

  visionLog((21, "Vision frame process complete"));

}


float ImageProcessor::getTrueFrontHeight() {
  return body_model_->abs_parts_[BodyPart::torso].translation.z;
}

int ImageProcessor::getTeamColor() {
  return robot_state_->team_;
}

void ImageProcessor::SetColorTable(unsigned char* table) {
  color_table_ = table;
  blob_detector_->SetColorTable(table);
}

float ImageProcessor::getHeadTilt() const {
  if (joint_angles_ == NULL)
    return 0;
  return -joint_angles_->values_[HeadTilt];
}

float ImageProcessor::getHeadPan() const {
  if (joint_angles_ == NULL)
    return 0;
  return joint_angles_->values_[HeadPan];
}

float ImageProcessor::getHeadChange() const {
  if (joint_angles_ == NULL)
    return 0;
  return joint_angles_->changes_[HeadPan];
}


std::vector<BallCandidate*> ImageProcessor::getBallCandidates() {
    std::vector<BallCandidate*> candidates;
    for(int i = 0; i < ball_detector_->ballCandCount; i++) {
        candidates.push_back(ball_detector_->candidates[i]);
    }
    return candidates;
}

BallCandidate* ImageProcessor::getBestBallCandidate() {
    if(ball_detector_->bestCandidateFound())
        return ball_detector_->candidates[ball_detector_->bestBallCandIndex];
    return 0;
}
