#include "RobotDetector.h"

#define IS_ROBOT_WHITE(c) ((c == c_WHITE) || (c == c_ROBOT_WHITE))

RobotDetector::RobotDetector(
                             WorldObjectBlock*& world_object_block,
                             RobotVisionBlock*& robot_vision_block,
                             Vision::TransformationProvider& transform,
                             Classifier*& classifier,
                             BlobDetector*& blob_detector,
                             Camera::Type& camera)
  :   world_objects_(world_object_block),
      robot_vision_(robot_vision_block),
      classifier_(classifier),
      blob_detector_(blob_detector),
      camera_(camera),
      transform_(transform)
{
}

float RobotDetector::estimateBandDistanceByWidth(Blob * band) {
  // get rid of warning for now`
  band = band;
  return 100000000.0; // hack to ignore robots that don't have green below them
}

float RobotDetector::estimateBandDistanceByGreenBelow(Blob * band) {
  int hstep, vstep;
  classifier_->getStepSize(hstep,vstep);

  int yStart = band->avgY;
  yStart = yStart - yStart % vstep;

  int xStart = band->xi + (10 * (band->xf - band->xi)) / 100;
  xStart = xStart + hstep - xStart % hstep;

  int xEnd = band->xf - (10 * (band->xf - band->xi)) / 100;
  xEnd = xEnd - xEnd % hstep;

  for (int y = yStart; y < IMAGE_HEIGHT; y += vstep) {
    int greenCounter = 0;
    int otherCounter = 0;

    for (int x = xStart; x <= xEnd; x += hstep) {

      unsigned char c = getSegPixelValueAt(x, y);
      bool isItGreen = c == c_FIELD_GREEN;
      greenCounter += isItGreen;
      otherCounter += !isItGreen;

    }
    //visionLog((39, "Checking green below at y = %i from xi=%i to xf=%i, Green counter: %i, Other Counter: %i", y, xStart, xEnd, greenCounter, otherCounter));
    if (greenCounter > otherCounter) {
      float tx, ty;
      //FastGetPointProjection(band->avgX, y, &tx, &tz);
      transform_.getPointProjection(band->avgX, y, tx, ty);
      return sqrtf(tx * tx + ty * ty);
    }
  }
  return 100000000.0;
}

float RobotDetector::estimateBandDistance(Blob * band) {

  float distance;

  float greenDistance = estimateBandDistanceByGreenBelow(band);
  float widthDistance = estimateBandDistanceByWidth(band);

  distance = (greenDistance < widthDistance) ? greenDistance : widthDistance;

  return distance;

}

void RobotDetector::sanitizeBandCandidate(Blob * band, int color) {
  /*
    if ((int16_t)band->yf - (int16_t)band->yi > 8 && (band->yf - band->yi )) {
    //visionLog((39,  "Robot: Blob at (%i, %i) thrown out based on size.", band->avgX, band->avgY));
    band->invalid = true;
    return;
    }
  */
  int yStart = band->avgY - 1 * band->avgWidth / 2;
  yStart = yStart - yStart % 2;

  int yEnd = band->avgY + 1 * band->avgWidth / 2;
  yEnd = yEnd + yEnd % 2;

  int xStart = band->xi;// - (10 * (band->xf - band->xi)) / 100;
  xStart = xStart - xStart % 4;

  int xEnd = band->xf;// + (10 * (band->xf - band->xi)) / 100;
  xEnd = xEnd + 4 - xEnd % 4;

  if (yStart < 0)
    yStart = 0;
  if (yEnd >= IMAGE_HEIGHT)
    yEnd = IMAGE_HEIGHT - 1;

  if (xStart < 0)
    xStart = 0;
  if (xEnd >= IMAGE_WIDTH)
    xEnd = IMAGE_WIDTH - 1;

  // Check above
  int whiteCounter = 0;
  int otherCounter = 0;

  int y;
  // move up from start until you reach a substantial amount of white (1/3 of total pixels in a row)
  for (y = yStart; y >= 0 && y >= yStart - 4 * band->avgWidth && whiteCounter < (xEnd - xStart) / 12.0; y -= 4 ) {
    whiteCounter = 0;
    for (int x = xStart; x <= xEnd; x += 4) {
      unsigned char c = getSegPixelValueAt(x, y);
      bool isWhite = IS_ROBOT_WHITE(c);
      whiteCounter += isWhite;
      otherCounter += !isWhite && (c != color) && (c != c_UNDEFINED);
    }
  }

  visionLog((39,  "Robot: Blob at (%i, %i) white at top found at %i, other pixels till then %i, width pixels = %i", band->avgX, band->avgY, y, otherCounter, (xEnd - xStart)));
  if (otherCounter > (xEnd - xStart) / 6.0) {
    visionLog((39, "  Blob invalid"));
    band->invalid = true;
    return;
  }

  y = y + 4;
  int yWhiteStart = y;
  whiteCounter = 0;
  otherCounter = 0;
  visionLog((39, "scanning from x = %i to %i, y = %i to %i", xStart, xEnd, 0, yWhiteStart));
  for (; y >= 0 && y >= yWhiteStart - 6; y -= 2) {
    for (int x = xStart; x <= xEnd; x += 4) {
      unsigned char c = getSegPixelValueAt(x, y);
      bool isItWhite = IS_ROBOT_WHITE(c);
      whiteCounter += isItWhite;
      otherCounter += !isItWhite;
    }
  }

  if (whiteCounter < 3 || 6 * whiteCounter < 2 * otherCounter) {
    visionLog((39,  "Robot: Blob at (%i, %i) thrown out because of too little white above (%i of %i).", band->avgX, band->avgY, whiteCounter, whiteCounter + otherCounter));
    //visionLog((39,  "    (%i, %i, %i).", yWhiteStart, xStart, xEnd));
    band->invalid = true;
    return;
  }

  whiteCounter = 0;
  otherCounter = 0;

  // move up from start until you reach a substantial amount of white (1/3 of total pixels in a row)
  for (y = yEnd; y < IMAGE_HEIGHT && y <= yEnd + 4 * band->avgWidth && whiteCounter <= (xEnd - xStart) / 12.0; y += 4 ) {
    whiteCounter = 0;
    for (int x = xStart; x <= xEnd; x += 4) {
      unsigned char c = getSegPixelValueAt(x, y);
      bool isWhite = IS_ROBOT_WHITE(c);
      whiteCounter += isWhite;
      otherCounter += !isWhite && c != color && c != c_UNDEFINED;
    }
  }

  visionLog((39,  "Robot: Blob at (%i, %i) white at top found at %i, other pixels till then %i, width pixels = %i", band->avgX, band->avgY, y, otherCounter, (xEnd - xStart)));
  if (otherCounter > (xEnd - xStart) / 6.0) {
    visionLog((39, "  Blob invalid"));
    band->invalid = true;
    return;
  }

  y = y - 4;
  yWhiteStart = y;
  whiteCounter = 0;
  otherCounter = 0;
  for (; y < IMAGE_HEIGHT && y <= yWhiteStart + 6; y += 2) {
    for (int x = xStart; x <= xEnd; x += 4) {
      unsigned char c = getSegPixelValueAt(x, y);
      bool isItWhite = IS_ROBOT_WHITE(c);
      whiteCounter += isItWhite;
      otherCounter += !isItWhite;
    }
  }

  if (whiteCounter < 3 || 6 * whiteCounter < 2 * otherCounter) {
    visionLog((39,  "Robot: Blob at (%i, %i) thrown out because of too little white below (%i of %i).", band->avgX, band->avgY, whiteCounter, whiteCounter + otherCounter));
    //visionLog((39,  "    (%i, %i, %i).", yWhiteStart, xStart, xEnd));
    band->invalid = true;
    return;
  }

  band->invalid = false;
}

void RobotDetector::sanitizeBandCandidateOnElevation(Blob * band, float y) {

  visionLog((39,  "Robot: Blob at (%i, %i) has z height %f", band->avgX, band->avgY, y));
  if (y > 250 || y < -250) {
    visionLog((39,  "Robot: Blob at (%i, %i) thrown out because of bad elevation %f", band->avgX, band->avgY, y));
    band->invalid = true;
    return;
  }

}

void RobotDetector::detectBlueRobots() {
  blueRobots_.clear();
  Blob** verticalBlob = blob_detector_->verticalBlob;
  uint32_t* verticalBlobCount = blob_detector_->verticalBlobCount;
  int mergeCount = 100;
  uint16_t mergeIndex[mergeCount];
  blob_detector_->mergeVerticalBlobs(c_BLUE,mergeIndex,mergeCount);

  int opp = WO_OPPONENT1;

  // Todd: make sure we start at first opponent not already filled in
  while (world_objects_->objects_[opp].seen && opp <= WO_OPPONENT4)
    opp++;

  //if (opp > WO_OPPONENT1)
  //  cout << "Camera " << camera_ << " start filling in blue robots at index " << opp << endl;

  if (opp > WO_OPPONENT4){
    // no more opponent slots
    return;
  }

  for (uint32_t i = 0; i < verticalBlobCount[c_BLUE]; i++) {
    Blob * band = &verticalBlob[c_BLUE][i];
    if(mergeIndex[i] == mergeCount) {
      band->invalid = false;
    }
    else {
      band->invalid = true;
      visionLog((39, "Robot: Band %i at x:(%i,%i) ~ %i, y:%i thrown out because it was merged with blob %i", i, band->xi, band->xf, band->avgX, band->avgY, mergeIndex[i]));
    }

    for (uint32_t j = 0; j < i; j++){
      if (verticalBlob[c_BLUE][j].invalid)
        continue;
      float dx = verticalBlob[c_BLUE][j].avgX - band->avgX;
      float dy = verticalBlob[c_BLUE][j].avgY - band->avgY;
      if (dx * dx + dy * dy < 100 * 100)
        {
          band->invalid = true;
          visionLog((39,"Robot: Blob at (%i, %i) thrown out - too close to blob at (%i, %i)",band->avgX,band->avgY,verticalBlob[c_BLUE][j].avgX,verticalBlob[c_BLUE][j].avgY));
          break;
        }
    }
    if (band->invalid) {
      continue;
    }
    visionLog((39, "Robot: Band %i at x:(%i,%i) ~ %i, y:%i selected for band check", i, band->xi, band->xf, band->avgX, band->avgY));
    /*
      TODO: Remove after investigating w/ Piyush. It doesn't look like these were being used at all, since the bluePosts elements
      were never assigned to after initialization. - JM 3/15/2012

      for (int goal = 0; goal < BluePostCounter; goal++) {
      if (!bluePosts[goal]->ValidLine)
      continue;

      float perpendicularSlope = -1 / bluePosts[goal]->Slope;
      float perpendicularOffset = band->avgY - perpendicularSlope * band->avgX;

      int xInt = (perpendicularOffset - bluePosts[goal]->Offset) / (bluePosts[goal]->Slope - perpendicularSlope);
      int yInt = perpendicularSlope * xInt + perpendicularOffset;

      float dist = sqrtf((band->avgX - xInt) * (band->avgX - xInt) + (band->avgY - yInt) * (band->avgY - yInt));

      //visionLog((39,  "Robot: Blob at (%i, %i) xInt: %i, yInt: %i, dist: %f", band->avgX, band->avgY, xInt, yInt, dist));
      if (dist < bluePosts[goal]->width + 10) {
      band->invalid = true;
      break;
      }
      }
      if (band->invalid)
      continue;
    */

    sanitizeBandCandidate(band, c_BLUE);
    if (band->invalid)
      continue;

    // Yay! A Detection. Fill in values;
    float distance = estimateBandDistance(band);

    if (distance < 100 || distance > 6000) { // distance invalid
      visionLog((39, "blob thrown out because of distance %5.3f", distance));
      continue;
    }

    float x, y, z;
    float bearing = transform_.calculateBearing(band->avgX);
    float elevation = transform_.calculateElevation(band->avgY);
    transform_.transform(distance, bearing, elevation, x, y, z);
    visionLog((39, "Transforming band distance (%5.3f), bearing (%5.3f), elevation (%5.3f)", distance, bearing, elevation));

    // Sanitize based on elevation
    sanitizeBandCandidateOnElevation(band, z);
    if (band->invalid)
      continue;

    // no more than 4, or we're detecting something crazy
    if (opp == WO_OPPONENT4 + 1) {
      //Remove all sightings
      world_objects_->objects_[WO_OPPONENT1].seen = false;
      world_objects_->objects_[WO_OPPONENT2].seen = false;
      world_objects_->objects_[WO_OPPONENT3].seen = false;
      world_objects_->objects_[WO_OPPONENT4].seen = false;
      break;
    }

    WorldObject *wo = &world_objects_->objects_[opp];
    wo->seen = true;
    transform_.transformPolar(distance, bearing, elevation, wo->visionDistance, wo->visionBearing, wo->visionElevation);
    wo->imageCenterX = band->avgX;
    wo->imageCenterY = band->avgY;
    wo->fromTopCamera = (camera_ == Camera::TOP);

    opp++;

    blueRobots_.push_back(band);
  }

}

void RobotDetector::detectPinkRobots() {
  pinkRobots_.clear();
  Blob** verticalBlob = blob_detector_->verticalBlob;
  uint32_t* verticalBlobCount = blob_detector_->verticalBlobCount;
  int mergeCount = 100;
  uint16_t mergeIndex[mergeCount];
  blob_detector_->mergeVerticalBlobs(c_PINK,mergeIndex,mergeCount);

  int opp = WO_OPPONENT1;

  // Todd: make sure we start at first opponent not already filled in
  while (world_objects_->objects_[opp].seen && opp <= WO_OPPONENT4)
    opp++;

  //if (opp > WO_OPPONENT1)
  //  cout << "Camera " << camera_ << " start filling in pink robots at index " << opp << endl;

  if (opp > WO_OPPONENT4){
    // no more opponent slots
    return;
  }

  for (uint32_t i = 0; i < verticalBlobCount[c_PINK]; i++) {
    Blob * band = &verticalBlob[c_PINK][i];
    if(mergeIndex[i] == mergeCount)
      band->invalid = false;
    else band->invalid = true;

    for (uint32_t j = 0; j < i; j++){
      if (verticalBlob[c_PINK][j].invalid)
        continue;
      float dx = verticalBlob[c_PINK][j].avgX - band->avgX;
      float dy = verticalBlob[c_PINK][j].avgY - band->avgY;
      if (dx * dx + dy * dy < 100 * 100)
        {
          band->invalid = true;
          //visionLog((39,"Robot: Blob at (%i, %i) thrown out - too close to blob at (%i, %i)",band->avgX,band->avgY,verticalBlob[c_PINK][j].avgX,verticalBlob[c_PINK][j].avgY));
          break;
        }
    }
    if (band->invalid)
      continue;

    sanitizeBandCandidate(band, c_PINK);
    if (band->invalid)
      continue;

    // Yay! A Detection. Fill in values;
    float distance = estimateBandDistance(band);

    float x, y, z;
    float bearing = transform_.calculateBearing(band->avgX);
    float elevation = transform_.calculateElevation(band->avgY);
    //FastTransformPoint(distance, band->avgX, band->avgY, &x, &y, &z);
    transform_.transform(distance, bearing, elevation, x, y, z);
    visionLog((39, "Transforming band distance (%5.3f), bearing (%5.3f), elevation (%5.3f)", distance, bearing, elevation));

    // Sanitize based on elevation
    sanitizeBandCandidateOnElevation(band, z);
    if (band->invalid)
      continue;

    // no more than 4, or we're detecting something crazy
    if (opp == WO_OPPONENT4 + 1) {
      //Remove all sightings
      world_objects_->objects_[WO_OPPONENT1].seen = false;
      world_objects_->objects_[WO_OPPONENT2].seen = false;
      world_objects_->objects_[WO_OPPONENT3].seen = false;
      world_objects_->objects_[WO_OPPONENT4].seen = false;
      break;
    }

    WorldObject &wo = world_objects_->objects_[opp];
    wo.seen = true;
    transform_.transformPolar(distance, bearing, elevation, wo.visionDistance, wo.visionBearing, wo.visionElevation);
    wo.imageCenterX = band->avgX;
    wo.imageCenterY = band->avgY;
    wo.fromTopCamera = (camera_ == Camera::TOP);

    opp++;

    pinkRobots_.push_back(band);
  }

}

void RobotDetector::detectRobotCluster() {
  visionLog((40,"Detection robot clusters"));
  // PARAMETERS for the cluster detection
  const int NUM_COLS = 4;
  const int NUM_ROWS = 4;
  const int MIN_CLUSTER_LENGTH = 2;
  const float MIN_WHITE_FRAC = 0.15;
  const float MIN_GREEN_FRAC = 0.06;
  const int END_GREEN_ROW = 1; // inclusive
  const int MIN_GREEN_COUNT = 2;

  const int COL_WIDTH = IMAGE_WIDTH / NUM_COLS;
  const int ROW_HEIGHT = IMAGE_HEIGHT / NUM_ROWS;
  const float COL_SIZE = IMAGE_HEIGHT * COL_WIDTH / 4; // /4 because every 4 pixels in seg
  const float CELL_SIZE = COL_SIZE / NUM_ROWS;
  WorldObject &wo = world_objects_->objects_[WO_ROBOT_CLUSTER];
  int whiteCounts[NUM_COLS];
  int greenCounts[NUM_ROWS][NUM_COLS];

  wo.seen = false; // assume we don't see it
  // count the white and green
  for (int col = 0; col < NUM_COLS; col++) {
    whiteCounts[col] = 0;
    for (int row = 0; row < NUM_ROWS; row++)
      greenCounts[row][col] = 0;
  }

  // check the left and right greens
  for (int x = 0; x < IMAGE_WIDTH; x += 4) {
    int col = (int)(x / COL_WIDTH);
    for (int y = 0; y < IMAGE_HEIGHT; y++) {
      int row = (int)(y / ROW_HEIGHT);

      unsigned char c = getSegPixelValueAt(x, y);
      whiteCounts[col] += IS_ROBOT_WHITE(c);
      greenCounts[row][col] += (c == c_FIELD_GREEN);
    }
  }

  float whiteFrac;
  bool whiteFound[NUM_COLS];
  float greenFracs[NUM_ROWS][NUM_COLS];
  // count up the white and green
  for (int col = 0; col < NUM_COLS; col++) {
    whiteFrac = whiteCounts[col] / COL_SIZE;
    visionLog((41,"  col %i, white: %f",col,whiteFrac));
    for (int row = 0; row < NUM_ROWS; row++) {
      greenFracs[row][col] = greenCounts[row][col] / CELL_SIZE;
      visionLog((41,"    section %i,%i, green: %f",col,row,greenFracs[row][col]));
    }
    whiteFound[col] = (whiteFrac > MIN_WHITE_FRAC);
  }

  // check for consecutive white sections
  int clusterStart = -1;
  int clusterLength = 0;
  // note: we're only checking for the first cluster of length >= 2
  for (int i = 0; i < NUM_COLS; i++) {
    if (whiteFound[i]) {
      if (clusterStart >= 0) {
        if (clusterStart + clusterLength == i) {
          // add to cluster
          clusterLength++;
        } else if (clusterLength < MIN_CLUSTER_LENGTH) {
          // start a new cluster
          clusterStart = i;
          clusterLength = 1;
        } // else ignore
      } else {
        // start a new cluster
        clusterStart = i;
        clusterLength = 1;
      }
    }
  }
  visionLog((41,"Best cluster: start: %i length: %i",clusterStart,clusterLength));
  if (clusterLength < MIN_CLUSTER_LENGTH) {
    visionLog((40,"Cluster too short: %i",clusterLength));
    return;
  }

  // now check the green
  int greenCount = 0;
  for (int col = 0; col < NUM_COLS; col++) {
    for (int row = 0; row <= END_GREEN_ROW; row++) {
      greenCount += (greenFracs[row][col] > MIN_GREEN_FRAC);
    }
  }
  if (greenCount < MIN_GREEN_COUNT) {
    visionLog((40,"Not enough green for cluster: %i",greenCount));
    return;
  }


  float clusterCenter = (clusterStart + clusterLength / 2.0) * COL_WIDTH;
  float dist = 200; // arbitrary I hope
  float elev = transform_.calculateElevation(IMAGE_HEIGHT_DIV2); // probably doesn't matter
  float bear = transform_.calculateBearing(clusterCenter);
  visionLog((40,"Raw cluster: %i %f %f",clusterStart,clusterCenter,bear));
  transform_.transformPolar(dist,bear,elev, wo.visionDistance, wo.visionBearing, wo.visionElevation);
  if (fabs(wo.visionBearing) > DEG_T_RAD * 20) {
    visionLog((40,"Cluster thrown out for too high of bearing: %f",wo.visionBearing));
    return;
  }
  wo.seen = true;
  wo.fromTopCamera = (camera_ == Camera::TOP);
  visionLog((40,"Cluster found with bearing: %f",wo.visionBearing));
}

std::list<Blob*> RobotDetector::getBlueRobots() {
  return blueRobots_;
}

std::list<Blob*> RobotDetector::getPinkRobots() {
  return pinkRobots_;
}
