#include <vision/TransformationProvider.h>
#include <common/RobotInfo.h>
#include <memory/BodyModelBlock.h>

namespace Vision {

  void TransformationProvider::calculateTransformation(float headZOffset, float headTilt, float headPan, float bodyTilt, float bodyRoll, float cameraTiltOffset, float cameraXOffset, float cameraZOffset, float trueFrontHeight) {

    // Reset transformation matrix
    transformationMatrix_ = Pose3D(0,0,0);

    RotationMatrix rotBodyTilt;
    rotBodyTilt.rotateY(bodyTilt);
    transformationMatrix_.rotate(rotBodyTilt);

    // headZOffset = NECKOFFSETZ + HIPOFFSETZ
    Vector3<float> translateHeadPos(0, 0, headZOffset);
    transformationMatrix_.translate(translateHeadPos);

    RotationMatrix rotHeadPan;
    rotHeadPan.rotateZ(headPan);
    transformationMatrix_.rotate(rotHeadPan);

    RotationMatrix rotHeadTilt;
    rotHeadTilt.rotateY(headTilt);
    transformationMatrix_.rotate(rotHeadTilt);

    Vector3<float> translateCamPos(cameraXOffset, 0, cameraZOffset);
    transformationMatrix_.translate(translateCamPos);

    RotationMatrix rotCamAngle;
    rotCamAngle.rotateY(cameraTiltOffset);
    transformationMatrix_.rotate(rotCamAngle);

    inverseTransformationMatrix_ = transformationMatrix_.invert();

    trueFrontHeight_ = trueFrontHeight;

  }

  void TransformationProvider::calculateTransformation(const Pose3D &forwardMatrix, float trueFrontHeight) {
    transformationMatrix_ = forwardMatrix;
    //int x = 140, y = 0, z = 2000 - 240;
    //std::cout << "final standard point: " << transformationMatrix_ * Vector3<float>(x, y, z) << "\n";
    inverseTransformationMatrix_ = transformationMatrix_.invert();
    trueFrontHeight_ = trueFrontHeight;
  }

  void TransformationProvider::calculateTransformation(const Pose3D &headMatrix, float cameraXOffset, float cameraZOffset, float cameraTiltOffset, float cameraRollOffset, float trueFrontHeight) {
    transformationMatrix_ = Pose3D(headMatrix)
      .translate(cameraXOffset, 0, cameraZOffset)
      .rotateY(cameraTiltOffset)
      .rotateX(cameraRollOffset);
    inverseTransformationMatrix_ = transformationMatrix_.invert();
    trueFrontHeight_ = trueFrontHeight;
  }

  void TransformationProvider::calculateTransformation(
    const Pose3D &torsoMatrix,
    float neckZOffset,
    float headPan,
    float headTilt,
    float cameraXOffset,
    float cameraZOffset,
    float cameraTiltOffset,
    float cameraRollOffset,
    float cameraYawOffset,
    float trueFrontHeight,
    float fovX,
    float fovY) {
        TAN_FOV_X_BY_2 = tanf(fovX / 2.0);
        TAN_FOV_Y_BY_2 = tanf(fovY / 2.0);
        transformationMatrix_ = Pose3D(0,0,0)
            .translate(0,0,trueFrontHeight)
            .rotate(torsoMatrix.rotation)
            .translate(0,0,neckZOffset)
            .rotateZ(headPan)
            .rotateY(headTilt)
            .translate(cameraXOffset, 0, cameraZOffset)
            .rotateX(cameraRollOffset)
            .rotateY(cameraTiltOffset)
            .rotateZ(cameraYawOffset);
            ;

        inverseTransformationMatrix_ = transformationMatrix_.invert();
        trueFrontHeight_ = trueFrontHeight;
  }


  void TransformationProvider::transform(float distance, float bearing, float elevation, float &x, float &y, float &z) const {

    //float cosBearing = cosf(bearing);
    //float sinBearing = sinf(bearing);
    //float cosElevation = cosf(elevation);
    //float sinElevation = sinf(elevation);

    //// Calculate the point in the space of the camera
    //float xCamOld = distance * cosBearing * cosElevation;
    //float yCamOld = distance * sinBearing;
    //float zCamOld = distance * cosBearing * sinElevation;

    float tanBearing = tanf(bearing);
    float tanElevation = tanf(elevation);
    float denom = sqrt(1 + tanBearing * tanBearing + tanElevation * tanElevation);
    float xCam = distance / denom;
    float yCam = xCam * tanBearing;
    float zCam = xCam * tanElevation;
    //printf("old %f %f %f  new %f %f %f denom %f\n",xCamOld,yCamOld,zCamOld,xCam,yCam,zCam,denom);
    //printf("  distance %f bearing %f  elevation %f\n",distance,bearing,elevation);

    Vector3<float> ptCam(xCam, yCam, zCam);

    // Apply transformation matrix to get point relative to body
    Vector3<float> transformedPt = transformationMatrix_ * ptCam;

    x = transformedPt.x;
    y = transformedPt.y;
    z = transformedPt.z;
  }

  void TransformationProvider::transformPolar(float distance, float bearing, float elevation, float &transformedDistance,float &transformedBearing,float &transformedElevation) const {

    float x, y, z;
    transform(distance, bearing, elevation, x, y, z);

    if (x == 0.0){
      if (y > 0.0) transformedBearing = (M_PI / 2);
      else transformedBearing = (- M_PI / 2);
    }
    else transformedBearing = atan2f(y, x);

    transformedElevation = atan2f(z, sqrtf(x * x + y * y));
    // NOTE: This distance is along the ground
    transformedDistance = sqrtf(x*x + y*y + z*z) * cosf(transformedElevation);
  }

  void TransformationProvider::inverseTransform(float x, float y, float z, float &distance, float &bearing, float &elevation) const {
    invertPoint(x,y,z,x,y,z);
    computeDistanceBearingElevation(x,y,z,distance,bearing,elevation);
  }

  void TransformationProvider::computeDistanceBearingElevation(float x, float y, float z, float& distance, float& bearing, float& elevation) const {
    distance = sqrtf(x * x + y * y + z * z);
    bearing = asin(y/distance);
    elevation = asin(z/(distance * cos(bearing)));
  }

  void TransformationProvider::invertPoint(float x, float y, float z, float& ix, float& iy, float& iz) const {
    Vector3<float> bodyPt(x, y, z);
    Vector3<float> transformedPt = inverseTransformationMatrix_ * bodyPt;
    ix = transformedPt.x;
    iy = transformedPt.y;
    iz = transformedPt.z;
  }
  void TransformationProvider::invertPoint(Vector3<float> in, Vector3<float>& out) const {
    invertPoint(in.x,in.y,in.z,out.x,out.y,out.z);
  }
  void TransformationProvider::getPointProjection(float imageX, float imageY, float &x, float &y) const {

    float elev = calculateElevation(imageY);
    float bear = calculateBearing(imageX);

    float dist1 = 2000;
    float x1, y1, z1;
    float dist2 = 20000;
    float x2, y2, z2;

    transform(dist1, bear, elev, x1, y1, z1);
    transform(dist2, bear, elev, x2, y2, z2);

    // Now that we have the equation of the line, set z = the height and voila
    float z;
    z = -trueFrontHeight_;
    float k = (z - z1) / (z2 - z1);
    x = (x2 - x1) * k + x1;
    y = (y2 - y1) * k + y1;

  }

  void TransformationProvider::getPolarProjection(float imageX, float imageY, float &transformedDistance, float &transformedBearing, float &transformedElevation) const {

    float x, y, z;
    z = -trueFrontHeight_;
    getPointProjection(imageX, imageY, x, y);

    if (x == 0.0){
      if (y > 0.0) transformedBearing = (M_PI / 2);
      else transformedBearing = (- M_PI / 2);
    }
    else transformedBearing = atan2f(y, x);

    transformedElevation = atan2f(z, sqrtf(x * x + y * y));
    // NOTE: This distance is along the ground
    transformedDistance = sqrtf(x*x + y*y + z*z) * cosf(transformedElevation);

  }

  void TransformationProvider::testPose3D() {
    transformationMatrix_ = Pose3D(0,0,0);

    Vector3<float> translateCamPos(100, 200, 500);
    transformationMatrix_.translate(translateCamPos);

    RotationMatrix rotCamAngle;
    rotCamAngle.rotateZ(M_PI / 6);
    transformationMatrix_.rotate(rotCamAngle);

    Vector3<float> ptCam(500, 0, 0);

    // Apply transformation matrix to get point relative to body
    Vector3<float> transformedPt = transformationMatrix_ * ptCam;

    std::cout << 500 * cosf(M_PI / 6) + 100 << " " << 500 * sinf(M_PI / 6) + 200 << " " << 0 + 500 << std::endl;
    std::cout << transformedPt.x << " " << transformedPt.y << " " << transformedPt.z << std::endl;

  }

  void TransformationProvider::rotatePoint(const float x, const float y, const float angle, float& rx, float& ry) const {
      Pose3D matrix = Pose3D(x,y,0);
      matrix.rotateZ(angle * DEG_T_RAD);
      rx = matrix.translation.x;
      ry = matrix.translation.y;
      std::cout << "(" << x << "," << y << ")" << "--(" << angle << ")-->" << "(" << rx << "," << ry << ")\n";
  }

  Point2D TransformationProvider::backProject(float x, float y, float z) const {
      return backProject(Vector3<float>(x,y,z));
  }

  Point2D TransformationProvider::backProject(const Vector3<float> v) const {
      float distance, bearing, elevation;
      inverseTransform(v.x, v.y, v.z, distance, bearing, elevation);
      float x = invertBearing(bearing), y =  invertElevation(elevation);
      return Point2D(x,y);
  }

  LineSegment TransformationProvider::backProject(const Vector3<float> v1, const Vector3<float> v2) const {
      Point2D p1 = backProject(v1), p2 = backProject(v2);
      LineSegment segment;
      segment.start = p1;
      segment.end = p2;
      return segment;
  }

  Point2D TransformationProvider::forwardProject(float imageX, float imageY) const {
      float x, y;
      getPointProjection(imageX, imageY, x, y);
      return Point2D(x,y);
  }

//#define REL_PARTS_FOR_MATRICES body_model->rel_parts_
// doing it without rotations now?
#define REL_PARTS_FOR_MATRICES body_model->rel_parts_no_rotations_

  Pose3D TransformationProvider::getTorsoMatrix(BodyModelBlock *body_model) {
    //Pose3D torsoRotation = Pose3D(0,0,0);
    //torsoRotation.rotateY(-(joint_angles_->values_[RHipPitch] + joint_angles_->values_[LHipPitch]) / 2 - 46.23 * DEG_T_RAD);
    //std::cout << "rhippitch: " << joint_angles_->values_[RHipPitch] << "\n";
    //std::cout << "lhippitch: " << joint_angles_->values_[LHipPitch] << "\n";
    //std::cout << (-(joint_angles_->values_[RHipPitch] + joint_angles_->values_[LHipPitch]) / 2 - 46.23 * DEG_T_RAD) * RAD_T_DEG << "\n";
    //return torsoRotation;
    if (body_model == NULL)
      return Pose3D(0,0,0);
    return REL_PARTS_FOR_MATRICES[BodyPart::torso];
  }

  Pose3D TransformationProvider::getCameraMatrix(BodyModelBlock *body_model,Camera::Type camera) {
    if (body_model == NULL)
      return Pose3D(0,0,0);
    if (camera == Camera::TOP)
      return REL_PARTS_FOR_MATRICES[BodyPart::top_camera];
    return REL_PARTS_FOR_MATRICES[BodyPart::bottom_camera];
  }
}
