#ifndef TRANSFORMATIONPROVIDER_G183NKTE
#define TRANSFORMATIONPROVIDER_G183NKTE

#include <math/Pose3D.h>
#include <math/Vector2.h>
#include <math/Geometry.h>
#include <common/RobotInfo.h>

class BodyModelBlock;

namespace Vision {
  /**
   * Provides transformation between the camera image into the 3d space
   * relative to the robot and vice versa. Implements a pinhole camera
   * model
   */
  class TransformationProvider {
  public:
    Pose3D transformationMatrix_;
    Pose3D inverseTransformationMatrix_;
    float trueFrontHeight_;
    float TAN_FOV_X_BY_2;
    float TAN_FOV_Y_BY_2;

  public:
    TransformationProvider() : TAN_FOV_X_BY_2(tanf(FOVx / 2.0)), TAN_FOV_Y_BY_2(tanf(FOVy / 2.0)) {}
    void calculateTransformation(float headZOffset, float headTilt, float headPan, float bodyTilt, float bodyRoll, float cameraTiltOffset, float cameraXOffset, float cameraZOffset, float trueFrontHeight);
    void calculateTransformation(const Pose3D &forwardMatrix, float trueFrontHeight);
    void calculateTransformation(const Pose3D &headMatrix, float cameraXOffset, float cameraZOffset, float cameraTiltOffset, float cameraRollOffset, float trueFrontHeight);
    void calculateTransformation(
        const Pose3D &torsoMatrix,
        float neckZOffset,
        float headPan,
        float headTilt,
        float cameraXOffset,
        float cameraZOffset,
        float cameraTiltOffset,
        float cameraRollOffset,
        float cameraYawOffset,
        float trueFrontHeight,
        float fovX,
        float fovY);

    void transform(float distance, float bearing, float elevation, float &x, float &y, float &z) const;
    void transformPolar(float distance, float bearing, float elevation, float &transformedDistance, float &transformedBearing, float &transformedElevation) const;

    void inverseTransform(float x, float y, float z, float &distance, float &bearing, float &elevation) const;
    void invertPoint(float,float,float,float&,float&,float&) const;
    void invertPoint(Vector3<float>,Vector3<float>&) const;
    void computeDistanceBearingElevation(float,float,float,float&,float&,float&) const;
    void getPointProjection(float imageX, float imageY, float &x, float &y) const;
    void getPolarProjection(float imageX, float imageY, float &transformedDistance, float &transformedBearing, float &transformedElevation) const;

    void rotatePoint(const float x, const float y, const float angle, float& rx, float& ry) const;
    Point2D backProject(float, float, float) const;
    Point2D backProject(const Vector3<float>) const;
    LineSegment backProject(const Vector3<float>,const Vector3<float>) const;
    Point2D forwardProject(float,float) const;

    inline float calculateBearing(float imageX) const {
      return atanf((IMAGE_X / 2.0 - imageX) / (IMAGE_X / (2.0 * TAN_FOV_X_BY_2)));
    }

    inline float calculateElevation(float imageY) const {
      return atanf((IMAGE_Y / 2.0 - imageY) / (IMAGE_Y / (2.0 * TAN_FOV_Y_BY_2)));
    }

    inline float invertBearing(float bearing) const {
      return (IMAGE_X / 2.0) * (1.0 - (tanf(bearing) / TAN_FOV_X_BY_2));
    }

    inline float invertElevation(float elevation) const {
      return (IMAGE_Y / 2.0) * (1.0 - (tanf(elevation) / TAN_FOV_Y_BY_2));
    }

    void testPose3D();

    Pose3D getTorsoMatrix(BodyModelBlock *body_model);
    Pose3D getCameraMatrix(BodyModelBlock *body_model,Camera::Type camera);

  };

}

#endif /* end of include guard: TRANSFORMATIONPROVIDER_G183NKTE */
