#include "VisionModule.h"
#include <memory/FrameInfoBlock.h>
#include <memory/JointBlock.h>
#include <memory/RobotVisionBlock.h>
#include <memory/SensorBlock.h>
#include <memory/ImageBlock.h>
#include <memory/BodyModelBlock.h>
#include <memory/RobotStateBlock.h>
#include <memory/CameraBlock.h>
#include <memory/WorldObjectBlock.h>
#include <memory/RobotInfoBlock.h>

#include <boost/lexical_cast.hpp>

void VisionModule::specifyMemoryDependency() {
  //last_frame_processed_ = 0;

  requiresMemoryBlock("vision_frame_info");
  requiresMemoryBlock("vision_joint_angles");
  requiresMemoryBlock("robot_vision");
  requiresMemoryBlock("vision_sensors");
  requiresMemoryBlock("raw_image");
  requiresMemoryBlock("robot_state");
  requiresMemoryBlock("vision_body_model");
  requiresMemoryBlock("camera_info");
  requiresMemoryBlock("robot_info");

  providesMemoryBlock("world_objects");
}

void VisionModule::specifyMemoryBlocks() {
  getOrAddMemoryBlock(vision_frame_info_,"vision_frame_info");
  getOrAddMemoryBlock(joint_angles_,"vision_joint_angles");
  getOrAddMemoryBlock(robot_vision_,"robot_vision");
  getOrAddMemoryBlock(sensors_,"vision_sensors");
  getOrAddMemoryBlock(image_,"raw_image");
  getOrAddMemoryBlock(world_objects_,"world_objects");
  getOrAddMemoryBlock(robot_state_,"robot_state");
  getOrAddMemoryBlock(body_model_,"vision_body_model");
  getOrAddMemoryBlock(camera_info_,"camera_info");

  getOrAddMemoryBlock(robot_info_,"robot_info");
}

void VisionModule::processFrame() {

  // reset world objects
  world_objects_->reset();

  //visionLog((30, "pre unknown line 1 seen %i, last seen %i", world_objects_->objects_[WO_UNKNOWN_FIELD_LINE_1].seen, world_objects_->objects_[WO_UNKNOWN_FIELD_LINE_1].frameLastSeen));

  // Todd: look for ball in bottom first, then top, then can do high res in top if need be
  visionLog((30, "Processing bottom camera"));
  bottom_processor_->processFrame();

  visionLog((30, "Processing top camera"));
  top_processor_->processFrame();

  //visionLog((30, "post unknown line 1 seen %i, last seen %i", world_objects_->objects_[WO_UNKNOWN_FIELD_LINE_1].seen, world_objects_->objects_[WO_UNKNOWN_FIELD_LINE_1].frameLastSeen));

}

void VisionModule::updateTransforms() {
    top_processor_->updateTransform();
    bottom_processor_->updateTransform();
}

bool VisionModule::useSimColorTable() {
  return (vision_frame_info_->source == MEMORY_SIM);
}

string VisionModule::getDataBase() {
  return memory_->data_path_;
}

int VisionModule::getRobotId() {
  return robot_state_->robot_id_;
}

void VisionModule::initSpecificModule() {
  loadColorTables();
  top_processor_->SetColorTable(topColorTable);
  bottom_processor_->SetColorTable(bottomColorTable);
  top_processor_->init(textlogger);
  bottom_processor_->init(textlogger);
  loadCalibration();
}

VisionModule::VisionModule() {

  printf("Creating vision system");
  fflush(stdout);

  bottomColorTable = new unsigned char [LUT_SIZE];
  topColorTable = new unsigned char [LUT_SIZE];
  bottomColorTableName = "none";
  topColorTableName = "none";
  memset(bottomColorTable, c_UNDEFINED, LUT_SIZE);
  memset(topColorTable, c_UNDEFINED, LUT_SIZE);
  top_processor_ = new ImageProcessor(world_objects_, body_model_, joint_angles_, image_, robot_vision_, vision_frame_info_, robot_state_, Camera::TOP);
  bottom_processor_ = new ImageProcessor(world_objects_, body_model_, joint_angles_, image_, robot_vision_, vision_frame_info_, robot_state_, Camera::BOTTOM);

  puts(" Done!");
  fflush(stdout);
}

VisionModule::~VisionModule() {
  if (topColorTable != bottomColorTable)
    delete [] topColorTable;
  delete [] bottomColorTable;
}

void VisionModule::loadCalibration() {
    std::string calibrationFilename = boost::lexical_cast<std::string>(getRobotId()) + "camera.cal";
    std::string file = getDataBase() + calibrationFilename;
    CameraCalibration cal;
    if(cal.loadFromFile(getDataBase() + calibrationFilename) || cal.loadFromFile(getDataBase() + "defaultcamera.cal")) {
        top_processor_->setCalibration(cal);
        bottom_processor_->setCalibration(cal);
    }
    else {
        std::cout << "ERROR: No camera calibration found\n";
    }

}

bool VisionModule::loadColorTables() {
  bool uniqueColorTablesAvailable = true;
  if (useSimColorTable()) {
    loadColorTable(Camera::BOTTOM, "sim.col");
    // just set top to match bottom (esp so both pointers are the same, and editing either edits both in tool)
    topColorTable = bottomColorTable;
    topColorTableName = bottomColorTableName;
  } else {

    // Attempt to load robot-specific color tables if available
    std::string bottomColorTableFile =
      boost::lexical_cast<std::string>(getRobotId()) +
      "bottom.col";
    std::string topColorTableFile =
      boost::lexical_cast<std::string>(getRobotId()) +
      "top.col";
    bool bottomOk = loadColorTable(Camera::BOTTOM, bottomColorTableFile.c_str(), false);
    bool topOk = loadColorTable(Camera::TOP, topColorTableFile.c_str(), false);

    // If not available, then load the default top and bottom color tables
    uniqueColorTablesAvailable = bottomOk && topOk;
    if (!bottomOk) {
      bottomOk = loadColorTable(Camera::BOTTOM, "defaultbottom.col");
    }
    if (!topOk) {
      topOk = loadColorTable(Camera::TOP, "defaulttop.col");
    }

    // If still not available, then load the single default color table
    if (!bottomOk) {
      bottomOk = loadColorTable(Camera::BOTTOM, "default.col");
    }
    if (!topOk) {
      topOk = loadColorTable(Camera::TOP, "default.col");
    }


  }
  return uniqueColorTablesAvailable;
}

bool VisionModule::loadColorTable(Camera::Type camera, std::string fileName, bool fullPath) {
  unsigned char* colorTable = bottomColorTable;
  if (camera == Camera::TOP) {
    colorTable=topColorTable;
  }
  string colorTableName;
  if (!fullPath) {
#ifdef TOOL
    colorTableName = getDataBase() + "current/" + fileName;
#else
    colorTableName = getDataBase() + fileName;
#endif
  } else {
    colorTableName = fileName;
  }

  FILE* f=fopen(colorTableName.c_str(), "rb");
  if (f==NULL) {
    std::cout << "Vision: *** ERROR can't load " << colorTableName << " *** for camera " << camera << std::endl << std::flush;
    colorTableName = "none";
    return false;
  }
  std::cout << "Vision: Loaded " << colorTableName << " ! for camera " << camera << std::endl << std::flush;
  bool ok = fread(colorTable,LUT_SIZE,1,f);
  fclose(f);

  if (camera==Camera::TOP) {
    topColorTableName = colorTableName;
  } else {
    bottomColorTableName = colorTableName;
  }

 return ok;
}
