#include "BehaviorSimulation.h"

// core
#include <VisionCore.h>
#include <communications/CommunicationModule.h>

#include <localization/UKFModule.h>

// memory
#include <memory/GameStateBlock.h>
#include <memory/WorldObjectBlock.h>
#include <memory/Memory.h>
#include <memory/BehaviorBlock.h>
#include <memory/TeamPacketsBlock.h>
#include <memory/FrameInfoBlock.h>
#include <memory/SensorBlock.h>
#include <memory/OdometryBlock.h>

#include <stdlib.h>

// lua
#include <lua/LuaModule.h>

BehaviorSimulation::BehaviorSimulation(int n, bool penaltyKick, bool lMode){
  for (int i = 1; i < 9; i++){
    sims[i] = NULL;
    simActive[i] = false;
    fallenTime[i] = 0;
  }

  locMode = lMode;
  lastKick = 0;
  lastKickX = 0;
  ballClearedFromCircle = false;

#ifdef TOOL
  simInfo = "";
#endif

  simBlueScore = 0;
  simRedScore = 0;
  simPenaltyKick =false;
  numHalves = 0;
  PRINT = false;
  simTimer = 45;
  // start at 655 to include the 5 init, 45 ready and 5 set seconds before kickoff
  halfTimer = 655;

  simPenaltyKick = penaltyKick;
  nplayers = n;
  timeInc = 1.0/30.0;

  // init our world object mem
  memory_ = new Memory(false, MemoryOwner::VISION, 0, 1);
  memory_->getOrAddBlockByName(worldObjects, "world_objects");
  worldObjects->init(TEAM_BLUE);
  worldObjects->reset();
  memory_->getOrAddBlockByName(gameState, "game_state");
  gameState->isPenaltyKick = simPenaltyKick;
  memory_->getOrAddBlockByName(frameInfo, "vision_frame_info");
  frameInfo->frame_id = 0;
  frameInfo->seconds_since_start = 0;

  // if we haven't initialized any before, init them
  for (int i = 1; i < 9; i++){
    if (nplayers == 8){
      simActive[i] = true;
    } else if (nplayers == 1) {
      if (i == 4)
        simActive[i] = true;
      else
        simActive[i] = false;
    } else if (nplayers == 2){
      if (i == 4 || i == 5)
        simActive[i] = true;
      else
        simActive[i] = false;
    }
    if (sims[i] == NULL && simActive[i]){
      int team = 0;
      if (i > 4) team = 1;
      int self = ((i-1) % 4) + 1;
      sims[i] = new SimulatedPlayer(team, self, i, locMode);
    }
  }

  gameState->state = INITIAL;
  simTimer = 5.0;

  // set initial positions
  worldObjects->objects_[WO_BALL].loc = Point2D(0,0);

  for (int i = 1; i < 9; i++){
    int self = i;
    if (self > 4) self -= 4;
    int teamsign = 1;
    if (i > 4) teamsign = -1;

    if (self == 1 || self == 3){
      worldObjects->objects_[i].loc.x = teamsign * -2000;
    } else {
      worldObjects->objects_[i].loc.x = teamsign * -1000;
    }
    if (self == 1 || self == 2){
      worldObjects->objects_[i].loc.y = teamsign * 2000;
    } else {
      worldObjects->objects_[i].loc.y = teamsign * -2000;
    }

    if (worldObjects->objects_[i].loc.y > 0){
      worldObjects->objects_[i].orientation = -M_PI/2.0;
    } else {
      worldObjects->objects_[i].orientation = M_PI/2.0;
    }

    // special locations for penalty kick
    if (simPenaltyKick) {
      if (i == 4){
        worldObjects->objects_[i].loc.x = 0;
        worldObjects->objects_[i].loc.y = 0;
        worldObjects->objects_[i].orientation = 0;
        worldObjects->objects_[WO_BALL].loc = Point2D(PENALTY_CROSS_X,0);
      } else {
        worldObjects->objects_[i].loc.x = FIELD_X/2.0 - 100;
        worldObjects->objects_[i].loc.y = 0;
        worldObjects->objects_[i].orientation = M_PI;
      }
    }

    // set way off field if not active
    if (sims[i] == NULL || !simActive[i]){
      worldObjects->objects_[i].loc.x = 10000;
      worldObjects->objects_[i].loc.y = 10000;
    }

  }

}



BehaviorSimulation::~BehaviorSimulation(){
  // core will delete memory for us
  delete worldObjects;
  delete gameState;
  delete memory_;
  for (int i = 1; i < 9; i++){
    if (sims[i] != NULL)
      delete sims[i];
    sims[i] = NULL;
  }
  worldObjects = NULL;
  gameState = NULL;
  memory_ = NULL;
}

void BehaviorSimulation::doPenaltyKickReset(){
  worldObjects->objects_[WO_BALL].loc = Point2D(PENALTY_CROSS_X,0);
  worldObjects->objects_[WO_BALL].absVel = Point2D(0,0);
  worldObjects->objects_[4].loc = Point2D(0,0);
  worldObjects->objects_[4].orientation = 0;
  worldObjects->objects_[5].loc = Point2D(FIELD_X/2.0-100,0);
  worldObjects->objects_[5].orientation = M_PI;
  changeSimulationState(READY);
  simTimer = 5.0;

  if (sims[4] != NULL){
    sims[4]->resetCounters();
    sims[4]->gameState->state = READY;
  }
  if (sims[5] != NULL){
    sims[5]->resetCounters();
    sims[5]->gameState->state = READY;
  }
}

void BehaviorSimulation::simulationStep(){

#ifdef TOOL
  QString oldSimInfo = simInfo;
  simInfo = "";
#endif

  // update the ball position from velocity for everyone
  Point2D ballVel = worldObjects->objects_[WO_BALL].absVel;
  Point2D ballLoc = worldObjects->objects_[WO_BALL].loc;
  Point2D disp = ballVel * timeInc;
  ballLoc += disp;
  ballVel *= 0.966;

  simTimer -= timeInc;
  halfTimer -= timeInc;
  gameState->secsRemaining = (int)halfTimer;
  frameInfo->seconds_since_start += timeInc;
  frameInfo->frame_id++;

#ifdef TOOL
  if (simTimer < 10 && simPenaltyKick && sims[4]->gameState->state == PLAYING){
    simInfo += QString::number(simTimer) + " seconds left in PK, ";
  }
#endif

  if (simTimer <= 0){
    // different for penalty mode
    if (simPenaltyKick){
      // go through ready,set,pen
      if (gameState->state == INITIAL){
        changeSimulationState(READY);
        simTimer = 5.0;
      } else if (gameState->state == READY){
        ballLoc = Point2D(PENALTY_CROSS_X,0);
        // go to set
        changeSimulationState(SET);
        simTimer = 5.0;
      } else if (gameState->state == SET){
        ballLoc = Point2D(PENALTY_CROSS_X,0);
        // go to playing
        changeSimulationState(PLAYING);
        simTimer = 60.0;
        halfTimer = 60.0;
      } else if (gameState->state == PENALISED){
        ballLoc = Point2D(PENALTY_CROSS_X,0);
        // go to playing
        changeSimulationState(PLAYING);
        simTimer = 60.0;
        halfTimer = 60.0;
      } else if (gameState->state == PLAYING){
        // TIME IS UP!
#ifdef TOOL
        simInfo += "PK TIME UP, NO GOAL SCORED!, ";
#endif
        doPenaltyKickReset();
      }
      ballClearedFromCircle = true;
    } else {
      if (gameState->state == INITIAL){
        // go to ready
        ballLoc = Point2D(0,0);
        ballVel = Point2D(0,0);
        changeSimulationState(READY);
        simTimer = 45.0;
      } else if (gameState->state == READY){
        ballLoc = Point2D(0,0);
        ballVel = Point2D(0,0);

        // go to set
        changeSimulationState(SET);
        simTimer = 5.0;
      } else if (gameState->state == SET){
        changeSimulationState(PLAYING);
        if (halfTimer > 600) halfTimer = 600;
        simTimer = halfTimer;
      }
      // back to ready
      else if (gameState->state == PLAYING){
        cout << "Half: " << numHalves << " Blue: " << simBlueScore << ", Red: " << simRedScore << endl;
        numHalves++;
        gameState->ourKickOff = numHalves % 2 == 0;

        ballLoc = Point2D(0,0);
        ballVel = Point2D(0,0);
        changeSimulationState(READY);
        simTimer = 45.0;
        halfTimer = 650.0;
      }
    }
  }

  // check for goal
  if (fabs(ballLoc.x) > FIELD_X/2.0 && fabs(ballLoc.y) < GOAL_Y/2.0){
    if (ballLoc.x > 0){
      // if from circle and our kickoff, not valid
      if (gameState->ourKickOff && !ballClearedFromCircle){
#ifdef TOOL
        simInfo += "KICKOFF SHOT BLUE, ";
#endif
      } else {
#ifdef TOOL
        simInfo += "GOAL BLUE, ";
#endif
        setSimScore(true);
        //cout << sims[3]->frameInfo->frame_id << " Goal blue. Score Blue: " << simBlueScore << " Red: " << simRedScore << endl;
      }
      gameState->ourKickOff = false;
    }
    else {
      // if from circle and our kickoff, not valid
      if (!gameState->ourKickOff && !ballClearedFromCircle){
#ifdef TOOL
        simInfo += "KICKOFF SHOT RED, ";
#endif
      } else {
#ifdef TOOL
        simInfo += "GOAL RED, ";
#endif
        setSimScore(false);
        //cout << sims[3]->frameInfo->frame_id << " Goal red. Score Blue: " << simBlueScore << " Red: " << simRedScore << endl;
      }
      gameState->ourKickOff = true;
    }
    ballLoc = Point2D(0,0);
    ballVel = Point2D(0,0);
    // go to ready
    simTimer = 45.0;
    changeSimulationState(READY);
    if (simPenaltyKick){
      doPenaltyKickReset();
      ballLoc = worldObjects->objects_[WO_BALL].loc;
      ballVel = worldObjects->objects_[WO_BALL].absVel;
    }
  }

  // check for out of bounds
  if (fabs(ballLoc.y) > FIELD_Y/2.0){
    if (PRINT) cout << "ball out at " << ballLoc << endl;
    ballVel = Point2D(0,0);
    if (ballLoc.y > 0)
      ballLoc.y = FIELD_Y/2.0 - 400;
    else
      ballLoc.y = -FIELD_Y/2.0 + 400;
    if (lastKick == TEAM_BLUE){
#ifdef TOOL
      simInfo += "OUT OF BOUNDS on BLUE, ";
#endif
      if (ballLoc.x < lastKickX) {
        ballLoc.x -= 1000;
      } else {
        ballLoc.x = lastKickX - 1000;
      }
    } else {
#ifdef TOOL
      simInfo += "OUT OF BOUNDS on RED, ";
#endif
      if (ballLoc.x > lastKickX) {
        ballLoc.x += 1000;
      } else {
        ballLoc.x = lastKickX + 1000;
      }
    }
    // max and min values
    if (ballLoc.x < -2000) ballLoc.x = -2000;
    if (ballLoc.x > 2000) ballLoc.x = 2000;
  }

  if (fabs(ballLoc.x) > FIELD_X/2.0){
    if (PRINT) cout << "ball out at " << ballLoc << endl;
    ballVel = Point2D(0,0);
    if (ballLoc.y > 0)
      ballLoc.y = FIELD_Y/2.0 - 400;
    else
      ballLoc.y = -FIELD_Y/2.0 + 400;
    if (lastKick == TEAM_BLUE){
#ifdef TOOL
      simInfo += "OUT OF BOUNDS on BLUE, ";
#endif
      if (ballLoc.x > 0){
        ballLoc.x = 0;
        if (PRINT) cout << "Ball out endline on blue" << endl;
      } else {
        ballLoc.x = -2000;
      }
    } else {
#ifdef TOOL
      simInfo += "OUT OF BOUNDS on RED, ";
#endif
      if (ballLoc.x > 0){
        ballLoc.x = 2000;
      } else {
        ballLoc.x = 0;
        if (PRINT) cout << "Ball out endline on red" << endl;
      }
    }
  }


  // see if ball hit anyone
  for (int i = 1; i < 9; i++){
    if (!simActive[i]) continue;
    float ballDist = 0;
    ballDist = ballLoc.getDistanceTo(worldObjects->objects_[i].loc);

    if (ballDist < 120){
      // slow down ball
      if (i < 5){
#ifdef TOOL
        simInfo += "BALL HIT BLUE, ";
#endif
        lastKick = TEAM_BLUE;
        lastKickX = worldObjects->objects_[i].loc.x;
      } else {
#ifdef TOOL
        simInfo += "BALL HIT RED, ";
#endif
        lastKick = TEAM_RED;
        lastKickX = -worldObjects->objects_[i].loc.x;
      }
      //cout << "ball hit " << lastKick << " at " << lastKickX << endl;
      ballVel *= 0.5;
    }

    // special check for keeper
    Point2D teamBallLoc = ballLoc;
    if (fabs(sims[i]->sensors->values_[angleY]) > (M_PI/2.0 - 0.08)){
      Point2D robotLoc = worldObjects->objects_[i].loc;
      bool yMatch = false;
      if (sims[i]->sensors->values_[angleY] > 0){
        yMatch = (teamBallLoc.y < robotLoc.y && teamBallLoc.y > robotLoc.y-480);
      } else {
        yMatch = (teamBallLoc.y > robotLoc.y && teamBallLoc.y < robotLoc.y+480);
      }
      if (yMatch && fabs(teamBallLoc.x - robotLoc.x) < 50){
        if (i < 5){
#ifdef TOOL
          simInfo += "BLUE KEEPER SAVE!, ";
#endif
          lastKick = TEAM_BLUE;
          lastKickX = worldObjects->objects_[i].loc.x;
        } else {
#ifdef TOOL
          simInfo += "RED KEEPER SAVE!, ";
#endif
          lastKick = TEAM_RED;
          lastKickX = worldObjects->objects_[i].loc.x;
        }
        //cout << "ball saved by " << lastKick << " at " << lastKickX << endl;
        if ((((float)rand())/((float)RAND_MAX)) < 0.5 && !simPenaltyKick){
          // keeper sweep
          if (ballLoc.x < 0){
            ballVel.x = 1000;
          } else {
            ballVel.x = -1000;
          }
        } else {
          // bounced off
          ballVel.x = -ballVel.x;
          ballVel *= 0.5;
        }
      }
    }
  }

  // save new loc and vel back to wo
  worldObjects->objects_[WO_BALL].absVel = ballVel;
  worldObjects->objects_[WO_BALL].loc = ballLoc;


  // step the simulation ahead
  for (int i = 1; i < 9; i++){
    if (sims[i] == NULL || !simActive[i]) continue;

    // check how long each robot has been down for
    if (fabs(sims[i]->sensors->values_[angleX]) > 0.8 || fabs(sims[i]->sensors->values_[angleY]) > 0.8)
      fallenTime[i] += timeInc;
    else
      fallenTime[i] = 0;

    // penalty if they have not started getting up after 5 seconds
    if (fallenTime[i] > 5.0 && sims[i]->odometry->getting_up_side_ == Getup::NONE){
#ifdef TOOL
      if (i < 5)
        simInfo += "INACTIVE ROBOT BLUE "+QString::number(sims[i]->self_)+", ";
      else
        simInfo += "INACTIVE ROBOT RED "+QString::number(sims[i]->self_)+", ";
#endif

      sims[i]->setPenalty(worldObjects);
    }

    // penalty if left field
    if (fabs(worldObjects->objects_[i].loc.x) > (GRASS_X/2.0 + 500.0) || fabs(worldObjects->objects_[i].loc.y) > (GRASS_Y/2.0 + 500.0)){
#ifdef TOOL
      if (i < 5)
        simInfo += "LEAVING FIELD BLUE "+QString::number(sims[i]->self_)+", ";
      else
        simInfo += "LEAVING FIELD RED "+QString::number(sims[i]->self_)+", ";
#endif
      sims[i]->setPenalty(worldObjects);
    }


    bool kicked = sims[i]->processFrame(worldObjects,gameState);
    if (kicked){
      lastKick = sims[i]->team_;
      lastKickX = worldObjects->objects_[i].loc.x;
      if (worldObjects->objects_[WO_BALL].loc.getMagnitude() > CIRCLE_RADIUS){
        ballClearedFromCircle = true;
      }
#ifdef TOOL
      if (lastKick == TEAM_BLUE){
        simInfo += "BLUE KICK, ";
      } else {
        simInfo += "RED KICK, ";
      }
#endif
      //cout << "ball kicked by " << lastKick << " at " << lastKickX << endl;

      // if its a penalty kick, this could be illegal
      if (simPenaltyKick) {
        if (lastKick == TEAM_BLUE){
          Point2D ballLoc = worldObjects->objects_[WO_BALL].loc;
          if (ballLoc.x > (FIELD_X/2.0 - PENALTY_X) && fabs(ballLoc.y < PENALTY_Y/2.0)){
#ifdef TOOL
            simInfo += "BLUE TOUCHED BALL INSIDE PEN BOX, NO GOAL, ";
#endif
            doPenaltyKickReset();
          }
        } else {
          Point2D ballLoc = sims[i]->worldObjects->objects_[WO_BALL].loc;
          if (ballLoc.x > (-FIELD_X/2.0 + PENALTY_X) || fabs(ballLoc.y > PENALTY_Y/2.0)){
#ifdef TOOL
            simInfo += "RED TOUCHED BALL OUTSIDE PEN BOX, GOAL, ";
#endif
            setSimScore(true);
            doPenaltyKickReset();
          }
        }
      } // pen kick
    } // any kick


    // get updates of where we are and where ball is, and update other bots with it
    // get new ball location
    ballLoc = worldObjects->objects_[WO_BALL].loc;
    ballVel = worldObjects->objects_[WO_BALL].absVel;

    // get new robot location
    WorldObject* robot = &(worldObjects->objects_[i]);

    // check if we bumped ball (dribble)
    if (ballLoc.getDistanceTo(robot->loc) < 60){
      // add up to 30 degree error
      float randPct = ((float)rand())/((float)RAND_MAX)-0.5;
      float headingError = randPct*60.0*DEG_T_RAD;
      // add up to 25 m/s vel error
      float randPct2 = ((float)rand())/((float)RAND_MAX)-0.5;
      float velError = randPct2*500.0;

      Point2D absVel(250+velError, robot->orientation+headingError, POLAR);
      Point2D disp(100, robot->orientation, POLAR);

      worldObjects->objects_[WO_BALL].loc += disp;
      worldObjects->objects_[WO_BALL].absVel = absVel;
      ballVel = absVel;
      ballLoc = worldObjects->objects_[WO_BALL].loc;
    }

    // check if we went into own penalty box
    if (sims[i]->self_ != 1 && i < 5 && robot->loc.x < (-FIELD_X/2.0 + PENALTY_X) && fabs(robot->loc.y) < PENALTY_Y/2.0){
      sims[i]->setPenalty(worldObjects);
#ifdef TOOL
      simInfo += "ILLEGAL DEFENDER BLUE " + QString::number(sims[i]->self_) + ", ";
#endif
    }
    if (sims[i]->self_ != 1 && i > 4 && robot->loc.x > (FIELD_X/2.0 - PENALTY_X) && fabs(robot->loc.y) < PENALTY_Y/2.0){
      sims[i]->setPenalty(worldObjects);
#ifdef TOOL
      simInfo += "ILLEGAL DEFENDER RED " + QString::number(sims[i]->self_) + ", ";
#endif
    }

    // get team packet we would have sent
    // and update other teammates
    TeamPacket* sent = &(sims[i]->teamPackets->tp[sims[i]->self_]);

    // check if we hit other robots
    // send team pkts to other robots
    for (int j = 1; j < 9; j++){
      if (i == j) continue;
      if (!simActive[j]) continue;
      // check if we ran into another robot
      float mateDist = robot->loc.getDistanceTo(worldObjects->objects_[j].loc);
      if (mateDist < 180 && sims[j]->gameState->state != PENALISED && sims[i]->gameState->state != PENALISED){
        sims[i]->setPenalty(worldObjects);
#ifdef TOOL
        if (sims[i]->team_ == TEAM_BLUE){
          simInfo += "PUSHING BLUE " + QString::number(sims[i]->self_) + ", ";
        } else {
          simInfo += "PUSHING RED " + QString::number(sims[i]->self_) + ", ";
        }
#endif
      }

      // basically we "send" our team packets across this way
      if (sims[i]->frameInfo->frame_id % 6 == 0 && sims[i]->team_ == sims[j]->team_){
        int robotNumber = sims[i]->self_;
        sims[j]->teamPackets->tp[robotNumber] = *sent;
        sims[j]->teamPackets->frameReceived[robotNumber] = sims[j]->frameInfo->frame_id;
        sims[j]->teamPackets->ballUpdated[robotNumber] = true;
        sims[j]->teamPackets->oppUpdated[robotNumber] = true;

        if (locMode){
          // Populate world objects for team mate position
          sims[j]->worldObjects->objects_[robotNumber].loc.x = sent->locData.robotX;
          sims[j]->worldObjects->objects_[robotNumber].loc.y = sent->locData.robotY;
          sims[j]->worldObjects->objects_[robotNumber].orientation = sent->locData.orient;

          sims[j]->worldObjects->objects_[robotNumber].sd.x = sent->locData.robotSDX;
          sims[j]->worldObjects->objects_[robotNumber].sd.y = sent->locData.robotSDY;
          sims[j]->worldObjects->objects_[robotNumber].sdOrientation = sent->locData.sdOrient;
        }
      }

    } // check with other robots
  } // simulate step for this robot

  // keep old text around until we have new text
#ifdef TOOL
  if (simInfo == "") simInfo = oldSimInfo;
#endif
}

Memory* BehaviorSimulation::getMemory(int index){
  if (index > 0 && index < 9 && simActive[index]) return sims[index]->getMemory();
  // if index 0, return truth memory
  if (index == 0) return memory_;
  cout << "ERROR: sim index is invalid: " << index << endl << flush;
  return NULL;
}

std::vector<std::string> BehaviorSimulation::getTextDebug(int index){
  if (index < 1 || index > 8 || !simActive[index]){
    std::vector<std::string> emptytext;
    return emptytext;
  }
  return sims[index]->getTextDebug();
}

void BehaviorSimulation::changeSimulationState(int state){
  gameState->state = state;
  ballClearedFromCircle = false;

  if (state == SET && !simPenaltyKick){
    worldObjects->objects_[WO_BALL].loc = Point2D(0,0);
    worldObjects->objects_[WO_BALL].absVel = Point2D(0,0);
    // check positions
    int blueCircleCount = 0;
    int redCircleCount = 0;
    for (int i = 8; i > 0; i--){
      if (!simActive[i]) continue;
      bool manual = false;
      float xLineDef = -PENALTY_CROSS_X - 50;
      float xLineKick = -50;

      if (i < 5){
        float xLine = xLineDef;
        if (gameState->ourKickOff) xLine = xLineKick;

        // past x barrier
        if (worldObjects->objects_[i].loc.x > xLine){
          manual = true;
        }
        // more than 1 robot inside circle
        if (worldObjects->objects_[i].loc.getMagnitude() < CIRCLE_RADIUS){
          blueCircleCount++;
          if (blueCircleCount > 1)
            manual = true;
        }
        // keeper not in box
        if (i == 1 && (worldObjects->objects_[i].loc.x > (-FIELD_X/2.0 + PENALTY_X) || fabs(worldObjects->objects_[i].loc.y) > (PENALTY_Y/2.0))){
          manual = true;
        }
        // other player in box
        if (i != 1 && (worldObjects->objects_[i].loc.x < (-FIELD_X/2.0 + PENALTY_X) && fabs(worldObjects->objects_[i].loc.y) < (PENALTY_Y/2.0))){
          manual = true;
        }
        // off field
        if (fabs(worldObjects->objects_[i].loc.x) > (FIELD_X/2.0) || fabs(worldObjects->objects_[i].loc.y) > (FIELD_Y/2.0)){
          manual = true;
        }
      } else { // team red
        float xLine = -xLineDef;
        if (!gameState->ourKickOff) xLine = -xLineKick;

        // past x barrier
        if (worldObjects->objects_[i].loc.x < xLine){
          manual = true;
        }
        // more than 1 robot inside circle
        if (worldObjects->objects_[i].loc.getMagnitude() < CIRCLE_RADIUS){
          redCircleCount++;
          if (redCircleCount > 1)
            manual = true;
        }
        // keeper not in box
        if (i == 5 && (worldObjects->objects_[i].loc.x < (FIELD_X/2.0 - PENALTY_X) || fabs(worldObjects->objects_[i].loc.y) > (PENALTY_Y/2.0))){
          manual = true;
        }
        // other player in box
        if (i != 5 && (worldObjects->objects_[i].loc.x > (FIELD_X/2.0 - PENALTY_X) && fabs(worldObjects->objects_[i].loc.y) < (PENALTY_Y/2.0))){
          manual = true;
        }
        // off field
        if (fabs(worldObjects->objects_[i].loc.x) > (FIELD_X/2.0) || fabs(worldObjects->objects_[i].loc.y) > (FIELD_Y/2.0)){
          manual = true;
        }
      } // team

      if (manual){
#ifdef TOOL
        if (i < 5)
          simInfo += "ILLEGAL POSITION BLUE" + QString::number(i) + ", ";
        else
          simInfo += "ILLEGAL POSITION RED" + QString::number(i-4) + ", ";
#endif
        worldObjects->objects_[i].orientation = 0;
        if (i == 1 || i == 5)
          worldObjects->objects_[i].loc = Point2D(-FIELD_X/2.0, 0);
        if ((i < 5 && gameState->ourKickOff) || (i > 4 && !gameState->ourKickOff)){
          if (i == 2 || i == 6)
            worldObjects->objects_[i].loc = Point2D(-PENALTY_CROSS_X, PENALTY_Y/2.0);
          if (i == 3 || i == 7)
            worldObjects->objects_[i].loc = Point2D(-FIELD_X/2.0+PENALTY_X+200, -PENALTY_Y/2.0);
          if (i == 4 || i == 8)
            worldObjects->objects_[i].loc = Point2D(-CIRCLE_RADIUS-100, 0);
        } else {
          // their kickoff
          float xPos = -FIELD_X/2.0+PENALTY_X+200;
          if (i == 2 || i == 6)
            worldObjects->objects_[i].loc = Point2D(xPos, (FIELD_Y+PENALTY_Y)/4.0);
          if (i == 3 || i == 7)
            worldObjects->objects_[i].loc = Point2D(xPos, (FIELD_Y+PENALTY_Y)/-4.0);
          if (i == 4 || i == 8)
            worldObjects->objects_[i].loc = Point2D(xPos, GOAL_Y/2.0);
        }

        if (i > 4){
          worldObjects->objects_[i].loc = -worldObjects->objects_[i].loc;
          worldObjects->objects_[i].orientation = M_PI;
        }
      } // manual
    } // loop
  }

  simTimer = 45.0;
  if (gameState->state == SET){
    simTimer = 5.0;
  } else if (gameState->state == PLAYING){
    if (simPenaltyKick){
      simTimer = 60.0;
      halfTimer = 60.0;
    } else {
      simTimer = halfTimer;
    }
  }

  if (gameState->state == INITIAL){
    for (int i = 1; i < 9; i++){
      if (sims[i] != NULL && simActive[i])
        sims[i]->resetCounters();
    }
  }
}


void BehaviorSimulation::moveRobot(int index, AngRad rotation, Point2D translation){
  if (index < 1 || index > 8) return;
  WorldObject* robot = &(worldObjects->objects_[index]);
  robot->loc += translation;
  robot->orientation += rotation;
}

void BehaviorSimulation::moveBall(Point2D movement){

  WorldObject* ball = &(worldObjects->objects_[WO_BALL]);
  ball->loc += movement;
}

void BehaviorSimulation::changeSimulationKickoff(){
  gameState->ourKickOff = !gameState->ourKickOff;
}

void BehaviorSimulation::restartSimulationLua(){
  for (int i = 1; i < 9; i++){
    if (sims[i] == NULL || !simActive[i]) continue;
    sims[i]->restartLua();
  }
}

void BehaviorSimulation::setSimScore(bool blue){
  if (blue) simBlueScore++;
  else simRedScore++;
  gameState->ourScore = simBlueScore;
  gameState->opponentScore = simRedScore;
}


void BehaviorSimulation::setPenalty(int index){
  if (index < 1 || index > 8) return;
  sims[index]->setPenalty(worldObjects);
}

void BehaviorSimulation::flipRobot(int index){
  if (index < 1 || index > 8) return;
  worldObjects->objects_[index].loc = -worldObjects->objects_[index].loc;
  worldObjects->objects_[index].orientation = normalizeAngle(worldObjects->objects_[index].orientation + M_PI);
}

#ifdef TOOL
QString BehaviorSimulation::getSimInfo(){
  return simInfo;
}
#endif

void BehaviorSimulation::setFallen(int index){
  if (index < 1 || index > 8) return;
  if (sims[index] == NULL || !simActive[index]) return;


  sims[index]->setFallen();
}

void BehaviorSimulation::kickBall(){
  WorldObject* ball = &(worldObjects->objects_[WO_BALL]);
  // random target within goal
  float targetY = 1400.0*(((float)rand())/((float)RAND_MAX)-0.5);
  float targetX = -3000;
  if (ball->loc.x > 0){
    targetX = 3000;
  }
  Point2D absVel(3500, ball->loc.getAngleTo(Point2D(targetX, targetY)), POLAR);
  ball->absVel = absVel;
}

// check if anyone's localization is off and by how much
int BehaviorSimulation::checkLocalizationErrors(){
  PRINT = true;

  for (int i = 1; i < 9; i++){
    if (!simActive[i] || sims[i] == NULL) continue;
    sims[i]->PRINT = true;

    // not while the robot is getting up
    if (sims[i]->odometry->getting_up_side_ != Getup::NONE) continue;
    // not if penalized
    if (sims[i]->gameState->state == PENALISED) continue;
    // not in set if keeper
    if (sims[i]->gameState->state == SET && sims[i]->self_ == 1) continue;

    WorldObject* trueRobot = &(worldObjects->objects_[i]);
    WorldObject* belRobot = &(sims[i]->worldObjects->objects_[sims[i]->self_]);

    float distance, orient;
    if (i < 5){
      distance = trueRobot->loc.getDistanceTo(belRobot->loc);
      orient   = fabs(normalizeAngle(trueRobot->orientation - belRobot->orientation));
      //cout << i << " true location: " << trueRobot->loc << " ori: " << RAD_T_DEG* trueRobot->orientation << " belief: " << belRobot->loc << " ori: " << RAD_T_DEG* belRobot->orientation << " error: "<< distance << ", " << orient *RAD_T_DEG << endl;

    } else {
      distance = trueRobot->loc.getDistanceTo(-belRobot->loc);
      orient   = fabs(normalizeAngle(trueRobot->orientation + M_PI - belRobot->orientation));
      //cout << i << " true location: " << trueRobot->loc << " ori: " << RAD_T_DEG* trueRobot->orientation << " belief: " << -belRobot->loc << " ori: " << RAD_T_DEG* normalizeAngle(M_PI+belRobot->orientation) << " error: "<< distance << ", " << orient *RAD_T_DEG << endl;

    }

    if (distance > 500 || orient > (DEG_T_RAD*25.0) || isnan(belRobot->loc.x) || isnan(belRobot->loc.y) || isnan(belRobot->sd.x) || isnan(belRobot->sd.y) || isnan(belRobot->orientation) || isnan(belRobot->sdOrientation)){
      cout << i << " error: "<< distance << ", " << orient *RAD_T_DEG << endl;
      return i;
    }
  }
  return 0;
}


// track out localization error
int BehaviorSimulation::measureLocalizationErrors(int redParam){
  PRINT = false;

  float blueTotalDistError = 0;
  float blueTotalOriError = 0;
  float redTotalDistError = 0;
  float redTotalOriError = 0;

  for (int i = 5; i < 9; i++){
    if (!simActive[i] || sims[i] == NULL) continue;
    sims[i]->core->localization_->ukfParams.ambig_model_min_ball_frames = redParam;
  }

  // run game for 80 halves
  while (numHalves < 80){
    simulationStep();

    for (int i = 1; i < 9; i++){
      if (!simActive[i] || sims[i] == NULL) continue;
      sims[i]->PRINT = false;

      // not while the robot is getting up
      if (sims[i]->odometry->getting_up_side_ != Getup::NONE) continue;
      // not if penalized
      if (sims[i]->gameState->state == PENALISED) continue;
      // not in set if keeper
      if (sims[i]->gameState->state == SET && sims[i]->self_ == 1) continue;

      WorldObject* trueRobot = &(worldObjects->objects_[i]);
      WorldObject* belRobot = &(sims[i]->worldObjects->objects_[sims[i]->self_]);

      float distance, orient;
      if (i < 5){
        distance = trueRobot->loc.getDistanceTo(belRobot->loc);
        orient   = fabs(normalizeAngle(trueRobot->orientation - belRobot->orientation));
        //cout << i << " true location: " << trueRobot->loc << " ori: " << RAD_T_DEG* trueRobot->orientation << " belief: " << belRobot->loc << " ori: " << RAD_T_DEG* belRobot->orientation << " error: "<< distance << ", " << orient *RAD_T_DEG << endl;
        blueTotalDistError += distance;
        blueTotalOriError += orient;

      } else {
        distance = trueRobot->loc.getDistanceTo(-belRobot->loc);
        orient   = fabs(normalizeAngle(trueRobot->orientation + M_PI - belRobot->orientation));
        //cout << i << " true location: " << trueRobot->loc << " ori: " << RAD_T_DEG* trueRobot->orientation << " belief: " << -belRobot->loc << " ori: " << RAD_T_DEG* normalizeAngle(M_PI+belRobot->orientation) << " error: "<< distance << ", " << orient *RAD_T_DEG << endl;

        redTotalDistError += distance;
        redTotalOriError += orient;

      }



    } // player check
  } // 80 halves loop

  cout << "Total Blue Dist Error: " << blueTotalDistError << ", totalOriError: "<< blueTotalOriError << endl;
  cout << "Total Red Dist Error: " << redTotalDistError << ", totalOriError: "<< redTotalOriError << endl;

  return 0;
}




//////////////////////////////
// behavior simulation stuff
//////////////////////////////


void BehaviorSimulation::runParamTests(){
  for (float diff = -10.0*DEG_T_RAD; diff < 12*DEG_T_RAD; diff += 5.0*DEG_T_RAD){

    cout << endl << endl << "TEST with diff " << diff*RAD_T_DEG << endl << endl;

    // set all players with this diff
    for (int i = 5; i < 9; i++){
      cout << "i is " << i << endl;
      if (sims[i] == NULL || !simActive[i]) continue;
      cout << " and setting " << i << endl;
      sims[i]->PRINT = false;
      // modify parameters
      // restart lua
      //sims[i]->restartLua();
    }

    // run param test
    runParamTest();

  }
}


void BehaviorSimulation::runKickTests(){
  // run param test
  runParamTest();
}


void BehaviorSimulation::runParamTest(){

  simBlueScore = 0;
  simRedScore = 0;
  simPenaltyKick =false;
  numHalves = 0;
  PRINT = false;
  simTimer = 45;
  halfTimer = 655;
  changeSimulationState(INITIAL);
  gameState->state = INITIAL;
  simTimer = 5.0;

  /*
  for (int i = 1; i < 9; i++){
    if (sims[i] == NULL || !simActive[i]) continue;
    cout << "Player " << i << " has max Arc Angle " << sims[i]->behaviorParams->mainStrategy.maxArcAngle << endl;
  }
  */

  // run game for 80 halves
  while (numHalves < 80)
    simulationStep();

}


/*


  void WorldGLWidget::compareParams(){
  if (sims[0]->behaviorParams->mainStrategy.forwardOpeningAngle != sims[4]->behaviorParams->mainStrategy.forwardOpeningAngle){
  cout << "forwardOpeningAngle differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.forwardOpeningAngle*RAD_T_DEG << ", RED: " << sims[4]->behaviorParams->mainStrategy.forwardOpeningAngle*RAD_T_DEG << endl;
  }
  if (sims[0]->behaviorParams->mainStrategy.maxArcAngle != sims[4]->behaviorParams->mainStrategy.maxArcAngle){
  cout << "maxArcAngle differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.maxArcAngle*RAD_T_DEG << ", RED: " << sims[4]->behaviorParams->mainStrategy.maxArcAngle*RAD_T_DEG << endl;
  }
  if (sims[0]->behaviorParams->mainStrategy.maxKickAngle != sims[4]->behaviorParams->mainStrategy.maxKickAngle){
  cout << "maxKickAngle differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.maxKickAngle*RAD_T_DEG << ", RED: " << sims[4]->behaviorParams->mainStrategy.maxKickAngle*RAD_T_DEG << endl;
  }
  if (sims[0]->behaviorParams->mainStrategy.minOpponentDist != sims[4]->behaviorParams->mainStrategy.minOpponentDist){
  cout << "minOpponentDist differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.minOpponentDist << ", RED: " << sims[4]->behaviorParams->mainStrategy.minOpponentDist << endl;
  }
  if (sims[0]->behaviorParams->mainStrategy.opponentWidth != sims[4]->behaviorParams->mainStrategy.opponentWidth){
  cout << "opponentWidth differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.opponentWidth << ", RED: " << sims[4]->behaviorParams->mainStrategy.opponentWidth << endl;
  }
  if (sims[0]->behaviorParams->mainStrategy.passDistance != sims[4]->behaviorParams->mainStrategy.passDistance){
  cout << "passDistance differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.passDistance << ", RED: " << sims[4]->behaviorParams->mainStrategy.passDistance << endl;
  }
  if (sims[0]->behaviorParams->mainStrategy.supportFwdDist != sims[4]->behaviorParams->mainStrategy.supportFwdDist){
  cout << "supportFwdDist differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.supportFwdDist << ", RED: " << sims[4]->behaviorParams->mainStrategy.supportFwdDist << endl;
  }
  if (sims[0]->behaviorParams->mainStrategy.defendBackDist != sims[4]->behaviorParams->mainStrategy.defendBackDist){
  cout << "defendBackDist differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.defendBackDist << ", RED: " << sims[4]->behaviorParams->mainStrategy.defendBackDist << endl;
  }
  if (sims[0]->behaviorParams->mainStrategy.forwardFwdDist != sims[4]->behaviorParams->mainStrategy.forwardFwdDist){
  cout << "forwardFwdDist differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.forwardFwdDist << ", RED: " << sims[4]->behaviorParams->mainStrategy.forwardFwdDist << endl;
  }
  if (sims[0]->behaviorParams->mainStrategy.clusterBehavior != sims[4]->behaviorParams->mainStrategy.clusterBehavior){
  cout << "clusterBehavior differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.clusterBehavior << ", RED: " << sims[4]->behaviorParams->mainStrategy.clusterBehavior << endl;
  }
  if (sims[0]->behaviorParams->mainStrategy.shootOnGoalRadius != sims[4]->behaviorParams->mainStrategy.shootOnGoalRadius){
  cout << "shootOnGoalRadius differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.shootOnGoalRadius << ", RED: " << sims[4]->behaviorParams->mainStrategy.shootOnGoalRadius << endl;
  }
  // kicks
  for (int j = 0; j < NUM_KICKS; j++){
  if (sims[0]->behaviorParams->mainStrategy.usableKicks[j] != sims[4]->behaviorParams->mainStrategy.usableKicks[j]){
  cout << kickNames[j] << " " << j << " differs. BLUE: " << sims[0]->behaviorParams->mainStrategy.usableKicks[j] << ", RED: " << sims[4]->behaviorParams->mainStrategy.usableKicks[j] << endl;
  }
  }
  }

  void WorldGLWidget::differParams(int param){
  for (int robot = 4; robot < 8; robot++){
  sims[robot]->behaviorParams->mainStrategy = sims[3]->behaviorParams->mainStrategy;
  switch (param){
  case 0:
  sims[robot]->behaviorParams->mainStrategy.forwardOpeningAngle = sims[3]->behaviorParams->mainStrategy.forwardOpeningAngle + DEG_T_RAD*20.0;
  break;
  case 1:
  sims[robot]->behaviorParams->mainStrategy.forwardOpeningAngle = sims[3]->behaviorParams->mainStrategy.forwardOpeningAngle - DEG_T_RAD*20.0;
  break;
  case 2:
  sims[robot]->behaviorParams->mainStrategy.maxArcAngle = sims[3]->behaviorParams->mainStrategy.maxArcAngle + DEG_T_RAD*20.0;
  break;
  case 3:
  sims[robot]->behaviorParams->mainStrategy.maxArcAngle = sims[3]->behaviorParams->mainStrategy.maxArcAngle - DEG_T_RAD*20.0;
  if (sims[robot]->behaviorParams->mainStrategy.maxArcAngle < 0) sims[robot]->behaviorParams->mainStrategy.maxArcAngle = 0;
  break;
  case 4:
  sims[robot]->behaviorParams->mainStrategy.minOpponentDist = sims[3]->behaviorParams->mainStrategy.minOpponentDist + 250;
  break;
  case 5:
  sims[robot]->behaviorParams->mainStrategy.minOpponentDist = sims[3]->behaviorParams->mainStrategy.minOpponentDist - 250;
  break;
  case 6:
  sims[robot]->behaviorParams->mainStrategy.opponentWidth = sims[3]->behaviorParams->mainStrategy.opponentWidth + 25;
  break;
  case 7:
  sims[robot]->behaviorParams->mainStrategy.opponentWidth = sims[3]->behaviorParams->mainStrategy.opponentWidth - 25;
  break;
  case 8:
  sims[robot]->behaviorParams->mainStrategy.passDistance = sims[3]->behaviorParams->mainStrategy.passDistance + 100;
  break;
  case 9:
  sims[robot]->behaviorParams->mainStrategy.passDistance = sims[3]->behaviorParams->mainStrategy.passDistance - 100;
  break;
  case 10:
  sims[robot]->behaviorParams->mainStrategy.supportFwdDist = sims[3]->behaviorParams->mainStrategy.supportFwdDist + 250;
  break;
  case 11:
  sims[robot]->behaviorParams->mainStrategy.supportFwdDist = sims[3]->behaviorParams->mainStrategy.supportFwdDist - 250;
  break;
  case 12:
  sims[robot]->behaviorParams->mainStrategy.defendBackDist = sims[3]->behaviorParams->mainStrategy.defendBackDist + 250;
  break;
  case 13:
  sims[robot]->behaviorParams->mainStrategy.defendBackDist = sims[3]->behaviorParams->mainStrategy.defendBackDist - 250;
  break;
  case 14:
  sims[robot]->behaviorParams->mainStrategy.forwardFwdDist = sims[3]->behaviorParams->mainStrategy.forwardFwdDist + 250;
  break;
  case 15:
  sims[robot]->behaviorParams->mainStrategy.forwardFwdDist = sims[3]->behaviorParams->mainStrategy.forwardFwdDist - 250;
  break;

  case 16:
  // turn off short kicks
  for (int kick = FwdShortStraightKick; kick <= FwdShortRightwardKick; kick++){
  sims[robot]->behaviorParams->mainStrategy.usableKicks[kick] = !sims[3]->behaviorParams->mainStrategy.usableKicks[kick];
  }
  break;
  case 17:
  // turn off pass kicks
  for (int kick = FwdPass4Kick; kick <= FwdPass2Kick; kick++){
  sims[robot]->behaviorParams->mainStrategy.usableKicks[kick] = !sims[3]->behaviorParams->mainStrategy.usableKicks[kick];
  }
  break;
  case 18:
  // turn off long kicks
  for (int kick = FwdLongStraightKick; kick <= FwdLongRightwardKick; kick++){
  sims[robot]->behaviorParams->mainStrategy.usableKicks[kick] = !sims[3]->behaviorParams->mainStrategy.usableKicks[kick];
  }
  break;
  case 19:
  // long kicks only (turn off pass,medium,short)
  for (int kick = FwdMediumStraightKick; kick <= FwdShortRightwardKick; kick++){
  sims[robot]->behaviorParams->mainStrategy.usableKicks[kick] = !sims[3]->behaviorParams->mainStrategy.usableKicks[kick];
  }
  for (int kick = FwdPass4Kick; kick <= FwdPass2Kick; kick++){
  sims[robot]->behaviorParams->mainStrategy.usableKicks[kick] = !sims[3]->behaviorParams->mainStrategy.usableKicks[kick];
  }
  break;
  case 20:
  // bigger max kick angle
  sims[robot]->behaviorParams->mainStrategy.maxKickAngle = sims[3]->behaviorParams->mainStrategy.maxKickAngle + DEG_T_RAD*10.0;
  break;
  case 21:
  // smaller max kick angle
  sims[robot]->behaviorParams->mainStrategy.maxKickAngle = sims[3]->behaviorParams->mainStrategy.maxKickAngle - DEG_T_RAD*10.0;
  if (sims[robot]->behaviorParams->mainStrategy.maxKickAngle < 0) sims[robot]->behaviorParams->mainStrategy.maxKickAngle = 0;
  break;
  case 22:
  // try dribbling in clusters
  sims[robot]->behaviorParams->mainStrategy.clusterBehavior = Cluster::DRIBBLE;
  break;
  case 23:
  // smaller shooting radius
  sims[robot]->behaviorParams->mainStrategy.shootOnGoalRadius = sims[3]->behaviorParams->mainStrategy.shootOnGoalRadius - 500.0;
  break;
  case 24:
  // bigger shooting radius
  sims[robot]->behaviorParams->mainStrategy.shootOnGoalRadius = sims[3]->behaviorParams->mainStrategy.shootOnGoalRadius + 500.0;
  break;
  case 25:
  // try even bigger arc
  sims[robot]->behaviorParams->mainStrategy.maxArcAngle = 45.0 *DEG_T_RAD;
  break;
  case 26:
  // try even bigger opponent dist
  sims[robot]->behaviorParams->mainStrategy.minOpponentDist = 1100.0;
  break;
  case 27:
  // try smaller pass distance
  sims[robot]->behaviorParams->mainStrategy.passDistance = 650.0;
  break;
  case 28:
  // try smaller arc
  sims[robot]->behaviorParams->mainStrategy.maxArcAngle = 20.0 *DEG_T_RAD;
  break;
  default:
  // no change
  break;
  }
  // re-init strategy
  sims[robot]->setStrategy();
  }
  }


  void WorldGLWidget::differBHumanParams(int param){
  for (int robot = 4; robot < 8; robot++){
  sims[robot]->behaviorParams->mainStrategy = sims[3]->behaviorParams->mainStrategy;
  switch (param){
  case 1:
  sims[robot]->behaviorParams->mainStrategy.maxArcAngle = sims[3]->behaviorParams->mainStrategy.maxArcAngle + DEG_T_RAD*10.0;
  break;
  case 2:
  sims[robot]->behaviorParams->mainStrategy.maxArcAngle = sims[3]->behaviorParams->mainStrategy.maxArcAngle + DEG_T_RAD*20.0;
  if (sims[robot]->behaviorParams->mainStrategy.maxArcAngle < 0) sims[robot]->behaviorParams->mainStrategy.maxArcAngle = 0;
  break;
  case 3:
  sims[robot]->behaviorParams->mainStrategy.minOpponentDist = sims[3]->behaviorParams->mainStrategy.minOpponentDist + 150;
  break;
  case 4:
  sims[robot]->behaviorParams->mainStrategy.minOpponentDist = sims[3]->behaviorParams->mainStrategy.minOpponentDist + 300;
  break;
  case 5:
  sims[robot]->behaviorParams->mainStrategy.minOpponentDist = sims[3]->behaviorParams->mainStrategy.minOpponentDist + 450;
  break;
  case 6:
  // bigger shooting radius
  sims[robot]->behaviorParams->mainStrategy.shootOnGoalRadius = sims[3]->behaviorParams->mainStrategy.shootOnGoalRadius + 500.0;
  break;
  case 7:
  // bigger shooting radius
  sims[robot]->behaviorParams->mainStrategy.shootOnGoalRadius = sims[3]->behaviorParams->mainStrategy.shootOnGoalRadius + 1000.0;
  break;
  case 8:
  // bigger shooting radius
  sims[robot]->behaviorParams->mainStrategy.shootOnGoalRadius = sims[3]->behaviorParams->mainStrategy.shootOnGoalRadius + 1500.0;
  break;
  case 9:
  // bigger shooting radius
  sims[robot]->behaviorParams->mainStrategy.shootOnGoalRadius = sims[3]->behaviorParams->mainStrategy.shootOnGoalRadius + 2000.0;
  break;
  case 10:
  // bigger orientation error factor
  sims[robot]->behaviorParams->mainStrategy.orientationErrorFactor = sims[3]->behaviorParams->mainStrategy.orientationErrorFactor * 2.0;
  break;
  case 11:
  // bigger orientation error factor
  sims[robot]->behaviorParams->mainStrategy.orientationErrorFactor = sims[3]->behaviorParams->mainStrategy.orientationErrorFactor * 5.0;
  break;
  case 12:
  // bigger orientation error factor
  sims[robot]->behaviorParams->mainStrategy.orientationErrorFactor = sims[3]->behaviorParams->mainStrategy.orientationErrorFactor * 10.0;
  break;
  case 13:
  // bigger orientation error factor
  sims[robot]->behaviorParams->mainStrategy.orientationErrorFactor = sims[3]->behaviorParams->mainStrategy.orientationErrorFactor * 20.0;
  break;


  case 14:
  sims[robot]->behaviorParams->mainStrategy.supportFwdDist = sims[3]->behaviorParams->mainStrategy.supportFwdDist + 500;
  break;
  case 15:
  sims[robot]->behaviorParams->mainStrategy.supportFwdDist = sims[3]->behaviorParams->mainStrategy.supportFwdDist - 500;
  break;
  case 16:
  sims[robot]->behaviorParams->mainStrategy.defendBackDist = sims[3]->behaviorParams->mainStrategy.defendBackDist + 500;
  break;
  case 17:
  sims[robot]->behaviorParams->mainStrategy.defendBackDist = sims[3]->behaviorParams->mainStrategy.defendBackDist - 500;
  break;
  case 18:
  sims[robot]->behaviorParams->mainStrategy.forwardFwdDist = sims[3]->behaviorParams->mainStrategy.forwardFwdDist + 500;
  break;
  case 19:
  sims[robot]->behaviorParams->mainStrategy.forwardFwdDist = sims[3]->behaviorParams->mainStrategy.forwardFwdDist - 500;
  break;

  // long kicks only (turn off pass,medium,short)
  for (int kick = FwdMediumStraightKick; kick <= FwdShortRightwardKick; kick++){
  sims[robot]->behaviorParams->mainStrategy.usableKicks[kick] = !sims[3]->behaviorParams->mainStrategy.usableKicks[kick];
  }
  for (int kick = FwdPass4Kick; kick <= FwdPass2Kick; kick++){
  sims[robot]->behaviorParams->mainStrategy.usableKicks[kick] = !sims[3]->behaviorParams->mainStrategy.usableKicks[kick];
  }
  break;

  case 20:
  // bigger max kick angle
  sims[robot]->behaviorParams->mainStrategy.maxKickAngle = sims[3]->behaviorParams->mainStrategy.maxKickAngle + DEG_T_RAD*10.0;
  break;
  case 21:
  // smaller max kick angle
  sims[robot]->behaviorParams->mainStrategy.maxKickAngle = sims[3]->behaviorParams->mainStrategy.maxKickAngle - DEG_T_RAD*10.0;
  if (sims[robot]->behaviorParams->mainStrategy.maxKickAngle < 0) sims[robot]->behaviorParams->mainStrategy.maxKickAngle = 0;
  break;


  default:
  // no change
  break;
  }
  // re-init strategy
  sims[robot]->setStrategy();
  }
  }

*/
