#include <QtGui>
#include "CalibrationWindow.h"

class SampleItem : public QListWidgetItem {
    private:
        int _x, _y;
    public:
        SampleItem();
        SampleItem(int,int);
};

SampleItem::SampleItem() : QListWidgetItem() {}

SampleItem::SampleItem(int x, int y){
    _x = x;
    _y = y;
    QString text = "X=" + QString::number(x) + ", Y=" + QString::number(y);
    setText(text);
}

void CalibrationWindow::addSample(Sample s){
    SampleItem* item = new SampleItem(s.x,s.y);
    if(s.camera == Camera::TOP)
        topSamplesList->addItem(item);
    else
        bottomSamplesList->addItem(item);
    samples_.push_back(s);
}

void CalibrationWindow::loadCalibration(const CameraCalibration& cal) {
    topTiltOffset->setValue(cal.topCameraTiltOffset);
    topRollOffset->setValue(cal.topCameraRollOffset);
    topYawOffset->setValue(cal.topCameraYawOffset);
    topZOffset->setValue(cal.topCameraZOffset);
    topXOffset->setValue(cal.topCameraXOffset);
    bottomTiltOffset->setValue(cal.bottomCameraTiltOffset);
    bottomRollOffset->setValue(cal.bottomCameraRollOffset);
    bottomYawOffset->setValue(cal.bottomCameraYawOffset);
    bottomZOffset->setValue(cal.bottomCameraZOffset);
    bottomXOffset->setValue(cal.bottomCameraXOffset);
    headTiltOffset->setValue(cal.headTiltOffset);
    headPanOffset->setValue(cal.headPanOffset);
    topFOVx->setValue(cal.topCameraFOVx);
    topFOVy->setValue(cal.topCameraFOVy);
    bottomFOVx->setValue(cal.bottomCameraFOVx);
    bottomFOVy->setValue(cal.bottomCameraFOVy);
}

CameraCalibration CalibrationWindow::getCalibration() const {
    CameraCalibration cal;
    cal.topCameraTiltOffset = topTiltOffset->value();
    cal.topCameraRollOffset = topRollOffset->value();
    cal.topCameraYawOffset = topYawOffset->value();
    cal.topCameraZOffset = topZOffset->value();
    cal.topCameraXOffset = topXOffset->value();
    cal.bottomCameraTiltOffset = bottomTiltOffset->value();
    cal.bottomCameraRollOffset = bottomRollOffset->value();
    cal.bottomCameraZOffset = bottomZOffset->value();
    cal.bottomCameraXOffset = bottomXOffset->value();
    cal.bottomCameraYawOffset = bottomYawOffset->value();
    cal.headTiltOffset = headTiltOffset->value();
    cal.headPanOffset = headPanOffset->value();
    cal.topCameraFOVx = topFOVx->value();
    cal.topCameraFOVy = topFOVy->value();
    cal.bottomCameraFOVx = bottomFOVx->value();
    cal.bottomCameraFOVy = bottomFOVy->value();
    return cal;
}

void CalibrationWindow::saveCalibration(std::string file) {
    CameraCalibration cal = getCalibration();
    cal.saveToFile(file);
}

void CalibrationWindow::loadCalibration(std::string file) {
    CameraCalibration cal;
    cal.loadFromFile(file);
    loadCalibration(cal);
}

void CalibrationWindow::resetParameters(){
    loadCalibration(calibration_file_);
}

void CalibrationWindow::clear(){
    topSamplesList->clear();
    bottomSamplesList->clear();
    samples_.clear();
}

void CalibrationWindow::save(){
    std::cout << "Saving calibration to file: " << calibration_file_ << "...";
    saveCalibration(calibration_file_);
    std::cout << "done!\n";
}

void CalibrationWindow::saveAs(){
  QString file = QFileDialog::getSaveFileName(this, tr("Save Camera Calibration"),
                                                  QString(getenv("NAO_HOME")) + "/data",
                                                  tr("Camera Calibration Files (*.cal)"));
  if (file.isEmpty()) return;
  calibration_file_ = file.toUtf8().constData();
  save();
}

void CalibrationWindow::load(){
    QString file = QFileDialog::getOpenFileName(this, tr("Open Camera Calibration"),
                                              QString(getenv("NAO_HOME")) + "/data",
                                              tr("Camera Calibration Files (*.cal)"));
    if (file.isEmpty())
    return;

    calibration_file_ = file.toUtf8().constData();
    std::cout << "Loading calibration from file: " << calibration_file_ << "...";
    loadCalibration(calibration_file_);
    std::cout << "done!\n";

}

CalibrationWindow::CalibrationWindow(QMainWindow* parent, QMainWindow *main) {
  setupUi(this);

  connect(updateButton, SIGNAL(clicked()), main, SLOT(updateCalibrationFile()));
  connect(getButton, SIGNAL(clicked()), main, SLOT(readCalibrationFile()));
  connect(optimizeButton, SIGNAL(clicked()), parent, SLOT(optimizeCalibration()));
  connect(clearButton,SIGNAL(clicked()), this, SLOT(clear()));
  connect(clearButton,SIGNAL(clicked()), parent, SLOT(clearSamples()));
  connect(resetButton,SIGNAL(clicked()), this, SLOT(resetParameters()));
  connect(saveButton,SIGNAL(clicked()), this, SLOT(save()));
  connect(saveAsButton,SIGNAL(clicked()), this, SLOT(saveAs()));
  connect(loadButton,SIGNAL(clicked()), this, SLOT(load()));

  connect(topTiltOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(topRollOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(topYawOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(topZOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(topXOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(bottomTiltOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(bottomRollOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(bottomYawOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(bottomZOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(bottomXOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(headPanOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(headTiltOffset,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(poseX,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(poseY,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(poseRotation,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));

  connect(topFOVx,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(topFOVy,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(bottomFOVx,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));
  connect(bottomFOVy,SIGNAL(valueChanged(double)), parent, SLOT(calibrationsUpdated()));

  calibration_file_ = (QString(getenv("NAO_HOME")) + "/data/defaultcamera.cal").toUtf8().constData();
  loadCalibration(calibration_file_);

}
