#ifndef MOTION_WIDGET_H
#define MOTION_WIDGET_H

#include <QGLViewer/qglviewer.h>
#include <QWidget>

#include <vector>

#include "utOpenGL/BasicGL.h"
#include "utOpenGL/RobotGL.h"

#include <memory/Memory.h>
//#include <memory/AbstractMemory.h>
#include <memory/BodyModelBlock.h>
#include <memory/SensorBlock.h>
#include <memory/FrameInfoBlock.h>
#include <memory/JointBlock.h>
#include <memory/JointCommandBlock.h>
#include <memory/WalkRequestBlock.h>
#include <memory/WalkParamBlock.h>
#include <memory/SensorCalibrationBlock.h>
#include <memory/OdometryBlock.h>
#include <memory/KickRequestBlock.h>
#include <memory/KickEngineBlock.h>
#include <memory/KickModuleBlock.h>
#include <memory/KickParamBlock.h>

#include <math/Vector2.h>
#include <math/Vector3.h>
#include <math/Spline3D.h>
#include <common/RobotDimensions.h>
#include <common/RobotInfo.h>
#include <common/RingQueue.h>

#include <common/MassCalibration.h>

#define TIME_TO_HEIGHT 200.0

class MotionSimulation;

class MotionWidget : public QGLViewer {
Q_OBJECT        // must include this if you use Qt signals/slots
 public:
  MotionWidget(QWidget* parent);

  enum {  // Display modes
    BODYMODELMODE,
    JOINTVALUESMODE,
    JOINTCOMMANDSMODE,
    WALKSIMMODE,
    WALKMODE,
    KICKMODE,
    KICKSIMMODE,
    GETUPSIMMODE,
    NUM_MODES
  };
  
  enum { // Display options
    SHOWBODYMODEL,
    SHOWJOINTVALUESMODEL,
    SHOWJOINTCOMMANDSMODEL,
    SHOWSENSORSTEXT,
    SHOWWALKTEXT,
    SHOWFRAMETEXT,
    SHOWWALKREQUESTTEXT,
    SHOWSTEPS,
    SHOWSTEPSTEXT,
    SHOWZMPREF,
    SHOWCURRENTZMP,
    SHOWDESIREDZMP,
    SHOWSENSEDZMP,
    SHOWCURRENTPEN,
    SHOWDESIREDPEN,
    SHOWZMPPENTEXT,
    SHOWSWINGFOOT,
    SHOWFEETTEXT,
    SHOWABSFEETTARGETS,
    SHOWODOMETRYTEXT,
    SHOWTARGETPT,
    SHOWKICKFOOTTARGET,
    SHOWKICKREQUESTTEXT,
    SHOWKICKTARGETTEXT,
    SHOWKICKINFO,
    SHOWKICKSPLINE,
    NUM_DISPLAY_OPTIONS
  };



  virtual void draw();
  virtual void init();

  QWidget* parent;
  void updateMemory(Memory* mem);

 private:

  BasicGL basicGL_;
  RobotGL robotGL_;
  
  Vector3<float> stick_figure_[BodyFrame::NUM_POINTS];

  BodyModelBlock* body_model_;
  SensorBlock* sensors_;
  FrameInfoBlock* frame_;
  JointBlock* joint_values_;
  JointCommandBlock* joint_commands_;
  WalkRequestBlock* walk_request_;
  WalkParamBlock* walk_param_;
  SensorCalibrationBlock* sensor_calibration_;
  OdometryBlock* odometry_;
  KickRequestBlock* kick_request_;
  KickEngineBlock* kick_engine_;
  KickModuleBlock* kick_module_;
  KickParamBlock* kick_params_;

  RobotDimensions dimensions_;
  MassCalibration mass_calibration_;

  int currentMode;
  void setMode(int mode);
  void setAllDisplayOptions(bool val);
  bool displayOptions[NUM_DISPLAY_OPTIONS];

  void calculateStickFigure(BodyModelBlock* bm);
  void drawBodyModel(BodyModelBlock* bodyModel, RGB color);
  void drawBodyModelFromJointCommands();
  void drawBodyModelFromJointValues();
  BodyModelBlock* getBodyModelFromJoints(float *joints);
  void drawSteps();
  void drawZmpTrajectory(RGB color, RingQueue<Vector2<float>, 100> zmp);
  void drawZmp(RGB color, Vector2<float> zmp, float time);
  void drawPen(RGB color, Vector2<float> pen, float time);
  void drawSwingFoot();
  void drawAbsFeetTargets();
  void drawTargetPoint();

  void drawKickFootTarget();
  void drawKickSpline(bool names = false);
  void addTimeToSwingSplinePoint(Vector3<float> &pt,float time_frac);

  Pose3D globalToDrawingFrame(Pose3D a);
  Pose3D globalToDrawingFrame(Pose2D a);
  Pose3D globalToDrawingFrame(Vector2<float> a);

  QString Vector2ToString(Vector2<float> val);
  QString Vector3ToString(Vector3<float> val);

  void textSensors();
  void textWalk();
  void textFrame();
  void textWalkRequest();
  void textSteps();
  void textZmpCom();
  void textFeet();
  void textOdometry();
  void textKickFeetTargets();
  void textKickRequest();
  void textKickInfo();

  // motion simulation
  MotionSimulation* motion_sim_;
  Memory* memory_;
  void initMotionSim(bool kick);
  void simulationStep();
  std::vector<Memory> simLog;

  void createKickSplines();
  void calcKickSplinePoints();
  void moveSplinePoint(int x, int y, int z);
  void removeSplinePoint();
  void addSplinePoint();
  Spline3D kick_stance_spline_;
  Spline3D kick_swing_spline_;
  std::vector<Vector3<float> > kick_stance_spline_pts_;
  std::vector<Vector3<float> > kick_swing_spline_pts_;
  KickState::State prev_kick_state_;

  void drawWithNames();
  qglviewer::Vec orig, dir, selectedPoint;

 protected:
  void keyPressEvent(QKeyEvent *event);
  
 signals:
  void prevSnapshot();
  void nextSnapshot();
  void play();
  void pause();
  void modeChanged(QString);
  
};

	
#endif
