#include <QtGui>

#include "PlotWindow.h"
#include <math/Pose3D.h>
#include <math/Geometry.h>
#include <memory/FrameInfoBlock.h>
#include <memory/GraphableBlock.h>
#include <qwt_plot_canvas.h>
#include <fstream>

PlotWindow::PlotWindow() :
  QWidget(),
  memory_log_(NULL)
{
  QGridLayout *layout = new QGridLayout;

  plot_.setCanvasBackground("white");
  layout->addWidget(&plot_, 0, 0, 5, 8);

  // add marker
  QPen pen;
  pen.setColor("black");
  pen.setWidth(2);
  plot_marker_.setLineStyle(QwtPlotMarker::VLine);
  plot_marker_.setLinePen(pen);
  plot_marker_.attach(&plot_);
  
  // add panner and magnifier
  panner_ = new QwtPlotPanner(plot_.canvas());
  magnifier_ = new QwtPlotMagnifier(plot_.canvas());
  // add legend
  legend_ = new QwtLegend;
  plot_.insertLegend(legend_,QwtPlot::TopLegend);
        
  setLayout(layout);
  resize(1000,600);
  setWindowTitle(tr("Plot System"));
}

PlotWindow::~PlotWindow() {
  for (unsigned int i = 0; i < plot_curves_.size(); i++)
    delete plot_curves_[i];
  delete panner_;
  delete magnifier_;
  delete legend_;
}

void PlotWindow::addLine(std::string name, std::string color) {
  GraphableBlock *graphable(NULL);
  double *data = new double[times_.size()];
  
  for (unsigned int i = 0; i < memory_log_->size(); i++) {
    (*memory_log_)[i].getBlockByName(graphable,"graphable");
    if (graphable == NULL) return;
    data[i] = graphable->getData(name.c_str());
  }
  
  QPen pen;
  pen.setWidth(3);
  pen.setColor(color.c_str());
  
  plot_curves_.push_back(new QwtPlotCurve(name.c_str()));
  int index = plot_curves_.size() - 1;
  plot_curves_[index]->setData(&times_[0],data,memory_log_->size());
  plot_curves_[index]->attach(&plot_);
  plot_curves_[index]->setPen(pen);

  delete [] data;
}

void PlotWindow::setMemoryLog(std::vector<Memory> *memory_log) {
  memory_log_ = memory_log;

  for (unsigned int i = 0; i < memory_log_->size(); i++)
    populateGraphableBlock((*memory_log_)[i]);

  // set up the time
  FrameInfoBlock *frame_info(NULL);

  // clear times and data for new log
  times_.clear();
  for (unsigned i = 0; i < plot_curves_.size(); i++){
    plot_curves_[i]->detach();
    delete plot_curves_[i];
  }
  plot_curves_.clear();

  for (unsigned int i = 0; i < memory_log_->size(); i++) {
    (*memory_log_)[i].getBlockByName(frame_info,"frame_info",false);
    if (frame_info == NULL)
      (*memory_log_)[i].getBlockByName(frame_info,"vision_frame_info",false);
    if (frame_info == NULL)
      continue;
    times_.push_back(frame_info->seconds_since_start);
  }
  
  std::string filename = std::string(getenv("NAO_HOME")) + "/tools/UTNaoTool/graph.csv";
  std::ifstream in(filename.c_str());
  std::string name;
  std::string color;
  if (!in.good()) {
    std::cout << "ERROR opening: " << filename << std::endl;
    return;
  }
  std::cout << "Graphing from: " << filename << std::endl;
  while (!in.eof()) {
    getline(in,name,',');
    if (in.eof())
      break;
    getline(in,color,'\n');
    if (in.eof())
      break;
    if (name[0] == '#')
      continue;
    addLine(name,color);
  }
  //addLine("com.x","green");
  //addLine("commanded com.x","blue");
  plot_.replot();
}

void PlotWindow::update(Memory* mem) {
  FrameInfoBlock *frame_info(NULL);
  mem->getBlockByName(frame_info,"frame_info",false);
  if (frame_info == NULL) return;
  plot_marker_.setXValue(frame_info->seconds_since_start);
  plot_.replot();
}

void PlotWindow::mousePressEvent(QMouseEvent *event) {
  if (event->button() != Qt::MidButton)
    return;
  
  int offset = plot_.canvas()->x() + plot_.x();
  int xpos = event->x() - offset;
  double desired_time =  plot_.invTransform(QwtPlot::xBottom,xpos);
  int snapshot_index = -1;

  for (unsigned int i = 0; i < times_.size(); i++) {
    if (times_[i] > desired_time) {
      snapshot_index = i;
      break;
    }
  }
  if (snapshot_index < 0)
    snapshot_index = times_.size() - 1;
  emit gotoSnapshot(snapshot_index);
}

void PlotWindow::keyPressEvent(QKeyEvent *event) {
  //bool ctrl = event->modifiers() & (Qt::ControlModifier);
  switch (event->key()) {
  case Qt::Key_Left:
    emit prevSnapshot();
    break;
  case Qt::Key_Right:
    emit nextSnapshot();
    break;
  case Qt::Key_Up:
    emit play();
    break;
  case Qt::Key_Down:
    emit pause();
    break;
  }
}

#include <memory/WalkParamBlock.h>
#include <memory/BodyModelBlock.h>
#include <memory/SensorBlock.h>
#include <memory/JointBlock.h>
#include <memory/JointCommandBlock.h>
#include <memory/OdometryBlock.h>
#include <memory/WorldObjectBlock.h>

void PlotWindow::populateGraphableBlock(Memory &memory) {
  GraphableBlock *graphable = NULL;
  memory.getOrAddBlockByName(graphable,"graphable");

  BodyModelBlock *body_model;
  memory.getBlockByName(body_model,"body_model",false);
  if (body_model != NULL) {
    graphable->addData("com.x",body_model->center_of_mass_.x);
    graphable->addData("com.y",body_model->center_of_mass_.y);
    graphable->addData("com.z",body_model->center_of_mass_.z);
     
    graphable->addData("sensor_tilt", RAD_T_DEG*body_model->sensors_tilt_roll_.tilt_);
    graphable->addData("sensor_roll", RAD_T_DEG*body_model->sensors_tilt_roll_.roll_);
    graphable->addData("left_foot_tilt", RAD_T_DEG*body_model->left_foot_body_tilt_roll_.tilt_);
    graphable->addData("left_foot_roll",RAD_T_DEG* body_model->left_foot_body_tilt_roll_.roll_);
    graphable->addData("right_foot_tilt", RAD_T_DEG*body_model->right_foot_body_tilt_roll_.tilt_);
    graphable->addData("right_foot_roll", RAD_T_DEG*body_model->right_foot_body_tilt_roll_.roll_);
  }

  SensorBlock* sensors;
  memory.getBlockByName(sensors,"processed_sensors",false);
  if (sensors != NULL){
    graphable->addData("sensor_accel.x",sensors->values_[accelX]);
    graphable->addData("sensor_accel.y",sensors->values_[accelY]);
    graphable->addData("sensor_accel.z",sensors->values_[accelZ]);
    graphable->addData("sensor_tilt",sensors->values_[angleY]);
    graphable->addData("sensor_roll",sensors->values_[angleX]);
  }
  
  SensorBlock* raw_sensors;
  memory.getBlockByName(raw_sensors,"raw_sensors",false);
  if (raw_sensors != NULL) {
    graphable->addData("raw_sensor_tilt",raw_sensors->values_[angleY]);
    graphable->addData("raw_sensor_roll",raw_sensors->values_[angleX]);
  }

  JointBlock *joint_angles;
  memory.getBlockByName(joint_angles,"processed_joint_angles",false);
  if (joint_angles != NULL) {
    graphable->addData("sensed_lhiproll",RAD_T_DEG * joint_angles->values_[LHipRoll]);
    graphable->addData("sensed_rhiproll",RAD_T_DEG * joint_angles->values_[RHipRoll]);
    graphable->addData("sensed_lhippitch",RAD_T_DEG * joint_angles->values_[LHipPitch]);
    graphable->addData("sensed_rhippitch",RAD_T_DEG * joint_angles->values_[RHipPitch]);

    graphable->addData("sensed_lanklepitch",RAD_T_DEG * joint_angles->values_[LAnklePitch]);
    graphable->addData("sensed_ranklepitch",RAD_T_DEG * joint_angles->values_[RAnklePitch]);
    graphable->addData("sensed_lankleroll",RAD_T_DEG * joint_angles->values_[LAnkleRoll]);
    graphable->addData("sensed_rankleroll",RAD_T_DEG * joint_angles->values_[RAnkleRoll]);
  }
  
  JointCommandBlock *joint_commands;
  memory.getBlockByName(joint_commands,"processed_joint_commands",false);
  if (joint_commands != NULL) {
    graphable->addData("commanded_lhiproll",RAD_T_DEG * joint_commands ->angles_[LHipRoll]);
    graphable->addData("commanded_rhiproll",RAD_T_DEG * joint_commands ->angles_[RHipRoll]);
    graphable->addData("commanded_lhippitch",RAD_T_DEG * joint_commands ->angles_[LHipPitch]);
    graphable->addData("commanded_rhippitch",RAD_T_DEG * joint_commands ->angles_[RHipPitch]);
    
    graphable->addData("commanded_lkneepitch",RAD_T_DEG * joint_commands ->angles_[LKneePitch]);
    graphable->addData("commanded_rkneepitch",RAD_T_DEG * joint_commands ->angles_[RKneePitch]);
    
    graphable->addData("commanded_lanklepitch",RAD_T_DEG * joint_commands->angles_[LAnklePitch]);
    graphable->addData("commanded_ranklepitch",RAD_T_DEG * joint_commands->angles_[RAnklePitch]);
    graphable->addData("commanded_lankleroll",RAD_T_DEG * joint_commands->angles_[LAnkleRoll]);
    graphable->addData("commanded_rankleroll",RAD_T_DEG * joint_commands->angles_[RAnkleRoll]);
  }
  
  OdometryBlock *odometry;
  memory.getBlockByName(odometry,"vision_odometry",false);
  if (odometry != NULL) {
    graphable->addData("odom.y",odometry->displacement.translation.y);
  }

  WorldObjectBlock *world_object;
  memory.getBlockByName(world_object,"world_objects",false);
  if (world_object != NULL) {
    WorldObject &ball = world_object->objects_[WO_BALL];
    graphable->addData("rel_ball.y",ball.relPos.y);
    graphable->addData("vision_rel_ball.y",ball.visionDistance * sin(ball.visionBearing));
  }
}
